@@ -21,7 +21,6 @@
 #![deny(clippy::trivially_copy_pass_by_ref)]
 #![feature(box_patterns)]
 #![feature(if_let_guard)]
-#![feature(let_chains)]
 
 pub mod alias;
 pub mod annotation;
@@ -36,7 +35,6 @@ pub mod lit_int;
 pub mod literal;
 pub mod module;
 pub mod param_spec;
-pub mod qname;
 pub mod quantified;
 pub mod read_only;
 pub mod simplify;
