@@ -5,6 +5,7 @@
  * LICENSE file in the root directory of this source tree.
  */
 
+use std::borrow::Cow;
 use std::cmp::Ordering;
 use std::fmt;
 use std::fmt::Display;
@@ -15,6 +16,7 @@ use parse_display::Display;
 use pyrefly_derive::TypeEq;
 use pyrefly_derive::Visit;
 use pyrefly_derive::VisitMut;
+use pyrefly_python::qname::QName;
 use pyrefly_util::assert_words;
 use pyrefly_util::display::commas_iter;
 use pyrefly_util::uniques::Unique;
@@ -41,7 +43,6 @@ use crate::keywords::KwCall;
 use crate::literal::Lit;
 use crate::module::ModuleType;
 use crate::param_spec::ParamSpec;
-use crate::qname::QName;
 use crate::quantified::Quantified;
 use crate::quantified::QuantifiedKind;
 use crate::simplify::unions;
@@ -77,6 +78,23 @@ impl Var {
     }
 }
 
+#[derive(PartialEq, Eq)]
+pub enum TParamsSource {
+    Class,
+    TypeAlias,
+    Function,
+}
+
+impl Display for TParamsSource {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::Class => write!(f, "class"),
+            Self::TypeAlias => write!(f, "type alias"),
+            Self::Function => write!(f, "function"),
+        }
+    }
+}
+
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
 #[derive(Visit, VisitMut, TypeEq)]
 pub struct TParam {
@@ -189,6 +207,10 @@ impl TArgs {
         self.0.0.iter().zip(self.0.1.iter())
     }
 
+    pub fn iter_paired_mut(&mut self) -> impl ExactSizeIterator<Item = (&TParam, &mut Type)> {
+        self.0.0.iter().zip(self.0.1.iter_mut())
+    }
+
     pub fn len(&self) -> usize {
         self.0.1.len()
     }
@@ -213,31 +235,44 @@ impl TArgs {
     /// This is mainly useful to take ancestors coming from the MRO (which are always in terms
     /// of the current class's type parameters) and re-express them in terms of the current
     /// class specialized with type arguments.
-    pub fn apply_substitution(&self, substitution: &Substitution) -> Self {
+    pub fn substitute_with(&self, substitution: &Substitution) -> Self {
         let tys = self
             .0
             .1
             .iter()
-            .map(|ty| substitution.substitute(ty.clone()))
+            .map(|ty| substitution.substitute_into(ty.clone()))
             .collect();
         Self::new(self.0.0.dupe(), tys)
     }
 
-    pub fn substitution<'a>(&'a self) -> Substitution<'a> {
+    pub fn substitution_map(&self) -> SmallMap<&Quantified, &Type> {
         let tparams = self.tparams();
         let tys = self.as_slice();
-        Substitution(tparams.quantifieds().zip(tys.iter()).collect())
+        tparams.quantifieds().zip(tys.iter()).collect()
     }
 
-    pub fn substitute(&self, ty: Type) -> Type {
-        self.substitution().substitute(ty)
+    pub fn substitution<'a>(&'a self) -> Substitution<'a> {
+        Substitution(self.substitution_map())
+    }
+
+    pub fn substitute_into_mut(&self, ty: &mut Type) {
+        self.substitution().substitute_into_mut(ty)
+    }
+
+    pub fn substitute_into(&self, mut ty: Type) -> Type {
+        self.substitute_into_mut(&mut ty);
+        ty
     }
 }
 
 pub struct Substitution<'a>(SmallMap<&'a Quantified, &'a Type>);
 
 impl<'a> Substitution<'a> {
-    pub fn substitute(&self, ty: Type) -> Type {
+    pub fn substitute_into_mut(&self, ty: &mut Type) {
+        ty.subst_mut(&self.0)
+    }
+
+    pub fn substitute_into(&self, ty: Type) -> Type {
         ty.subst(&self.0)
     }
 }
@@ -289,17 +324,23 @@ pub struct TypeAlias {
     pub name: Box<Name>,
     ty: Box<Type>,
     pub style: TypeAliasStyle,
+    annotated_metadata: Box<[Type]>,
 }
 
 impl TypeAlias {
-    pub fn new(name: Name, ty: Type, style: TypeAliasStyle) -> Self {
+    pub fn new(name: Name, ty: Type, style: TypeAliasStyle, annotated_metadata: Vec<Type>) -> Self {
         Self {
             name: Box::new(name),
             ty: Box::new(ty),
             style,
+            annotated_metadata: annotated_metadata.into_boxed_slice(),
         }
     }
 
+    pub fn annotated_metadata(&self) -> &[Type] {
+        &self.annotated_metadata
+    }
+
     /// Gets the type contained within the type alias for use in a value
     /// position - for example, for a function call or attribute access.
     pub fn as_value(&self, stdlib: &Stdlib) -> Type {
@@ -358,17 +399,22 @@ pub enum CalleeKind {
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
 #[derive(Visit, VisitMut, TypeEq)]
 pub struct BoundMethod {
+    /// Type of the self/cls argument,
     pub obj: Type,
+    /// Type of the function.
     pub func: BoundMethodType,
 }
 
 impl BoundMethod {
-    pub fn drop_self(&self) -> Option<Type> {
-        self.as_function().drop_first_param_of_unbound_callable()
+    pub fn with_bound_object(&self, obj: Type) -> Self {
+        Self {
+            obj,
+            func: self.func.clone(),
+        }
     }
 
-    pub fn as_function(&self) -> Type {
-        self.func.as_type()
+    pub fn as_type(self) -> Type {
+        Type::BoundMethod(Box::new(self))
     }
 }
 
@@ -381,30 +427,21 @@ pub enum BoundMethodType {
 }
 
 impl BoundMethodType {
-    pub fn as_type(&self) -> Type {
+    pub fn as_type(self) -> Type {
         match self {
-            Self::Function(func) => Type::Function(Box::new(func.clone())),
-            Self::Forall(forall) => {
-                Forallable::Function(forall.body.clone()).forall(forall.tparams.clone())
-            }
-            Self::Overload(overload) => Type::Overload(overload.clone()),
+            Self::Function(func) => Type::Function(Box::new(func)),
+            Self::Forall(forall) => Forallable::Function(forall.body).forall(forall.tparams),
+            Self::Overload(overload) => Type::Overload(overload),
         }
     }
 
-    pub fn subst_self_type_mut(
-        &mut self,
-        replacement: &Type,
-        is_subset: &dyn Fn(&Type, &Type) -> bool,
-    ) {
+    pub fn subst_self_type_mut(&mut self, replacement: &Type) {
         match self {
-            Self::Function(func) => func.signature.subst_self_type_mut(replacement, is_subset),
-            Self::Forall(forall) => forall
-                .body
-                .signature
-                .subst_self_type_mut(replacement, is_subset),
+            Self::Function(func) => func.signature.subst_self_type_mut(replacement),
+            Self::Forall(forall) => forall.body.signature.subst_self_type_mut(replacement),
             Self::Overload(overload) => {
                 for sig in overload.signatures.iter_mut() {
-                    sig.subst_self_type_mut(replacement, is_subset)
+                    sig.subst_self_type_mut(replacement)
                 }
             }
         }
@@ -455,44 +492,37 @@ impl Overload {
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
 #[derive(Visit, VisitMut, TypeEq)]
 pub enum OverloadType {
-    Callable(Function),
+    Function(Function),
     Forall(Forall<Function>),
 }
 
 impl OverloadType {
     pub fn as_type(&self) -> Type {
         match self {
-            Self::Callable(f) => Type::Callable(Box::new(f.signature.clone())),
+            Self::Function(f) => Type::Function(Box::new(f.clone())),
             Self::Forall(forall) => {
                 Forallable::Function(forall.body.clone()).forall(forall.tparams.clone())
             }
         }
     }
 
-    fn subst_self_type_mut(
-        &mut self,
-        replacement: &Type,
-        is_subset: &dyn Fn(&Type, &Type) -> bool,
-    ) {
+    fn subst_self_type_mut(&mut self, replacement: &Type) {
         match self {
-            Self::Callable(f) => f.signature.subst_self_type_mut(replacement, is_subset),
-            Self::Forall(forall) => forall
-                .body
-                .signature
-                .subst_self_type_mut(replacement, is_subset),
+            Self::Function(f) => f.signature.subst_self_type_mut(replacement),
+            Self::Forall(forall) => forall.body.signature.subst_self_type_mut(replacement),
         }
     }
 
     fn is_typeguard(&self) -> bool {
         match self {
-            Self::Callable(f) => f.signature.is_typeguard(),
+            Self::Function(f) => f.signature.is_typeguard(),
             Self::Forall(forall) => forall.body.signature.is_typeguard(),
         }
     }
 
     fn is_typeis(&self) -> bool {
         match self {
-            Self::Callable(f) => f.signature.is_typeis(),
+            Self::Function(f) => f.signature.is_typeis(),
             Self::Forall(forall) => forall.body.signature.is_typeis(),
         }
     }
@@ -507,7 +537,7 @@ pub struct Forall<T> {
 
 impl Forall<Forallable> {
     pub fn apply_targs(self, targs: TArgs) -> Type {
-        targs.substitute(self.body.as_type())
+        targs.substitute_into(self.body.as_type())
     }
 }
 
@@ -517,6 +547,7 @@ impl Forall<Forallable> {
 pub enum Forallable {
     TypeAlias(TypeAlias),
     Function(Function),
+    Callable(Callable),
 }
 
 impl Forallable {
@@ -531,23 +562,26 @@ impl Forallable {
         }
     }
 
-    pub fn name(&self) -> Name {
+    pub fn name(&self) -> Cow<'_, Name> {
         match self {
-            Self::Function(func) => func.metadata.kind.as_func_id().func,
-            Self::TypeAlias(ta) => (*ta.name).clone(),
+            Self::Function(func) => func.metadata.kind.function_name(),
+            Self::Callable(_) => Cow::Owned(Name::new_static("<callable>")),
+            Self::TypeAlias(ta) => Cow::Borrowed(&ta.name),
         }
     }
 
     pub fn as_type(self) -> Type {
         match self {
             Self::Function(func) => Type::Function(Box::new(func)),
-            Self::TypeAlias(ta) => Type::TypeAlias(ta),
+            Self::Callable(callable) => Type::Callable(Box::new(callable)),
+            Self::TypeAlias(ta) => Type::TypeAlias(Box::new(ta)),
         }
     }
 
     fn is_typeguard(&self) -> bool {
         match self {
             Self::Function(func) => func.signature.is_typeguard(),
+            Self::Callable(callable) => callable.is_typeguard(),
             Self::TypeAlias(_) => false,
         }
     }
@@ -555,6 +589,7 @@ impl Forallable {
     fn is_typeis(&self) -> bool {
         match self {
             Self::Function(func) => func.signature.is_typeis(),
+            Self::Callable(callable) => callable.is_typeis(),
             Self::TypeAlias(_) => false,
         }
     }
@@ -567,7 +602,7 @@ impl Forallable {
 #[derive(Visit, VisitMut, TypeEq)]
 pub enum SuperObj {
     Instance(ClassType),
-    Class(Class),
+    Class(ClassType),
 }
 
 // Note: The fact that Literal and LiteralString are at the front is important for
@@ -581,8 +616,7 @@ pub enum Type {
     /// A function declared using the `def` keyword.
     /// Note that the FunctionKind metadata doesn't participate in subtyping, and thus two types with distinct metadata are still subtypes.
     Function(Box<Function>),
-    /// A method of a class. The first `Box<Type>` is the self/cls argument,
-    /// and the second is the function.
+    /// A method of a class.
     BoundMethod(Box<BoundMethod>),
     /// An overloaded function.
     Overload(Overload),
@@ -616,7 +650,12 @@ pub enum Type {
     Module(ModuleType),
     Forall(Box<Forall<Forallable>>),
     Var(Var),
-    Quantified(Quantified),
+    /// The type of a value which is annotated with a type var.
+    Quantified(Box<Quantified>),
+    /// The type of type var _value_ itself, after it has been bound to a function or a class.
+    /// This is equivalent to Type::TypeVar/ParamSpec/TypeVarTuple as a value, but when used
+    /// in a type annotation, it becomes Type::Quantified.
+    QuantifiedValue(Box<Quantified>),
     TypeGuard(Box<Type>),
     TypeIs(Box<Type>),
     Unpack(Box<Type>),
@@ -626,16 +665,24 @@ pub enum Type {
     SpecialForm(SpecialForm),
     Concatenate(Box<[Type]>, Box<Type>),
     ParamSpecValue(ParamList),
-    /// Used to represent `P.args`. The spec describes it as an annotation,
-    /// but it's easier to think of it as a type that can't occur in nested positions.
-    Args(Quantified),
-    Kwargs(Quantified),
+    /// The type of a value which is annotated with `P.args`.
+    Args(Box<Quantified>),
+    /// The type of a value which is annotated with `P.kwargs`.
+    Kwargs(Box<Quantified>),
+    /// The type of the _value_ `P.args`.
+    /// This is equivalent to `typing.ParamSpecArgs`, but when used in a type annotation it
+    /// becomes Type::Args.
+    ArgsValue(Box<Quantified>),
+    /// The type of the _value_ `P.kwargs`.
+    /// This is equivalent to `typing.ParamSpecKwargs`, but when used in a type annotation it
+    /// becomes Type::Kwargs.
+    KwargsValue(Box<Quantified>),
     /// Used to represent a type that has a value representation, e.g. a class
     Type(Box<Type>),
     Ellipsis,
     Any(AnyStyle),
     Never(NeverStyle),
-    TypeAlias(TypeAlias),
+    TypeAlias(Box<TypeAlias>),
     /// Represents the result of a super() call. The first ClassType is the point in the MRO that attribute lookup
     /// on the super instance should start at (*not* the class passed to the super() call), and the second
     /// ClassType is the second argument (implicit or explicit) to the super() call. For example, in:
@@ -654,6 +701,14 @@ pub enum Type {
     /// Wraps the result of a function call whose keyword arguments have typing effects, like
     /// `typing.dataclass_transform(...)`.
     KwCall(Box<KwCall>),
+    /// All possible materializations of Any. A subset check with Type::Materialization succeeds
+    /// only if it would succeed with any type. This behaves like top (`object`) in one direction
+    /// and bottom (`Never`) in the other:
+    /// * `Materialization` <: `T` succeeds iff `object` <: `T` would succeed
+    /// * `T` <: `Materialization` succeeds iff `T` <: `Never` would succeed
+    ///
+    /// See https://typing.python.org/en/latest/spec/glossary.html#term-materialize.
+    Materialization,
     None,
 }
 
@@ -677,6 +732,7 @@ impl Visit for Type {
             Type::Forall(x) => x.visit(f),
             Type::Var(x) => x.visit(f),
             Type::Quantified(x) => x.visit(f),
+            Type::QuantifiedValue(x) => x.visit(f),
             Type::TypeGuard(x) => x.visit(f),
             Type::TypeIs(x) => x.visit(f),
             Type::Unpack(x) => x.visit(f),
@@ -688,6 +744,8 @@ impl Visit for Type {
             Type::ParamSpecValue(x) => x.visit(f),
             Type::Args(x) => x.visit(f),
             Type::Kwargs(x) => x.visit(f),
+            Type::ArgsValue(x) => x.visit(f),
+            Type::KwargsValue(x) => x.visit(f),
             Type::Type(x) => x.visit(f),
             Type::Ellipsis => {}
             Type::Any(x) => x.visit(f),
@@ -696,7 +754,7 @@ impl Visit for Type {
             Type::SuperInstance(x) => x.visit(f),
             Type::SelfType(x) => x.visit(f),
             Type::KwCall(x) => x.visit(f),
-            Type::None => {}
+            Type::Materialization | Type::None => {}
         }
     }
 }
@@ -721,6 +779,7 @@ impl VisitMut for Type {
             Type::Forall(x) => x.visit_mut(f),
             Type::Var(x) => x.visit_mut(f),
             Type::Quantified(x) => x.visit_mut(f),
+            Type::QuantifiedValue(x) => x.visit_mut(f),
             Type::TypeGuard(x) => x.visit_mut(f),
             Type::TypeIs(x) => x.visit_mut(f),
             Type::Unpack(x) => x.visit_mut(f),
@@ -732,6 +791,8 @@ impl VisitMut for Type {
             Type::ParamSpecValue(x) => x.visit_mut(f),
             Type::Args(x) => x.visit_mut(f),
             Type::Kwargs(x) => x.visit_mut(f),
+            Type::ArgsValue(x) => x.visit_mut(f),
+            Type::KwargsValue(x) => x.visit_mut(f),
             Type::Type(x) => x.visit_mut(f),
             Type::Ellipsis => {}
             Type::Any(x) => x.visit_mut(f),
@@ -740,7 +801,7 @@ impl VisitMut for Type {
             Type::SuperInstance(x) => x.visit_mut(f),
             Type::SelfType(x) => x.visit_mut(f),
             Type::KwCall(x) => x.visit_mut(f),
-            Type::None => {}
+            Type::Materialization | Type::None => {}
         }
     }
 }
@@ -820,13 +881,17 @@ impl Type {
     }
 
     pub fn any_tuple() -> Self {
-        Type::Tuple(Tuple::Unbounded(Box::new(Type::Any(AnyStyle::Implicit))))
+        Type::Tuple(Tuple::unbounded(Type::Any(AnyStyle::Implicit)))
     }
 
     pub fn is_any(&self) -> bool {
         matches!(self, Type::Any(_))
     }
 
+    pub fn is_typed_dict(&self) -> bool {
+        matches!(self, Type::TypedDict(_) | Type::PartialTypedDict(_))
+    }
+
     pub fn is_error(&self) -> bool {
         matches!(self, Type::Any(AnyStyle::Error))
     }
@@ -889,41 +954,6 @@ impl Type {
         }
     }
 
-    /// If this is an unbound callable (i.e., a callable that is not BoundMethod), strip the first parameter.
-    /// TODO: Does not handle generics.
-    pub fn drop_first_param_of_unbound_callable(&self) -> Option<Type> {
-        match self {
-            Type::Callable(callable) => callable
-                .drop_first_param()
-                .map(|callable| Type::Callable(Box::new(callable))),
-            Type::Function(func) => func.signature.drop_first_param().map(|callable| {
-                Type::Function(Box::new(Function {
-                    signature: callable,
-                    metadata: func.metadata.clone(),
-                }))
-            }),
-            Type::Overload(overload) => overload
-                .signatures
-                .try_mapped_ref(|x| match x {
-                    OverloadType::Callable(f) => {
-                        f.signature.drop_first_param().ok_or(()).map(|c| Function {
-                            signature: c,
-                            metadata: f.metadata.clone(),
-                        })
-                    }
-                    _ => Err(()),
-                })
-                .ok()
-                .map(|signatures| {
-                    Type::Overload(Overload {
-                        signatures: signatures.mapped(OverloadType::Callable),
-                        metadata: overload.metadata.clone(),
-                    })
-                }),
-            _ => None,
-        }
-    }
-
     pub fn is_none(&self) -> bool {
         matches!(self, Type::None)
     }
@@ -940,21 +970,29 @@ impl Type {
         }
     }
 
-    pub fn subst(mut self, mp: &SmallMap<&Quantified, &Type>) -> Self {
+    pub fn subst_mut_fn(&mut self, mp: &mut dyn FnMut(&Quantified) -> Option<Type>) {
         // We are looking up Quantified in a map, and Quantified may contain a Quantified within it.
         // Therefore, to make sure we still get matches, work top-down (not using `transform`).
-        fn f(ty: &mut Type, mp: &SmallMap<&Quantified, &Type>) {
+        fn f(ty: &mut Type, mp: &mut dyn FnMut(&Quantified) -> Option<Type>) {
             if let Type::Quantified(x) = ty {
-                if let Some(w) = mp.get(x) {
-                    *ty = (*w).clone();
+                if let Some(w) = mp(x) {
+                    *ty = w;
                 }
             } else {
                 ty.recurse_mut(&mut |x| f(x, mp));
             }
         }
+        f(self, mp);
+    }
+
+    pub fn subst_mut(&mut self, mp: &SmallMap<&Quantified, &Type>) {
         if !mp.is_empty() {
-            f(&mut self, mp);
+            self.subst_mut_fn(&mut |x| mp.get(x).map(|t| (*t).clone()));
         }
+    }
+
+    pub fn subst(mut self, mp: &SmallMap<&Quantified, &Type>) -> Self {
+        self.subst_mut(mp);
         self
     }
 
@@ -966,13 +1004,9 @@ impl Type {
         });
     }
 
-    pub fn subst_self_type_mut(
-        &mut self,
-        replacement: &Type,
-        is_subset: &dyn Fn(&Type, &Type) -> bool,
-    ) {
+    pub fn subst_self_type_mut(&mut self, replacement: &Type) {
         self.transform_mut(&mut |t| {
-            if matches!(t, Type::SelfType(_)) && is_subset(replacement, t) {
+            if matches!(t, Type::SelfType(_)) {
                 *t = replacement.clone();
             }
         })
@@ -1015,14 +1049,27 @@ impl Type {
                 body: Forallable::Function(func),
             })
             | Type::BoundMethod(box BoundMethod {
-                func: BoundMethodType::Function(func),
+                func:
+                    BoundMethodType::Function(func)
+                    | BoundMethodType::Forall(Forall {
+                        tparams: _,
+                        body: func,
+                    }),
                 ..
             }) => check(&func.metadata),
-            Type::Overload(overload) => check(&overload.metadata),
+            Type::Overload(overload)
+            | Type::BoundMethod(box BoundMethod {
+                func: BoundMethodType::Overload(overload),
+                ..
+            }) => check(&overload.metadata),
             _ => T::default(),
         }
     }
 
+    pub fn is_abstract_method(&self) -> bool {
+        self.check_toplevel_func_metadata(&|meta| meta.flags.is_abstract_method)
+    }
+
     pub fn is_override(&self) -> bool {
         self.check_toplevel_func_metadata(&|meta| meta.flags.is_override)
     }
@@ -1035,6 +1082,10 @@ impl Type {
         self.check_toplevel_func_metadata(&|meta| meta.flags.is_property_getter)
     }
 
+    pub fn is_cached_property(&self) -> bool {
+        self.check_toplevel_func_metadata(&|meta| meta.flags.is_cached_property)
+    }
+
     pub fn is_property_setter_decorator(&self) -> bool {
         self.check_toplevel_func_metadata(&|meta| meta.flags.is_property_setter_decorator)
     }
@@ -1047,7 +1098,7 @@ impl Type {
         self.check_toplevel_func_metadata(&|meta| meta.flags.is_overload)
     }
 
-    pub fn is_deprecated(&self) -> bool {
+    pub fn is_deprecated_function(&self) -> bool {
         self.check_toplevel_func_metadata(&|meta| meta.flags.is_deprecated)
     }
 
@@ -1059,6 +1110,13 @@ impl Type {
         self.check_toplevel_func_metadata(&|meta| meta.flags.dataclass_transform_metadata.clone())
     }
 
+    /// If a Protocol method lacks an implementation and does not come from a `.pyi` file, then it cannot be called
+    pub fn is_non_callable_protocol_method(&self) -> bool {
+        self.check_toplevel_func_metadata(&|meta| {
+            meta.flags.lacks_implementation && !meta.flags.defined_in_stub_file
+        })
+    }
+
     /// Transforms this type's function metadata, if it is a function. Note that we do *not*
     /// recurse into the type to find nested function types.
     pub fn transform_toplevel_func_metadata(&mut self, mut f: impl FnMut(&mut FuncMetadata)) {
@@ -1082,6 +1140,10 @@ impl Type {
     pub fn visit_toplevel_callable<'a>(&'a self, mut f: impl FnMut(&'a Callable)) {
         match self {
             Type::Callable(callable) => f(callable),
+            Type::Forall(box Forall {
+                body: Forallable::Callable(callable),
+                ..
+            }) => f(callable),
             Type::Function(box func)
             | Type::Forall(box Forall {
                 body: Forallable::Function(func),
@@ -1102,7 +1164,7 @@ impl Type {
             }) => {
                 for x in overload.signatures.iter() {
                     match x {
-                        OverloadType::Callable(function) => f(&function.signature),
+                        OverloadType::Function(function) => f(&function.signature),
                         OverloadType::Forall(forall) => f(&forall.body.signature),
                     }
                 }
@@ -1111,11 +1173,15 @@ impl Type {
         }
     }
 
-    /// Transform this type if it is a callable. Note that we do *not* recurse into the type to
+    /// Apply `f` to this type if it is a callable. Note that we do *not* recurse into the type to
     /// find nested callable types.
-    fn transform_toplevel_callable(&mut self, mut f: impl FnMut(&mut Callable)) {
+    pub fn visit_toplevel_callable_mut<'a>(&'a mut self, mut f: impl FnMut(&'a mut Callable)) {
         match self {
             Type::Callable(callable) => f(callable),
+            Type::Forall(box Forall {
+                body: Forallable::Callable(callable),
+                ..
+            }) => f(callable),
             Type::Function(box func)
             | Type::Forall(box Forall {
                 body: Forallable::Function(func),
@@ -1136,7 +1202,51 @@ impl Type {
             }) => {
                 for x in overload.signatures.iter_mut() {
                     match x {
-                        OverloadType::Callable(function) => f(&mut function.signature),
+                        OverloadType::Function(function) => f(&mut function.signature),
+                        OverloadType::Forall(forall) => f(&mut forall.body.signature),
+                    }
+                }
+            }
+            _ => {}
+        }
+    }
+
+    fn is_toplevel_callable(&self) -> bool {
+        let mut is_callable = false;
+        self.visit_toplevel_callable(&mut |_| is_callable = true);
+        is_callable
+    }
+
+    /// Transform this type if it is a callable. Note that we do *not* recurse into the type to
+    /// find nested callable types.
+    fn transform_toplevel_callable(&mut self, mut f: impl FnMut(&mut Callable)) {
+        match self {
+            Type::Callable(callable) => f(callable),
+            Type::Forall(box Forall {
+                body: Forallable::Callable(callable),
+                ..
+            }) => f(callable),
+            Type::Function(box func)
+            | Type::Forall(box Forall {
+                body: Forallable::Function(func),
+                ..
+            })
+            | Type::BoundMethod(box BoundMethod {
+                func: BoundMethodType::Function(func),
+                ..
+            })
+            | Type::BoundMethod(box BoundMethod {
+                func: BoundMethodType::Forall(Forall { body: func, .. }),
+                ..
+            }) => f(&mut func.signature),
+            Type::Overload(overload)
+            | Type::BoundMethod(box BoundMethod {
+                func: BoundMethodType::Overload(overload),
+                ..
+            }) => {
+                for x in overload.signatures.iter_mut() {
+                    match x {
+                        OverloadType::Function(function) => f(&mut function.signature),
                         OverloadType::Forall(forall) => f(&mut forall.body.signature),
                     }
                 }
@@ -1188,11 +1298,23 @@ impl Type {
         sigs
     }
 
-    pub fn promote_literals(self, stdlib: &Stdlib) -> Type {
-        self.transform(&mut |ty| match &ty {
+    pub fn promote_literals(mut self, stdlib: &Stdlib) -> Type {
+        fn g(ty: &mut Type, f: &mut dyn FnMut(&mut Type)) {
+            // This isn't quite right: we should decide whether to promote a literal based on
+            // whether it is inferred or annotated. But we don't have an easy way to track that
+            // right now, and promoting literals in callable signatures is always wrong, so let's
+            // special-case callables for now.
+            if !ty.is_toplevel_callable() {
+                ty.recurse_mut(&mut |ty| g(ty, f));
+                f(ty);
+            }
+        }
+        g(&mut self, &mut |ty| match &ty {
             Type::Literal(lit) => *ty = lit.general_class_type(stdlib).clone().to_type(),
+            Type::LiteralString => *ty = stdlib.str().clone().to_type(),
             _ => {}
-        })
+        });
+        self
     }
 
     // Attempt at a function that will convert @ to Any for now.
@@ -1231,6 +1353,15 @@ impl Type {
         })
     }
 
+    /// type[a | b] -> type[a] | type[b]
+    pub fn distribute_type_over_union(self) -> Self {
+        self.transform(&mut |ty| {
+            if let Type::Type(box Type::Union(members)) = ty {
+                *ty = unions(members.drain(..).map(Type::type_form).collect());
+            }
+        })
+    }
+
     pub fn anon_callables(self) -> Self {
         self.transform(&mut |mut ty| {
             if let Type::Function(func) = ty {
@@ -1258,6 +1389,16 @@ impl Type {
         })
     }
 
+    pub fn promote_typevar_values(self, stdlib: &Stdlib) -> Self {
+        self.transform(&mut |ty| match &ty {
+            Type::TypeVar(_) => *ty = stdlib.type_var().clone().to_type(),
+            Type::ParamSpec(_) => *ty = stdlib.param_spec().clone().to_type(),
+            Type::TypeVarTuple(_) => *ty = stdlib.type_var_tuple().clone().to_type(),
+            Type::QuantifiedValue(q) => *ty = q.class_type(stdlib).clone().to_type(),
+            _ => {}
+        })
+    }
+
     pub fn sort_unions(self) -> Self {
         self.transform(&mut |ty| {
             if let Type::Union(ts) = ty {
@@ -1304,7 +1445,7 @@ impl Type {
 
     pub fn as_quantified(&self) -> Option<Quantified> {
         match self {
-            Type::Quantified(q) => Some(q.clone()),
+            Type::Quantified(q) => Some((**q).clone()),
             _ => None,
         }
     }
@@ -1347,6 +1488,7 @@ impl Type {
             Type::TypeVarTuple(t) => Some(t.qname()),
             Type::ParamSpec(t) => Some(t.qname()),
             Type::SelfType(cls) => Some(cls.qname()),
+            Type::Literal(Lit::Enum(e)) => Some(e.class.qname()),
             _ => None,
         }
     }
@@ -1376,6 +1518,50 @@ impl Type {
             _ => None,
         }
     }
+
+    pub fn to_callable(self) -> Option<Callable> {
+        match self {
+            Type::Callable(callable) => Some(*callable),
+            Type::Function(function) => Some(function.signature),
+            Type::BoundMethod(bound_method) => match bound_method.func {
+                BoundMethodType::Function(function) => Some(function.signature),
+                BoundMethodType::Forall(forall) => Some(forall.body.signature),
+                BoundMethodType::Overload(_) => None,
+            },
+            _ => None,
+        }
+    }
+
+    /// Return the FunctionKind if this type corresponds to a function or method.
+    pub fn to_func_kind(&self) -> Option<&FunctionKind> {
+        match &self {
+            Type::Function(f) => Some(&f.metadata.kind),
+            Type::BoundMethod(m) => Some(&m.func.metadata().kind),
+            Type::Overload(o) => Some(&o.metadata.kind),
+            _ => None,
+        }
+    }
+
+    pub fn materialize(&self) -> Self {
+        self.clone().transform(&mut |ty| {
+            if ty.is_any() {
+                *ty = Type::Materialization;
+            }
+            ty.transform_toplevel_callable(&mut |callable: &mut Callable| {
+                if matches!(callable.params, Params::Ellipsis) {
+                    callable.params = Params::Materialization;
+                }
+            })
+        })
+    }
+
+    /// Is this an instance of `warnings.deprecated` or its backport `typing_extensions.deprecated`?
+    pub fn is_deprecation_marker(&self) -> bool {
+        let Type::ClassType(cls) = self else {
+            return false;
+        };
+        cls.has_qname("warnings", "deprecated") || cls.has_qname("typing_extensions", "deprecated")
+    }
 }
 
 #[cfg(test)]
