@@ -12,14 +12,12 @@ use std::path::PathBuf;
 use std::sync::Arc;
 
 use dupe::Dupe;
-use dupe::OptionDupedExt;
 use pyrefly_python::module::TextRangeWithModule;
 use pyrefly_python::module_name::ModuleName;
 use pyrefly_python::module_path::ModulePath;
 use pyrefly_util::display::DisplayWith;
 use pyrefly_util::display::DisplayWithCtx;
 use pyrefly_util::lock::Mutex;
-use pyrefly_util::recurser::Recurser;
 use pyrefly_util::uniques::UniqueFactory;
 use pyrefly_util::visit::VisitMut;
 use ruff_python_ast::name::Name;
@@ -48,6 +46,7 @@ use crate::graph::index::Idx;
 use crate::graph::index_map::IndexMap;
 use crate::module::module_info::ModuleInfo;
 use crate::solver::solver::Solver;
+use crate::solver::solver::VarRecurser;
 use crate::state::ide::IntermediateDefinition;
 use crate::state::ide::key_to_intermediate_definition;
 use crate::table;
@@ -77,10 +76,15 @@ pub struct Index {
 }
 
 #[derive(Debug)]
-struct OverloadedCallee {
-    all_overloads: Vec<Callable>,
-    closest_overload: Callable,
-    is_closest_overload_chosen: bool,
+enum OverloadedCallee {
+    Resolved {
+        callable: Callable,
+    },
+    Candidates {
+        all: Vec<Callable>,
+        closest: Callable,
+        is_closest_chosen: bool,
+    },
 }
 
 #[derive(Debug, Default)]
@@ -88,6 +92,8 @@ pub struct Traces {
     types: SmallMap<TextRange, Arc<Type>>,
     /// A map from (range of callee, overload information)
     overloaded_callees: SmallMap<TextRange, OverloadedCallee>,
+    /// A map of text ranges that correspond to 'b' portion in expressions a.b where b is a property access -> getter type
+    invoked_properties: SmallMap<TextRange, Arc<Type>>,
 }
 
 /// Invariants:
@@ -325,7 +331,7 @@ impl Solutions {
 }
 
 pub trait LookupAnswer: Sized {
-    /// Look up the value. If present, the `path` is a hint which can optimise certain cases.
+    /// Look up the value. If present, the `path` is a hint which can optimize certain cases.
     ///
     /// Return None if the file is undergoing concurrent modification.
     fn get<K: Solve<Self> + Exported>(
@@ -428,7 +434,7 @@ impl Answers {
                 }
             }
         }
-        let recurser = &Recurser::new();
+        let recurser = &VarRecurser::new();
         let thread_state = &ThreadState::new();
         let answers_solver = AnswersSolver::new(
             answers,
@@ -520,7 +526,7 @@ impl Answers {
         AnswerTable: TableKeyed<K, Value = AnswerEntry<K>>,
         BindingTable: TableKeyed<K, Value = BindingEntry<K>>,
     {
-        let recurser = &Recurser::new();
+        let recurser = &VarRecurser::new();
         let solver = AnswersSolver::new(
             answers,
             self,
@@ -545,26 +551,42 @@ impl Answers {
         self.table.get::<K>().get(k)?.get()
     }
 
-    pub fn for_display(&self, t: Type) -> Type {
-        self.solver.for_display(t)
+    fn deep_force(&self, t: Type) -> Type {
+        self.solver.deep_force(t)
     }
 
     pub fn solver(&self) -> &Solver {
         &self.solver
     }
 
-    pub fn get_type_trace(&self, range: TextRange) -> Option<Arc<Type>> {
-        let lock = self.trace.as_ref()?.lock();
-        lock.types.get(&range).duped()
+    pub fn get_type_at(&self, idx: Idx<Key>) -> Option<Type> {
+        Some(self.deep_force(self.get_idx(idx)?.arc_clone_ty()))
     }
 
-    pub fn get_chosen_overload_trace(&self, range: TextRange) -> Option<Callable> {
+    pub fn get_type_trace(&self, range: TextRange) -> Option<Type> {
         let lock = self.trace.as_ref()?.lock();
-        let overloaded_callee = lock.overloaded_callees.get(&range)?;
-        if overloaded_callee.is_closest_overload_chosen {
-            Some(overloaded_callee.closest_overload.clone())
-        } else {
-            None
+        Some(self.deep_force(lock.types.get(&range)?.as_ref().clone()))
+    }
+
+    pub fn try_get_getter_for_range(&self, range: TextRange) -> Option<Type> {
+        let lock = self.trace.as_ref()?.lock();
+        Some(self.deep_force(lock.invoked_properties.get(&range)?.as_ref().clone()))
+    }
+
+    pub fn get_chosen_overload_trace(&self, range: TextRange) -> Option<Type> {
+        let lock = self.trace.as_ref()?.lock();
+        match lock.overloaded_callees.get(&range)? {
+            OverloadedCallee::Resolved { callable } => {
+                Some(self.deep_force(Type::Callable(Box::new(callable.clone()))))
+            }
+            OverloadedCallee::Candidates {
+                closest,
+                is_closest_chosen,
+                ..
+            } if *is_closest_chosen => {
+                Some(self.deep_force(Type::Callable(Box::new(closest.clone()))))
+            }
+            _ => None,
         }
     }
 
@@ -574,23 +596,13 @@ impl Answers {
         range: TextRange,
     ) -> Option<(Vec<Callable>, Option<usize>)> {
         let lock = self.trace.as_ref()?.lock();
-        let overloaded_callee = lock.overloaded_callees.get(&range)?;
-        let chosen_overload_index =
-            overloaded_callee
-                .all_overloads
-                .iter()
-                .enumerate()
-                .find_map(|(index, signature)| {
-                    if signature == &overloaded_callee.closest_overload {
-                        Some(index)
-                    } else {
-                        None
-                    }
-                });
-        Some((
-            overloaded_callee.all_overloads.clone(),
-            chosen_overload_index,
-        ))
+        match lock.overloaded_callees.get(&range)? {
+            OverloadedCallee::Resolved { callable } => Some((vec![callable.clone()], Some(0))),
+            OverloadedCallee::Candidates { all, closest, .. } => {
+                let chosen_index = all.iter().position(|signature| signature == closest);
+                Some((all.clone(), chosen_index))
+            }
+        }
     }
 }
 
@@ -616,6 +628,17 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         &self.current().solver
     }
 
+    pub fn record_resolved_trace(&self, loc: TextRange, ty: Type) {
+        if let Some(trace) = &self.current().trace
+            && let Some(callable) = ty.to_callable()
+        {
+            trace
+                .lock()
+                .overloaded_callees
+                .insert(loc, OverloadedCallee::Resolved { callable });
+        }
+    }
+
     /// Record all the overloads and the chosen overload.
     /// The trace will be used to power signature help and hover for overloaded functions.
     pub fn record_overload_trace(
@@ -628,13 +651,10 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         if let Some(trace) = &self.current().trace {
             trace.lock().overloaded_callees.insert(
                 loc,
-                OverloadedCallee {
-                    all_overloads: all_overloads
-                        .into_iter()
-                        .map(|func| (*func).clone())
-                        .collect(),
-                    closest_overload: closest_overload.clone(),
-                    is_closest_overload_chosen,
+                OverloadedCallee::Candidates {
+                    all: all_overloads.into_iter().cloned().collect(),
+                    closest: closest_overload.clone(),
+                    is_closest_chosen: is_closest_overload_chosen,
                 },
             );
         }
@@ -650,7 +670,10 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             for AttrInfo {
                 name: _,
                 ty: _,
+                is_deprecated: _,
                 definition,
+                docstring_range: _,
+                is_reexport: _,
             } in self.completions(base.clone(), Some(attribute_name), false)
             {
                 match definition {
@@ -680,8 +703,19 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         }
     }
 
-    pub fn record_type_trace(&self, loc: TextRange, ty: &Type) {
+    pub fn record_property_getter(&self, loc: TextRange, getter_ty: &Type) {
         if let Some(trace) = &self.current().trace {
+            trace
+                .lock()
+                .invoked_properties
+                .insert(loc, Arc::new(getter_ty.clone()));
+        }
+    }
+
+    pub fn record_type_trace(&self, loc: TextRange, ty: &Type) {
+        if let Some(trace) = &self.current().trace
+            && !loc.is_empty()
+        {
             trace.lock().types.insert(loc, Arc::new(ty.clone()));
         }
     }
