@@ -19,7 +19,7 @@ use pyrefly_python::module_name::ModuleName;
 use pyrefly_python::module_path::ModulePath;
 use pyrefly_util::display::DisplayWithCtx;
 use pyrefly_util::display::commas_iter;
-use pyrefly_util::recurser::Recurser;
+use pyrefly_util::recurser::Guard;
 use pyrefly_util::uniques::UniqueFactory;
 use ruff_text_size::TextRange;
 use starlark_map::Hashed;
@@ -51,6 +51,7 @@ use crate::graph::calculation::Calculation;
 use crate::graph::calculation::ProposalResult;
 use crate::graph::index::Idx;
 use crate::module::module_info::ModuleInfo;
+use crate::solver::solver::VarRecurser;
 use crate::solver::type_order::TypeOrder;
 use crate::types::class::Class;
 use crate::types::stdlib::Stdlib;
@@ -251,7 +252,7 @@ impl Cycle {
         cycle
     }
 
-    /// Do a pre-calculation check, to handle progress recurively traversing
+    /// Do a pre-calculation check, to handle progress recursively traversing
     /// the cycle until we reach the second instance of `break_at`.
     ///
     /// For each cycle participant we encounter, we move it from the
@@ -374,6 +375,8 @@ impl Cycles {
 pub struct ThreadState {
     cycles: Cycles,
     stack: CalcStack,
+    /// For debugging only: thread-global that allows us to control debug logging across components.
+    debug: RefCell<bool>,
 }
 
 impl ThreadState {
@@ -381,6 +384,7 @@ impl ThreadState {
         Self {
             cycles: Cycles::new(),
             stack: CalcStack::new(),
+            debug: RefCell::new(false),
         }
     }
 }
@@ -396,7 +400,7 @@ pub struct AnswersSolver<'a, Ans: LookupAnswer> {
     bindings: &'a Bindings,
     pub exports: &'a dyn LookupExport,
     pub uniques: &'a UniqueFactory,
-    pub recurser: &'a Recurser<Var>,
+    pub recurser: &'a VarRecurser,
     pub stdlib: &'a Stdlib,
 }
 
@@ -408,7 +412,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         bindings: &'a Bindings,
         exports: &'a dyn LookupExport,
         uniques: &'a UniqueFactory,
-        recurser: &'a Recurser<Var>,
+        recurser: &'a VarRecurser,
         stdlib: &'a Stdlib,
         thread_state: &'a ThreadState,
     ) -> AnswersSolver<'a, Ans> {
@@ -425,6 +429,17 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         }
     }
 
+    /// Is the debug flag set? Intended to support print debugging.
+    pub fn is_debug(&self) -> bool {
+        *self.thread_state.debug.borrow()
+    }
+
+    /// Set the debug flag. Intended to support print debugging.
+    #[allow(dead_code)]
+    pub fn set_debug(&self, value: bool) {
+        *self.thread_state.debug.borrow_mut() = value;
+    }
+
     pub fn current(&self) -> &Answers {
         self.current
     }
@@ -453,7 +468,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         self.solver().for_display(t)
     }
 
-    pub fn type_order(&self) -> TypeOrder<Ans> {
+    pub fn type_order(&self) -> TypeOrder<'_, Ans> {
         TypeOrder::new(self)
     }
 
@@ -548,16 +563,16 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         let answer = K::solve(self, binding, self.base_errors);
         let (v, rec) = calculation.record_value(answer);
         // If this was the first write to a Calculation that had a recursive placeholder,
-        // we need to record the placeholder => final answer correspondance.
+        // we need to record the placeholder => final answer correspondence.
         if let Some(r) = rec {
             let k = self.bindings().idx_to_key(idx).range();
             // Always force recursive Vars as soon as we produce the final answer. This limits
             // nondeterminism by ensuring that nothing downstream of the cycle can pin the type
             // once the cycle has finished (although there can still be data races where the
-            // Var excapes the cycle in another thread before it has finished computing).
+            // Var escapes the cycle in another thread before it has finished computing).
             //
             // `Var::ZERO` is just a dummy value used by a few of the `K: Solve`
-            // implementations that doen't actually use the Var, so we have to skip it.
+            // implementations that doesn't actually use the Var, so we have to skip it.
             K::record_recursive(self, k, &v, r, self.base_errors);
             if r != Var::ZERO {
                 self.solver().force_var(r);
@@ -615,8 +630,11 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         BindingTable: TableKeyed<K, Value = BindingEntry<K>>,
         SolutionsTable: TableKeyed<K, Value = SolutionsEntry<K>>,
     {
-        if module == self.module().name() && path.is_none_or(|path| path == self.module().path()) {
-            Some(self.get(k))
+        if module == self.module().name() && path == Some(self.module().path()) {
+            // We are working in our own module, so don't have to go back to the `LookupAnswer` trait.
+            // But even though we are looking at our own module, we might be using our own type via an import
+            // from a mutually recursive module, so have to deal with key_to_idx finding nothing due to incremental.
+            Some(self.get_idx(self.bindings().key_to_idx_hashed_opt(Hashed::new(k))?))
         } else {
             self.answers.get(module, path, k, self.thread_state)
         }
@@ -672,14 +690,19 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     }
 
     pub fn create_recursive(&self, binding: &Binding) -> Var {
-        let t = if let Binding::Default(default, _) = binding {
-            self.get_calculation(*default)
-                .get()
-                .map(|t| t.arc_clone_ty().promote_literals(self.stdlib))
-        } else {
-            None
-        };
-        self.solver().fresh_recursive(self.uniques, t)
+        match binding {
+            Binding::LoopPhi(prior_idx, _) => self.solver().fresh_loop_recursive(
+                self.uniques,
+                self.get_idx(*prior_idx)
+                    .arc_clone_ty()
+                    .promote_literals(self.stdlib),
+            ),
+            _ => self.solver().fresh_recursive(self.uniques),
+        }
+    }
+
+    pub fn recurse(&'a self, var: Var) -> Option<Guard<'a, Var>> {
+        self.solver().recurse(var, self.recurser)
     }
 
     pub fn record_recursive(
@@ -693,52 +716,57 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             .record_recursive::<Ans>(recursive, answer, self.type_order(), errors, loc);
     }
 
-    /// Check if `want` matches `got` returning `want` if the check fails.
+    /// Check if `got` matches `want`, returning `want` if the check fails.
     pub fn check_and_return_type_info(
         &self,
-        want: &Type,
         got: TypeInfo,
+        want: &Type,
         loc: TextRange,
         errors: &ErrorCollector,
         tcc: &dyn Fn() -> TypeCheckContext,
     ) -> TypeInfo {
-        if self.check_type(want, got.ty(), loc, errors, tcc) {
+        if self.check_type(got.ty(), want, loc, errors, tcc) {
             got
         } else {
             got.with_ty(want.clone())
         }
     }
 
-    /// Check if `want` matches `got` returning `want` if the check fails.
+    /// Check if `got` matches `want`, returning `want` if the check fails.
     pub fn check_and_return_type(
         &self,
-        want: &Type,
         got: Type,
+        want: &Type,
         loc: TextRange,
         errors: &ErrorCollector,
         tcc: &dyn Fn() -> TypeCheckContext,
     ) -> Type {
-        if self.check_type(want, &got, loc, errors, tcc) {
+        if self.check_type(&got, want, loc, errors, tcc) {
             got
         } else {
             want.clone()
         }
     }
 
-    /// Check if `want` matches `got`, returning `true` on success and `false` on failure.
+    /// Check if `got` matches `want`, returning `true` on success and `false` on failure.
     pub fn check_type(
         &self,
-        want: &Type,
         got: &Type,
+        want: &Type,
         loc: TextRange,
         errors: &ErrorCollector,
         tcc: &dyn Fn() -> TypeCheckContext,
     ) -> bool {
-        if got.is_error() || self.is_subset_eq(got, want) {
+        if got.is_error() {
             true
         } else {
-            self.solver().error(want, got, errors, loc, tcc);
-            false
+            match self.is_subset_eq_with_reason(got, want) {
+                Ok(()) => true,
+                Err(error) => {
+                    self.solver().error(got, want, errors, loc, tcc, error);
+                    false
+                }
+            }
         }
     }
 
@@ -776,7 +804,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     Type::Type(box Type::Union(tys)) if !in_type => {
                         tys.iter().for_each(|ty| self.go(ty, true))
                     }
-                    Type::Var(v) if let Some(_guard) = self.me.recurser.recurse(*v) => {
+                    Type::Var(v) if let Some(_guard) = self.me.recurse(*v) => {
                         self.go(&self.me.solver().force_var(*v), in_type)
                     }
                     _ if in_type => (self.f)(&Type::Type(Box::new(ty.clone()))),
