@@ -6,10 +6,16 @@
  */
 
 use itertools::Itertools;
+use pyrefly_python::dunder;
+use pyrefly_types::callable::FunctionKind;
+use pyrefly_types::typed_dict::ExtraItems;
+use pyrefly_types::types::TArgs;
+use pyrefly_types::types::TParams;
 use pyrefly_util::display::count;
 use pyrefly_util::owner::Owner;
 use pyrefly_util::prelude::SliceExt;
 use pyrefly_util::prelude::VecExt;
+use pyrefly_util::visit::VisitMut;
 use ruff_python_ast::Expr;
 use ruff_python_ast::Identifier;
 use ruff_python_ast::Keyword;
@@ -24,6 +30,7 @@ use crate::alt::answers::LookupAnswer;
 use crate::alt::answers_solver::AnswersSolver;
 use crate::alt::expr::TypeOrExpr;
 use crate::alt::solve::Iterable;
+use crate::alt::unwrap::HintRef;
 use crate::config::error_kind::ErrorKind;
 use crate::error::collector::ErrorCollector;
 use crate::error::context::ErrorContext;
@@ -31,8 +38,8 @@ use crate::error::context::ErrorInfo;
 use crate::error::context::TypeCheckContext;
 use crate::error::context::TypeCheckKind;
 use crate::error::display::function_suffix;
+use crate::solver::solver::QuantifiedHandle;
 use crate::types::callable::Callable;
-use crate::types::callable::FuncId;
 use crate::types::callable::Param;
 use crate::types::callable::ParamList;
 use crate::types::callable::Params;
@@ -43,7 +50,7 @@ use crate::types::types::Type;
 use crate::types::types::Var;
 
 /// Structure to turn TypeOrExprs into Types.
-/// This is used to avoid re-infering types for arguments multiple types.
+/// This is used to avoid re-inferring types for arguments multiple types.
 ///
 /// Implemented by keeping an `Owner` to hand out references to `Type`.
 pub struct CallWithTypes(Owner<Type>);
@@ -139,6 +146,36 @@ impl<'a> CallKeyword<'a> {
             value: TypeOrExpr::Expr(&x.value),
         }
     }
+
+    pub fn materialize<Ans: LookupAnswer>(
+        &self,
+        solver: &AnswersSolver<Ans>,
+        errors: &ErrorCollector,
+        owner: &'a Owner<Type>,
+    ) -> (Self, bool) {
+        let transformation = |ty: &Type| {
+            if self.arg.is_none() && ty.is_any() {
+                // See test::overload::test_kwargs_materialization - we need to turn this
+                // into Mapping[str, Any] to correctly materialize the `**kwargs` type.
+                solver
+                    .stdlib
+                    .mapping(solver.stdlib.str().clone().to_type(), ty.clone())
+                    .to_type()
+                    .materialize()
+            } else {
+                ty.materialize()
+            }
+        };
+        let (materialized, changed) = self.value.transform(solver, errors, owner, transformation);
+        (
+            Self {
+                range: self.range,
+                arg: self.arg,
+                value: materialized,
+            },
+            changed,
+        )
+    }
 }
 
 #[derive(Clone, Debug)]
@@ -176,6 +213,33 @@ impl<'a> CallArg<'a> {
         }
     }
 
+    pub fn materialize<Ans: LookupAnswer>(
+        &self,
+        solver: &AnswersSolver<Ans>,
+        errors: &ErrorCollector,
+        owner: &'a Owner<Type>,
+    ) -> (Self, bool) {
+        match self {
+            Self::Arg(value) => {
+                let (materialized, changed) =
+                    value.transform(solver, errors, owner, |ty| ty.materialize());
+                (Self::Arg(materialized), changed)
+            }
+            Self::Star(value, range) => {
+                let (materialized, changed) = value.transform(solver, errors, owner, |ty| {
+                    if ty.is_any() {
+                        // See test::overload::test_varargs_materialization - we need to turn this
+                        // into Iterable[Any] to correctly materialize the `*args` type.
+                        solver.stdlib.iterable(ty.clone()).to_type().materialize()
+                    } else {
+                        ty.materialize()
+                    }
+                });
+                (Self::Star(materialized, *range), changed)
+            }
+        }
+    }
+
     // Splat arguments might be fixed-length tuples, which are handled precisely, or have unknown
     // length. This function evaluates splat args to determine how many params should be consumed,
     // but does not evaluate other expressions, which might be contextually typed.
@@ -183,13 +247,13 @@ impl<'a> CallArg<'a> {
         &self,
         solver: &AnswersSolver<Ans>,
         arg_errors: &ErrorCollector,
-    ) -> CallArgPreEval {
+    ) -> CallArgPreEval<'_> {
         match self {
             Self::Arg(TypeOrExpr::Type(ty, _)) => CallArgPreEval::Type(ty, false),
             Self::Arg(TypeOrExpr::Expr(e)) => CallArgPreEval::Expr(e, false),
             Self::Star(e, range) => {
                 let ty = e.infer(solver, arg_errors);
-                let iterables = solver.iterate(&ty, *range, arg_errors);
+                let iterables = solver.iterate(&ty, *range, arg_errors, None);
                 // If we have a union of iterables, use a fixed length only if every iterable is
                 // fixed and has the same length. Otherwise, use star.
                 let mut fixed_lens = Vec::new();
@@ -214,14 +278,7 @@ impl<'a> CallArg<'a> {
                     let tys = fixed_tys.into_map(|tys| solver.unions(tys));
                     CallArgPreEval::Fixed(tys, 0)
                 } else {
-                    let mut star_tys = Vec::new();
-                    for x in iterables {
-                        match x {
-                            Iterable::OfType(ty) => star_tys.push(ty.clone()),
-                            Iterable::FixedLen(tys) => star_tys.extend(tys),
-                        }
-                    }
-                    let ty = solver.unions(star_tys);
+                    let ty = solver.get_produced_type(iterables);
                     CallArgPreEval::Star(ty, false)
                 }
             }
@@ -230,7 +287,7 @@ impl<'a> CallArg<'a> {
 }
 
 // Pre-evaluated args are iterable. Type/Expr/Star variants iterate once (tracked via bool field),
-// Fixed variant iterates over the the vec (tracked via usize field).
+// Fixed variant iterates over the vec (tracked via usize field).
 #[derive(Clone, Debug)]
 enum CallArgPreEval<'a> {
     Type(&'a Type, bool),
@@ -254,7 +311,7 @@ impl CallArgPreEval<'_> {
     fn post_check<Ans: LookupAnswer>(
         &mut self,
         solver: &AnswersSolver<Ans>,
-        callable_name: Option<&FuncId>,
+        callable_name: Option<&FunctionKind>,
         hint: &Type,
         param_name: Option<&Name>,
         vararg: bool,
@@ -274,22 +331,22 @@ impl CallArgPreEval<'_> {
         match self {
             Self::Type(ty, done) => {
                 *done = true;
-                solver.check_type(hint, ty, range, call_errors, tcc);
+                solver.check_type(ty, hint, range, call_errors, tcc);
             }
             Self::Expr(x, done) => {
                 *done = true;
                 solver.expr_with_separate_check_errors(
                     x,
-                    Some((hint, tcc, call_errors)),
+                    Some((hint, call_errors, tcc)),
                     arg_errors,
                 );
             }
             Self::Star(ty, done) => {
                 *done = vararg;
-                solver.check_type(hint, ty, range, call_errors, tcc);
+                solver.check_type(ty, hint, range, call_errors, tcc);
             }
             Self::Fixed(tys, i) => {
-                solver.check_type(hint, &tys[*i], range, call_errors, tcc);
+                solver.check_type(&tys[*i], hint, range, call_errors, tcc);
                 *i += 1;
             }
         }
@@ -332,33 +389,33 @@ enum PosParamKind {
 }
 
 /// Helps track matching of arguments against positional parameters in AnswersSolver::callable_infer_params.
-struct PosParam {
-    ty: Type,
-    name: Option<Name>,
+struct PosParam<'a> {
+    ty: &'a Type,
+    name: Option<&'a Name>,
     kind: PosParamKind,
 }
 
-impl PosParam {
-    fn new(p: &Param) -> Option<Self> {
+impl<'a> PosParam<'a> {
+    fn new(p: &'a Param) -> Option<Self> {
         match p {
             Param::PosOnly(name, ty, _required) => Some(Self {
-                ty: ty.clone(),
-                name: name.clone(),
+                ty,
+                name: name.as_ref(),
                 kind: PosParamKind::PositionalOnly,
             }),
             Param::Pos(name, ty, _required) => Some(Self {
-                ty: ty.clone(),
-                name: Some(name.clone()),
+                ty,
+                name: Some(name),
                 kind: PosParamKind::Positional,
             }),
             Param::VarArg(name, Type::Unpack(ty)) => Some(Self {
-                ty: (**ty).clone(),
-                name: name.clone(),
+                ty: &**ty,
+                name: name.as_ref(),
                 kind: PosParamKind::Unpacked,
             }),
             Param::VarArg(name, ty) => Some(Self {
-                ty: ty.clone(),
-                name: name.clone(),
+                ty,
+                name: name.as_ref(),
                 kind: PosParamKind::Variadic,
             }),
             Param::KwOnly(..) | Param::Kwargs(..) => None,
@@ -367,12 +424,12 @@ impl PosParam {
 }
 
 impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
-    fn is_param_spec_args(&self, x: &CallArg, q: Quantified, errors: &ErrorCollector) -> bool {
+    fn is_param_spec_args(&self, x: &CallArg, q: &Quantified, errors: &ErrorCollector) -> bool {
         match x {
             CallArg::Star(x, _) => {
                 let mut ty = x.infer(self, errors);
-                self.expand_type_mut(&mut ty);
-                ty == Type::Args(q)
+                self.expand_vars_mut(&mut ty);
+                matches!(ty, Type::Args(q2) if &*q2 == q)
             }
             _ => false,
         }
@@ -381,18 +438,18 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     fn is_param_spec_kwargs(
         &self,
         x: &CallKeyword,
-        q: Quantified,
+        q: &Quantified,
         errors: &ErrorCollector,
     ) -> bool {
         let mut ty = x.value.infer(self, errors);
-        self.expand_type_mut(&mut ty);
-        ty == Type::Kwargs(q)
+        self.expand_vars_mut(&mut ty);
+        matches!(ty, Type::Kwargs(q2) if &*q2 == q)
     }
 
     // See comment on `callable_infer` about `arg_errors` and `call_errors`.
     fn callable_infer_params(
         &self,
-        callable_name: Option<FuncId>,
+        callable_name: Option<&FunctionKind>,
         params: &ParamList,
         // A ParamSpec Var (if any) that comes at the end of the parameter list.
         // See test::paramspec::test_paramspec_twice for an example of this.
@@ -405,6 +462,12 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         call_errors: &ErrorCollector,
         context: Option<&dyn Fn() -> ErrorContext>,
     ) {
+        // We want to work mostly with references, but some things are taken from elsewhere,
+        // so have some owners to capture them.
+        let param_list_owner = Owner::new();
+        let name_owner = Owner::new();
+        let type_owner = Owner::new();
+
         let error = |errors, range, kind, msg: String| {
             self.error(
                 errors,
@@ -413,20 +476,21 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 format!(
                     "{}{}",
                     msg,
-                    function_suffix(callable_name.as_ref(), self.module().name())
+                    function_suffix(callable_name, self.module().name())
                 ),
             )
         };
         let iargs = self_arg.iter().chain(args.iter());
         // Creates a reversed copy of the parameters that we iterate through from back to front,
         // so that we can easily peek at and pop from the end.
-        let mut rparams = params.items().iter().cloned().rev().collect::<Vec<_>>();
-        let mut num_positional_params = 0;
-        let mut extra_positional_args = Vec::new();
-        let mut seen_names = SmallMap::new();
-        let mut extra_arg_pos = None;
-        let mut unpacked_vararg = None;
-        let mut unpacked_vararg_matched_args = Vec::new();
+        let mut rparams: Vec<&Param> = params.items().iter().rev().collect::<Vec<_>>();
+        let mut num_positional_params: usize = 0;
+        let mut extra_positional_args: Vec<TextRange> = Vec::new();
+        let mut seen_names: SmallMap<&Name, &Type> = SmallMap::new();
+        let mut extra_arg_pos: Option<TextRange> = None;
+        let mut unpacked_vararg: Option<(Option<&Name>, &Type)> = None;
+        let mut unpacked_vararg_matched_args: Vec<CallArgPreEval<'_>> = Vec::new();
+
         let var_to_rparams = |var| {
             let ps = match self.solver().force_var(var) {
                 Type::ParamSpecValue(ps) => ps,
@@ -446,7 +510,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     ParamList::everything()
                 }
             };
-            ps.items().iter().cloned().rev().collect()
+            param_list_owner.push(ps).items().iter().rev().collect()
         };
         for arg in iargs {
             let mut arg_pre = arg.pre_eval(self, arg_errors);
@@ -471,18 +535,18 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     }) => {
                         num_positional_params += 1;
                         rparams.pop();
-                        if let Some(name) = &name
+                        if let Some(name) = name
                             && kind == PosParamKind::Positional
                         {
                             // Remember names of positional parameters to detect duplicates.
                             // We ignore positional-only parameters because they can't be passed in by name.
-                            seen_names.insert(name.clone(), ty.clone());
+                            seen_names.insert(name, ty);
                         }
                         arg_pre.post_check(
                             self,
-                            callable_name.as_ref(),
-                            &ty,
-                            name.as_ref(),
+                            callable_name,
+                            ty,
+                            name,
                             false,
                             arg.range(),
                             arg_errors,
@@ -507,9 +571,9 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                         kind: PosParamKind::Variadic,
                     }) => arg_pre.post_check(
                         self,
-                        callable_name.as_ref(),
-                        &ty,
-                        name.as_ref(),
+                        callable_name,
+                        ty,
+                        name,
                         true,
                         arg.range(),
                         arg_errors,
@@ -565,29 +629,29 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     }
                 }
             }
-            let unpacked_args_ty = match middle.as_slice() {
-                [] => Type::tuple(prefix),
-                [middle] => Type::Tuple(Tuple::unpacked(
+            let unpacked_args_ty = match middle.len() {
+                0 => Type::tuple(prefix),
+                1 => Type::Tuple(Tuple::unpacked(
                     prefix,
-                    Type::Tuple(Tuple::unbounded(middle.clone())),
+                    Type::Tuple(Tuple::unbounded(middle.pop().unwrap())),
                     suffix,
                 )),
                 _ => Type::Tuple(Tuple::unpacked(
                     prefix,
-                    Type::Tuple(Tuple::Unbounded(Box::new(self.unions(middle)))),
+                    Type::Tuple(Tuple::unbounded(self.unions(middle))),
                     suffix,
                 )),
             };
             self.check_type(
-                &unpacked_param_ty,
                 &unpacked_args_ty,
+                unpacked_param_ty,
                 range,
                 arg_errors,
                 &|| TypeCheckContext {
                     kind: TypeCheckKind::CallVarArgs(
                         true,
-                        unpacked_name.clone(),
-                        callable_name.clone(),
+                        unpacked_name.cloned(),
+                        callable_name.cloned(),
                     ),
                     context: context.map(|ctx| ctx()),
                 },
@@ -596,11 +660,11 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         // Missing positional-only arguments, split by whether the corresponding parameters
         // in the callable have names. E.g., functions declared with `def` have named posonly
         // parameters and `typing.Callable`s have unnamed ones.
-        let mut missing_unnamed_posonly = 0;
-        let mut missing_named_posonly = SmallSet::new();
-        let mut kwparams = OrderedMap::new();
-        let mut kwargs = None;
-        let mut kwargs_is_unpack = false;
+        let mut missing_unnamed_posonly: usize = 0;
+        let mut missing_named_posonly: SmallSet<&Name> = SmallSet::new();
+        let mut kwparams: OrderedMap<&Name, (&Type, bool)> = OrderedMap::new();
+        let mut kwargs: Option<(Option<&Name>, &Type)> = None;
+        let mut kwargs_is_unpack: bool = false;
         loop {
             let p = match rparams.pop() {
                 Some(p) => p,
@@ -616,7 +680,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             };
             match p {
                 Param::PosOnly(name, _, required) => {
-                    if required == Required::Required {
+                    if required == &Required::Required {
                         if let Some(name) = name {
                             missing_named_posonly.insert(name);
                         } else {
@@ -626,18 +690,26 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 }
                 Param::VarArg(..) => {}
                 Param::Pos(name, ty, required) | Param::KwOnly(name, ty, required) => {
-                    kwparams.insert(name.clone(), (ty, required == Required::Required));
+                    kwparams.insert(name, (ty, required == &Required::Required));
                 }
-                Param::Kwargs(_, Type::Unpack(box Type::TypedDict(typed_dict))) => {
-                    self.typed_dict_fields(&typed_dict)
+                Param::Kwargs(name, Type::Unpack(box Type::TypedDict(typed_dict))) => {
+                    self.typed_dict_fields(typed_dict)
                         .into_iter()
                         .for_each(|(name, field)| {
-                            kwparams.insert(name, (field.ty, field.required));
+                            kwparams.insert(
+                                name_owner.push(name),
+                                (type_owner.push(field.ty), field.required),
+                            );
                         });
+                    if let ExtraItems::Extra(extra) =
+                        self.typed_dict_extra_items(typed_dict.class_object())
+                    {
+                        kwargs = Some((name.as_ref(), type_owner.push(extra.ty)))
+                    }
                     kwargs_is_unpack = true;
                 }
                 Param::Kwargs(name, ty) => {
-                    kwargs = Some((name, ty));
+                    kwargs = Some((name.as_ref(), ty));
                 }
             }
         }
@@ -664,8 +736,9 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 None => {
                     let ty = kw.value.infer(self, arg_errors);
                     if let Type::TypedDict(typed_dict) = ty {
-                        for (name, field) in self.typed_dict_fields(&typed_dict).iter() {
-                            let mut hint = kwargs.as_ref().map(|(_, ty)| ty.clone());
+                        for (name, field) in self.typed_dict_fields(&typed_dict).into_iter() {
+                            let name = name_owner.push(name);
+                            let mut hint = kwargs.as_ref().map(|(_, ty)| *ty);
                             if let Some(ty) = seen_names.get(name) {
                                 error(
                                     call_errors,
@@ -673,9 +746,9 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                                     ErrorKind::BadKeywordArgument,
                                     format!("Multiple values for argument `{name}`"),
                                 );
-                                hint = Some(ty.clone());
+                                hint = Some(*ty);
                             } else if let Some((ty, required)) = kwparams.get(name) {
-                                seen_names.insert(name.clone(), ty.clone());
+                                seen_names.insert(name, *ty);
                                 if *required && !field.required {
                                     error(
                                         call_errors,
@@ -684,16 +757,16 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                                         format!("Expected key `{name}` to be required"),
                                     );
                                 }
-                                hint = Some(ty.clone())
+                                hint = Some(*ty)
                             } else if kwargs.is_none() && !kwargs_is_unpack {
                                 unexpected_keyword_error(name, kw.range);
                             }
                             if let Some(want) = &hint {
-                                self.check_type(want, &field.ty, kw.range, call_errors, &|| {
+                                self.check_type(&field.ty, want, kw.range, call_errors, &|| {
                                     TypeCheckContext {
                                         kind: TypeCheckKind::CallArgument(
                                             Some(name.clone()),
-                                            callable_name.clone(),
+                                            callable_name.cloned(),
                                         ),
                                         context: context.map(|ctx| ctx()),
                                     }
@@ -706,15 +779,15 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                                 if self.is_subset_eq(&key, &self.stdlib.str().clone().to_type()) {
                                     if let Some((name, want)) = kwargs.as_ref() {
                                         self.check_type(
-                                            want,
                                             &value,
+                                            want,
                                             kw.range,
                                             call_errors,
                                             &|| TypeCheckContext {
                                                 kind: TypeCheckKind::CallKwArgs(
                                                     None,
-                                                    name.clone(),
-                                                    callable_name.clone(),
+                                                    name.cloned(),
+                                                    callable_name.cloned(),
                                                 ),
                                                 context: context.map(|ctx| ctx()),
                                             },
@@ -748,7 +821,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     }
                 }
                 Some(id) => {
-                    let mut hint = kwargs.as_ref().map(|(_, ty)| ty.clone());
+                    let mut hint = kwargs.as_ref().map(|(_, ty)| *ty);
                     let mut has_matching_param = false;
                     if let Some(ty) = seen_names.get(&id.id) {
                         error(
@@ -757,23 +830,23 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                             ErrorKind::BadKeywordArgument,
                             format!("Multiple values for argument `{}`", id.id),
                         );
-                        hint = Some(ty.clone());
+                        hint = Some(*ty);
                         has_matching_param = true;
                     } else if let Some((ty, _)) = kwparams.get(&id.id) {
-                        seen_names.insert(id.id.clone(), ty.clone());
-                        hint = Some(ty.clone());
+                        seen_names.insert(&id.id, *ty);
+                        hint = Some(*ty);
                         has_matching_param = true;
                     } else if kwargs.is_none() {
                         unexpected_keyword_error(&id.id, id.range);
                     }
                     let tcc: &dyn Fn() -> TypeCheckContext = &|| TypeCheckContext {
                         kind: if has_matching_param {
-                            TypeCheckKind::CallArgument(Some(id.id.clone()), callable_name.clone())
+                            TypeCheckKind::CallArgument(Some(id.id.clone()), callable_name.cloned())
                         } else {
                             TypeCheckKind::CallKwArgs(
                                 Some(id.id.clone()),
-                                kwargs.as_ref().and_then(|(name, _)| name.clone()),
-                                callable_name.clone(),
+                                kwargs.as_ref().and_then(|(name, _)| name.cloned()),
+                                callable_name.cloned(),
                             )
                         },
                         context: context.map(|ctx| ctx()),
@@ -782,7 +855,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                         TypeOrExpr::Expr(x) => {
                             self.expr_with_separate_check_errors(
                                 x,
-                                hint.as_ref().map(|ty| (ty, tcc, call_errors)),
+                                hint.map(|ty| (ty, call_errors, tcc)),
                                 arg_errors,
                             );
                         }
@@ -790,7 +863,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                             if let Some(hint) = &hint
                                 && !hint.is_any()
                             {
-                                self.check_type(hint, x, range, call_errors, tcc);
+                                self.check_type(x, hint, range, call_errors, tcc);
                             }
                         }
                     }
@@ -852,8 +925,11 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     }
                 }
                 for (ty, range) in &splat_kwargs {
-                    self.check_type(want, ty, *range, call_errors, &|| TypeCheckContext {
-                        kind: TypeCheckKind::CallUnpackKwArg(name.clone(), callable_name.clone()),
+                    self.check_type(ty, want, *range, call_errors, &|| TypeCheckContext {
+                        kind: TypeCheckKind::CallUnpackKwArg(
+                            (*name).clone(),
+                            callable_name.cloned(),
+                        ),
                         context: context.map(|ctx| ctx()),
                     });
                 }
@@ -895,16 +971,53 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     pub fn callable_infer(
         &self,
         callable: Callable,
-        callable_name: Option<FuncId>,
-        self_obj: Option<Type>,
-        args: &[CallArg],
+        callable_name: Option<&FunctionKind>,
+        tparams: Option<&TParams>,
+        mut self_obj: Option<Type>,
+        mut args: &[CallArg],
         keywords: &[CallKeyword],
         range: TextRange,
         arg_errors: &ErrorCollector,
         call_errors: &ErrorCollector,
         context: Option<&dyn Fn() -> ErrorContext>,
-        _hint: Option<&Type>,
+        hint: Option<HintRef>,
+        mut ctor_targs: Option<&mut TArgs>,
     ) -> Type {
+        let (qs, mut callable) = if let Some(tparams) = tparams {
+            // If we have a hint, we want to try to instantiate against it first, so we can contextually type
+            // arguments. If we don't match the hint, we need to throw away any instantiations we might have made.
+            // By invariant, hint will be None if we are calling a constructor.
+            if let Some(hint) = hint {
+                let (qs_, callable_) = self.instantiate_fresh_callable(tparams, callable.clone());
+                if self.is_subset_eq(&callable_.ret, hint.ty())
+                    && !self.solver().has_instantiation_errors(&qs_)
+                {
+                    (qs_, callable_)
+                } else {
+                    self.instantiate_fresh_callable(tparams, callable)
+                }
+            } else {
+                self.instantiate_fresh_callable(tparams, callable)
+            }
+        } else {
+            (QuantifiedHandle::empty(), callable)
+        };
+        if let Some(targs) = ctor_targs.as_mut() {
+            self.solver().freshen_class_targs(targs, self.uniques);
+            let mp = targs.substitution_map();
+            callable.params.visit_mut(&mut |t| t.subst_mut(&mp));
+            if let Some(obj) = self_obj.as_mut() {
+                obj.subst_mut(&mp);
+            } else if let Some(id) = callable_name
+                && id.function_name().as_ref() == &dunder::NEW
+                && let Some((first, rest)) = args.split_first()
+                && let CallArg::Arg(TypeOrExpr::Type(obj, _)) = first
+            {
+                // hack: we inserted a class type into the args list, but we need to substitute it
+                self_obj = Some((*obj).clone().subst(&mp));
+                args = rest;
+            }
+        }
         let self_arg = self_obj.as_ref().map(|ty| CallArg::ty(ty, range));
         match callable.params {
             Params::List(params) => {
@@ -921,14 +1034,14 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     context,
                 );
             }
-            Params::Ellipsis => {
+            Params::Ellipsis | Params::Materialization => {
                 // Deal with Callable[..., R]
                 for arg in self_arg.iter().chain(args.iter()) {
                     arg.pre_eval(self, arg_errors).post_infer(self, arg_errors)
                 }
             }
             Params::ParamSpec(concatenate, p) => {
-                let p = self.solver().expand(p);
+                let p = self.solver().expand_vars(p);
                 match p {
                     Type::ParamSpecValue(params) => self.callable_infer_params(
                         callable_name,
@@ -946,7 +1059,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     // Before we match an argument to `f`, we don't know what `P` is, so we don't have an answer for the Var yet.
                     Type::Var(var) => self.callable_infer_params(
                         callable_name,
-                        &ParamList::new_types(&concatenate),
+                        &ParamList::new_types(concatenate.into_vec()),
                         Some(var),
                         self_arg,
                         args,
@@ -959,10 +1072,10 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     Type::Quantified(q) => {
                         if !args
                             .last()
-                            .is_some_and(|x| self.is_param_spec_args(x, q.clone(), arg_errors))
-                            || !keywords.last().is_some_and(|x| {
-                                self.is_param_spec_kwargs(x, q.clone(), arg_errors)
-                            })
+                            .is_some_and(|x| self.is_param_spec_args(x, &q, arg_errors))
+                            || !keywords
+                                .last()
+                                .is_some_and(|x| self.is_param_spec_kwargs(x, &q, arg_errors))
                         {
                             self.error(
                                 call_errors,
@@ -977,7 +1090,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                         } else {
                             self.callable_infer_params(
                                 callable_name,
-                                &ParamList::new_types(&concatenate),
+                                &ParamList::new_types(concatenate.into_vec()),
                                 None,
                                 self_arg,
                                 &args[0..args.len() - 1],
@@ -1002,6 +1115,20 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 }
             }
         };
-        self.solver().expand(callable.ret)
+        if let Some(targs) = ctor_targs {
+            self.solver().generalize_class_targs(targs);
+        }
+        if let Err(e) = self.solver().finish_quantified(qs) {
+            for e in e {
+                let kind = TypeCheckKind::TypeVarSpecialization(e.name);
+                self.error(
+                    call_errors,
+                    range,
+                    ErrorInfo::new(kind.as_error_kind(), context),
+                    kind.format_error(&e.got, &e.want, self.module().name()),
+                );
+            }
+        }
+        self.solver().finish_function_return(callable.ret)
     }
 }
