@@ -7,9 +7,11 @@
 
 use std::sync::Arc;
 
+use dupe::Dupe;
 use pyrefly_python::dunder;
 use pyrefly_util::prelude::SliceExt;
 use ruff_python_ast::Arguments;
+use ruff_python_ast::Expr::EllipsisLiteral;
 use ruff_python_ast::name::Name;
 use ruff_text_size::TextRange;
 use starlark_map::small_map::SmallMap;
@@ -18,6 +20,7 @@ use vec1::Vec1;
 
 use crate::alt::answers::LookupAnswer;
 use crate::alt::answers_solver::AnswersSolver;
+use crate::alt::call::TargetWithTParams;
 use crate::alt::callable::CallArg;
 use crate::alt::callable::CallKeyword;
 use crate::alt::class::class_field::ClassField;
@@ -26,6 +29,10 @@ use crate::alt::types::class_metadata::ClassMetadata;
 use crate::alt::types::class_metadata::ClassSynthesizedField;
 use crate::alt::types::class_metadata::ClassSynthesizedFields;
 use crate::alt::types::class_metadata::DataclassMetadata;
+use crate::binding::pydantic::GE;
+use crate::binding::pydantic::GT;
+use crate::binding::pydantic::LT;
+use crate::binding::pydantic::STRICT;
 use crate::config::error_kind::ErrorKind;
 use crate::error::collector::ErrorCollector;
 use crate::error::context::ErrorInfo;
@@ -39,7 +46,6 @@ use crate::types::callable::ParamList;
 use crate::types::callable::Params;
 use crate::types::callable::Required;
 use crate::types::class::Class;
-use crate::types::class::ClassType;
 use crate::types::display::ClassDisplayContext;
 use crate::types::keywords::DataclassFieldKeywords;
 use crate::types::keywords::TypeMap;
@@ -53,7 +59,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     pub fn get_dataclass_fields(
         &self,
         cls: &Class,
-        bases_with_metadata: &[(ClassType, Arc<ClassMetadata>)],
+        bases_with_metadata: &[(Class, Arc<ClassMetadata>)],
     ) -> SmallSet<Name> {
         let mut all_fields = SmallSet::new();
         for (_, metadata) in bases_with_metadata.iter().rev() {
@@ -77,11 +83,16 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         let metadata = self.get_metadata_for_class(cls);
         let dataclass = metadata.dataclass_metadata()?;
         let mut fields = SmallMap::new();
+
         if dataclass.kws.init {
-            fields.insert(
-                dunder::INIT,
-                self.get_dataclass_init(cls, dataclass, errors),
-            );
+            let init_method = if let Some((root_model_type, has_strict)) =
+                self.get_pydantic_root_model_type_via_mro(cls, &metadata)
+            {
+                self.get_pydantic_root_model_init(cls, root_model_type, has_strict)
+            } else {
+                self.get_dataclass_init(cls, dataclass, dataclass.kws.strict, errors)
+            };
+            fields.insert(dunder::INIT, init_method);
         }
         let dataclass_fields_type = self.stdlib.dict(
             self.stdlib.str().clone().to_type(),
@@ -129,10 +140,10 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         &self,
         cls: &Class,
         dataclass_metadata: &DataclassMetadata,
-        bases_with_metadata: &[(ClassType, Arc<ClassMetadata>)],
+        bases_with_metadata: &[(Class, Arc<ClassMetadata>)],
         errors: &ErrorCollector,
     ) {
-        for (base_type, base_metadata) in bases_with_metadata {
+        for (base, base_metadata) in bases_with_metadata {
             if let Some(base_dataclass_metadata) = base_metadata.dataclass_metadata() {
                 let is_base_frozen = base_dataclass_metadata.kws.frozen;
                 let is_current_frozen = dataclass_metadata.kws.frozen;
@@ -149,7 +160,6 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                         "non-frozen"
                     };
 
-                    let base = base_type.class_object();
                     let ctx = ClassDisplayContext::new(&[cls, base]);
                     self.error(
                         errors,
@@ -181,14 +191,14 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         let mut params = Vec::new();
         for (name, field, _) in self.iter_fields(cls, dataclass_metadata, true) {
             if field.is_init_var() {
-                params.push(field.as_param(&name, false, false, None));
+                params.push(self.as_param(&field, &name, false, false, true, None, errors));
             }
         }
         let want = Type::Callable(Box::new(Callable::list(
             ParamList::new(params),
             self.stdlib.object().clone().to_type(),
         )));
-        self.check_type(&want, &post_init, range, errors, &|| {
+        self.check_type(&post_init, &want, range, errors, &|| {
             TypeCheckContext::of_kind(TypeCheckKind::PostInit)
         });
     }
@@ -197,9 +207,11 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         &self,
         func: &Type,
         args: &Arguments,
+        dataclass_metadata: &DataclassMetadata,
         errors: &ErrorCollector,
     ) -> DataclassFieldKeywords {
         let mut map = TypeMap::new();
+        let alias_keyword = &dataclass_metadata.alias_keyword;
         for kw in args.keywords.iter() {
             if let Some(name) = &kw.arg {
                 map.0
@@ -207,17 +219,35 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             }
         }
         let mut init = map.get_bool(&DataclassFieldKeywords::INIT);
-        let default = [
+        let mut default = [
             &DataclassFieldKeywords::DEFAULT,
             &DataclassFieldKeywords::DEFAULT_FACTORY,
             &DataclassFieldKeywords::FACTORY,
         ]
         .iter()
         .any(|k| map.0.contains_key(*k));
+
+        if !default && dataclass_metadata.default_can_be_positional {
+            // Check whether a default was passed positionally. This is needed for `pydantic.Field`.
+            default = !matches!(args.args.first(), Some(EllipsisLiteral(_)) | None);
+        }
+
         let mut kw_only = map.get_bool(&DataclassFieldKeywords::KW_ONLY);
-        let mut alias = map
-            .get_string(&DataclassFieldKeywords::ALIAS)
-            .map(Name::new);
+
+        let mut alias = if dataclass_metadata.init_defaults.init_by_alias {
+            map.get_string(alias_keyword)
+                .or_else(|| map.get_string(&DataclassFieldKeywords::ALIAS))
+                .map(Name::new)
+        } else {
+            None
+        };
+
+        let gt = map.0.get(&GT).cloned();
+        let lt = map.0.get(&LT).cloned();
+        let ge = map.0.get(&GE).cloned();
+
+        let strict: Option<bool> = map.0.get(&STRICT).and_then(|v| v.as_bool());
+
         let mut converter_param = map
             .0
             .get(&DataclassFieldKeywords::CONVERTER)
@@ -229,6 +259,11 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 func,
                 args,
                 errors,
+                if dataclass_metadata.init_defaults.init_by_alias {
+                    Some(alias_keyword)
+                } else {
+                    None
+                },
                 &mut init,
                 &mut kw_only,
                 &mut alias,
@@ -239,7 +274,12 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             init: init.unwrap_or(true),
             default,
             kw_only,
-            alias,
+            init_by_name: dataclass_metadata.init_defaults.init_by_name || alias.is_none(),
+            init_by_alias: alias,
+            lt,
+            gt,
+            ge,
+            strict,
             converter_param,
         }
     }
@@ -250,6 +290,8 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         func: &Type,
         args: &Arguments,
         errors: &ErrorCollector,
+        // The name of the function parameter from which to fill in an alias keyword value
+        alias_keyword: Option<&Name>,
         init: &mut Option<bool>,
         kw_only: &mut Option<bool>,
         alias: &mut Option<Name>,
@@ -262,10 +304,17 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             && let Type::Overload(overload) = func
         {
             // Overloaded function. Call it to see which signature is actually used.
+            // TODO: sigs could contain unbound type parameters, because `callable_signatures`
+            // looks through foralls. Overload selection might fail spuriously.
             self.call_overloads(
-                Vec1::try_from_vec(sigs.map(|x| Function {
-                    signature: ((*x).clone()),
-                    metadata: *overload.metadata.clone(),
+                Vec1::try_from_vec(sigs.map(|x| {
+                    TargetWithTParams(
+                        None,
+                        Function {
+                            signature: ((*x).clone()),
+                            metadata: *overload.metadata.clone(),
+                        },
+                    )
                 }))
                 .unwrap(),
                 (*overload.metadata).clone(),
@@ -276,6 +325,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 errors,
                 None,
                 None,
+                None,
             )
             .1
         } else {
@@ -299,7 +349,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 if name == &DataclassFieldKeywords::KW_ONLY {
                     self.fill_in_literal(kw_only, ty, default_ty, |ty| ty.as_bool());
                 }
-                if alias.is_none() && name == &DataclassFieldKeywords::ALIAS {
+                if alias.is_none() && Some(name) == alias_keyword {
                     self.fill_in_literal(alias, ty, default_ty, |ty| match ty {
                         Type::Literal(Lit::Str(s)) => Some(Name::new(s)),
                         _ => None,
@@ -334,12 +384,13 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     fn get_converter_param(&self, converter: &Type) -> Type {
         let converter = {
             if let Type::ClassDef(cls) = converter
-                && let Type::ClassType(instance) = self.instantiate_fresh(cls)
+                && let Type::ClassType(instance) = self.promote_silently(cls)
             {
                 let callable = self.constructor_to_callable(&instance);
                 &self.distribute_over_union(&callable, |ty| {
                     if let Type::BoundMethod(m) = ty {
-                        m.drop_self().unwrap_or_else(|| ty.clone())
+                        self.bind_boundmethod(m, &mut |a, b| self.is_subset_eq(a, b))
+                            .unwrap_or_else(|| ty.clone())
                     } else {
                         ty.clone()
                     }
@@ -402,13 +453,16 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         &self,
         cls: &Class,
         dataclass: &DataclassMetadata,
+        strict_default: bool,
         errors: &ErrorCollector,
     ) -> ClassSynthesizedField {
         let mut params = vec![self.class_self_param(cls, false)];
         let mut has_seen_default = false;
         for (name, field, field_flags) in self.iter_fields(cls, dataclass, true) {
+            let strict = field_flags.strict.unwrap_or(strict_default);
             if field_flags.init {
-                let has_default = field_flags.default;
+                let has_default = field_flags.default
+                    || (field_flags.init_by_name && field_flags.init_by_alias.is_some());
                 let is_kw_only = field_flags.is_kw_only();
                 if !is_kw_only {
                     if !has_default
@@ -416,30 +470,49 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                         && let Some(range) = cls.field_decl_range(&name)
                     {
                         self.error(
-                                errors,
-                                range,
-                                ErrorInfo::Kind(ErrorKind::BadClassDefinition),
-                                format!(
-                                    "Dataclass field `{name}` without a default may not follow dataclass field with a default"
-                                ),
-                            );
+                            errors,
+                            range,
+                            ErrorInfo::Kind(ErrorKind::BadClassDefinition),
+                            format!(
+                                "Dataclass field `{name}` without a default may not follow dataclass field with a default"
+                            ),
+                        );
                     }
                     if has_default {
                         has_seen_default = true;
                     }
                 }
-                params.push(field.as_param(
-                    &field_flags.alias.unwrap_or(name),
-                    has_default,
-                    is_kw_only,
-                    field_flags.converter_param,
-                ));
+                if field_flags.init_by_name {
+                    params.push(self.as_param(
+                        &field,
+                        &name,
+                        has_default,
+                        is_kw_only,
+                        strict,
+                        field_flags.converter_param.clone(),
+                        errors,
+                    ));
+                }
+                if let Some(alias) = &field_flags.init_by_alias {
+                    params.push(self.as_param(
+                        &field,
+                        alias,
+                        has_default,
+                        is_kw_only,
+                        strict,
+                        field_flags.converter_param.clone(),
+                        errors,
+                    ));
+                }
             }
         }
+        if dataclass.kws.extra {
+            params.push(Param::Kwargs(None, Type::Any(AnyStyle::Implicit)));
+        }
 
         let ty = Type::Function(Box::new(Function {
             signature: Callable::list(ParamList::new(params), Type::None),
-            metadata: FuncMetadata::def(self.module().name(), cls.name().clone(), dunder::INIT),
+            metadata: FuncMetadata::def(self.module().dupe(), cls.dupe(), dunder::INIT),
         }));
         ClassSynthesizedField::new(ty)
     }
@@ -508,11 +581,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                         } else {
                             callable.clone()
                         },
-                        metadata: FuncMetadata::def(
-                            self.module().name(),
-                            cls.name().clone(),
-                            name.clone(),
-                        ),
+                        metadata: FuncMetadata::def(self.module().dupe(), cls.dupe(), name.clone()),
                     }))),
                 )
             })
@@ -524,7 +593,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         let ret = self.stdlib.int().clone().to_type();
         ClassSynthesizedField::new(Type::Function(Box::new(Function {
             signature: Callable::list(ParamList::new(params), ret),
-            metadata: FuncMetadata::def(self.module().name(), cls.name().clone(), dunder::HASH),
+            metadata: FuncMetadata::def(self.module().dupe(), cls.dupe(), dunder::HASH),
         })))
     }
 }
