@@ -5,14 +5,15 @@
  * LICENSE file in the root directory of this source tree.
  */
 
-pub mod base_class;
 pub mod class_field;
 pub mod class_metadata;
 pub mod classdef;
 pub mod dataclass;
+pub mod django;
 pub mod enums;
 pub mod named_tuple;
 pub mod new_type;
+pub mod pydantic;
 pub mod targs;
 pub mod total_ordering;
 pub mod tparams;
