@@ -8,24 +8,32 @@
 use std::sync::Arc;
 
 use dupe::Dupe;
+use pyrefly_types::callable::Callable;
+use pyrefly_types::callable::Function;
 use pyrefly_util::display::count;
 use pyrefly_util::prelude::SliceExt;
 use ruff_python_ast::name::Name;
 use ruff_text_size::TextRange;
 use starlark_map::small_map::SmallMap;
+use vec1::vec1;
 
 use crate::alt::answers::LookupAnswer;
 use crate::alt::answers_solver::AnswersSolver;
 use crate::config::error_kind::ErrorKind;
 use crate::error::collector::ErrorCollector;
 use crate::error::context::ErrorInfo;
+use crate::error::context::TypeCheckContext;
+use crate::error::context::TypeCheckKind;
+use crate::solver::solver::QuantifiedHandle;
 use crate::types::callable::Param;
 use crate::types::callable::ParamList;
 use crate::types::callable::Required;
 use crate::types::class::Class;
 use crate::types::class::ClassType;
+use crate::types::quantified::Quantified;
 use crate::types::quantified::QuantifiedKind;
 use crate::types::tuple::Tuple;
+use crate::types::type_var::PreInferenceVariance;
 use crate::types::typed_dict::TypedDict;
 use crate::types::types::Forall;
 use crate::types::types::Forallable;
@@ -33,7 +41,6 @@ use crate::types::types::TArgs;
 use crate::types::types::TParam;
 use crate::types::types::TParams;
 use crate::types::types::Type;
-use crate::types::types::Var;
 
 impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     /// Silently promotes a Class to a ClassType, using default type arguments. It is up to the
@@ -46,6 +53,40 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         )
     }
 
+    fn specialize_impl(
+        &self,
+        cls: &Class,
+        targs: Vec<Type>,
+        range: TextRange,
+        validate_restriction: bool,
+        errors: &ErrorCollector,
+    ) -> Type {
+        let metadata = self.get_metadata_for_class(cls);
+        let tparams = self.get_class_tparams(cls);
+
+        // We didn't find any type parameters for this class, but it may have ones we don't know about if:
+        // - the class inherits from Any, or
+        // - the class inherits from Generic[...] or Protocol [...]. We probably dropped the type
+        //   arguments because we found an error in them.
+        let has_unknown_tparams =
+            tparams.is_empty() && (metadata.has_base_any() || metadata.has_generic_base_class());
+
+        let targs = if !targs.is_empty() && has_unknown_tparams {
+            // Accept any number of arguments (by ignoring them).
+            TArgs::default()
+        } else {
+            self.create_targs(
+                cls.name(),
+                tparams,
+                targs,
+                range,
+                validate_restriction,
+                errors,
+            )
+        };
+        self.type_of_instance(cls, targs)
+    }
+
     /// Given a class or typed dictionary and some (explicit) type arguments, construct a `Type`
     /// that represents the type of an instance of the class or typed dictionary with those `targs`.
     ///
@@ -60,19 +101,36 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         range: TextRange,
         errors: &ErrorCollector,
     ) -> Type {
-        let targs = if !targs.is_empty() && self.get_metadata_for_class(cls).has_unknown_tparams() {
-            // Accept any number of arguments (by ignoring them).
-            TArgs::default()
-        } else {
-            self.check_and_create_targs(
-                cls.name(),
-                self.get_class_tparams(cls),
-                targs,
-                range,
-                errors,
-            )
-        };
-        self.type_of_instance(cls, targs)
+        self.specialize_impl(cls, targs, range, true, errors)
+    }
+
+    pub fn specialize_in_base_class(
+        &self,
+        cls: &Class,
+        targs: Vec<Type>,
+        range: TextRange,
+        errors: &ErrorCollector,
+    ) -> Type {
+        self.specialize_impl(cls, targs, range, false, errors)
+    }
+
+    fn specialize_forall_impl(
+        &self,
+        forall: Forall<Forallable>,
+        targs: Vec<Type>,
+        range: TextRange,
+        validate_restriction: bool,
+        errors: &ErrorCollector,
+    ) -> Type {
+        let targs = self.create_targs(
+            &forall.body.name(),
+            forall.tparams.dupe(),
+            targs,
+            range,
+            validate_restriction,
+            errors,
+        );
+        forall.apply_targs(targs)
     }
 
     pub fn specialize_forall(
@@ -82,14 +140,17 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         range: TextRange,
         errors: &ErrorCollector,
     ) -> Type {
-        let targs = self.check_and_create_targs(
-            &forall.body.name(),
-            forall.tparams.dupe(),
-            targs,
-            range,
-            errors,
-        );
-        forall.apply_targs(targs)
+        self.specialize_forall_impl(forall, targs, range, true, errors)
+    }
+
+    pub fn specialize_forall_in_base_class(
+        &self,
+        forall: Forall<Forallable>,
+        targs: Vec<Type>,
+        range: TextRange,
+        errors: &ErrorCollector,
+    ) -> Type {
+        self.specialize_forall_impl(forall, targs, range, false, errors)
     }
 
     /// Given a class or typed dictionary, create a `Type` that represents to an instance annotated
@@ -105,13 +166,30 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     /// specialize(list, [int]) == list[int]
     /// promote(list) == list[Any]
     /// instantiate(list) == list[T]
-    pub fn promote(&self, cls: &Class, range: TextRange) -> Type {
-        let targs = self.create_default_targs(self.get_class_tparams(cls), Some(range));
+    pub fn promote(&self, cls: &Class, range: TextRange, errors: &ErrorCollector) -> Type {
+        let targs = self.create_default_targs(
+            self.get_class_tparams(cls),
+            Some(&|tparam: &TParam| {
+                errors.add(
+                    range,
+                    ErrorInfo::Kind(ErrorKind::ImplicitAny),
+                    vec1![
+                        format!(
+                            "Cannot determine the type parameter `{}` for generic class `{}`",
+                            tparam.name(),
+                            cls.name(),
+                        ),
+                        "Either specify the type argument explicitly, or specify a default for the type variable.".to_owned(),
+                    ],
+                );
+            }),
+        );
         self.type_of_instance(cls, targs)
     }
 
-    pub fn promote_forall(&self, forall: Forall<Forallable>, range: TextRange) -> Type {
-        let targs = self.create_default_targs(forall.tparams.dupe(), Some(range));
+    pub fn promote_forall(&self, forall: Forall<Forallable>, _range: TextRange) -> Type {
+        // TODO(grievejia): We probably want to error here as well
+        let targs = self.create_default_targs(forall.tparams.dupe(), None);
         forall.apply_targs(targs)
     }
 
@@ -156,7 +234,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     }
 
     /// Instantiates a class or typed dictionary with fresh variables for its type parameters.
-    pub fn instantiate_fresh(&self, cls: &Class) -> Type {
+    pub fn instantiate_fresh_class(&self, cls: &Class) -> Type {
         self.solver()
             .fresh_quantified(
                 &self.get_class_tparams(cls),
@@ -166,32 +244,84 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             .1
     }
 
-    pub fn instantiate_forall(&self, forall: Forall<Forallable>) -> (Vec<Var>, Type) {
+    pub fn instantiate_fresh_tuple(&self) -> Type {
+        let quantified = Quantified::type_var_tuple(Name::new_static("Ts"), self.uniques, None);
+        let tparams = TParams::new(vec![TParam {
+            quantified: quantified.clone(),
+            variance: PreInferenceVariance::PCovariant,
+        }]);
+        let tuple_ty = Type::Tuple(Tuple::Unpacked(Box::new((
+            Vec::new(),
+            Type::Quantified(Box::new(quantified)),
+            Vec::new(),
+        ))));
+        self.solver()
+            .fresh_quantified(&tparams, tuple_ty, self.uniques)
+            .1
+    }
+
+    pub fn instantiate_fresh_forall(&self, forall: Forall<Forallable>) -> (QuantifiedHandle, Type) {
         self.solver()
             .fresh_quantified(&forall.tparams, forall.body.as_type(), self.uniques)
     }
 
+    pub fn instantiate_fresh_function(
+        &self,
+        tparams: &TParams,
+        func: Function,
+    ) -> (QuantifiedHandle, Function) {
+        let (qs, t) =
+            self.solver()
+                .fresh_quantified(tparams, Type::Function(Box::new(func)), self.uniques);
+        match t {
+            Type::Function(func) => (qs, *func),
+            // We passed a Function to fresh_quantified(), so we know we get a Function back out.
+            _ => unreachable!(),
+        }
+    }
+
+    pub fn instantiate_fresh_callable(
+        &self,
+        tparams: &TParams,
+        c: Callable,
+    ) -> (QuantifiedHandle, Callable) {
+        let (qs, t) =
+            self.solver()
+                .fresh_quantified(tparams, Type::Callable(Box::new(c)), self.uniques);
+        match t {
+            Type::Callable(c) => (qs, *c),
+            // We passed a Function to fresh_quantified(), so we know we get a Function back out.
+            _ => unreachable!(),
+        }
+    }
+
     /// Creates default type arguments for a class, falling back to Any for type parameters without defaults.
     fn create_default_targs(
         &self,
         tparams: Arc<TParams>,
-        // Placeholder for strict mode: we want to force callers to pass a range so
-        // that we don't refactor in a way where none is available, but this is unused
-        // because we do not have a strict mode yet.
-        _range: Option<TextRange>,
+        on_fallback_to_gradual: Option<&dyn Fn(&TParam)>,
     ) -> TArgs {
         if tparams.is_empty() {
             TArgs::default()
         } else {
-            // TODO(stroxler): We should error here, but the error needs to be
-            // configurable in the long run, and also suppressed in dependencies
-            // no matter what the configuration is.
-            //
-            // Our plumbing isn't ready for that yet, so for now we are silently
-            // using gradual type arguments.
             let tys = tparams
                 .iter()
-                .map(|x| x.quantified.as_gradual_type())
+                .map(|x| {
+                    // TODO(grievejia): This is actually not a 100% accurate way of detecting graudal fallbacks:
+                    // it will trigger when the tparam doesn't have a default, but won't trigger if
+                    // - The tparam has a default, but that default is another type var without a default.
+                    // - The tparam has a default, but part of that default type requires another fallback (e.g. the
+                    //   default is `list[Foo]`, where `Foo` is a generic class whose tparam doesn't have a default).
+                    //
+                    // To make it 100% accurate, we actually need to hook the callback into `as_graudal_type()`. It's doable
+                    // but could add a lot of complexities so let's keep it as an exercise in the future.
+                    if let Some(f) = on_fallback_to_gradual
+                        && x.quantified.default().is_none()
+                    {
+                        f(x);
+                    }
+                    x.quantified.as_gradual_type()
+                })
                 .collect();
             TArgs::new(tparams, tys)
         }
@@ -206,12 +336,13 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         }
     }
 
-    fn check_and_create_targs(
+    fn create_targs(
         &self,
         name: &Name,
         tparams: Arc<TParams>,
         targs: Vec<Type>,
         range: TextRange,
+        validate_restriction: bool,
         errors: &ErrorCollector,
     ) -> TArgs {
         let nparams = tparams.len();
@@ -257,7 +388,13 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                         targ_idx += 1;
                     }
                     QuantifiedKind::TypeVar => {
-                        checked_targs.push(self.create_next_typevar_arg(param, arg, range, errors));
+                        checked_targs.push(self.create_next_typevar_arg(
+                            param,
+                            arg,
+                            range,
+                            validate_restriction,
+                            errors,
+                        ));
                         targ_idx += 1;
                     }
                 }
@@ -344,7 +481,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 }
                 Type::Unpack(t) => {
                     if !suffix.is_empty() {
-                        middle.push(Type::Tuple(Tuple::Unbounded(Box::new(self.unions(suffix)))));
+                        middle.push(Type::Tuple(Tuple::unbounded(self.unions(suffix))));
                         suffix = Vec::new();
                     } else {
                         middle.push((**t).clone())
@@ -384,7 +521,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     .collect();
                 Type::Tuple(Tuple::unpacked(
                     prefix,
-                    Type::Tuple(Tuple::Unbounded(Box::new(self.unions(middle_types)))),
+                    Type::Tuple(Tuple::unbounded(self.unions(middle_types))),
                     suffix,
                 ))
             }
@@ -423,6 +560,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         param: &TParam,
         arg: &Type,
         range: TextRange,
+        validate_restriction: bool,
         errors: &ErrorCollector,
     ) -> Type {
         match arg {
@@ -451,6 +589,31 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                         "`ParamSpec` cannot be used for type parameter".to_owned(),
                     )
                 } else {
+                    let restriction = param.restriction();
+                    if validate_restriction && restriction.is_restricted() {
+                        let tcc = &|| {
+                            TypeCheckContext::of_kind(TypeCheckKind::TypeVarSpecialization(
+                                param.name().clone(),
+                            ))
+                        };
+                        // In a legacy type alias, one old-style TypeVar can be specialized with
+                        // another, which we handle by checking their upper bounds against each other.
+                        let arg_for_check = {
+                            let arg = arg.clone();
+                            arg.transform(&mut |x| {
+                                if let Type::TypeVar(tv) = x {
+                                    *x = tv.restriction().as_type(self.stdlib);
+                                }
+                            })
+                        };
+                        self.check_type(
+                            &arg_for_check,
+                            &restriction.as_type(self.stdlib),
+                            range,
+                            errors,
+                            tcc,
+                        );
+                    }
                     arg.clone()
                 }
             }
