@@ -8,6 +8,7 @@
 use crate::alt::answers::LookupAnswer;
 use crate::alt::answers_solver::AnswersSolver;
 use crate::types::class::ClassType;
+use crate::types::simplify::simplify_tuples;
 use crate::types::tuple::Tuple;
 use crate::types::types::Type;
 
@@ -18,16 +19,16 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         }
         if cls.class_object().is_builtin("tuple") && cls.targs().as_slice().len() == 1 {
             let tuple_targ: Type = cls.targs().as_slice()[0].clone();
-            return Some(Tuple::Unbounded(Box::new(tuple_targ)));
+            return Some(Tuple::unbounded(tuple_targ));
         }
-        let class_metadata = self.get_metadata_for_class(cls.class_object());
-        if let Some(Type::Tuple(tuple)) = class_metadata
-            .tuple_base()
+        let class_bases = self.get_base_types_for_class(cls.class_object());
+        if let Some(Type::Tuple(tuple)) = class_bases
+            .tuple_ancestor()
             .cloned()
             .map(Type::Tuple)
-            .map(|ty| cls.targs().substitute(ty))
+            .map(|ty| cls.targs().substitute_into(ty))
         {
-            return Some(tuple);
+            return Some(simplify_tuples(tuple));
         }
         None
     }
