@@ -6,32 +6,37 @@
  */
 
 use num_traits::ToPrimitive;
+use pyrefly_python::ast::Ast;
+use pyrefly_types::type_info::JoinStyle;
 use pyrefly_util::prelude::SliceExt;
 use ruff_python_ast::Arguments;
 use ruff_python_ast::AtomicNodeIndex;
 use ruff_python_ast::Expr;
 use ruff_python_ast::ExprNumberLiteral;
-use ruff_python_ast::ExprStringLiteral;
 use ruff_python_ast::Int;
 use ruff_python_ast::Number;
-use ruff_python_ast::StringLiteral;
-use ruff_python_ast::StringLiteralFlags;
-use ruff_python_ast::StringLiteralValue;
 use ruff_python_ast::name::Name;
 use ruff_text_size::Ranged;
 use ruff_text_size::TextRange;
+use ruff_text_size::TextSize;
+use vec1::Vec1;
 
 use crate::alt::answers::LookupAnswer;
 use crate::alt::answers_solver::AnswersSolver;
+use crate::alt::call::CallTargetLookup;
 use crate::alt::callable::CallArg;
 use crate::alt::callable::CallKeyword;
 use crate::binding::narrow::AtomicNarrowOp;
+use crate::binding::narrow::FacetOrigin;
+use crate::binding::narrow::FacetSubject;
 use crate::binding::narrow::NarrowOp;
 use crate::error::collector::ErrorCollector;
+use crate::error::style::ErrorStyle;
 use crate::types::callable::FunctionKind;
 use crate::types::class::ClassType;
 use crate::types::facet::FacetChain;
 use crate::types::facet::FacetKind;
+use crate::types::lit_int::LitInt;
 use crate::types::literal::Lit;
 use crate::types::tuple::Tuple;
 use crate::types::type_info::TypeInfo;
@@ -99,7 +104,8 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         self.intersect_with_fallback(left, right, Type::never)
     }
 
-    fn intersects(&self, ts: &[Type]) -> Type {
+    /// Calculate the intersection of a number of types
+    pub fn intersects(&self, ts: &[Type]) -> Type {
         match ts {
             [] => Type::ClassType(self.stdlib.object().clone()),
             [ty] => ty.clone(),
@@ -174,10 +180,18 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         self.intersects(&res)
     }
 
-    fn narrow_issubclass(&self, left: &Type, right: &Type, range: TextRange) -> Type {
+    fn narrow_issubclass(
+        &self,
+        left: &Type,
+        right: &Type,
+        range: TextRange,
+        errors: &ErrorCollector,
+    ) -> Type {
         let mut res = Vec::new();
         for right in self.as_class_info(right.clone()) {
-            if let Some(left) = self.untype_opt(left.clone(), range)
+            if matches!(left, Type::ClassDef(_)) && matches!(right, Type::ClassDef(_)) {
+                res.push(self.intersect(left, &right))
+            } else if let Some(left) = self.untype_opt(left.clone(), range, errors)
                 && let Some(right) = self.unwrap_class_object_silently(&right)
             {
                 res.push(Type::type_form(self.intersect(&left, &right)))
@@ -188,10 +202,18 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         self.unions(res)
     }
 
-    fn narrow_is_not_subclass(&self, left: &Type, right: &Type, range: TextRange) -> Type {
+    fn narrow_is_not_subclass(
+        &self,
+        left: &Type,
+        right: &Type,
+        range: TextRange,
+        errors: &ErrorCollector,
+    ) -> Type {
         let mut res = Vec::new();
         for right in self.as_class_info(right.clone()) {
-            if let Some(left) = self.untype_opt(left.clone(), range)
+            if matches!(left, Type::ClassDef(_)) && matches!(right, Type::ClassDef(_)) {
+                res.push(self.subtract(left, &right))
+            } else if let Some(left) = self.untype_opt(left.clone(), range, errors)
                 && let Some(right) = self.unwrap_class_object_silently(&right)
             {
                 res.push(Type::type_form(self.subtract(&left, &right)))
@@ -239,7 +261,103 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         })
     }
 
-    pub fn atomic_narrow(
+    // Try to narrow a type based on the type of its facet.
+    // For example, if we have a `x.y == 0` check and `x` is some union,
+    // we can eliminate cases from the union where `x.y` is some other
+    // literal.
+    pub fn atomic_narrow_for_facet(
+        &self,
+        base: &Type,
+        facet: &FacetKind,
+        op: &AtomicNarrowOp,
+        range: TextRange,
+        errors: &ErrorCollector,
+    ) -> Option<Type> {
+        match op {
+            AtomicNarrowOp::Is(v) => {
+                let right = self.expr_infer(v, errors);
+                Some(self.distribute_over_union(base, |t| {
+                    let base_info = TypeInfo::of_ty(t.clone());
+                    let facet_ty = self.get_facet_chain_type(
+                        &base_info,
+                        &FacetChain::new(Vec1::new(facet.clone())),
+                        range,
+                    );
+                    match right {
+                        Type::None | Type::Literal(Lit::Bool(_)) | Type::Literal(Lit::Enum(_)) => {
+                            if self.is_subset_eq(&right, &facet_ty) {
+                                t.clone()
+                            } else {
+                                Type::never()
+                            }
+                        }
+                        _ => t.clone(),
+                    }
+                }))
+            }
+            AtomicNarrowOp::IsNot(v) => {
+                let right = self.expr_infer(v, errors);
+                Some(self.distribute_over_union(base, |t| {
+                    let base_info = TypeInfo::of_ty(t.clone());
+                    let facet_ty = self.get_facet_chain_type(
+                        &base_info,
+                        &FacetChain::new(Vec1::new(facet.clone())),
+                        range,
+                    );
+                    match (&facet_ty, &right) {
+                        (
+                            Type::None | Type::Literal(Lit::Bool(_)) | Type::Literal(Lit::Enum(_)),
+                            Type::None | Type::Literal(Lit::Bool(_)) | Type::Literal(Lit::Enum(_)),
+                        ) if right == facet_ty => Type::never(),
+                        _ => t.clone(),
+                    }
+                }))
+            }
+            AtomicNarrowOp::Eq(v) => {
+                let right = self.expr_infer(v, errors);
+                Some(self.distribute_over_union(base, |t| {
+                    let base_info = TypeInfo::of_ty(t.clone());
+                    let facet_ty = self.get_facet_chain_type(
+                        &base_info,
+                        &FacetChain::new(Vec1::new(facet.clone())),
+                        range,
+                    );
+                    match right {
+                        Type::None | Type::Literal(_) => {
+                            if self.is_subset_eq(&right, &facet_ty) {
+                                t.clone()
+                            } else {
+                                Type::never()
+                            }
+                        }
+                        _ => t.clone(),
+                    }
+                }))
+            }
+            AtomicNarrowOp::NotEq(v) => {
+                let right = self.expr_infer(v, errors);
+                Some(self.distribute_over_union(base, |t| {
+                    let base_info = TypeInfo::of_ty(t.clone());
+                    let facet_ty = self.get_facet_chain_type(
+                        &base_info,
+                        &FacetChain::new(Vec1::new(facet.clone())),
+                        range,
+                    );
+                    match (&facet_ty, &right) {
+                        (Type::None | Type::Literal(_), Type::None | Type::Literal(_))
+                            if right == facet_ty =>
+                        {
+                            Type::never()
+                        }
+                        _ => t.clone(),
+                    }
+                }))
+            }
+            _ => None,
+        }
+    }
+
+    fn atomic_narrow(
         &self,
         ty: &Type,
         op: &AtomicNarrowOp,
@@ -471,17 +589,23 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             AtomicNarrowOp::TypeNotEq(_) => ty.clone(),
             AtomicNarrowOp::IsSubclass(v) => {
                 let right = self.expr_infer(v, errors);
-                self.narrow_issubclass(ty, &right, v.range())
+                self.narrow_issubclass(ty, &right, v.range(), errors)
             }
             AtomicNarrowOp::IsNotSubclass(v) => {
                 let right = self.expr_infer(v, errors);
-                self.narrow_is_not_subclass(ty, &right, v.range())
+                self.narrow_is_not_subclass(ty, &right, v.range(), errors)
             }
+            // `hasattr` and `getattr` are handled in `narrow`
+            AtomicNarrowOp::HasAttr(_) => ty.clone(),
+            AtomicNarrowOp::NotHasAttr(_) => ty.clone(),
+            AtomicNarrowOp::GetAttr(_, _) => ty.clone(),
+            AtomicNarrowOp::NotGetAttr(_, _) => ty.clone(),
             AtomicNarrowOp::TypeGuard(t, arguments) => {
-                if let Some(call_target) = self.as_call_target(t.clone()) {
+                if let CallTargetLookup::Ok(call_target) = self.as_call_target(t.clone()) {
                     let args = arguments.args.map(CallArg::expr_maybe_starred);
                     let kws = arguments.keywords.map(CallKeyword::new);
-                    let ret = self.call_infer(call_target, &args, &kws, range, errors, None, None);
+                    let ret =
+                        self.call_infer(call_target, &args, &kws, range, errors, None, None, None);
                     if let Type::TypeGuard(t) = ret {
                         return *t;
                     }
@@ -490,10 +614,11 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             }
             AtomicNarrowOp::NotTypeGuard(_, _) => ty.clone(),
             AtomicNarrowOp::TypeIs(t, arguments) => {
-                if let Some(call_target) = self.as_call_target(t.clone()) {
+                if let CallTargetLookup::Ok(call_target) = self.as_call_target(t.clone()) {
                     let args = arguments.args.map(CallArg::expr_maybe_starred);
                     let kws = arguments.keywords.map(CallKeyword::new);
-                    let ret = self.call_infer(call_target, &args, &kws, range, errors, None, None);
+                    let ret =
+                        self.call_infer(call_target, &args, &kws, range, errors, None, None, None);
                     if let Type::TypeIs(t) = ret {
                         return self.intersect(ty, &t);
                     }
@@ -501,10 +626,11 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 ty.clone()
             }
             AtomicNarrowOp::NotTypeIs(t, arguments) => {
-                if let Some(call_target) = self.as_call_target(t.clone()) {
+                if let CallTargetLookup::Ok(call_target) = self.as_call_target(t.clone()) {
                     let args = arguments.args.map(CallArg::expr_maybe_starred);
                     let kws = arguments.keywords.map(CallKeyword::new);
-                    let ret = self.call_infer(call_target, &args, &kws, range, errors, None, None);
+                    let ret =
+                        self.call_infer(call_target, &args, &kws, range, errors, None, None, None);
                     if let Type::TypeIs(t) = ret {
                         return self.subtract(ty, &t);
                     }
@@ -514,13 +640,32 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             AtomicNarrowOp::IsTruthy | AtomicNarrowOp::IsFalsy => {
                 self.distribute_over_union(ty, |t| {
                     let boolval = matches!(op, AtomicNarrowOp::IsTruthy);
-                    if t.as_bool() == Some(!boolval) {
-                        Type::never()
-                    } else if matches!(t, Type::ClassType(cls) if cls.is_builtin("bool")) {
-                        Type::Literal(Lit::Bool(boolval))
-                    } else {
-                        t.clone()
+                    // Do not emit errors here: the narrowed range doesn't always correspond to a valid expression
+                    // For example, narrowing generated for implicit else branches.
+                    if self.as_bool(
+                        t,
+                        range,
+                        &ErrorCollector::new(errors.module().clone(), ErrorStyle::Never),
+                    ) == Some(!boolval)
+                    {
+                        return Type::never();
+                    } else if let Type::ClassType(cls) = t {
+                        if cls.is_builtin("bool") {
+                            return Type::Literal(Lit::Bool(boolval));
+                        }
+                        if !boolval {
+                            if cls.is_builtin("int") {
+                                return Type::Literal(Lit::Int(LitInt::new(0)));
+                            } else if cls.is_builtin("str") {
+                                return Type::Literal(Lit::Str("".into()));
+                            } else if cls.is_builtin("bytes") {
+                                let empty = Vec::new();
+                                return Type::Literal(Lit::Bytes(empty.into_boxed_slice()));
+                            }
+                        }
                     }
+
+                    t.clone()
                 })
             }
             AtomicNarrowOp::Eq(v) => {
@@ -596,7 +741,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         match first_facet {
             FacetKind::Attribute(first_attr_name) => match remaining_facets.split_first() {
                 None => match base.type_at_facet(first_facet) {
-                    Some(ty) => ty.clone(),
+                    Some(ty) => self.force_for_narrowing(ty),
                     None => self.narrowable_for_attr(base.ty(), first_attr_name, range, errors),
                 },
                 Some((next_name, remaining_facets)) => {
@@ -612,15 +757,15 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             },
             FacetKind::Index(idx) => {
                 // We synthesize a slice expression for the subscript here
-                // The range doesn't matter, since narrowing logic swallows type errors
+                // Use a synthesized fake range to avoid overwriting typing traces
                 let synthesized_slice = Expr::NumberLiteral(ExprNumberLiteral {
                     node_index: AtomicNodeIndex::dummy(),
-                    range,
+                    range: TextRange::empty(TextSize::from(0)),
                     value: Number::Int(Int::from(*idx as u64)),
                 });
                 match remaining_facets.split_first() {
                     None => match base.type_at_facet(first_facet) {
-                        Some(ty) => ty.clone(),
+                        Some(ty) => self.force_for_narrowing(ty),
                         None => self.subscript_infer_for_type(
                             base.ty(),
                             &synthesized_slice,
@@ -642,20 +787,11 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             }
             FacetKind::Key(key) => {
                 // We synthesize a slice expression for the subscript here
-                // The range doesn't matter, since narrowing logic swallows type errors
-                let synthesized_slice = Expr::StringLiteral(ExprStringLiteral {
-                    node_index: AtomicNodeIndex::dummy(),
-                    range,
-                    value: StringLiteralValue::single(StringLiteral {
-                        node_index: AtomicNodeIndex::dummy(),
-                        range,
-                        value: key.clone().into_boxed_str(),
-                        flags: StringLiteralFlags::empty(),
-                    }),
-                });
+                // Use a synthesized fake range to avoid overwriting typing traces
+                let synthesized_slice = Ast::str_expr(key, TextRange::empty(TextSize::from(0)));
                 match remaining_facets.split_first() {
                     None => match base.type_at_facet(first_facet) {
-                        Some(ty) => ty.clone(),
+                        Some(ty) => self.force_for_narrowing(ty),
                         None => self.subscript_infer_for_type(
                             base.ty(),
                             &synthesized_slice,
@@ -686,19 +822,119 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         errors: &ErrorCollector,
     ) -> TypeInfo {
         match op {
-            NarrowOp::Atomic(None, op) => {
-                type_info
-                    .clone()
-                    .with_ty(self.atomic_narrow(type_info.ty(), op, range, errors))
+            NarrowOp::Atomic(subject, AtomicNarrowOp::HasAttr(attr)) => {
+                let base_ty = match subject {
+                    Some(facet_subject) => {
+                        self.get_facet_chain_type(type_info, &facet_subject.chain, range)
+                    }
+                    None => self.force_for_narrowing(type_info.ty()),
+                };
+                // We only narrow the attribute to `Any` if the attribute does not exist
+                if !self.has_attr(&base_ty, attr) {
+                    let attr_facet = FacetKind::Attribute(attr.clone());
+                    let facets = match subject {
+                        Some(facet_subject) => {
+                            let mut new_facets = facet_subject.chain.facets().clone();
+                            new_facets.push(attr_facet);
+                            new_facets
+                        }
+                        None => Vec1::new(attr_facet),
+                    };
+                    type_info.with_narrow(&facets, Type::any_implicit())
+                } else {
+                    type_info.clone()
+                }
             }
-            NarrowOp::Atomic(Some(facet_chain), op) => {
+            NarrowOp::Atomic(subject, AtomicNarrowOp::GetAttr(attr, default)) => {
+                let suppress_errors =
+                    ErrorCollector::new(errors.module().clone(), ErrorStyle::Never);
+                let default_ty = default
+                    .as_ref()
+                    .map_or(Type::None, |v| self.expr_infer(v, &suppress_errors));
+                // We can't narrow the type if the specified default is not falsy
+                if self.as_bool(&default_ty, range, &suppress_errors) != Some(false) {
+                    return type_info.clone();
+                }
+                let base_ty = match subject {
+                    Some(facet_subject) => {
+                        self.get_facet_chain_type(type_info, &facet_subject.chain, range)
+                    }
+                    None => self.force_for_narrowing(type_info.ty()),
+                };
+                let attr_ty =
+                    self.attr_infer_for_type(&base_ty, attr, range, &suppress_errors, None);
+                let attr_facet = FacetKind::Attribute(attr.clone());
+                let facets = match subject {
+                    Some(facet_subject) => {
+                        let mut new_facets = facet_subject.chain.facets().clone();
+                        new_facets.push(attr_facet);
+                        new_facets
+                    }
+                    None => Vec1::new(attr_facet),
+                };
+                // Given that the default is falsy:
+                // If the attribute does not exist we narrow to `Any`
+                // If the attribute exists we narrow it to be truthy
+                if attr_ty == Type::any_error() {
+                    type_info.with_narrow(&facets, Type::any_implicit())
+                } else {
+                    let narrowed_ty = self.atomic_narrow(
+                        &attr_ty,
+                        &AtomicNarrowOp::IsTruthy,
+                        range,
+                        &suppress_errors,
+                    );
+                    type_info.with_narrow(&facets, narrowed_ty)
+                }
+            }
+            NarrowOp::Atomic(None, op) => {
                 let ty = self.atomic_narrow(
-                    &self.get_facet_chain_type(type_info, facet_chain, range),
+                    &self.force_for_narrowing(type_info.ty()),
                     op,
                     range,
                     errors,
                 );
-                type_info.with_narrow(facet_chain.facets(), ty)
+                type_info.clone().with_ty(ty)
+            }
+            NarrowOp::Atomic(Some(facet_subject), op) => {
+                if facet_subject.origin == FacetOrigin::GetMethod
+                    && !self.supports_dict_get_subject(type_info, facet_subject, range)
+                {
+                    return type_info.clone();
+                }
+                let ty = self.atomic_narrow(
+                    &self.get_facet_chain_type(type_info, &facet_subject.chain, range),
+                    op,
+                    range,
+                    errors,
+                );
+                let mut narrowed = type_info.with_narrow(facet_subject.chain.facets(), ty);
+                // For certain types of narrows, we can also narrow the parent of the current subject
+                if let Some((last, prefix)) = facet_subject.chain.facets().split_last() {
+                    match Vec1::try_from(prefix) {
+                        Ok(prefix_facets) => {
+                            let prefix_chain = FacetChain::new(prefix_facets);
+                            let base_ty =
+                                self.get_facet_chain_type(type_info, &prefix_chain, range);
+                            if let Some(narrowed_ty) =
+                                self.atomic_narrow_for_facet(&base_ty, last, op, range, errors)
+                                && narrowed_ty != base_ty
+                            {
+                                narrowed = narrowed.with_narrow(prefix_chain.facets(), narrowed_ty);
+                            }
+                        }
+                        _ => {
+                            let base_ty = type_info.ty();
+                            if let Some(narrowed_ty) =
+                                self.atomic_narrow_for_facet(base_ty, last, op, range, errors)
+                                && narrowed_ty != *base_ty
+                            {
+                                narrowed = narrowed.clone().with_ty(narrowed_ty);
+                            }
+                        }
+                    };
+                }
+                narrowed
             }
             NarrowOp::And(ops) => {
                 let mut ops_iter = ops.iter();
@@ -715,7 +951,37 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             NarrowOp::Or(ops) => TypeInfo::join(
                 ops.map(|op| self.narrow(type_info, op, range, errors)),
                 &|tys| self.unions(tys),
+                &|got, want| self.is_subset_eq(got, want),
+                JoinStyle::SimpleMerge,
             ),
         }
     }
+
+    /// We only narrow `x.get("key")` if `x` resolves to a `dict`
+    fn supports_dict_get_subject(
+        &self,
+        type_info: &TypeInfo,
+        subject: &FacetSubject,
+        range: TextRange,
+    ) -> bool {
+        let base_ty = if subject.chain.facets().len() == 1 {
+            type_info.ty().clone()
+        } else {
+            let prefix: Vec<_> = subject
+                .chain
+                .facets()
+                .iter()
+                .take(subject.chain.facets().len() - 1)
+                .cloned()
+                .collect();
+            match Vec1::try_from_vec(prefix) {
+                Ok(vec1) => {
+                    let prefix_chain = FacetChain::new(vec1);
+                    self.get_facet_chain_type(type_info, &prefix_chain, range)
+                }
+                Err(_) => return false,
+            }
+        };
+        self.is_dict_like(&base_ty)
+    }
 }
