@@ -11,6 +11,7 @@
  * file contains the implementations of a few special calls that need to be hard-coded.
  */
 
+use pyrefly_types::callable::FuncMetadata;
 use pyrefly_util::visit::Visit;
 use pyrefly_util::visit::VisitMut;
 use ruff_python_ast::Expr;
@@ -21,7 +22,10 @@ use ruff_text_size::TextRange;
 
 use crate::alt::answers::LookupAnswer;
 use crate::alt::answers_solver::AnswersSolver;
+use crate::alt::callable::CallArg;
+use crate::alt::callable::CallKeyword;
 use crate::alt::solve::TypeFormContext;
+use crate::alt::unwrap::HintRef;
 use crate::config::error_kind::ErrorKind;
 use crate::error::collector::ErrorCollector;
 use crate::error::context::ErrorInfo;
@@ -41,7 +45,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         args: &[Expr],
         keywords: &[Keyword],
         range: TextRange,
-        hint: Option<&Type>,
+        hint: Option<HintRef>,
         errors: &ErrorCollector,
     ) -> Type {
         let ret = if args.len() == 2 {
@@ -49,24 +53,26 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             let expr_b = &args[1];
             let a = self.expr_infer_with_hint(expr_a, hint, errors);
             let b = self.expr_untype(expr_b, TypeFormContext::FunctionArgument, errors);
-            let mut a = self
-                .canonicalize_all_class_types(self.solver().deep_force(a), expr_a.range())
-                .explicit_any()
-                .noreturn_to_never()
-                .anon_callables();
-            let mut b = self
-                .canonicalize_all_class_types(self.solver().deep_force(b), expr_b.range())
-                .explicit_any()
-                .noreturn_to_never()
-                .anon_callables();
-            // Make assert_type(Self@SomeClass, typing.Self) work.
             let self_form = Type::SpecialForm(SpecialForm::SelfType);
-            a.subst_self_type_mut(&self_form, &|_, _| true);
-            b.subst_self_type_mut(&self_form, &|_, _| true);
-
-            // Re-sort unions. Make sure to keep this as the final step before comparison.
-            a = a.sort_unions();
-            b = b.sort_unions();
+            let normalize_type = |ty: Type, expr: &Expr| {
+                let mut ty = self
+                    .canonicalize_all_class_types(
+                        self.solver().deep_force(ty),
+                        expr.range(),
+                        errors,
+                    )
+                    .promote_typevar_values(self.stdlib)
+                    .explicit_any()
+                    .noreturn_to_never()
+                    .anon_callables()
+                    .distribute_type_over_union();
+                // Make assert_type(Self@SomeClass, typing.Self) work.
+                ty.subst_self_type_mut(&self_form);
+                // Re-sort unions. Make sure to keep this as the final step before comparison.
+                ty.sort_unions()
+            };
+            let a = normalize_type(a, expr_a);
+            let b = normalize_type(b, expr_b);
             if a != b {
                 self.error(
                     errors,
@@ -114,7 +120,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         args: &[Expr],
         keywords: &[Keyword],
         range: TextRange,
-        hint: Option<&Type>,
+        hint: Option<HintRef>,
         errors: &ErrorCollector,
     ) -> Type {
         let ret = if args.len() == 1 {
@@ -223,7 +229,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
             );
         }
         let ret = if let Some(t) = typ {
-            match self.untype_opt(self.expr_infer(t, errors), range) {
+            match self.untype_opt(self.expr_infer(t, errors), range, errors) {
                 Some(t) => t,
                 None => self.error(
                     errors,
@@ -248,6 +254,21 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 "`typing.cast` missing required argument `val`".to_owned(),
             );
         }
+        if let Some(val_expr) = val {
+            let val_type = self.expr_infer(val_expr, errors);
+            if !val_type.is_any() && val_type == ret {
+                self.error(
+                    errors,
+                    range,
+                    ErrorInfo::Kind(ErrorKind::RedundantCast),
+                    format!(
+                        "Redundant cast: `{}` is the same type as `{}`",
+                        val_type.deterministic_printing(),
+                        ret.clone().deterministic_printing()
+                    ),
+                );
+            }
+        }
         ret
     }
 
@@ -272,14 +293,14 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     ) -> Type {
         // Verify that the `cls` argument has type `type`.
         self.check_type(
-            &self.stdlib.builtins_type().clone().to_type(),
             &self.expr_infer(cls, errors),
+            &self.stdlib.builtins_type().clone().to_type(),
             cls.range(),
             errors,
             &|| {
                 TypeCheckContext::of_kind(TypeCheckKind::CallArgument(
                     Some(Name::new_static("cls")),
-                    Some(FunctionKind::IsSubclass.as_func_id()),
+                    Some(FunctionKind::IsSubclass),
                 ))
             },
         );
@@ -291,16 +312,22 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         &self,
         ty: Type,
         contains_subscript: bool,
-        contains_any: bool,
         range: TextRange,
         func_kind: &FunctionKind,
         errors: &ErrorCollector,
     ) {
         for ty in self.as_class_info(ty) {
             if let Type::ClassDef(cls) = &ty {
+                if cls.has_toplevel_qname("typing", "Any") {
+                    self.error(
+                        errors,
+                        range,
+                        ErrorInfo::Kind(ErrorKind::InvalidArgument),
+                        "Expected class object, got `Any`".to_owned(),
+                    );
+                }
                 let metadata = self.get_metadata_for_class(cls);
-                let func_display =
-                    || format!("{}()", func_kind.as_func_id().format(self.module().name()));
+                let func_display = || format!("{}()", func_kind.format(self.module().name()));
                 if metadata.is_new_type() {
                     self.error(
                         errors,
@@ -357,16 +384,6 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                         self.for_display(ty)
                     ),
                 );
-            } else if contains_any && matches!(&ty, Type::Type(box Type::Any(AnyStyle::Explicit))) {
-                // If the raw expression contains something that structurally looks like `A[T]` and
-                // part of the expression resolves to a parameterized class type, then we likely have a
-                // literal parameterized type, which is a runtime exception.
-                self.error(
-                    errors,
-                    range,
-                    ErrorInfo::Kind(ErrorKind::InvalidArgument),
-                    "Expected class object, got `Any`".to_owned(),
-                );
             } else if self.unwrap_class_object_silently(&ty).is_none() {
                 self.error(
                     errors,
@@ -376,14 +393,14 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 );
             } else {
                 self.check_type(
-                    &self.stdlib.builtins_type().clone().to_type(),
                     &ty,
+                    &self.stdlib.builtins_type().clone().to_type(),
                     range,
                     errors,
                     &|| {
                         TypeCheckContext::of_kind(TypeCheckKind::CallArgument(
                             Some(Name::new_static("class_or_tuple")),
-                            Some(func_kind.as_func_id()),
+                            Some(func_kind.clone()),
                         ))
                     },
                 );
@@ -426,20 +443,15 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     ) {
         let arg_class_type = self.expr_infer(arg_expr, errors);
         let mut contains_subscript = false;
-        let mut contains_any = false;
         arg_expr.visit(&mut |e| {
             if matches!(e, Expr::Subscript(_)) {
                 contains_subscript = true;
             }
-            if matches!(e, Expr::Name(x) if x.id.as_str() == "Any") {
-                contains_any = true;
-            }
         });
 
         self.check_type_is_class_object(
             arg_class_type,
             contains_subscript,
-            contains_any,
             arg_expr.range(),
             func_kind,
             errors,
@@ -450,7 +462,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
     pub fn as_class_info(&self, ty: Type) -> Vec<Type> {
         fn f<'a, Ans: LookupAnswer>(me: &AnswersSolver<'a, Ans>, t: Type, res: &mut Vec<Type>) {
             match t {
-                Type::Var(v) if let Some(_guard) = me.recurser.recurse(v) => {
+                Type::Var(v) if let Some(_guard) = me.recurse(v) => {
                     f(me, me.solver().force_var(v), res)
                 }
                 Type::ClassType(ref c)
@@ -491,4 +503,31 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
         f(self, ty, &mut res);
         res
     }
+
+    pub fn maybe_apply_function_decorator(
+        &self,
+        callee: &Type,
+        args: &[CallArg],
+        kws: &[CallKeyword],
+        errors: &ErrorCollector,
+    ) -> Option<Type> {
+        let special_decorator = self.get_special_decorator(callee)?;
+        // Does this call have a single positional argument?
+        // If not, it cannot be a decorator application.
+        if kws.is_empty()
+            && let [CallArg::Arg(arg)] = args
+        {
+            let mut arg_ty = arg.infer(self, errors);
+            // Try to apply the decorator to arg_ty. Does nothing if the decorator does not have known
+            // typing effects or if arg_ty is not a function.
+            let mut applied = false;
+            arg_ty.transform_toplevel_func_metadata(|meta: &mut FuncMetadata| {
+                applied |=
+                    self.set_flag_from_special_decorator(&mut meta.flags, &special_decorator);
+            });
+            if applied { Some(arg_ty) } else { None }
+        } else {
+            None
+        }
+    }
 }
