@@ -231,7 +231,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     {
                         literals.push(e.to_type())
                     }
-                    Type::Any(AnyStyle::Error) => literals.push(Type::any_error()),
+                    ty @ Type::Any(AnyStyle::Error) => literals.push(ty),
                     _ => {
                         self.error(
                             errors,
@@ -251,7 +251,7 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                 let ty = self.expr_infer(x, errors);
                 self.map_over_union(&ty, |ty| match ty {
                     Type::Type(box lit @ Type::Literal(_)) => literals.push(lit.clone()),
-                    Type::Any(AnyStyle::Error) => literals.push(Type::any_error()),
+                    ty @ Type::Any(AnyStyle::Error) => literals.push(ty.clone()),
                     _ => {
                         self.error(
                             errors,
@@ -485,7 +485,11 @@ impl<'a, Ans: LookupAnswer> AnswersSolver<'a, Ans> {
                     arguments.len()
                 ),
             ),
-            SpecialForm::Annotated if arguments.len() > 1 => self.expr_infer(&arguments[0], errors),
+            SpecialForm::Annotated if arguments.len() > 1 => Type::type_form(self.expr_untype(
+                &arguments[0],
+                TypeFormContext::TypeArgument,
+                errors,
+            )),
             // Keep this in sync with `SpecialForm::can_be_subscripted``
             SpecialForm::SelfType
             | SpecialForm::LiteralString
