@@ -14,6 +14,7 @@ use std::sync::Arc;
 use dupe::Dupe;
 use pyrefly_derive::TypeEq;
 use pyrefly_derive::VisitMut;
+use pyrefly_types::typed_dict::ExtraItems;
 use pyrefly_util::display::commas_iter;
 use pyrefly_util::visit::VisitMut;
 use ruff_python_ast::name::Name;
@@ -24,6 +25,7 @@ use vec1::Vec1;
 use vec1::vec1;
 
 use crate::alt::class::class_field::ClassField;
+use crate::alt::types::pydantic::PydanticModelKind;
 use crate::config::error_kind::ErrorKind;
 use crate::error::collector::ErrorCollector;
 use crate::error::context::ErrorInfo;
@@ -33,7 +35,6 @@ use crate::types::display::ClassDisplayContext;
 use crate::types::keywords::DataclassKeywords;
 use crate::types::keywords::DataclassTransformKeywords;
 use crate::types::stdlib::Stdlib;
-use crate::types::tuple::Tuple;
 use crate::types::types::CalleeKind;
 use crate::types::types::Type;
 
@@ -46,18 +47,19 @@ pub struct ClassMetadata {
     enum_metadata: Option<EnumMetadata>,
     protocol_metadata: Option<ProtocolMetadata>,
     dataclass_metadata: Option<DataclassMetadata>,
-    tuple_base: Option<Tuple>,
-    bases_with_metadata: Vec<(ClassType, Arc<ClassMetadata>)>,
+    extends_abc: bool,
+    bases: Vec<Class>,
+    has_generic_base_class: bool,
     has_base_any: bool,
     is_new_type: bool,
     is_final: bool,
-    /// Is it possible for this class to have type parameters that we don't know about?
-    /// This can happen if, e.g., a class inherits from Any.
-    has_unknown_tparams: bool,
+    is_deprecated: bool,
     total_ordering_metadata: Option<TotalOrderingMetadata>,
     /// If this class is decorated with `typing.dataclass_transform(...)`, the keyword arguments
     /// that were passed to the `dataclass_transform` call.
     dataclass_transform_metadata: Option<DataclassTransformKeywords>,
+    pydantic_model_kind: Option<PydanticModelKind>,
+    django_model_metadata: Option<DjangoModelMetadata>,
 }
 
 impl VisitMut<Type> for ClassMetadata {
@@ -75,38 +77,44 @@ impl Display for ClassMetadata {
 
 impl ClassMetadata {
     pub fn new(
-        bases_with_metadata: Vec<(ClassType, Arc<ClassMetadata>)>,
-        metaclass: Option<ClassType>,
+        bases: Vec<Class>,
+        metaclass: Metaclass,
         keywords: Vec<(Name, Type)>,
         typed_dict_metadata: Option<TypedDictMetadata>,
         named_tuple_metadata: Option<NamedTupleMetadata>,
         enum_metadata: Option<EnumMetadata>,
         protocol_metadata: Option<ProtocolMetadata>,
         dataclass_metadata: Option<DataclassMetadata>,
-        tuple_base: Option<Tuple>,
+        extends_abc: bool,
+        has_generic_base_class: bool,
         has_base_any: bool,
         is_new_type: bool,
         is_final: bool,
-        has_unknown_tparams: bool,
+        is_deprecated: bool,
         total_ordering_metadata: Option<TotalOrderingMetadata>,
         dataclass_transform_metadata: Option<DataclassTransformKeywords>,
+        pydantic_model_kind: Option<PydanticModelKind>,
+        django_model_metadata: Option<DjangoModelMetadata>,
     ) -> ClassMetadata {
         ClassMetadata {
-            metaclass: Metaclass(metaclass),
+            metaclass,
             keywords: Keywords(keywords),
             typed_dict_metadata,
             named_tuple_metadata,
             enum_metadata,
             protocol_metadata,
             dataclass_metadata,
-            tuple_base,
-            bases_with_metadata,
+            extends_abc,
+            bases,
+            has_generic_base_class,
             has_base_any,
             is_new_type,
             is_final,
-            has_unknown_tparams,
+            is_deprecated,
             total_ordering_metadata,
             dataclass_transform_metadata,
+            pydantic_model_kind,
+            django_model_metadata,
         }
     }
 
@@ -119,19 +127,33 @@ impl ClassMetadata {
             enum_metadata: None,
             protocol_metadata: None,
             dataclass_metadata: None,
-            tuple_base: None,
-            bases_with_metadata: Vec::new(),
+            extends_abc: false,
+            bases: Vec::new(),
+            has_generic_base_class: false,
             has_base_any: false,
             is_new_type: false,
             is_final: false,
-            has_unknown_tparams: false,
+            is_deprecated: false,
             total_ordering_metadata: None,
             dataclass_transform_metadata: None,
+            pydantic_model_kind: None,
+            django_model_metadata: None,
         }
     }
 
-    pub fn metaclass(&self) -> Option<&ClassType> {
-        self.metaclass.0.as_ref()
+    /// The class's custom (non-`type`) metaclass, if it has one.
+    pub fn custom_metaclass(&self) -> Option<&ClassType> {
+        self.metaclass.get()
+    }
+
+    pub fn custom_metaclass_raw(&self) -> &Metaclass {
+        &self.metaclass
+    }
+
+    /// The class's metaclass.
+    pub fn metaclass<'a>(&'a self, stdlib: &'a Stdlib) -> &'a ClassType {
+        self.custom_metaclass()
+            .unwrap_or_else(|| stdlib.builtins_type())
     }
 
     #[allow(dead_code)] // This is used in tests now, and will be needed later in production.
@@ -143,16 +165,53 @@ impl ClassMetadata {
         self.typed_dict_metadata.is_some()
     }
 
+    pub fn is_pydantic_base_model(&self) -> bool {
+        self.pydantic_model_kind.is_some()
+    }
+
+    pub fn is_django_model(&self) -> bool {
+        self.django_model_metadata.is_some()
+    }
+
+    pub fn pydantic_model_kind(&self) -> Option<PydanticModelKind> {
+        self.pydantic_model_kind.clone()
+    }
+
     pub fn is_final(&self) -> bool {
         self.is_final
     }
 
-    pub fn has_base_any(&self) -> bool {
-        self.has_base_any
+    pub fn extends_abc(&self) -> bool {
+        self.extends_abc
     }
 
-    pub fn has_unknown_tparams(&self) -> bool {
-        self.has_unknown_tparams
+    pub fn is_explicitly_abstract(&self) -> bool {
+        for base in self.base_class_objects() {
+            if base.has_toplevel_qname("abc", "ABC") {
+                return true;
+            }
+        }
+        // Only check the metaclass if it's directly specified on this class
+        if let Metaclass::Direct(metaclass) = self.custom_metaclass_raw()
+            && metaclass
+                .class_object()
+                .has_toplevel_qname("abc", "ABCMeta")
+        {
+            return true;
+        }
+        false
+    }
+
+    pub fn is_deprecated(&self) -> bool {
+        self.is_deprecated
+    }
+
+    pub fn has_generic_base_class(&self) -> bool {
+        self.has_generic_base_class
+    }
+
+    pub fn has_base_any(&self) -> bool {
+        self.has_base_any
     }
 
     pub fn typed_dict_metadata(&self) -> Option<&TypedDictMetadata> {
@@ -163,16 +222,12 @@ impl ClassMetadata {
         self.named_tuple_metadata.as_ref()
     }
 
-    pub fn tuple_base(&self) -> Option<&Tuple> {
-        self.tuple_base.as_ref()
-    }
-
     pub fn enum_metadata(&self) -> Option<&EnumMetadata> {
         self.enum_metadata.as_ref()
     }
 
-    pub fn bases_with_metadata(&self) -> &[(ClassType, Arc<ClassMetadata>)] {
-        &self.bases_with_metadata
+    pub fn base_class_objects(&self) -> &[Class] {
+        &self.bases
     }
 
     pub fn is_protocol(&self) -> bool {
@@ -212,8 +267,16 @@ impl ClassMetadata {
     pub fn dataclass_transform_metadata(&self) -> Option<&DataclassTransformKeywords> {
         self.dataclass_transform_metadata.as_ref()
     }
+
+    pub fn django_model_metadata(&self) -> Option<&DjangoModelMetadata> {
+        self.django_model_metadata.as_ref()
+    }
 }
 
+/// A field that we synthesize and add to a class. Note that if a non-synthesized field already
+/// exists on the class, it will take precedence over the synthesized field in attribute lookup.
+/// If you want to modify the type of a non-synthesized field, see
+/// AnswersSolver::get_special_class_field_type() in class_field.rs.
 #[derive(Clone, Debug, TypeEq, PartialEq, Eq)]
 pub struct ClassSynthesizedField {
     pub inner: Arc<ClassField>,
@@ -271,6 +334,10 @@ impl ClassSynthesizedFields {
         }
         self
     }
+
+    pub fn fields(&self) -> impl ExactSizeIterator<Item = (&Name, &ClassSynthesizedField)> {
+        self.0.iter()
+    }
 }
 
 impl Display for ClassSynthesizedFields {
@@ -288,14 +355,36 @@ impl Display for ClassSynthesizedFields {
 
 /// A struct representing a class's metaclass. A value of `None` indicates
 /// no explicit metaclass, in which case the default metaclass is `type`.
-#[derive(Clone, Debug, TypeEq, PartialEq, Eq, Default)]
-struct Metaclass(Option<ClassType>);
+#[derive(Clone, Debug, TypeEq, PartialEq, Eq)]
+pub enum Metaclass {
+    Direct(ClassType),
+    Inherited(ClassType),
+    None,
+}
 
 impl Display for Metaclass {
     fn fmt(&self, f: &mut Formatter) -> fmt::Result {
-        match &self.0 {
-            Some(metaclass) => write!(f, "{metaclass}"),
-            None => write!(f, "type"),
+        match &self {
+            Self::Direct(metaclass) => write!(f, "{metaclass}"),
+            Self::Inherited(metaclass) => write!(f, "inherited({metaclass})"),
+            Self::None => write!(f, "type"),
+        }
+    }
+}
+
+impl Default for Metaclass {
+    fn default() -> Self {
+        Self::None
+    }
+}
+
+impl Metaclass {
+    /// Convenience function to get the metaclass as a ClassType, regardless of its origin
+    pub fn get(&self) -> Option<&ClassType> {
+        match self {
+            Self::Direct(metaclass) => Some(metaclass),
+            Self::Inherited(metaclass) => Some(metaclass),
+            Self::None => None,
         }
     }
 }
@@ -322,6 +411,7 @@ impl Display for Keywords {
 pub struct TypedDictMetadata {
     /// Field name to the value of the `total` keyword in the defining class.
     pub fields: SmallMap<Name, bool>,
+    pub extra_items: ExtraItems,
 }
 
 #[derive(Clone, Debug, TypeEq, PartialEq, Eq)]
@@ -331,6 +421,8 @@ pub struct EnumMetadata {
     pub is_flag: bool,
     /// Is there any `_value_` field present.
     pub has_value: bool,
+    /// Whether this is a special Django enum.
+    pub is_django: bool,
 }
 
 #[derive(Clone, Debug, TypeEq, PartialEq, Eq)]
@@ -338,12 +430,40 @@ pub struct NamedTupleMetadata {
     pub elements: SmallSet<Name>,
 }
 
+/// Defaults for `init_by_name` and `init_by_default`, per-field flags that control the name of
+/// a field's corresponding `__init__` parameter. See DataclassFieldKeywords for more information.
+#[derive(Clone, Debug, TypeEq, PartialEq, Eq)]
+pub struct InitDefaults {
+    pub init_by_name: bool,
+    pub init_by_alias: bool,
+}
+
+impl Default for InitDefaults {
+    fn default() -> Self {
+        Self {
+            init_by_name: false,
+            init_by_alias: true,
+        }
+    }
+}
+
 #[derive(Clone, Debug, TypeEq, PartialEq, Eq)]
 pub struct DataclassMetadata {
     /// The dataclass fields, e.g., `{'x'}` for `@dataclass class C: x: int`.
     pub fields: SmallSet<Name>,
     pub kws: DataclassKeywords,
     pub field_specifiers: Vec<CalleeKind>,
+    pub alias_keyword: Name,
+    pub init_defaults: InitDefaults,
+    /// Whether a default can be passed positionally to field specifier calls
+    pub default_can_be_positional: bool,
+}
+
+#[derive(Clone, Debug, TypeEq, PartialEq, Eq)]
+pub struct DjangoModelMetadata {
+    /// The name of the field that has primary_key=True, if any.
+    /// If None, the model uses the default auto-generated `id` field.
+    pub custom_primary_key_field: Option<Name>,
 }
 
 #[derive(Clone, Debug, TypeEq, PartialEq, Eq)]
@@ -376,7 +496,7 @@ pub struct TotalOrderingMetadata {
 /// If a class is present in multiple places of the inheritance tree (and is
 /// linearizable using C3 linearization), it is possible it appears with
 /// different type arguments. The type arguments computed here will always be
-/// those coming from the instance that was selected during lineariation.
+/// those coming from the instance that was selected during linearization.
 #[derive(Clone, Debug, VisitMut, TypeEq, PartialEq, Eq)]
 pub enum ClassMro {
     Resolved(Vec<ClassType>),
@@ -497,7 +617,7 @@ impl Linearization {
                 ClassMro::Resolved(ancestors) => {
                     let ancestors_through_base = ancestors
                         .iter()
-                        .map(|ancestor| ancestor.substitute(&base.substitution()))
+                        .map(|ancestor| ancestor.substitute_with(&base.substitution()))
                         .rev()
                         .collect::<Vec<_>>();
                     ancestor_chains.push(AncestorChain::from_base_and_ancestors(
