@@ -5,7 +5,10 @@
  * LICENSE file in the root directory of this source tree.
  */
 
+pub mod abstract_class;
+pub mod class_bases;
 pub mod class_metadata;
 pub mod decorated_function;
 pub mod legacy_lookup;
+pub mod pydantic;
 pub mod yields;
