@@ -8,9 +8,10 @@
 use std::mem;
 use std::sync::LazyLock;
 
+use dupe::Dupe as _;
 use pyrefly_python::ast::Ast;
 use pyrefly_python::docstring::Docstring;
-use pyrefly_python::module_name::ModuleName;
+use pyrefly_python::nesting_context::NestingContext;
 use pyrefly_python::short_identifier::ShortIdentifier;
 use pyrefly_util::prelude::SliceExt;
 use regex::Regex;
@@ -26,43 +27,48 @@ use ruff_python_ast::name::Name;
 use ruff_text_size::Ranged;
 use ruff_text_size::TextRange;
 use starlark_map::small_map::SmallMap;
-use starlark_map::small_set::SmallSet;
 
-use crate::alt::class::base_class::BaseClass;
+use crate::binding::base_class::BaseClass;
+use crate::binding::base_class::BaseClassGeneric;
 use crate::binding::binding::AnnotationTarget;
 use crate::binding::binding::Binding;
+use crate::binding::binding::BindingAbstractClassCheck;
 use crate::binding::binding::BindingAnnotation;
 use crate::binding::binding::BindingClass;
+use crate::binding::binding::BindingClassBaseType;
 use crate::binding::binding::BindingClassField;
 use crate::binding::binding::BindingClassMetadata;
 use crate::binding::binding::BindingClassMro;
 use crate::binding::binding::BindingClassSynthesizedFields;
+use crate::binding::binding::BindingConsistentOverrideCheck;
+use crate::binding::binding::BindingExpect;
 use crate::binding::binding::BindingTParams;
 use crate::binding::binding::BindingVariance;
 use crate::binding::binding::ClassBinding;
 use crate::binding::binding::ClassFieldDefinition;
 use crate::binding::binding::ExprOrBinding;
 use crate::binding::binding::Key;
+use crate::binding::binding::KeyAbstractClassCheck;
 use crate::binding::binding::KeyAnnotation;
 use crate::binding::binding::KeyClass;
+use crate::binding::binding::KeyClassBaseType;
 use crate::binding::binding::KeyClassField;
 use crate::binding::binding::KeyClassMetadata;
 use crate::binding::binding::KeyClassMro;
 use crate::binding::binding::KeyClassSynthesizedFields;
+use crate::binding::binding::KeyConsistentOverrideCheck;
+use crate::binding::binding::KeyExpect;
 use crate::binding::binding::KeyTParams;
 use crate::binding::binding::KeyVariance;
 use crate::binding::bindings::BindingsBuilder;
 use crate::binding::bindings::CurrentIdx;
-use crate::binding::bindings::LegacyTParamBuilder;
-use crate::binding::scope::ClassFieldInBody;
+use crate::binding::bindings::LegacyTParamCollector;
+use crate::binding::pydantic::PydanticConfigDict;
 use crate::binding::scope::ClassIndices;
 use crate::binding::scope::FlowStyle;
-use crate::binding::scope::InstanceAttribute;
 use crate::binding::scope::Scope;
-use crate::binding::scope::ScopeKind;
 use crate::config::error_kind::ErrorKind;
 use crate::error::context::ErrorInfo;
-use crate::graph::index::Idx;
 use crate::types::class::ClassDefIndex;
 use crate::types::class::ClassFieldProperties;
 use crate::types::types::Type;
@@ -93,10 +99,14 @@ impl<'a> BindingsBuilder<'a> {
         let class_indices = ClassIndices {
             def_index,
             class_idx: self.idx_for_promise(KeyClass(ShortIdentifier::new(class_name))),
+            base_type_idx: self.idx_for_promise(KeyClassBaseType(def_index)),
             metadata_idx: self.idx_for_promise(KeyClassMetadata(def_index)),
             mro_idx: self.idx_for_promise(KeyClassMro(def_index)),
             synthesized_fields_idx: self.idx_for_promise(KeyClassSynthesizedFields(def_index)),
             variance_idx: self.idx_for_promise(KeyVariance(def_index)),
+            consistent_override_check_idx: self
+                .idx_for_promise(KeyConsistentOverrideCheck(def_index)),
+            abstract_class_check_idx: self.idx_for_promise(KeyAbstractClassCheck(def_index)),
         };
         // The user - used for first-usage tracking of any expressions we analyze in a class definition -
         // is the `Idx<Key>` of the class object bound to the class name.
@@ -105,23 +115,12 @@ impl<'a> BindingsBuilder<'a> {
         (class_object, class_indices)
     }
 
-    pub fn class_def(&mut self, mut x: StmtClassDef) {
-        if self.module_info.name() == ModuleName::typing() && x.name.as_str() == "Any" {
-            // We special case the definition of `Any`, because it isn't a `SpecialForm`,
-            // but an ordinary `class`.
-            self.bind_definition(
-                &x.name,
-                Binding::Type(Type::type_form(Type::any_explicit())),
-                FlowStyle::Other,
-            );
-            return;
-        }
-
+    pub fn class_def(&mut self, mut x: StmtClassDef, parent: &NestingContext) {
         let (mut class_object, class_indices) = self.class_object_and_indices(&x.name);
-        let mut key_class_fields: SmallSet<Idx<KeyClassField>> = SmallSet::new();
-
+        let mut pydantic_config_dict = PydanticConfigDict::default();
         let docstring_range = Docstring::range_from_stmts(x.body.as_slice());
         let body = mem::take(&mut x.body);
+        let field_docstrings = self.extract_field_docstrings(&body);
         let decorators_with_ranges = self.ensure_and_bind_decorators_with_ranges(
             mem::take(&mut x.decorator_list),
             class_object.usage(),
@@ -133,7 +132,7 @@ impl<'a> BindingsBuilder<'a> {
             self.type_params(x);
         });
 
-        let mut legacy = Some(LegacyTParamBuilder::new(x.type_params.is_some()));
+        let mut legacy = Some(LegacyTParamCollector::new(x.type_params.is_some()));
         let bases = x.bases().map(|base| {
             let mut base = base.clone();
             // Forward refs are fine *inside* of a base expression in the type arguments,
@@ -160,7 +159,25 @@ impl<'a> BindingsBuilder<'a> {
                 &mut legacy
             };
             self.ensure_type(&mut base, legacy);
-            base
+
+            let base_class = self.base_class_of(base.clone());
+            // NOTE(grievejia): If any of the class base is a specialized generic class (e.g. `Foo[Bar]`), and if the tparam of the
+            // generic class has a bound or constraint, we won't be validating the type `Bar` against that bound/constraint of the
+            // tparam eagerly in order to avoid dependency cycle. But the validation needs to happen somewhere.
+            //
+            // Since we can't create "delayed" computations on-the-fly in the answer phase, we'll create a static computation here
+            // trying to type check all valid base class expressions. We are duplicating a lot of work with class base calculation,
+            // which is sad. Hence we were able to figure out a better places to insert those checks we should migrate.
+            //
+            // Also note that there's no risk of first-usage tracking issues here because `ensure_type` does not participate in first
+            // usage tracking.
+            if matches!(base_class, BaseClass::BaseClassExpr(..)) {
+                self.insert_binding(
+                    KeyExpect(base.range()),
+                    BindingExpect::TypeCheckBaseClassExpr(base),
+                );
+            }
+            base_class
         });
 
         let mut keywords = Vec::new();
@@ -183,14 +200,11 @@ impl<'a> BindingsBuilder<'a> {
         }
 
         self.insert_binding_idx(
-            class_indices.metadata_idx,
-            BindingClassMetadata {
+            class_indices.base_type_idx,
+            BindingClassBaseType {
                 class_idx: class_indices.class_idx,
-                bases: bases.clone().into_boxed_slice(),
-                keywords: keywords.into_boxed_slice(),
-                decorators: decorators_with_ranges.clone().into_boxed_slice(),
                 is_new_type: false,
-                special_base: None,
+                bases: bases.clone().into_boxed_slice(),
             },
         );
         self.insert_binding_idx(
@@ -204,8 +218,8 @@ impl<'a> BindingsBuilder<'a> {
             BindingClassSynthesizedFields(class_indices.class_idx),
         );
 
-        let legacy_tparam_builder = legacy.unwrap();
-        legacy_tparam_builder.add_name_definitions(self);
+        let legacy_tparam_collector = legacy.unwrap();
+        self.add_name_definitions(&legacy_tparam_collector);
 
         self.scopes.push(Scope::class_body(
             x.range,
@@ -213,106 +227,62 @@ impl<'a> BindingsBuilder<'a> {
             x.name.clone(),
         ));
         self.init_static_scope(&body, false);
-        self.stmts(body);
+        self.stmts(
+            body,
+            &NestingContext::class(ShortIdentifier::new(&x.name), parent.dupe()),
+        );
+        let field_definitions = self.scopes.finish_class_and_get_field_definitions();
 
-        let last_scope = self.scopes.pop();
-        self.scopes.pop(); // annotation scope
-        let mut fields = SmallMap::with_capacity(last_scope.stat.0.len());
-        for (name, info) in last_scope.flow.info.iter_hashed() {
-            // Ignore a name not in the current flow's static. This can happen because operations
-            // like narrows can change the local flow info for a name defined in some parent scope.
-            if let Some(stat_info) = last_scope.stat.0.get_hashed(name) {
-                let (definition, is_initialized_on_class) = {
-                    if let FlowStyle::FunctionDef(_, has_return_annotation) = info.style {
-                        (
-                            ClassFieldDefinition::MethodLike {
-                                definition: info.key,
-                                has_return_annotation,
-                            },
-                            true,
-                        )
-                    } else {
-                        match info.as_initial_value() {
-                            ClassFieldInBody::InitializedByAssign(e) => (
-                                ClassFieldDefinition::AssignedInBody {
-                                    value: ExprOrBinding::Expr(e.clone()),
-                                    annotation: stat_info.annot,
-                                },
-                                true,
-                            ),
-                            ClassFieldInBody::InitializedWithoutAssign => (
-                                ClassFieldDefinition::DefinedWithoutAssign {
-                                    definition: info.key,
-                                },
-                                true,
-                            ),
-                            ClassFieldInBody::Uninitialized => {
-                                let annotation = stat_info.annot.unwrap_or_else(
-                                    || panic!("A class field known in the body but uninitialized always has an annotation.")
-                                );
-                                (
-                                    ClassFieldDefinition::DeclaredByAnnotation { annotation },
-                                    false,
-                                )
-                            }
-                        }
-                    }
-                };
-                let binding = BindingClassField {
-                    class_idx: class_indices.class_idx,
-                    name: name.into_key().clone(),
-                    range: stat_info.loc,
-                    definition,
-                };
-                fields.insert_hashed(
-                    name.cloned(),
-                    ClassFieldProperties::new(
-                        stat_info.annot.is_some(),
-                        is_initialized_on_class,
-                        stat_info.loc,
-                    ),
-                );
-                let key_field = KeyClassField(class_indices.def_index, name.into_key().clone());
-                key_class_fields.insert(self.idx_for_promise(key_field.clone()));
-                self.insert_binding(key_field, binding);
-            }
-        }
-        if let ScopeKind::Class(class_scope) = last_scope.kind {
-            for (name, method, InstanceAttribute(value, annotation, range)) in
-                class_scope.method_defined_attributes()
+        let mut django_primary_key_field: Option<Name> = None;
+        let mut fields = SmallMap::with_capacity(field_definitions.len());
+        for (name, (definition, range)) in field_definitions.into_iter_hashed() {
+            if let ClassFieldDefinition::AssignedInBody {
+                value: ExprOrBinding::Expr(e),
+                ..
+            } = &definition
             {
-                if !fields.contains_key_hashed(name.as_ref()) {
-                    fields.insert_hashed(
-                        name.clone(),
-                        ClassFieldProperties::new(annotation.is_some(), false, range),
-                    );
+                self.extract_pydantic_config_dict(e, &name, &mut pydantic_config_dict);
 
-                    let key_field = KeyClassField(class_indices.def_index, name.key().clone());
-                    key_class_fields.insert(self.idx_for_promise(key_field.clone()));
-
-                    self.insert_binding(
-                        key_field,
-                        BindingClassField {
-                            class_idx: class_indices.class_idx,
-                            name: name.into_key(),
-                            range,
-                            definition: ClassFieldDefinition::DefinedInMethod {
-                                value,
-                                annotation,
-                                method,
-                            },
-                        },
-                    );
+                if self.extract_django_primary_key(e) {
+                    django_primary_key_field = Some(name.clone().into_key());
                 }
             }
-        } else {
-            unreachable!("Expected class body scope, got {:?}", last_scope.kind);
+            let (is_initialized_on_class, is_annotated) = match &definition {
+                ClassFieldDefinition::DefinedInMethod { annotation, .. } => {
+                    (false, annotation.is_some())
+                }
+                ClassFieldDefinition::DeclaredByAnnotation { .. } => (false, true),
+                ClassFieldDefinition::AssignedInBody { annotation, .. } => {
+                    (true, annotation.is_some())
+                }
+                _ => (true, false),
+            };
+
+            let docstring_range = field_docstrings.get(&range).copied();
+
+            fields.insert_hashed(
+                name.clone(),
+                ClassFieldProperties::new(
+                    is_annotated,
+                    is_initialized_on_class,
+                    range,
+                    docstring_range,
+                ),
+            );
+            let key_field = KeyClassField(class_indices.def_index, name.clone().into_key());
+            let binding = BindingClassField {
+                class_idx: class_indices.class_idx,
+                name: name.into_key(),
+                range,
+                definition,
+            };
+            self.insert_binding(key_field, binding);
         }
 
         let decorator_keys = decorators_with_ranges
             .map(|(idx, _)| *idx)
             .into_boxed_slice();
-        self.bind_definition_current(
+        self.bind_current_as(
             &x.name,
             class_object,
             Binding::ClassDef(class_indices.class_idx, decorator_keys),
@@ -321,7 +291,7 @@ impl<'a> BindingsBuilder<'a> {
 
         // Insert a `KeyTParams` / `BindingTParams` pair, but only if there is at least
         // one generic base class - otherwise, it is not possible that legacy tparams are used.
-        let legacy_tparams = legacy_tparam_builder.lookup_keys();
+        let legacy_tparams = legacy_tparam_collector.lookup_keys();
         let tparams_require_binding = !legacy_tparams.is_empty();
         if tparams_require_binding {
             let scoped_type_params = mem::take(&mut x.type_params);
@@ -330,7 +300,14 @@ impl<'a> BindingsBuilder<'a> {
                 BindingTParams {
                     name: x.name.clone(),
                     scoped_type_params,
-                    bases: bases.clone().into_boxed_slice(),
+                    generic_bases: bases
+                        .iter()
+                        .filter_map(|base| match base {
+                            BaseClass::Generic(x) => Some(x),
+                            _ => None,
+                        })
+                        .cloned()
+                        .collect::<Box<[BaseClassGeneric]>>(),
                     legacy_tparams: legacy_tparams.into_boxed_slice(),
                 },
             );
@@ -342,6 +319,7 @@ impl<'a> BindingsBuilder<'a> {
             BindingClass::ClassDef(ClassBinding {
                 def_index: class_indices.def_index,
                 def: x,
+                parent: parent.dupe(),
                 fields,
                 tparams_require_binding,
                 docstring_range,
@@ -354,6 +332,68 @@ impl<'a> BindingsBuilder<'a> {
                 class_key: class_indices.class_idx,
             },
         );
+        self.insert_binding_idx(
+            class_indices.consistent_override_check_idx,
+            BindingConsistentOverrideCheck {
+                class_key: class_indices.class_idx,
+            },
+        );
+        self.insert_binding_idx(
+            class_indices.metadata_idx,
+            BindingClassMetadata {
+                class_idx: class_indices.class_idx,
+                bases: bases.clone().into_boxed_slice(),
+                keywords: keywords.into_boxed_slice(),
+                decorators: decorators_with_ranges.clone().into_boxed_slice(),
+                is_new_type: false,
+                pydantic_config_dict,
+                django_primary_key_field,
+            },
+        );
+        self.insert_binding_idx(
+            class_indices.abstract_class_check_idx,
+            BindingAbstractClassCheck {
+                class_idx: class_indices.class_idx,
+            },
+        );
+    }
+
+    /// Extracts docstrings for each field, mapping the field's range to the docstring's range.
+    fn extract_field_docstrings(
+        &self,
+        body: &[ruff_python_ast::Stmt],
+    ) -> SmallMap<TextRange, TextRange> {
+        use ruff_python_ast::Expr;
+        use ruff_python_ast::Stmt;
+
+        let mut field_docstrings = SmallMap::new();
+        let mut i = 0;
+
+        while i < body.len() {
+            let stmt = &body[i];
+
+            let is_field = matches!(stmt, Stmt::AnnAssign(_) | Stmt::Assign(_));
+
+            if let Stmt::FunctionDef(func_def) = stmt {
+                if let Some(docstring_range) = Docstring::range_from_stmts(&func_def.body) {
+                    field_docstrings.insert(func_def.name.range, docstring_range);
+                }
+            } else if let Stmt::ClassDef(class_def) = stmt {
+                if let Some(docstring_range) = Docstring::range_from_stmts(&class_def.body) {
+                    field_docstrings.insert(class_def.name.range, docstring_range);
+                }
+            } else if is_field
+                && let Some(next_stmt) = body.get(i + 1)
+                && let Stmt::Expr(expr_stmt) = next_stmt
+                && matches!(&*expr_stmt.value, Expr::StringLiteral(_))
+            {
+                field_docstrings.insert(stmt.range(), next_stmt.range());
+            }
+
+            i += 1;
+        }
+
+        field_docstrings
     }
 
     fn extract_string_literals(
@@ -421,6 +461,7 @@ impl<'a> BindingsBuilder<'a> {
         class_name: Identifier,
         class_object: CurrentIdx,
         class_indices: ClassIndices,
+        parent: &NestingContext,
         base: Option<Expr>,
         keywords: Box<[(Name, Expr)]>,
         // name, position, annotation, value
@@ -428,23 +469,32 @@ impl<'a> BindingsBuilder<'a> {
         illegal_identifier_handling: IllegalIdentifierHandling,
         force_class_initialization: bool,
         class_kind: SynthesizedClassKind,
-        special_base: Option<Box<BaseClass>>,
+        special_base: Option<BaseClass>,
     ) {
-        let mut key_class_fields: SmallSet<Idx<KeyClassField>> = SmallSet::new();
-
+        let base_classes = base
+            .into_iter()
+            .map(|base| self.base_class_of(base))
+            .chain(special_base)
+            .collect::<Vec<_>>();
+        let is_new_type = class_kind == SynthesizedClassKind::NewType;
+        self.insert_binding_idx(
+            class_indices.base_type_idx,
+            BindingClassBaseType {
+                class_idx: class_indices.class_idx,
+                is_new_type,
+                bases: base_classes.clone().into_boxed_slice(),
+            },
+        );
         self.insert_binding_idx(
             class_indices.metadata_idx,
             BindingClassMetadata {
                 class_idx: class_indices.class_idx,
-                bases: base
-                    .clone()
-                    .into_iter()
-                    .collect::<Vec<_>>()
-                    .into_boxed_slice(),
+                bases: base_classes.into_boxed_slice(),
                 keywords,
                 decorators: Box::new([]),
-                is_new_type: class_kind == SynthesizedClassKind::NewType,
-                special_base,
+                is_new_type,
+                pydantic_config_dict: PydanticConfigDict::default(),
+                django_primary_key_field: None,
             },
         );
         self.insert_binding_idx(
@@ -510,6 +560,7 @@ impl<'a> BindingsBuilder<'a> {
                     member_annotation.is_some() || class_kind == SynthesizedClassKind::NamedTuple,
                     member_value.is_some(),
                     range,
+                    None, // Synthesized fields don't have docstrings
                 ),
             );
             let annotation = member_annotation.map(|annotation_expr| {
@@ -539,7 +590,7 @@ impl<'a> BindingsBuilder<'a> {
                     None => ClassFieldDefinition::DeclaredWithoutAnnotation,
                 },
             };
-            let idx = self.insert_binding(
+            self.insert_binding(
                 KeyClassField(class_indices.def_index, member_name.clone()),
                 BindingClassField {
                     class_idx: class_indices.class_idx,
@@ -548,9 +599,8 @@ impl<'a> BindingsBuilder<'a> {
                     definition,
                 },
             );
-            key_class_fields.insert(idx);
         }
-        self.bind_definition_current(
+        self.bind_current_as(
             &class_name,
             class_object,
             Binding::ClassDef(class_indices.class_idx, Box::new([])),
@@ -558,7 +608,12 @@ impl<'a> BindingsBuilder<'a> {
         );
         self.insert_binding_idx(
             class_indices.class_idx,
-            BindingClass::FunctionalClassDef(class_indices.def_index, class_name, fields),
+            BindingClass::FunctionalClassDef(
+                class_indices.def_index,
+                class_name,
+                parent.dupe(),
+                fields,
+            ),
         );
 
         self.insert_binding_idx(
@@ -567,11 +622,24 @@ impl<'a> BindingsBuilder<'a> {
                 class_key: class_indices.class_idx,
             },
         );
+        self.insert_binding_idx(
+            class_indices.consistent_override_check_idx,
+            BindingConsistentOverrideCheck {
+                class_key: class_indices.class_idx,
+            },
+        );
+        self.insert_binding_idx(
+            class_indices.abstract_class_check_idx,
+            BindingAbstractClassCheck {
+                class_idx: class_indices.class_idx,
+            },
+        );
     }
 
     pub fn synthesize_enum_def(
         &mut self,
         name: &ExprName,
+        parent: &NestingContext,
         func: &mut Expr,
         arg_name: &mut Expr,
         members: &mut [Expr],
@@ -603,6 +671,11 @@ impl<'a> BindingsBuilder<'a> {
                 }
                 // Enum('Color', 'RED', 'GREEN', 'BLUE')
                 [Expr::StringLiteral(_), ..] => self.extract_string_literals(members),
+                // Enum('Color', []), Enum('Color', ())
+                [
+                    Expr::List(ExprList { elts, .. }) | Expr::Tuple(ExprTuple { elts, .. }),
+                    ..,
+                ] if elts.is_empty() => Vec::new(),
                 // Enum('Color', ['RED', 'GREEN', 'BLUE'])
                 [Expr::List(ExprList { elts, .. })]
                     if matches!(elts.as_slice(), [Expr::StringLiteral(_), ..]) =>
@@ -668,6 +741,7 @@ impl<'a> BindingsBuilder<'a> {
             class_name,
             class_object,
             class_indices,
+            parent,
             Some(func.clone()),
             Box::new([]),
             member_definitions,
@@ -683,6 +757,7 @@ impl<'a> BindingsBuilder<'a> {
     pub fn synthesize_collections_named_tuple_def(
         &mut self,
         name: &ExprName,
+        parent: &NestingContext,
         func: &mut Expr,
         arg_name: &Expr,
         members: &mut [Expr],
@@ -708,6 +783,12 @@ impl<'a> BindingsBuilder<'a> {
                         .collect()
                 }
             }
+            // namedtuple('Point', []), namedtuple('Point', ())
+            [Expr::List(ExprList { elts, .. }) | Expr::Tuple(ExprTuple { elts, .. })]
+                if elts.is_empty() =>
+            {
+                Vec::new()
+            }
             // namedtuple('Point', ['x', 'y'])
             [Expr::List(ExprList { elts, .. })]
                 if matches!(elts.as_slice(), [Expr::StringLiteral(_), ..]) =>
@@ -743,7 +824,8 @@ impl<'a> BindingsBuilder<'a> {
                 }
             } else if let Some(name) = &kw.arg
                 && name.id == "defaults"
-                && let Expr::Tuple(ExprTuple { elts, .. }) = &kw.value
+                && let Expr::Tuple(ExprTuple { elts, .. }) | Expr::List(ExprList { elts, .. }) =
+                    &kw.value
             {
                 let n_defaults = elts.len();
                 if n_defaults > n_members {
@@ -760,10 +842,15 @@ impl<'a> BindingsBuilder<'a> {
                     defaults.splice(n_members - n_defaults.., elts.map(|x| Some(x.clone())));
                 }
             } else {
+                let maybe_name = if let Some(name) = &kw.arg {
+                    format!(" `{name}`")
+                } else {
+                    "".to_owned()
+                };
                 self.error(
-                    kw.value.range(),
+                    kw.range(),
                     ErrorInfo::Kind(ErrorKind::InvalidArgument),
-                    "Unrecognized argument for typed dictionary definition".to_owned(),
+                    format!("Unrecognized argument{maybe_name} for named tuple definition"),
                 );
             }
         }
@@ -778,13 +865,14 @@ impl<'a> BindingsBuilder<'a> {
             class_name,
             class_object,
             class_indices,
+            parent,
             None,
             Box::new([]),
             member_definitions_with_defaults,
             illegal_identifier_handling,
             false,
             SynthesizedClassKind::NamedTuple,
-            Some(Box::new(BaseClass::NamedTuple(range))),
+            Some(BaseClass::NamedTuple(range)),
         );
     }
 
@@ -792,6 +880,7 @@ impl<'a> BindingsBuilder<'a> {
     pub fn synthesize_typing_named_tuple_def(
         &mut self,
         name: &ExprName,
+        parent: &NestingContext,
         func: &mut Expr,
         arg_name: &Expr,
         members: &[Expr],
@@ -802,6 +891,12 @@ impl<'a> BindingsBuilder<'a> {
         self.check_functional_definition_name(&name.id, arg_name);
         let member_definitions: Vec<(String, TextRange, Option<Expr>, Option<Expr>)> =
             match members {
+                // NamedTuple('Point', []), NamedTuple('Point', ())
+                [Expr::List(ExprList { elts, .. }) | Expr::Tuple(ExprTuple { elts, .. })]
+                    if elts.is_empty() =>
+                {
+                    Vec::new()
+                }
                 // NamedTuple('Point', [('x', int), ('y', int)])
                 [Expr::List(ExprList { elts, .. })]
                     if matches!(elts.as_slice(), [Expr::Tuple(_), ..]) =>
@@ -837,6 +932,7 @@ impl<'a> BindingsBuilder<'a> {
             class_name,
             class_object,
             class_indices,
+            parent,
             Some(func.clone()),
             Box::new([]),
             member_definitions,
@@ -851,6 +947,7 @@ impl<'a> BindingsBuilder<'a> {
     pub fn synthesize_typing_new_type(
         &mut self,
         name: &ExprName,
+        parent: &NestingContext,
         new_type_name: &mut Expr,
         base: &mut Expr,
     ) {
@@ -863,6 +960,7 @@ impl<'a> BindingsBuilder<'a> {
             class_name,
             class_object,
             class_indices,
+            parent,
             Some(base.clone()),
             Box::new([]),
             Vec::new(),
@@ -876,6 +974,7 @@ impl<'a> BindingsBuilder<'a> {
     pub fn synthesize_typed_dict_def(
         &mut self,
         name: &ExprName,
+        parent: &NestingContext,
         func: &mut Expr,
         arg_name: &Expr,
         args: &mut [Expr],
@@ -885,19 +984,28 @@ impl<'a> BindingsBuilder<'a> {
         let (mut class_object, class_indices) = self.class_object_and_indices(&class_name);
         self.ensure_expr(func, class_object.usage());
         self.check_functional_definition_name(&name.id, arg_name);
-        let mut base_class_keywords: Box<[(Name, Expr)]> = Box::new([]);
+        let mut base_class_keywords = Vec::new();
         for kw in keywords {
             self.ensure_expr(&mut kw.value, class_object.usage());
-            if let Some(name) = &kw.arg
-                && name.id == "total"
-                && matches!(kw.value, Expr::BooleanLiteral(_))
-            {
-                base_class_keywords = Box::new([(name.id.clone(), kw.value.clone())])
+            let recognized_kw = match (kw.arg.as_ref().map(|id| &id.id), &kw.value) {
+                (Some(name), Expr::BooleanLiteral(_)) if name == "total" || name == "closed" => {
+                    Some(name)
+                }
+                (Some(name), _) if name == "extra_items" => Some(name),
+                _ => None,
+            };
+            if let Some(kw_name) = recognized_kw {
+                base_class_keywords.push((kw_name.clone(), kw.value.clone()));
             } else {
+                let maybe_name = if let Some(name) = &kw.arg {
+                    format!(" `{name}`")
+                } else {
+                    "".to_owned()
+                };
                 self.error(
-                    kw.value.range(),
+                    kw.range(),
                     ErrorInfo::Kind(ErrorKind::InvalidArgument),
-                    "Unrecognized argument for typed dictionary definition".to_owned(),
+                    format!("Unrecognized argument{maybe_name} for typed dictionary definition"),
                 );
             }
         }
@@ -946,8 +1054,9 @@ impl<'a> BindingsBuilder<'a> {
             class_name,
             class_object,
             class_indices,
+            parent,
             Some(func.clone()),
-            base_class_keywords,
+            base_class_keywords.into_boxed_slice(),
             member_definitions,
             IllegalIdentifierHandling::Allow,
             false,
