@@ -39,15 +39,12 @@ use crate::binding::binding::KeyYieldFrom;
 use crate::binding::binding::LinkedKey;
 use crate::binding::binding::SuperStyle;
 use crate::binding::bindings::BindingsBuilder;
-use crate::binding::bindings::LegacyTParamBuilder;
-use crate::binding::bindings::LookupError;
-use crate::binding::bindings::LookupKind;
+use crate::binding::bindings::LegacyTParamCollector;
+use crate::binding::bindings::LegacyTParamId;
+use crate::binding::bindings::NameLookupResult;
 use crate::binding::narrow::AtomicNarrowOp;
 use crate::binding::narrow::NarrowOps;
-use crate::binding::scope::Flow;
 use crate::binding::scope::Scope;
-use crate::binding::scope::ScopeClass;
-use crate::binding::scope::ScopeKind;
 use crate::config::error_kind::ErrorKind;
 use crate::error::context::ErrorInfo;
 use crate::export::special::SpecialExport;
@@ -55,6 +52,12 @@ use crate::graph::index::Idx;
 use crate::types::callable::unexpected_keyword;
 use crate::types::types::Type;
 
+/// Match on an expression by name. Should be used only for special names that we essentially treat like keywords,
+/// like reveal_type.
+fn is_special_name(name: &str) -> bool {
+    matches!(name, "reveal_type" | "assert_type")
+}
+
 /// Looking up names in an expression requires knowing the identity of the binding
 /// we are computing for usage tracking.
 ///
@@ -71,14 +74,32 @@ pub enum Usage {
     /// - Because narrowing introduces duplicate expressions, it is difficult
     ///   to ensure unpinned Vars cannot leak into the binding graph and cause
     ///   nondeterminism.
-    Narrowing,
+    ///
+    /// It carries an optional current idx so we could detect mutliple usages to
+    /// the same key within the same binding.
+    Narrowing(Option<Idx<Key>>),
     /// I'm a usage in some context (a type variable declaration, an annotation,
     /// a cast, etc) where we are dealing with static types. I will not pin
     /// any placeholder types.
     StaticTypeInformation,
-    /// I'm a usage in a mutable lookup (a reassignment or delete). I will not
-    /// pin any placeholder types.
-    MutableLookup,
+}
+
+impl Usage {
+    pub fn narrowing_from(other: &Self) -> Self {
+        match other {
+            Self::CurrentIdx(idx, _) => Self::Narrowing(Some(*idx)),
+            Self::Narrowing(idx) => Self::Narrowing(*idx),
+            Self::StaticTypeInformation => Self::Narrowing(None),
+        }
+    }
+
+    pub fn current_idx(&self) -> Option<Idx<Key>> {
+        match self {
+            Self::CurrentIdx(idx, _) => Some(*idx),
+            Self::Narrowing(idx) => *idx,
+            Self::StaticTypeInformation => None,
+        }
+    }
 }
 
 enum TestAssertion {
@@ -211,6 +232,47 @@ impl TestAssertion {
 }
 
 impl<'a> BindingsBuilder<'a> {
+    /// Ensure the name in an `ExprName`. Note that unlike `ensure_expr`, it
+    /// does not require a mutable ref.
+    pub fn ensure_expr_name(&mut self, x: &ExprName, usage: &mut Usage) -> Idx<Key> {
+        let name = Ast::expr_name_identifier(x.clone());
+        self.ensure_name(&name, usage, &mut None)
+    }
+
+    fn ensure_name(
+        &mut self,
+        name: &Identifier,
+        usage: &mut Usage,
+        tparams_builder: &mut Option<LegacyTParamCollector>,
+    ) -> Idx<Key> {
+        self.ensure_name_impl(
+            name,
+            usage,
+            tparams_builder
+                .as_mut()
+                .map(|tparams_builder| (tparams_builder, LegacyTParamId::Name(name.clone()))),
+        )
+    }
+
+    fn ensure_simple_attr(
+        &mut self,
+        value: &Identifier,
+        attr: &Identifier,
+        usage: &mut Usage,
+        tparams_builder: &mut Option<LegacyTParamCollector>,
+    ) -> Idx<Key> {
+        self.ensure_name_impl(
+            value,
+            usage,
+            tparams_builder.as_mut().map(|tparams_builder| {
+                (
+                    tparams_builder,
+                    LegacyTParamId::Attr(value.clone(), attr.clone()),
+                )
+            }),
+        )
+    }
+
     /// Given a name appearing in an expression, create a `Usage` key for that
     /// name at the current location. The binding will indicate how to compute
     /// the type if we found that name in scope; if we do not find the name we
@@ -222,19 +284,21 @@ impl<'a> BindingsBuilder<'a> {
     /// - Information about what binding it is being used in, which is used both
     ///   - to track first-use to get deterministic inference of placeholder
     ///     types like empty list
-    ///   - to determin when we are in a static typing usage
+    ///   - to determine when we are in a static typing usage
     /// - The lookup kind, which is used to distinguish between normal lookups,
     ///   which allow uses of nonlocals, versus mutable lookups that do not
     ///   (unless the nonlocal was explicitly mutably captured by a `global`
     ///   or `nonlocal` statement).
-    /// - An optional `tparams_builder`, which intercepts names - but only
-    ///   in static type contexts - that map to legacy type variables.
-    pub fn ensure_name(
+    /// - An optional `tparams_lookup`, which intercepts names - but only
+    ///   in static type contexts - that map to legacy type variables. It
+    ///   is a flexible callback in order to handle not only bare name type
+    ///   variables, but also `<module>.<name>` type variables, which have
+    ///   to be modeled as attribute narrows of the module at solve time.
+    fn ensure_name_impl(
         &mut self,
         name: &Identifier,
-        lookup_kind: LookupKind,
         usage: &mut Usage,
-        tparams_builder: &mut Option<LegacyTParamBuilder>,
+        tparams_lookup: Option<(&mut LegacyTParamCollector, LegacyTParamId)>,
     ) -> Idx<Key> {
         let key = Key::BoundName(ShortIdentifier::new(name));
         if name.is_empty() {
@@ -249,61 +313,68 @@ impl<'a> BindingsBuilder<'a> {
             return self.insert_binding_overwrite(key, Binding::Type(Type::any_error()));
         }
         let used_in_static_type = matches!(usage, Usage::StaticTypeInformation);
-        let value = if used_in_static_type && let Some(tparams_builder) = tparams_builder {
-            tparams_builder
-                .intercept_lookup(self, name)
-                .ok_or(LookupError::NotFound)
-        } else {
-            self.lookup_name(Hashed::new(&name.id), lookup_kind, usage)
-                .map(Binding::Forward)
-        };
-        match value {
-            Ok(value) => {
-                if !self.module_info.path().is_interface() {
-                    // Don't check flow for global/nonlocal lookups
-                    if let Some(error_message) = self
-                        .scopes
-                        .get_flow_style(&name.id, used_in_static_type)
-                        .uninitialized_error_message(name)
-                    {
-                        self.error(
-                            name.range,
-                            ErrorInfo::Kind(ErrorKind::UnboundName),
-                            error_message,
-                        );
-                    }
+        let lookup_result =
+            if used_in_static_type && let Some((tparams_collector, tparam_id)) = tparams_lookup {
+                self.intercept_lookup(tparams_collector, tparam_id)
+            } else {
+                self.lookup_name(Hashed::new(&name.id), usage)
+            };
+        match lookup_result {
+            NameLookupResult::Found {
+                idx: value,
+                uninitialized: is_initialized,
+            } => {
+                // Uninitialized local errors are only reported when we are neither in a stub
+                // nor a static type context.
+                if !used_in_static_type
+                    && !self.module_info.path().is_interface()
+                    && let Some(error_message) = is_initialized.as_error_message(&name.id)
+                {
+                    self.error(
+                        name.range,
+                        ErrorInfo::Kind(ErrorKind::UnboundName),
+                        error_message,
+                    );
                 }
-                self.insert_binding(key, value)
+                self.insert_binding(key, Binding::Forward(value))
             }
-            Err(error) => {
-                // Record a type error and fall back to `Any`.
-                self.error(
-                    name.range,
-                    ErrorInfo::Kind(ErrorKind::UnknownName),
-                    error.message(name),
-                );
+            NameLookupResult::NotFound => {
+                if is_special_name(name.id.as_str()) {
+                    self.error(
+                        name.range,
+                        ErrorInfo::Kind(ErrorKind::UnknownName),
+                        format!(
+                            "`{}` must be imported from `typing` for runtime usage",
+                            name
+                        ),
+                    );
+                } else {
+                    // Record a type error and fall back to `Any`.
+                    self.error(
+                        name.range,
+                        ErrorInfo::Kind(ErrorKind::UnknownName),
+                        format!("Could not find name `{name}`"),
+                    );
+                }
                 self.insert_binding(key, Binding::Type(Type::any_error()))
             }
         }
     }
 
-    pub fn ensure_mutable_name(&mut self, x: &ExprName, usage: &mut Usage) -> Idx<Key> {
-        let name = Ast::expr_name_identifier(x.clone());
-        self.ensure_name(&name, LookupKind::Mutable, usage, &mut None)
-    }
-
     fn bind_comprehensions(
         &mut self,
         range: TextRange,
         comprehensions: &mut [Comprehension],
         usage: &mut Usage,
     ) {
-        self.scopes.push(Scope::comprehension(range));
-        for comp in comprehensions {
+        for (i, comp) in comprehensions.iter_mut().enumerate() {
             // Resolve the type of the iteration value *before* binding the target of the iteration.
             // This is necessary so that, e.g. `[x for x in x]` correctly uses the outer scope for
             // the `in x` lookup.
             self.ensure_expr(&mut comp.iter, usage);
+            if i == 0 {
+                self.scopes.push(Scope::comprehension(range));
+            }
             // Incomplete nested comprehensions can have identical iterators
             // for inner and outer loops. It is safe to overwrite it because it literally the same.
             let iterable_value_idx = self.insert_binding_overwrite(
@@ -317,15 +388,15 @@ impl<'a> BindingsBuilder<'a> {
                 Binding::Forward(iterable_value_idx)
             });
             for x in comp.ifs.iter_mut() {
-                self.ensure_expr(x, &mut Usage::Narrowing);
+                self.ensure_expr(x, &mut Usage::narrowing_from(usage));
                 let narrow_ops = NarrowOps::from_expr(self, Some(x));
-                self.bind_narrow_ops(&narrow_ops, comp.range);
+                self.bind_narrow_ops(&narrow_ops, comp.range, usage);
             }
         }
     }
 
     pub fn bind_lambda(&mut self, lambda: &mut ExprLambda, usage: &mut Usage) {
-        self.scopes.push(Scope::function(lambda.range));
+        self.scopes.push(Scope::lambda(lambda.range, false));
         if let Some(parameters) = &lambda.parameters {
             for x in parameters {
                 self.bind_lambda_param(x.name());
@@ -338,7 +409,7 @@ impl<'a> BindingsBuilder<'a> {
         // TODO: We should properly handle `yield` and `yield from`; lambdas can be generators.
         // One example of this is in the standard library, in `_collections_abc.pyi`:
         // https://github.com/python/cpython/blob/965662ee4a986605b60da470d9e7c1e9a6f922b3/Lib/_collections_abc.py#L92
-        let (yields_and_returns, _) = self.scopes.pop_function_scope();
+        let (yields_and_returns, _, _) = self.scopes.pop_function_scope();
         for (idx, y) in yields_and_returns.yields {
             self.insert_binding_idx(idx, BindingYield::Invalid(y));
         }
@@ -347,39 +418,11 @@ impl<'a> BindingsBuilder<'a> {
         }
     }
 
-    /// Helper to clean up an expression that does type narrowing. We merge flows for the narrowing
-    /// operation and its negation, so that narrowing is limited to the body of the expression but
-    /// newly defined names persist.
-    fn negate_and_merge_flow(
-        &mut self,
-        base: Flow,
-        ops: &NarrowOps,
-        orelse: Option<&mut Expr>,
-        range: TextRange,
-        usage: &mut Usage,
-    ) {
-        let if_branch = self.scopes.replace_current_flow(base);
-        self.bind_narrow_ops(&ops.negate(), range);
-        self.ensure_expr_opt(orelse, usage);
-        // Swap them back again, to make sure that the merge order is if, then else
-        let else_branch = self.scopes.replace_current_flow(if_branch);
-        self.merge_branches_into_current(vec![else_branch], range);
-    }
-
-    fn enclosing_class_name(&self) -> Option<&Identifier> {
-        for scope in self.scopes.iter_rev() {
-            if let ScopeKind::Class(ScopeClass { name, .. }) = &scope.kind {
-                return Some(name);
-            }
-        }
-        None
-    }
-
     // We want to special-case `self.assertXXX()` methods in unit tests.
     // The logic is intentionally syntax-based as we want to avoid checking whether the base type
     // is `unittest.TestCase` on every single method invocation.
     fn as_assert_in_test(&self, func: &Expr) -> Option<TestAssertion> {
-        if let Some(class_name) = self.enclosing_class_name() {
+        if let Some(class_name) = self.scopes.enclosing_class_name() {
             let class_name_str = class_name.as_str();
             if !(class_name_str.contains("test") || class_name_str.contains("Test")) {
                 return None;
@@ -413,7 +456,7 @@ impl<'a> BindingsBuilder<'a> {
     }
 
     fn record_yield(&mut self, mut x: ExprYield) {
-        let mut yield_link = self.declare_current_idx(Key::UsageLink(x.range));
+        let mut yield_link = self.declare_current_idx(Key::YieldLink(x.range));
         let idx = self.idx_for_promise(KeyYield(x.range));
         self.ensure_expr_opt(x.value.as_deref_mut(), yield_link.usage());
         if let Err(oops_top_level) = self.scopes.record_or_reject_yield(idx, x) {
@@ -423,7 +466,7 @@ impl<'a> BindingsBuilder<'a> {
     }
 
     fn record_yield_from(&mut self, mut x: ExprYieldFrom) {
-        let mut yield_from_link = self.declare_current_idx(Key::UsageLink(x.range));
+        let mut yield_from_link = self.declare_current_idx(Key::YieldLink(x.range));
         let idx = self.idx_for_promise(KeyYieldFrom(x.range));
         self.ensure_expr(&mut x.value, yield_from_link.usage());
         if let Err(oops_top_level) = self.scopes.record_or_reject_yield_from(idx, x) {
@@ -440,13 +483,18 @@ impl<'a> BindingsBuilder<'a> {
         match x {
             Expr::If(x) => {
                 // Ternary operation. We treat it like an if/else statement.
-                let base = self.scopes.clone_current_flow();
-                self.ensure_expr(&mut x.test, &mut Usage::Narrowing);
+                self.start_fork_and_branch(x.range);
+                self.ensure_expr(&mut x.test, &mut Usage::narrowing_from(usage));
                 let narrow_ops = NarrowOps::from_expr(self, Some(&x.test));
-                self.bind_narrow_ops(&narrow_ops, x.body.range());
+                self.bind_narrow_ops(&narrow_ops, x.body.range(), usage);
                 self.ensure_expr(&mut x.body, usage);
-                let range = x.range();
-                self.negate_and_merge_flow(base, &narrow_ops, Some(&mut x.orelse), range, usage);
+                // Negate the narrow ops for the `orelse`, then merge the Flows.
+                // TODO(stroxler): We eventually want to drop all narrows but merge values.
+                self.next_branch();
+                self.bind_narrow_ops(&narrow_ops.negate(), x.range, usage);
+                self.ensure_expr(&mut x.orelse, usage);
+                self.finish_branch();
+                self.finish_exhaustive_fork();
             }
             Expr::BoolOp(ExprBoolOp {
                 node_index: _,
@@ -454,24 +502,40 @@ impl<'a> BindingsBuilder<'a> {
                 op,
                 values,
             }) => {
-                let base = self.scopes.clone_current_flow();
-                let mut narrow_ops = NarrowOps::new();
-                for value in values {
-                    self.bind_narrow_ops(&narrow_ops, value.range());
-                    self.ensure_expr(value, &mut Usage::Narrowing);
-                    let new_narrow_ops = NarrowOps::from_expr(self, Some(value));
+                let mut values = values.iter_mut();
+                fn get_narrow_ops(myself: &BindingsBuilder, expr: &Expr, op: BoolOp) -> NarrowOps {
+                    let raw_narrow_ops = NarrowOps::from_expr(myself, Some(expr));
                     match op {
                         BoolOp::And => {
                             // Every subsequent value is evaluated only if all previous values were truthy.
-                            narrow_ops.and_all(new_narrow_ops);
+                            raw_narrow_ops
                         }
                         BoolOp::Or => {
                             // Every subsequent value is evaluated only if all previous values were falsy.
-                            narrow_ops.and_all(new_narrow_ops.negate());
+                            raw_narrow_ops.negate()
                         }
                     }
                 }
-                self.negate_and_merge_flow(base, &narrow_ops, None, *range, usage);
+                if let Some(value) = values.next() {
+                    // The first operation runs unconditionally, so any walrus-defined
+                    // names will be added to the base flow.
+                    self.ensure_expr(value, &mut Usage::narrowing_from(usage));
+                    self.start_fork_and_branch(*range);
+                    let mut narrow_ops = get_narrow_ops(self, value, *op);
+                    for value in values {
+                        self.bind_narrow_ops(&narrow_ops, value.range(), usage);
+                        self.ensure_expr(value, &mut Usage::narrowing_from(usage));
+                        let new_narrow_ops = get_narrow_ops(self, value, *op);
+                        narrow_ops.and_all(new_narrow_ops);
+                    }
+                    // Negate the narrow ops in the base flow and merge.
+                    // TODO(stroxler): We eventually want to drop all narrows but merge values.
+                    // Once we have a way to do that, the negation will be unnecessary.
+                    self.next_branch();
+                    self.bind_narrow_ops(&narrow_ops.negate(), *range, usage);
+                    self.finish_branch();
+                    self.finish_bool_op_fork();
+                }
             }
             Expr::Call(ExprCall {
                 node_index: _,
@@ -545,25 +609,9 @@ impl<'a> BindingsBuilder<'a> {
                 }
                 let nargs = posargs.len();
                 let style = if nargs == 0 {
-                    let mut method_name = None;
-                    let mut class_key = None;
-                    for scope in self.scopes.iter_rev() {
-                        match &scope.kind {
-                            ScopeKind::Method(method_scope) => {
-                                method_name = Some(method_scope.name.clone());
-                            }
-                            ScopeKind::Class(class_scope) if method_name.is_some() => {
-                                class_key = Some(class_scope.indices.class_idx);
-                                break;
-                            }
-                            _ => {}
-                        }
-                    }
-                    match (class_key, method_name) {
-                        (Some(class_idx), Some(method)) => {
-                            SuperStyle::ImplicitArgs(class_idx, method)
-                        }
-                        _ => {
+                    match self.scopes.current_method_and_class() {
+                        Some((method, class_idx)) => SuperStyle::ImplicitArgs(class_idx, method),
+                        None => {
                             self.error(
                                 *range,
                                 ErrorInfo::Kind(ErrorKind::InvalidSuperCall),
@@ -614,12 +662,12 @@ impl<'a> BindingsBuilder<'a> {
             {
                 self.ensure_expr(func, usage);
                 for arg in arguments.args.iter_mut() {
-                    self.ensure_expr(arg, &mut Usage::Narrowing);
+                    self.ensure_expr(arg, &mut Usage::narrowing_from(usage));
                 }
                 for kw in arguments.keywords.iter_mut() {
                     self.ensure_expr(&mut kw.value, usage);
                 }
-                self.bind_narrow_ops(&narrow_op, *range);
+                self.bind_narrow_ops(&narrow_op, *range, usage);
             }
             Expr::Named(x) => {
                 // For scopes defined in terms of Definitions, we should normally already have the name in Static, but
@@ -665,7 +713,7 @@ impl<'a> BindingsBuilder<'a> {
             }
             Expr::Name(x) => {
                 let name = Ast::expr_name_identifier(x.clone());
-                self.ensure_name(&name, LookupKind::Regular, usage, &mut None);
+                self.ensure_name(&name, usage, &mut None);
             }
             Expr::Yield(x) => {
                 self.record_yield(x.clone());
@@ -673,6 +721,16 @@ impl<'a> BindingsBuilder<'a> {
             Expr::YieldFrom(x) => {
                 self.record_yield_from(x.clone());
             }
+            Expr::Await(x) => {
+                self.ensure_expr(&mut x.value, usage);
+                if !self.scopes.is_in_async_def() && !self.module_info.path().is_notebook() {
+                    self.error(
+                        x.range(),
+                        ErrorInfo::Kind(ErrorKind::InvalidSyntax),
+                        "`await` can only be used inside an async function".to_owned(),
+                    );
+                }
+            }
             _ => {
                 x.recurse_mut(&mut |x| self.ensure_expr(x, usage));
             }
@@ -687,19 +745,27 @@ impl<'a> BindingsBuilder<'a> {
     }
 
     /// Execute through the expr, ensuring every name has a binding.
-    pub fn ensure_type(&mut self, x: &mut Expr, tparams_builder: &mut Option<LegacyTParamBuilder>) {
+    pub fn ensure_type(
+        &mut self,
+        x: &mut Expr,
+        tparams_builder: &mut Option<LegacyTParamCollector>,
+    ) {
+        self.ensure_type_impl(x, tparams_builder, false);
+    }
+
+    fn ensure_type_impl(
+        &mut self,
+        x: &mut Expr,
+        tparams_builder: &mut Option<LegacyTParamCollector>,
+        in_string_literal: bool,
+    ) {
         self.track_potential_typing_self(x);
         // We do not treat static types as usage for the purpose of first-usage-based type inference.
         let static_type_usage = &mut Usage::StaticTypeInformation;
         match x {
             Expr::Name(x) => {
                 let name = Ast::expr_name_identifier(x.clone());
-                self.ensure_name(
-                    &name,
-                    LookupKind::Regular,
-                    static_type_usage,
-                    tparams_builder,
-                );
+                self.ensure_name(&name, static_type_usage, tparams_builder);
             }
             Expr::Subscript(ExprSubscript { value, .. })
                 if self.as_special_export(value) == Some(SpecialExport::Literal) =>
@@ -712,28 +778,28 @@ impl<'a> BindingsBuilder<'a> {
                     && matches!(&**slice, Expr::Tuple(tup) if !tup.is_empty()) =>
             {
                 // Only go inside the first argument to Annotated, the rest are non-type metadata.
-                self.ensure_type(&mut *value, tparams_builder);
+                self.ensure_type_impl(&mut *value, tparams_builder, in_string_literal);
                 // We can't bind a mut box in the guard (sadly), so force unwrapping it here
                 let tup = slice.as_tuple_expr_mut().unwrap();
-                self.ensure_type(&mut tup.elts[0], tparams_builder);
+                self.ensure_type_impl(&mut tup.elts[0], tparams_builder, in_string_literal);
                 for e in tup.elts[1..].iter_mut() {
                     self.ensure_expr(e, static_type_usage);
                 }
             }
-            Expr::StringLiteral(literal) if let Some(literal) = literal.as_single_part_string() => {
+            Expr::Subscript(ExprSubscript { value, slice, .. }) => {
+                self.ensure_type_impl(&mut *value, tparams_builder, in_string_literal);
+                self.ensure_type_impl(&mut *slice, tparams_builder, in_string_literal);
+            }
+            Expr::StringLiteral(literal)
+                if !in_string_literal && let Some(literal) = literal.as_single_part_string() =>
+            {
                 match Ast::parse_type_literal(literal) {
                     Ok(expr) => {
                         *x = expr;
-                        // TODO: Remember if we have already done a parse_type_literal, so we could properly
-                        // reject annotations of the form `"'T'"`.
-                        self.ensure_type(x, tparams_builder);
+                        self.ensure_type_impl(x, tparams_builder, true);
                     }
-                    Err(e) => {
-                        self.error(
-                            literal.range,
-                            ErrorInfo::Kind(ErrorKind::ParseError),
-                            format!("Could not parse type string: {}, got {e}", literal.value),
-                        );
+                    Err(_) => {
+                        // We don't need to emit errors here, because the solving logic expects the expression to resolve to a type, and it will fail.
                     }
                 }
             }
@@ -743,7 +809,22 @@ impl<'a> BindingsBuilder<'a> {
             // test::class_super::test_super_in_base_classes for an example of a SuperInstance
             // binding that we crash looking for if we don't do this.
             Expr::Call(_) => self.ensure_expr(x, static_type_usage),
-            _ => x.recurse_mut(&mut |x| self.ensure_type(x, tparams_builder)),
+            Expr::Attribute(ExprAttribute { value, attr, .. })
+                if let Expr::Name(value) = &**value
+                // We assume "args" and "kwargs" are ParamSpec attributes rather than imported TypeVars.
+                    && attr.id != "args" && attr.id != "kwargs" =>
+            {
+                // We intercept <name>.<name> to check if this is an imported legacy type parameter.
+                self.ensure_simple_attr(
+                    &Ast::expr_name_identifier(value.clone()),
+                    attr,
+                    static_type_usage,
+                    tparams_builder,
+                );
+            }
+            _ => {
+                x.recurse_mut(&mut |x| self.ensure_type_impl(x, tparams_builder, in_string_literal))
+            }
         }
     }
 
@@ -755,7 +836,8 @@ impl<'a> BindingsBuilder<'a> {
     fn track_potential_typing_self(&mut self, x: &Expr) {
         match self.as_special_export(x) {
             Some(SpecialExport::SelfType) => {
-                if let Some((current_class_idx, _)) = self.scopes.current_class_and_metadata_keys()
+                if let Some((current_class_idx, _)) =
+                    self.scopes.enclosing_class_and_metadata_keys()
                 {
                     self.insert_binding(
                         Key::SelfTypeLiteral(x.range()),
@@ -771,7 +853,7 @@ impl<'a> BindingsBuilder<'a> {
     pub fn ensure_type_opt(
         &mut self,
         x: Option<&mut Expr>,
-        tparams_builder: &mut Option<LegacyTParamBuilder>,
+        tparams_builder: &mut Option<LegacyTParamCollector>,
     ) {
         if let Some(x) = x {
             self.ensure_type(x, tparams_builder);
@@ -782,12 +864,12 @@ impl<'a> BindingsBuilder<'a> {
         &mut self,
         decorators: Vec<Decorator>,
         usage: &mut Usage,
-    ) -> Vec<Idx<Key>> {
+    ) -> Vec<(Idx<Key>, TextRange)> {
         let mut decorator_keys = Vec::with_capacity(decorators.len());
         for mut x in decorators {
             self.ensure_expr(&mut x.expression, usage);
             let k = self.insert_binding(Key::Anon(x.range), Binding::Decorator(x.expression));
-            decorator_keys.push(k);
+            decorator_keys.push((k, x.range));
         }
         decorator_keys
     }
