@@ -7,21 +7,25 @@
 
 use std::mem;
 
+use dupe::Dupe as _;
 use pyrefly_python::ast::Ast;
 use pyrefly_python::docstring::Docstring;
+use pyrefly_python::nesting_context::NestingContext;
 use pyrefly_python::short_identifier::ShortIdentifier;
 use pyrefly_python::sys_info::SysInfo;
 use pyrefly_util::prelude::VecExt;
 use pyrefly_util::visit::Visit;
-use ruff_python_ast::AnyParameterRef;
 use ruff_python_ast::Decorator;
 use ruff_python_ast::ExceptHandler;
 use ruff_python_ast::Expr;
+use ruff_python_ast::ExprCall;
 use ruff_python_ast::Identifier;
 use ruff_python_ast::Parameters;
 use ruff_python_ast::Stmt;
 use ruff_python_ast::StmtExpr;
 use ruff_python_ast::StmtFunctionDef;
+use ruff_python_ast::StmtRaise;
+use ruff_python_ast::StmtReturn;
 use ruff_python_ast::name::Name;
 use ruff_text_size::Ranged;
 use ruff_text_size::TextRange;
@@ -30,7 +34,8 @@ use starlark_map::small_map::SmallMap;
 use crate::binding::binding::AnnotationTarget;
 use crate::binding::binding::Binding;
 use crate::binding::binding::BindingAnnotation;
-use crate::binding::binding::BindingFunction;
+use crate::binding::binding::BindingDecoratedFunction;
+use crate::binding::binding::BindingUndecoratedFunction;
 use crate::binding::binding::BindingYield;
 use crate::binding::binding::BindingYieldFrom;
 use crate::binding::binding::FunctionStubOrImpl;
@@ -39,17 +44,19 @@ use crate::binding::binding::Key;
 use crate::binding::binding::KeyAnnotation;
 use crate::binding::binding::KeyClass;
 use crate::binding::binding::KeyLegacyTypeParam;
+use crate::binding::binding::KeyUndecoratedFunction;
 use crate::binding::binding::LastStmt;
 use crate::binding::binding::ReturnExplicit;
 use crate::binding::binding::ReturnImplicit;
 use crate::binding::binding::ReturnType;
 use crate::binding::binding::ReturnTypeKind;
 use crate::binding::bindings::BindingsBuilder;
-use crate::binding::bindings::LegacyTParamBuilder;
+use crate::binding::bindings::LegacyTParamCollector;
 use crate::binding::expr::Usage;
 use crate::binding::scope::FlowStyle;
 use crate::binding::scope::InstanceAttribute;
 use crate::binding::scope::Scope;
+use crate::binding::scope::UnusedParameter;
 use crate::binding::scope::YieldsAndReturns;
 use crate::config::base::UntypedDefBehavior;
 use crate::export::special::SpecialExport;
@@ -60,7 +67,8 @@ struct Decorators {
     has_no_type_check: bool,
     is_overload: bool,
     is_abstract_method: bool,
-    decorators: Box<[Idx<Key>]>,
+    is_override: bool,
+    decorators: Box<[(Idx<Key>, TextRange)]>,
 }
 
 pub struct SelfAssignments {
@@ -169,7 +177,12 @@ impl<'a> SelfAttrNames<'a> {
 }
 
 impl<'a> BindingsBuilder<'a> {
-    fn parameters(&mut self, x: &mut Parameters, class_key: Option<Idx<KeyClass>>) {
+    fn parameters(
+        &mut self,
+        x: &mut Parameters,
+        undecorated_idx: Idx<KeyUndecoratedFunction>,
+        class_key: Option<Idx<KeyClass>>,
+    ) {
         let mut self_name = None;
         for x in x.iter_non_variadic_params() {
             if class_key.is_some() && self_name.is_none() {
@@ -177,22 +190,28 @@ impl<'a> BindingsBuilder<'a> {
             }
             self.bind_function_param(
                 AnnotationTarget::Param(x.parameter.name.id.clone()),
-                AnyParameterRef::NonVariadic(x),
+                &x.parameter,
+                undecorated_idx,
                 class_key,
+                false,
             );
         }
         if let Some(args) = &x.vararg {
             self.bind_function_param(
                 AnnotationTarget::ArgsParam(args.name.id.clone()),
-                AnyParameterRef::Variadic(args),
+                args,
+                undecorated_idx,
                 class_key,
+                true,
             );
         }
         if let Some(kwargs) = &x.kwarg {
             self.bind_function_param(
                 AnnotationTarget::KwargsParam(kwargs.name.id.clone()),
-                AnyParameterRef::Variadic(kwargs),
+                kwargs,
+                undecorated_idx,
                 class_key,
+                true,
             );
         }
         self.scopes.set_self_name_if_applicable(self_name);
@@ -203,7 +222,7 @@ impl<'a> BindingsBuilder<'a> {
         mut x: Expr,
         func_name: &Identifier,
         class_key: Option<Idx<KeyClass>>,
-        tparams_builder: &mut Option<LegacyTParamBuilder>,
+        tparams_builder: &mut Option<LegacyTParamCollector>,
     ) -> (TextRange, Idx<KeyAnnotation>) {
         self.ensure_type(&mut x, tparams_builder);
         (
@@ -234,7 +253,7 @@ impl<'a> BindingsBuilder<'a> {
             .as_mut()
             .map(|tparams| self.type_params(tparams));
 
-        let mut legacy = Some(LegacyTParamBuilder::new(tparams.is_some()));
+        let mut legacy = Some(LegacyTParamCollector::new(tparams.is_some()));
 
         // We need to bind all the parameters expressions _after_ the type params, but before the parameter names,
         // which might shadow some types.
@@ -248,9 +267,9 @@ impl<'a> BindingsBuilder<'a> {
         let return_ann_with_range = mem::take(&mut x.returns)
             .map(|e| self.to_return_annotation_with_range(*e, func_name, class_key, &mut legacy));
 
-        let legacy_tparam_builder = legacy.unwrap();
-        legacy_tparam_builder.add_name_definitions(self);
-        let legacy_tparams = legacy_tparam_builder.lookup_keys();
+        let legacy_tparam_collector = legacy.unwrap();
+        self.add_name_definitions(&legacy_tparam_collector);
+        let legacy_tparams = legacy_tparam_collector.lookup_keys();
         (return_ann_with_range, legacy_tparams)
     }
 
@@ -264,13 +283,23 @@ impl<'a> BindingsBuilder<'a> {
         body: Vec<Stmt>,
         range: TextRange,
         func_name: &Identifier,
+        parent: &NestingContext,
+        undecorated_idx: Idx<KeyUndecoratedFunction>,
         class_key: Option<Idx<KeyClass>>,
-    ) -> (YieldsAndReturns, Option<SelfAssignments>) {
+        is_async: bool,
+    ) -> (
+        YieldsAndReturns,
+        Option<SelfAssignments>,
+        Vec<UnusedParameter>,
+    ) {
         self.scopes
-            .push_function_scope(range, func_name, class_key.is_some());
-        self.parameters(parameters, class_key);
+            .push_function_scope(range, func_name, class_key.is_some(), is_async);
+        self.parameters(parameters, undecorated_idx, class_key);
         self.init_static_scope(&body, false);
-        self.stmts(body);
+        self.stmts(
+            body,
+            &NestingContext::function(ShortIdentifier::new(func_name), parent.dupe()),
+        );
         self.scopes.pop_function_scope()
     }
 
@@ -280,14 +309,16 @@ impl<'a> BindingsBuilder<'a> {
         body: Vec<Stmt>,
         range: TextRange,
         func_name: &Identifier,
+        undecorated_idx: Idx<KeyUndecoratedFunction>,
         class_key: Option<Idx<KeyClass>>,
+        is_async: bool,
     ) -> Option<SelfAssignments> {
         // Push a scope to create the parameter keys (but do nothing else with it).
         self.scopes
-            .push_function_scope(range, func_name, class_key.is_some());
-        self.parameters(parameters, class_key);
+            .push_function_scope(range, func_name, class_key.is_some(), is_async);
+        self.parameters(parameters, undecorated_idx, class_key);
         self.scopes.pop();
-        // If we are in a class, use a simple visiter to find `self.<attr>` assignments.
+        // If we are in a class, use a simple visitor to find `self.<attr>` assignments.
         if class_key.is_some() {
             SelfAttrNames::find(func_name, parameters, body)
         } else {
@@ -312,19 +343,16 @@ impl<'a> BindingsBuilder<'a> {
     }
 
     /// Handles both checking yield / return expressions and binding the return type.
-    ///
-    /// The `implicit_return_if_inferring_return_type` argument should be None when
-    /// return type inference is disabled; it must be `Some(implicit_return_key)` to
-    /// get return type inference.
     fn analyze_return_type(
         &mut self,
         func_name: &Identifier,
         is_async: bool,
         yields_and_returns: YieldsAndReturns,
         return_ann_with_range: Option<(TextRange, Idx<KeyAnnotation>)>,
-        implicit_return_if_inferring_return_type: Option<Idx<Key>>,
+        implicit_return: Option<Idx<Key>>,
+        should_infer_return_type: bool,
         stub_or_impl: FunctionStubOrImpl,
-        decorators: Box<[Idx<Key>]>,
+        decorators: Box<[(Idx<Key>, TextRange)]>,
     ) {
         let is_generator =
             !(yields_and_returns.yields.is_empty() && yields_and_returns.yield_froms.is_empty());
@@ -341,6 +369,7 @@ impl<'a> BindingsBuilder<'a> {
                         expr: x.value,
                         is_generator,
                         is_async,
+                        range: x.range,
                     }),
                 )
             })
@@ -354,22 +383,22 @@ impl<'a> BindingsBuilder<'a> {
         let yield_from_keys = yields_and_returns
             .yield_froms
             .into_map(|(idx, x)| {
-                self.insert_binding_idx(idx, BindingYieldFrom::YieldFrom(return_ann, x))
+                self.insert_binding_idx(
+                    idx,
+                    BindingYieldFrom::YieldFrom(return_ann, IsAsync::new(is_async), x),
+                )
             })
             .into_boxed_slice();
 
         let return_type_binding = {
-            let kind = match (
-                return_ann_with_range,
-                implicit_return_if_inferring_return_type,
-            ) {
+            let kind = match (return_ann_with_range, implicit_return) {
                 (Some((range, annotation)), Some(implicit_return)) => {
                     // We have an explicit return annotation and we want to validate it.
                     ReturnTypeKind::ShouldValidateAnnotation {
                         range,
                         annotation,
                         stub_or_impl,
-                        decorators,
+                        decorators: decorators.into_iter().map(|x| x.0).collect(),
                         implicit_return,
                         is_generator: !(yield_keys.is_empty() && yield_from_keys.is_empty()),
                         has_explicit_return: !return_keys.is_empty(),
@@ -382,7 +411,7 @@ impl<'a> BindingsBuilder<'a> {
                         is_generator: !(yield_keys.is_empty() && yield_from_keys.is_empty()),
                     }
                 }
-                (None, Some(implicit_return)) => {
+                (None, Some(implicit_return)) if should_infer_return_type => {
                     // We don't have an explicit return annotation, but we want to infer it.
                     ReturnTypeKind::ShouldInferType {
                         returns: return_keys,
@@ -391,8 +420,9 @@ impl<'a> BindingsBuilder<'a> {
                         yield_froms: yield_from_keys,
                     }
                 }
-                (None, None) => {
-                    // We don't have an explicit return annotation, and we want to just treat it as returning `Any`.
+                (None, _) => {
+                    // We don't have an explicit return annotation, or we don't want to infer return type.
+                    // Just treat the return type as `Any`.
                     ReturnTypeKind::ShouldReturnAny {
                         is_generator: !(yield_keys.is_empty() && yield_from_keys.is_empty()),
                     }
@@ -417,11 +447,13 @@ impl<'a> BindingsBuilder<'a> {
 
     fn decorators(&mut self, decorator_list: Vec<Decorator>, usage: &mut Usage) -> Decorators {
         let mut is_overload = false;
+        let mut is_override = false;
         let mut has_no_type_check = false;
         let mut is_abstract_method = false;
         for d in &decorator_list {
             let special_export = self.as_special_export(&d.expression);
             is_overload = is_overload || matches!(special_export, Some(SpecialExport::Overload));
+            is_override = is_override || matches!(special_export, Some(SpecialExport::Override));
             is_abstract_method =
                 is_abstract_method || matches!(special_export, Some(SpecialExport::AbstractMethod));
             has_no_type_check =
@@ -434,6 +466,7 @@ impl<'a> BindingsBuilder<'a> {
             has_no_type_check,
             is_overload,
             is_abstract_method,
+            is_override,
             decorators,
         }
     }
@@ -447,6 +480,8 @@ impl<'a> BindingsBuilder<'a> {
         is_async: bool,
         return_ann_with_range: Option<(TextRange, Idx<KeyAnnotation>)>,
         func_name: &Identifier,
+        parent: &NestingContext,
+        undecorated_idx: Idx<KeyUndecoratedFunction>,
         class_key: Option<Idx<KeyClass>>,
     ) -> (FunctionStubOrImpl, Option<SelfAssignments>) {
         let stub_or_impl = if (body.first().is_some_and(is_docstring)
@@ -459,24 +494,97 @@ impl<'a> BindingsBuilder<'a> {
             FunctionStubOrImpl::Impl
         };
 
+        let body_is_trivial = match body.as_slice() {
+            [Stmt::Pass(_)] => true,
+            // raise NotImplementedError(...)
+            [
+                Stmt::Raise(StmtRaise {
+                    exc: Some(box Expr::Call(ExprCall { box func, .. })),
+                    ..
+                }),
+            ] if self.as_special_export(func) == Some(SpecialExport::NotImplementedError) => true,
+            // return NotImplemented
+            [
+                Stmt::Return(StmtReturn {
+                    value: Some(box val),
+                    ..
+                }),
+            ] if self.as_special_export(val) == Some(SpecialExport::NotImplemented) => true,
+            _ => false,
+        };
+        let should_report_unused_parameters = stub_or_impl == FunctionStubOrImpl::Impl
+            && !body_is_trivial
+            && !decorators.is_overload
+            && !decorators.is_override
+            && !decorators.is_abstract_method
+            && !is_ellipse(&body);
+
         let self_assignments = if decorators.has_no_type_check
             || (self.untyped_def_behavior == UntypedDefBehavior::SkipAndInferReturnAny
                 && !is_annotated(&return_ann_with_range, parameters))
         {
             self.mark_as_returns_any(func_name);
-            self.unchecked_function_body_scope(parameters, body, range, func_name, class_key)
+            self.unchecked_function_body_scope(
+                parameters,
+                body,
+                range,
+                func_name,
+                undecorated_idx,
+                class_key,
+                is_async,
+            )
         } else {
             match self.untyped_def_behavior {
-                UntypedDefBehavior::SkipAndInferReturnAny
-                | UntypedDefBehavior::CheckAndInferReturnAny => {
-                    let (yields_and_returns, self_assignments) =
-                        self.function_body_scope(parameters, body, range, func_name, class_key);
+                UntypedDefBehavior::SkipAndInferReturnAny => {
+                    let (yields_and_returns, self_assignments, unused_parameters) = self
+                        .function_body_scope(
+                            parameters,
+                            body,
+                            range,
+                            func_name,
+                            parent,
+                            undecorated_idx,
+                            class_key,
+                            is_async,
+                        );
+                    if should_report_unused_parameters {
+                        self.record_unused_parameters(unused_parameters);
+                    }
                     self.analyze_return_type(
                         func_name,
                         is_async,
                         yields_and_returns,
                         return_ann_with_range,
-                        None, // this disables return type inference
+                        None,
+                        false, // this disables return type inference
+                        stub_or_impl,
+                        decorators.decorators.clone(),
+                    );
+                    self_assignments
+                }
+                UntypedDefBehavior::CheckAndInferReturnAny => {
+                    let implicit_return = self.implicit_return(&body, func_name);
+                    let (yields_and_returns, self_assignments, unused_parameters) = self
+                        .function_body_scope(
+                            parameters,
+                            body,
+                            range,
+                            func_name,
+                            parent,
+                            undecorated_idx,
+                            class_key,
+                            is_async,
+                        );
+                    if should_report_unused_parameters {
+                        self.record_unused_parameters(unused_parameters);
+                    }
+                    self.analyze_return_type(
+                        func_name,
+                        is_async,
+                        yields_and_returns,
+                        return_ann_with_range,
+                        Some(implicit_return),
+                        false, // this disables return type inference
                         stub_or_impl,
                         decorators.decorators.clone(),
                     );
@@ -484,14 +592,27 @@ impl<'a> BindingsBuilder<'a> {
                 }
                 UntypedDefBehavior::CheckAndInferReturnType => {
                     let implicit_return = self.implicit_return(&body, func_name);
-                    let (yields_and_returns, self_assignments) =
-                        self.function_body_scope(parameters, body, range, func_name, class_key);
+                    let (yields_and_returns, self_assignments, unused_parameters) = self
+                        .function_body_scope(
+                            parameters,
+                            body,
+                            range,
+                            func_name,
+                            parent,
+                            undecorated_idx,
+                            class_key,
+                            is_async,
+                        );
+                    if should_report_unused_parameters {
+                        self.record_unused_parameters(unused_parameters);
+                    }
                     self.analyze_return_type(
                         func_name,
                         is_async,
                         yields_and_returns,
                         return_ann_with_range,
                         Some(implicit_return),
+                        true,
                         stub_or_impl,
                         decorators.decorators.clone(),
                     );
@@ -503,13 +624,16 @@ impl<'a> BindingsBuilder<'a> {
         (stub_or_impl, self_assignments)
     }
 
-    pub fn function_def(&mut self, mut x: StmtFunctionDef) {
+    pub fn function_def(&mut self, mut x: StmtFunctionDef, parent: &NestingContext) {
         let func_name = x.name.clone();
         let mut def_idx =
             self.declare_current_idx(Key::Definition(ShortIdentifier::new(&func_name)));
 
+        let undecorated_idx =
+            self.idx_for_promise(KeyUndecoratedFunction(ShortIdentifier::new(&func_name)));
+
         // Get preceding function definition, if any. Used for building an overload type.
-        let (function_idx, pred_idx) = self.create_function_index(&x.name);
+        let (function_idx, pred_idx) = self.create_function_index(&func_name);
 
         let (class_key, metadata_key) = match self.scopes.current_class_and_metadata_keys() {
             Some((class_key, metadata_key)) => (Some(class_key), Some(metadata_key)),
@@ -531,6 +655,8 @@ impl<'a> BindingsBuilder<'a> {
             x.is_async,
             return_ann_with_range,
             &func_name,
+            parent,
+            undecorated_idx,
             class_key,
         );
 
@@ -539,21 +665,28 @@ impl<'a> BindingsBuilder<'a> {
         self.scopes.pop();
         self.scopes
             .record_self_assignments_if_applicable(self_assignments);
-
-        self.insert_binding_idx(
-            function_idx,
-            BindingFunction {
+        let undecorated_idx = self.insert_binding_idx(
+            undecorated_idx,
+            BindingUndecoratedFunction {
                 def: x,
                 stub_or_impl,
                 class_key,
                 decorators: decorators.decorators,
                 legacy_tparams: legacy_tparams.into_boxed_slice(),
+                module_style: self.module_info.path().style(),
+            },
+        );
+
+        self.insert_binding_idx(
+            function_idx,
+            BindingDecoratedFunction {
+                undecorated_idx,
                 successor: None,
                 docstring_range,
             },
         );
 
-        self.bind_definition_current(
+        self.bind_current_as(
             &func_name,
             def_idx,
             Binding::Function(function_idx, pred_idx, metadata_key),
@@ -584,6 +717,42 @@ fn function_last_expressions<'a>(
                 }
                 f(sys_info, &x.body, res)?;
             }
+            Stmt::While(x) => {
+                // Infinite loops with no breaks cannot fall through
+                if sys_info.evaluate_bool(&x.test) != Some(true) {
+                    return None;
+                }
+                let mut has_break = false;
+                fn f(stmt: &Stmt, res: &mut bool) {
+                    match stmt {
+                        Stmt::Break(_) => {
+                            *res = true;
+                        }
+                        Stmt::While(_) | Stmt::For(_) => {}
+                        _ => stmt.recurse(&mut |stmt| f(stmt, res)),
+                    }
+                }
+                x.body.visit(&mut |stmt| f(stmt, &mut has_break));
+                if has_break {
+                    return None;
+                }
+            }
+            Stmt::For(x) => {
+                let mut has_break = false;
+                fn f(stmt: &Stmt, res: &mut bool) {
+                    match stmt {
+                        Stmt::Break(_) => {
+                            *res = true;
+                        }
+                        Stmt::While(_) | Stmt::For(_) => {}
+                        _ => stmt.recurse(&mut |stmt| f(stmt, res)),
+                    }
+                }
+                x.body.visit(&mut |stmt| f(stmt, &mut has_break));
+                if has_break {
+                    return None;
+                }
+            }
             Stmt::If(x) => {
                 let mut last_test = None;
                 for (test, body) in sys_info.pruned_if_branches(x) {
@@ -596,7 +765,13 @@ fn function_last_expressions<'a>(
                 }
             }
             Stmt::Try(x) => {
-                if !x.finalbody.is_empty() {
+                // If final body is not empty, _and_ contains a return statement,
+                // process it.
+                if !x.finalbody.is_empty()
+                    && x.finalbody
+                        .iter()
+                        .any(|stmt| matches!(stmt, Stmt::Return(_)))
+                {
                     f(sys_info, &x.finalbody, res)?;
                 } else {
                     if x.orelse.is_empty() {
