@@ -5,14 +5,16 @@
  * LICENSE file in the root directory of this source tree.
  */
 
+pub mod base_class;
 pub mod binding;
 pub mod bindings;
 pub mod class;
+pub mod django;
 pub mod expr;
-pub mod flow_ops;
 pub mod function;
 pub mod narrow;
 pub mod pattern;
+pub mod pydantic;
 pub mod scope;
 pub mod stmt;
 pub mod table;
