@@ -31,7 +31,9 @@ use ruff_python_ast::UnaryOp;
 use ruff_python_ast::name::Name;
 use ruff_text_size::Ranged;
 use ruff_text_size::TextRange;
+use starlark_map::small_map::Entry;
 use starlark_map::small_map::SmallMap;
+use starlark_map::small_set::SmallSet;
 use vec1::Vec1;
 
 use crate::binding::bindings::BindingsBuilder;
@@ -42,7 +44,7 @@ use crate::types::facet::FacetKind;
 use crate::types::types::Type;
 
 assert_words!(AtomicNarrowOp, 11);
-assert_words!(NarrowOp, 12);
+assert_words!(NarrowOp, 13);
 
 #[derive(Clone, Debug)]
 pub enum AtomicNarrowOp {
@@ -54,6 +56,10 @@ pub enum AtomicNarrowOp {
     IsNotInstance(Expr),
     IsSubclass(Expr),
     IsNotSubclass(Expr),
+    HasAttr(Name),
+    NotHasAttr(Name),
+    GetAttr(Name, Option<Box<Expr>>),
+    NotGetAttr(Name, Option<Box<Expr>>),
     TypeGuard(Type, Arguments),
     NotTypeGuard(Type, Arguments),
     TypeIs(Type, Arguments),
@@ -86,7 +92,7 @@ pub enum AtomicNarrowOp {
 
 #[derive(Clone, Debug)]
 pub enum NarrowOp {
-    Atomic(Option<FacetChain>, AtomicNarrowOp),
+    Atomic(Option<FacetSubject>, AtomicNarrowOp),
     And(Vec<NarrowOp>),
     Or(Vec<NarrowOp>),
 }
@@ -106,6 +112,24 @@ impl DisplayWith<ModuleInfo> for AtomicNarrowOp {
             AtomicNarrowOp::IsNotSubclass(expr) => {
                 write!(f, "IsNotSubclass({})", expr.display_with(ctx))
             }
+            AtomicNarrowOp::HasAttr(attr) => {
+                write!(f, "HasAttr({attr})")
+            }
+            AtomicNarrowOp::NotHasAttr(attr) => {
+                write!(f, "NotHasAttr({attr})")
+            }
+            AtomicNarrowOp::GetAttr(attr, Some(default)) => {
+                write!(f, "GetAttr({}, {})", attr, default.display_with(ctx))
+            }
+            AtomicNarrowOp::NotGetAttr(attr, Some(default)) => {
+                write!(f, "NotGetAttr({}, {})", attr, default.display_with(ctx))
+            }
+            AtomicNarrowOp::GetAttr(attr, None) => {
+                write!(f, "GetAttr({attr}, None)")
+            }
+            AtomicNarrowOp::NotGetAttr(attr, None) => {
+                write!(f, "NotGetAttr({attr}, None)")
+            }
             AtomicNarrowOp::TypeGuard(t, arguments) => {
                 write!(f, "TypeGuard({t}, {})", arguments.display_with(ctx))
             }
@@ -152,7 +176,7 @@ impl DisplayWith<ModuleInfo> for NarrowOp {
         match self {
             Self::Atomic(prop, op) => match prop {
                 None => write!(f, "{}", op.display_with(ctx)),
-                Some(prop) => write!(f, "[{prop}] {}", op.display_with(ctx)),
+                Some(prop) => write!(f, "[{}] {}", prop.chain, op.display_with(ctx)),
             },
             Self::And(ops) => {
                 write!(
@@ -181,6 +205,10 @@ impl AtomicNarrowOp {
             Self::IsNotInstance(v) => Self::IsInstance(v.clone()),
             Self::IsSubclass(v) => Self::IsNotSubclass(v.clone()),
             Self::IsNotSubclass(v) => Self::IsSubclass(v.clone()),
+            Self::HasAttr(attr) => Self::NotHasAttr(attr.clone()),
+            Self::NotHasAttr(attr) => Self::HasAttr(attr.clone()),
+            Self::GetAttr(attr, default) => Self::NotGetAttr(attr.clone(), default.clone()),
+            Self::NotGetAttr(attr, default) => Self::GetAttr(attr.clone(), default.clone()),
             Self::Eq(v) => Self::NotEq(v.clone()),
             Self::NotEq(v) => Self::Eq(v.clone()),
             Self::In(v) => Self::NotIn(v.clone()),
@@ -206,19 +234,45 @@ impl AtomicNarrowOp {
     }
 }
 
+#[derive(Clone, Copy, Debug, PartialEq, Eq)]
+pub enum FacetOrigin {
+    // This facet is a direct access, like `x.y`, `x[0]`, `x["key"]`
+    Direct,
+    // This facet came from a call to a `get` method, like `x.get("key")`
+    GetMethod,
+}
+
+#[derive(Clone, Debug)]
+pub struct FacetSubject {
+    pub chain: FacetChain,
+    pub origin: FacetOrigin,
+}
+
 #[derive(Clone, Debug)]
 pub enum NarrowingSubject {
     Name(Name),
-    Facets(Name, FacetChain),
+    Facets(Name, FacetSubject),
 }
 
 impl NarrowingSubject {
     pub fn with_facet(&self, prop: FacetKind) -> Self {
         match self {
-            Self::Name(name) => Self::Facets(name.clone(), FacetChain::new(Vec1::new(prop))),
-            Self::Facets(name, props) => {
-                let props = Vec1::from_vec_push(props.facets().to_vec(), prop);
-                Self::Facets(name.clone(), FacetChain::new(props))
+            Self::Name(name) => Self::Facets(
+                name.clone(),
+                FacetSubject {
+                    chain: FacetChain::new(Vec1::new(prop)),
+                    origin: FacetOrigin::Direct,
+                },
+            ),
+            Self::Facets(name, facets) => {
+                let props = Vec1::from_vec_push(facets.chain.facets().to_vec(), prop);
+                Self::Facets(
+                    name.clone(),
+                    FacetSubject {
+                        chain: FacetChain::new(props),
+                        origin: facets.origin,
+                    },
+                )
             }
         }
     }
@@ -279,9 +333,31 @@ impl NarrowOps {
     }
 
     pub fn and_all(&mut self, other: Self) {
+        let mut seen = SmallSet::new();
         for (name, (op, range)) in other.0 {
+            seen.insert(name.clone());
             self.and(name, op, range);
         }
+        // For names present in `self` but not `other`, `And` their narrows with a placeholder
+        let unmerged_names: Vec<_> = self
+            .0
+            .keys()
+            .filter_map(|name| {
+                if seen.contains(name) {
+                    None
+                } else {
+                    Some(name.clone())
+                }
+            })
+            .collect();
+        for name in unmerged_names {
+            if let Entry::Occupied(mut entry) = self.0.entry(name) {
+                entry
+                    .get_mut()
+                    .0
+                    .and(NarrowOp::Atomic(None, AtomicNarrowOp::Placeholder));
+            }
+        }
     }
 
     fn or(&mut self, name: Name, op: NarrowOp, range: TextRange) {
@@ -290,9 +366,31 @@ impl NarrowOps {
     }
 
     pub fn or_all(&mut self, other: Self) {
+        let mut seen = SmallSet::new();
         for (name, (op, range)) in other.0 {
+            seen.insert(name.clone());
             self.or(name, op, range);
         }
+        // For names present in `self` but not `other`, `Or` their narrows with a placeholder
+        let unmerged_names: Vec<_> = self
+            .0
+            .keys()
+            .filter_map(|name| {
+                if seen.contains(name) {
+                    None
+                } else {
+                    Some(name.clone())
+                }
+            })
+            .collect();
+        for name in unmerged_names {
+            if let Entry::Occupied(mut entry) = self.0.entry(name) {
+                entry
+                    .get_mut()
+                    .0
+                    .or(NarrowOp::Atomic(None, AtomicNarrowOp::Placeholder));
+            }
+        }
     }
 
     pub fn from_single_narrow_op(left: &Expr, op: AtomicNarrowOp, range: TextRange) -> Self {
@@ -300,7 +398,7 @@ impl NarrowOps {
         for subject in expr_to_subjects(left) {
             let (name, prop) = match subject {
                 NarrowingSubject::Name(name) => (name, None),
-                NarrowingSubject::Facets(name, prop) => (name, Some(prop)),
+                NarrowingSubject::Facets(name, facets) => (name, Some(facets)),
             };
             if let Some((existing, _)) = narrow_ops.0.get_mut(&name) {
                 existing.and(NarrowOp::Atomic(prop, op.clone()));
@@ -321,7 +419,7 @@ impl NarrowOps {
         let mut narrow_ops = Self::new();
         let (name, prop) = match subject {
             NarrowingSubject::Name(name) => (name, None),
-            NarrowingSubject::Facets(name, prop) => (name, Some(prop)),
+            NarrowingSubject::Facets(name, facets) => (name, Some(facets)),
         };
         if let Some((existing, _)) = narrow_ops.0.get_mut(&name) {
             existing.and(NarrowOp::Atomic(prop, op.clone()));
@@ -342,8 +440,10 @@ impl NarrowOps {
                 ops: cmp_ops,
                 comparators,
             })) => {
-                // If the left expression is a call to len(), we're narrowing the argument
+                // If the left expression is a call to `len()` or `getattr()`, we're narrowing the first argument
                 let mut left = &**left;
+                // If the left expression is a call to `getattr()` we store attribute name and default
+                let mut getattr_name = None;
                 let mut special_export = None;
                 if let Expr::Call(ExprCall {
                     func, arguments, ..
@@ -354,23 +454,49 @@ impl NarrowOps {
                     special_export = builder.as_special_export(func);
                     if matches!(
                         special_export,
-                        Some(SpecialExport::Len | SpecialExport::Type)
+                        Some(SpecialExport::Len | SpecialExport::BuiltinsType)
                     ) {
                         left = &arguments.args[0];
                     }
                 }
+                // If we have something like `getattr(x, "attr") != None` or `getattr(x, "attr", None) is not None`
+                // we can perform a `hasattr(x, "attr")` narrow.
+                if let Expr::Call(ExprCall {
+                    func, arguments, ..
+                }) = left
+                    && arguments.keywords.is_empty()
+                    && (arguments.args.len() == 2
+                        || (arguments.args.len() == 3
+                            && matches!(arguments.args[2], Expr::NoneLiteral(_))))
+                    && let Expr::StringLiteral(ExprStringLiteral { value, .. }) = &arguments.args[1]
+                {
+                    special_export = builder.as_special_export(func);
+                    if matches!(special_export, Some(SpecialExport::GetAttr)) {
+                        left = &arguments.args[0];
+                        getattr_name = Some(Name::new(value.to_string()));
+                    }
+                }
                 let mut ops = cmp_ops
                     .iter()
                     .zip(comparators)
                     .filter_map(|(cmp_op, right)| {
                         let range = right.range();
                         let op = match (cmp_op, special_export) {
-                            (CmpOp::Is | CmpOp::Eq, Some(SpecialExport::Type)) => {
+                            (CmpOp::Is | CmpOp::Eq, Some(SpecialExport::BuiltinsType)) => {
                                 AtomicNarrowOp::TypeEq(right.clone())
                             }
-                            (CmpOp::IsNot | CmpOp::NotEq, Some(SpecialExport::Type)) => {
+                            (CmpOp::IsNot | CmpOp::NotEq, Some(SpecialExport::BuiltinsType)) => {
                                 AtomicNarrowOp::TypeNotEq(right.clone())
                             }
+                            (CmpOp::IsNot | CmpOp::NotEq, Some(SpecialExport::GetAttr))
+                                if matches!(right, Expr::NoneLiteral(_))
+                                    && let Some(attr) = &getattr_name =>
+                            {
+                                AtomicNarrowOp::HasAttr(attr.clone())
+                            }
+                            (_, Some(SpecialExport::GetAttr)) => {
+                                return None;
+                            }
                             (CmpOp::Is, None) => AtomicNarrowOp::Is(right.clone()),
                             (CmpOp::IsNot, None) => AtomicNarrowOp::IsNot(right.clone()),
                             (CmpOp::Eq, Some(SpecialExport::Len)) => {
@@ -446,6 +572,53 @@ impl NarrowOps {
             {
                 Self::from_single_narrow_op(&arguments.args[0], AtomicNarrowOp::IsTruthy, *range)
             }
+            Some(Expr::Call(ExprCall {
+                node_index: _,
+                range,
+                func,
+                arguments,
+            })) if builder.as_special_export(func) == Some(SpecialExport::HasAttr)
+                && arguments.args.len() == 2
+                && arguments.keywords.is_empty()
+                && let Expr::StringLiteral(ExprStringLiteral { value, .. }) =
+                    &arguments.args[1] =>
+            {
+                Self::from_single_narrow_op(
+                    &arguments.args[0],
+                    AtomicNarrowOp::HasAttr(Name::new(value.to_string())),
+                    *range,
+                )
+            }
+            Some(Expr::Call(ExprCall {
+                node_index: _,
+                range,
+                func,
+                arguments,
+            })) if builder.as_special_export(func) == Some(SpecialExport::GetAttr)
+                && (arguments.args.len() == 2 || arguments.args.len() == 3)
+                && arguments.keywords.is_empty()
+                && let Expr::StringLiteral(ExprStringLiteral { value, .. }) =
+                    &arguments.args[1] =>
+            {
+                Self::from_single_narrow_op(
+                    &arguments.args[0],
+                    AtomicNarrowOp::GetAttr(
+                        Name::new(value.to_string()),
+                        if arguments.args.len() == 2 {
+                            None
+                        } else {
+                            Some(Box::new(arguments.args[2].clone()))
+                        },
+                    ),
+                    *range,
+                )
+            }
+            Some(e @ Expr::Call(call)) if dict_get_subject_for_call_expr(call).is_some() => {
+                // When the guard is something like `x.get("key")`, we narrow it like `x["key"]` if `x` resolves to a dict
+                // in the answers step.
+                // This cannot be a TypeGuard/TypeIs function call, since the first argument is a string literal
+                Self::from_single_narrow_op(e, AtomicNarrowOp::IsTruthy, e.range())
+            }
             Some(Expr::Call(ExprCall {
                 node_index: _,
                 range,
@@ -460,6 +633,30 @@ impl NarrowOps {
                     *range,
                 )
             }
+            Some(Expr::Named(named)) => {
+                let mut target_narrow = Self::from_single_narrow_op(
+                    &named.target,
+                    AtomicNarrowOp::IsTruthy,
+                    named.target.range(),
+                );
+                let value_narrow = Self::from_expr(builder, Some(*named.value.clone()).as_ref());
+                // Merge the entries from the two `NarrowOps`
+                // We don't use `and_all` because it always generates placeholders when the entry is not present.
+                // This causes `Or` ops to be generated when the narrowing is negated, which is correct for
+                // unrelated narrows but undesirable here because we know these two narrows are either both true or both false.
+                for (name, (op, range)) in value_narrow.0 {
+                    let existing_entry = target_narrow.0.entry(name);
+                    match existing_entry {
+                        Entry::Occupied(mut entry) => {
+                            entry.get_mut().0.and(op.clone());
+                        }
+                        Entry::Vacant(entry) => {
+                            entry.insert((op, range));
+                        }
+                    };
+                }
+                target_narrow
+            }
             Some(e) => Self::from_single_narrow_op(e, AtomicNarrowOp::IsTruthy, e.range()),
             None => Self::new(),
         }
@@ -605,16 +802,63 @@ pub fn identifier_and_chain_prefix_for_expr(expr: &Expr) -> Option<(Identifier,
     f(expr, Vec::new())
 }
 
-fn subject_for_expr(expr: &Expr) -> Option<NarrowingSubject> {
-    identifier_and_chain_for_expr(expr)
-        .map(|(identifier, attr)| NarrowingSubject::Facets(identifier.id, attr))
+// Handle narrowing on `dict.get("key")`. During solving, if the resolved
+// type of the object is not a subtype of `dict`, we will not perform any narrowing.
+fn dict_get_subject_for_call_expr(call_expr: &ExprCall) -> Option<NarrowingSubject> {
+    let func = &call_expr.func;
+    let arguments = &call_expr.arguments;
+    if arguments.keywords.is_empty()
+        && arguments.args.len() == 1
+        && let Some(first_arg) = arguments.args.first()
+        && let Expr::Attribute(attr) = &**func
+        && attr.attr.id.as_str() == "get"
+        && let Expr::StringLiteral(ExprStringLiteral { value, .. }) = first_arg
+    {
+        let key = value.to_string();
+        if let Some((identifier, facets)) = identifier_and_chain_for_expr(&attr.value) {
+            // x.y.z.get("key")
+            let props = Vec1::from_vec_push(facets.facets().to_vec(), FacetKind::Key(key.clone()));
+            return Some(NarrowingSubject::Facets(
+                identifier.id,
+                FacetSubject {
+                    chain: FacetChain::new(props),
+                    origin: FacetOrigin::GetMethod,
+                },
+            ));
+        } else if let Expr::Name(name) = &*attr.value {
+            // x.get("key")
+            return Some(NarrowingSubject::Facets(
+                name.id.clone(),
+                FacetSubject {
+                    chain: FacetChain::new(Vec1::new(FacetKind::Key(key))),
+                    origin: FacetOrigin::GetMethod,
+                },
+            ));
+        }
+    }
+    None
 }
 
 pub fn expr_to_subjects(expr: &Expr) -> Vec<NarrowingSubject> {
     fn f(expr: &Expr, res: &mut Vec<NarrowingSubject>) {
         match expr {
             Expr::Name(name) => res.push(NarrowingSubject::Name(name.id.clone())),
-            Expr::Attribute(_) | Expr::Subscript(_) => res.extend(subject_for_expr(expr)),
+            Expr::Attribute(_) | Expr::Subscript(_) => {
+                if let Some((identifier, facets)) = identifier_and_chain_for_expr(expr) {
+                    res.push(NarrowingSubject::Facets(
+                        identifier.id,
+                        FacetSubject {
+                            chain: facets,
+                            origin: FacetOrigin::Direct,
+                        },
+                    ));
+                }
+            }
+            Expr::Call(call) => {
+                if let Some(subject) = dict_get_subject_for_call_expr(call) {
+                    res.push(subject);
+                }
+            }
             Expr::Named(ExprNamed { target, value, .. }) => {
                 f(target, res);
                 f(value, res);
