@@ -6,6 +6,7 @@
  */
 
 use pyrefly_python::ast::Ast;
+use pyrefly_python::nesting_context::NestingContext;
 use ruff_python_ast::AtomicNodeIndex;
 use ruff_python_ast::Expr;
 use ruff_python_ast::ExprNumberLiteral;
@@ -33,6 +34,7 @@ use crate::binding::narrow::expr_to_subjects;
 use crate::binding::scope::FlowStyle;
 use crate::config::error_kind::ErrorKind;
 use crate::error::context::ErrorInfo;
+use crate::export::special::SpecialExport;
 use crate::graph::index::Idx;
 use crate::types::facet::FacetKind;
 
@@ -46,7 +48,7 @@ impl<'a> BindingsBuilder<'a> {
     ) -> NarrowOps {
         // In typical code, match patterns are more like static types than normal values, so
         // we ignore match patterns for first-usage tracking.
-        let narrowing_usage = &mut Usage::Narrowing;
+        let narrowing_usage = &mut Usage::Narrowing(None);
         match pattern {
             Pattern::MatchValue(mut p) => {
                 self.ensure_expr(&mut p.value, narrowing_usage);
@@ -240,6 +242,34 @@ impl<'a> BindingsBuilder<'a> {
                 } else {
                     NarrowOps::new()
                 };
+
+                // Check if this is a single-positional-slot builtin type
+                // These types (bool, bytearray, bytes, dict, float, frozenset, int, list, set, str, tuple)
+                // bind the entire narrowed value when used with a single positional pattern
+                let is_single_slot_builtin = if let Expr::Name(name) = x.cls.as_ref() {
+                    SpecialExport::new(&name.id)
+                        .map(|se| se.is_single_positional_slot_builtin())
+                        .unwrap_or(false)
+                } else {
+                    false
+                };
+
+                // Handle positional patterns
+                if is_single_slot_builtin
+                    && x.arguments.patterns.len() == 1
+                    && x.arguments.keywords.is_empty()
+                {
+                    // For single-positional-slot builtins with exactly one positional pattern,
+                    // bind the pattern directly to the narrowed subject (like MatchAs)
+                    let pattern = x.arguments.patterns.into_iter().next().unwrap();
+                    narrow_ops.and_all(self.bind_pattern(
+                        match_subject.clone(),
+                        pattern,
+                        subject_idx,
+                    ));
+                    return narrow_ops;
+                }
+                // Normal MatchClass handling
                 // TODO: narrow class type vars based on pattern arguments
                 x.arguments
                     .patterns
@@ -279,18 +309,17 @@ impl<'a> BindingsBuilder<'a> {
             }
             Pattern::MatchOr(x) => {
                 let mut narrow_ops: Option<NarrowOps> = None;
-                let range = x.range;
-                let mut branches = Vec::new();
+                self.start_fork(x.range);
                 let n_subpatterns = x.patterns.len();
                 for (idx, pattern) in x.patterns.into_iter().enumerate() {
+                    self.start_branch();
                     if pattern.is_irrefutable() && idx != n_subpatterns - 1 {
                         self.error(
                             pattern.range(),
-                            ErrorInfo::Kind(ErrorKind::MatchError),
+                            ErrorInfo::Kind(ErrorKind::BadMatch),
                             "Only the last subpattern in MatchOr may be irrefutable".to_owned(),
                         )
                     }
-                    let mut base = self.scopes.clone_current_flow();
                     let new_narrow_ops =
                         self.bind_pattern(match_subject.clone(), pattern, subject_idx);
                     if let Some(ref mut ops) = narrow_ops {
@@ -298,25 +327,23 @@ impl<'a> BindingsBuilder<'a> {
                     } else {
                         narrow_ops = Some(new_narrow_ops);
                     }
-                    self.scopes.swap_current_flow_with(&mut base);
-                    branches.push(base);
+                    self.finish_branch();
                 }
-                self.set_current_flow_to_merged_branches(branches, range);
+                self.finish_match_or_fork();
                 narrow_ops.unwrap_or_default()
             }
             Pattern::MatchStar(_) => NarrowOps::new(),
         }
     }
 
-    pub fn stmt_match(&mut self, mut x: StmtMatch) {
+    pub fn stmt_match(&mut self, mut x: StmtMatch, parent: &NestingContext) {
         let mut subject = self.declare_current_idx(Key::Anon(x.subject.range()));
         self.ensure_expr(&mut x.subject, subject.usage());
         let subject_idx =
             self.insert_binding_current(subject, Binding::Expr(None, *x.subject.clone()));
         let match_narrowing_subject = expr_to_subjects(&x.subject).first().cloned();
         let mut exhaustive = false;
-        let range = x.range;
-        let mut branches = Vec::new();
+        self.start_fork(x.range);
         // Type narrowing operations that are carried over from one case to the next. For example, in:
         //   match x:
         //     case None:
@@ -327,33 +354,29 @@ impl<'a> BindingsBuilder<'a> {
         // is carried over to the fallback case.
         let mut negated_prev_ops = NarrowOps::new();
         for case in x.cases {
-            let mut base = self.scopes.clone_current_flow();
+            self.start_branch();
             if case.pattern.is_wildcard() || case.pattern.is_irrefutable() {
                 exhaustive = true;
             }
-            let new_narrow_ops =
+            self.bind_narrow_ops(&negated_prev_ops, case.range, &Usage::Narrowing(None));
+            let mut new_narrow_ops =
                 self.bind_pattern(match_narrowing_subject.clone(), case.pattern, subject_idx);
-            self.bind_narrow_ops(&negated_prev_ops, case.range);
-            self.bind_narrow_ops(&new_narrow_ops, case.range);
-            negated_prev_ops.and_all(new_narrow_ops.negate());
+            self.bind_narrow_ops(&new_narrow_ops, case.range, &Usage::Narrowing(None));
             if let Some(mut guard) = case.guard {
-                self.ensure_expr(&mut guard, &mut Usage::Narrowing);
-                let narrow_ops = NarrowOps::from_expr(self, Some(guard.as_ref()));
-                self.bind_narrow_ops(&narrow_ops, case.range);
+                self.ensure_expr(&mut guard, &mut Usage::Narrowing(None));
+                let guard_narrow_ops = NarrowOps::from_expr(self, Some(guard.as_ref()));
+                self.bind_narrow_ops(&guard_narrow_ops, case.range, &Usage::Narrowing(None));
                 self.insert_binding(Key::Anon(guard.range()), Binding::Expr(None, *guard));
+                new_narrow_ops.and_all(guard_narrow_ops)
             }
-            self.stmts(case.body);
-            self.scopes.swap_current_flow_with(&mut base);
-            branches.push(base);
+            negated_prev_ops.and_all(new_narrow_ops.negate());
+            self.stmts(case.body, parent);
+            self.finish_branch();
         }
-        // If the match branches cover all possibilities, then the flow after the match
-        // is just the merged branch flows.
-        //
-        // Otherwise, we need to merge the branches with the original `base` flow (which is current).
         if exhaustive {
-            self.set_current_flow_to_merged_branches(branches, range);
+            self.finish_exhaustive_fork();
         } else {
-            self.merge_branches_into_current(branches, range);
+            self.finish_non_exhaustive_fork(&negated_prev_ops);
         }
     }
 }
