@@ -5,7 +5,7 @@
  * LICENSE file in the root directory of this source tree.
  */
 
-use std::collections::HashMap;
+use std::collections::HashSet;
 use std::fmt;
 use std::fmt::Display;
 use std::fs::File;
@@ -23,23 +23,24 @@ use anstream::stdout;
 use anyhow::Context as _;
 use clap::Parser;
 use clap::ValueEnum;
-use dupe::Dupe;
+use dupe::Dupe as _;
+use pyrefly_build::handle::Handle;
 use pyrefly_config::args::ConfigOverrideArgs;
+use pyrefly_config::config::ConfigFile;
+use pyrefly_config::finder::ConfigError;
 use pyrefly_python::module_name::ModuleName;
 use pyrefly_python::module_path::ModulePath;
-use pyrefly_python::module_path::ModulePathDetails;
-use pyrefly_python::sys_info::SysInfo;
+use pyrefly_util::arc_id::ArcId;
 use pyrefly_util::display;
 use pyrefly_util::display::count;
 use pyrefly_util::display::number_thousands;
 use pyrefly_util::events::CategorizedEvents;
 use pyrefly_util::forgetter::Forgetter;
 use pyrefly_util::fs_anyhow;
-use pyrefly_util::globs::FilteredGlobs;
-use pyrefly_util::lined_buffer::LineNumber;
+use pyrefly_util::includes::Includes;
 use pyrefly_util::memory::MemoryUsageTrace;
-use pyrefly_util::prelude::SliceExt;
 use pyrefly_util::watcher::Watcher;
+use ruff_text_size::Ranged;
 use starlark_map::small_map::SmallMap;
 use starlark_map::small_set::SmallSet;
 use tracing::debug;
@@ -52,12 +53,11 @@ use crate::config::finder::ConfigFinder;
 use crate::error::error::Error;
 use crate::error::error::print_error_counts;
 use crate::error::legacy::LegacyErrors;
-use crate::error::summarise::print_error_summary;
+use crate::error::summarize::print_error_summary;
 use crate::error::suppress;
-use crate::module::from_path::module_from_path;
 use crate::module::typeshed::stdlib_search_path;
 use crate::report;
-use crate::state::handle::Handle;
+use crate::state::load::FileContents;
 use crate::state::require::Require;
 use crate::state::state::State;
 use crate::state::state::Transaction;
@@ -72,35 +72,32 @@ pub struct FullCheckArgs {
     pub files: FilesArgs,
 
     /// Watch for file changes and re-check them.
+    /// (Warning: This mode is highly experimental!)
     #[arg(long, conflicts_with = "check_all")]
     watch: bool,
 
     /// Type checking arguments and configuration
     #[command(flatten)]
-    pub args: CheckArgs,
+    args: CheckArgs,
+
+    /// Configuration override options
+    #[command(flatten, next_help_heading = "Config Overrides")]
+    pub config_override: ConfigOverrideArgs,
 }
 
 impl FullCheckArgs {
-    pub async fn run(self, allow_forget: bool) -> anyhow::Result<CommandExitStatus> {
-        self.args.config_override.validate()?;
-        let (files_to_check, config_finder) = self.files.resolve(&self.args.config_override)?;
-        run_check(
-            self.args,
-            self.watch,
-            files_to_check,
-            config_finder,
-            allow_forget,
-        )
-        .await
+    pub async fn run(self) -> anyhow::Result<CommandExitStatus> {
+        self.config_override.validate()?;
+        let (files_to_check, config_finder) = self.files.resolve(self.config_override)?;
+        run_check(self.args, self.watch, files_to_check, config_finder).await
     }
 }
 
 async fn run_check(
     args: CheckArgs,
     watch: bool,
-    files_to_check: FilteredGlobs,
+    files_to_check: Box<dyn Includes>,
     config_finder: ConfigFinder,
-    allow_forget: bool,
 ) -> anyhow::Result<CommandExitStatus> {
     if watch {
         let roots = files_to_check.roots();
@@ -113,7 +110,7 @@ async fn run_check(
             .await?;
         Ok(CommandExitStatus::Success)
     } else {
-        match args.run_once(files_to_check, config_finder, allow_forget) {
+        match args.run_once(files_to_check, config_finder) {
             Ok((status, _)) => Ok(status),
             Err(e) => Err(e),
         }
@@ -143,9 +140,6 @@ pub struct CheckArgs {
     /// Behavior-related configuration options
     #[command(flatten, next_help_heading = "Behavior")]
     behavior: BehaviorArgs,
-    /// Configuration override options
-    #[command(flatten, next_help_heading = "Config Overrides")]
-    pub config_override: ConfigOverrideArgs,
 }
 
 /// Arguments for snippet checking (excludes behavior args that don't apply to snippets)
@@ -171,8 +165,8 @@ pub struct SnippetCheckArgs {
 }
 
 impl SnippetCheckArgs {
-    pub async fn run(self, allow_forget: bool) -> anyhow::Result<CommandExitStatus> {
-        let (_, config_finder) = FilesArgs::get(vec![], self.config, &self.config_override)?;
+    pub async fn run(self) -> anyhow::Result<CommandExitStatus> {
+        let (_, config_finder) = FilesArgs::get(vec![], self.config, self.config_override)?;
         let check_args = CheckArgs {
             output: self.output,
             behavior: BehaviorArgs {
@@ -180,10 +174,10 @@ impl SnippetCheckArgs {
                 suppress_errors: false,
                 expectations: false,
                 remove_unused_ignores: false,
+                all: false,
             },
-            config_override: self.config_override,
         };
-        match check_args.run_once_with_snippet(self.code, config_finder, allow_forget) {
+        match check_args.run_once_with_snippet(self.code, config_finder) {
             Ok((status, _)) => Ok(status),
             Err(e) => Err(e),
         }
@@ -239,6 +233,16 @@ struct OutputArgs {
     )]
     summarize_errors: Option<usize>,
 
+    /// Filter the error summary to show only a specific error kind (e.g., bad-assignment, missing-return, etc.).
+    /// Must be used in conjunction with --summarize-errors.
+    #[arg(
+        long,
+        value_enum,
+        value_name = "ERROR_KIND",
+        hide_possible_values = true
+    )]
+    only: Option<crate::config::error_kind::ErrorKind>,
+
     /// By default show the number of errors. Pass `--summary` to show information about lines checked and time/memory,
     /// or `--summary=none` to hide the summary line entirely.
     #[arg(
@@ -251,10 +255,18 @@ struct OutputArgs {
     )]
     summary: Summary,
 
-    /// Omit the summary in the last line of the output.
-    /// Deprecated: will be removed in the next release. Use `--summary=none` instead.
+    /// When specified, strip this prefix from any paths in the output.
+    /// Pass "" to show absolute paths. When omitted, we will use the current working directory.
     #[arg(long)]
-    no_summary: bool,
+    relative_to: Option<String>,
+
+    /// Path to baseline file for comparing type errors
+    #[arg(long, value_name = "BASELINE_FILE")]
+    baseline: Option<PathBuf>,
+
+    /// When specified, emit a sorted/formatted JSON of the errors to the baseline file
+    #[arg(long, requires("baseline"))]
+    update_baseline: bool,
 }
 
 #[derive(Clone, Debug, ValueEnum, Default, PartialEq, Eq)]
@@ -281,69 +293,94 @@ struct BehaviorArgs {
     /// Remove unused ignores from the input files.
     #[arg(long)]
     remove_unused_ignores: bool,
+    /// If we are removing unused ignores, should we remove all unused ignores or only Pyrefly specific `pyrefly: ignore`s?
+    #[arg(long, requires("remove_unused_ignores"))]
+    all: bool,
 }
 
 impl OutputFormat {
     fn write_error_text_to_file(
         path: &Path,
+        relative_to: &Path,
         errors: &[Error],
         verbose: bool,
     ) -> anyhow::Result<()> {
         let mut file = BufWriter::new(File::create(path)?);
         for e in errors {
-            e.write_line(&mut file, verbose)?;
+            e.write_line(&mut file, relative_to, verbose)?;
         }
         file.flush()?;
         Ok(())
     }
 
-    fn write_error_text_to_console(errors: &[Error], verbose: bool) -> anyhow::Result<()> {
+    fn write_error_text_to_console(
+        relative_to: &Path,
+        errors: &[Error],
+        verbose: bool,
+    ) -> anyhow::Result<()> {
         for error in errors {
-            error.print_colors(verbose);
+            error.print_colors(relative_to, verbose);
         }
         Ok(())
     }
 
-    fn write_error_json(writer: &mut impl Write, errors: &[Error]) -> anyhow::Result<()> {
-        let legacy_errors = LegacyErrors::from_errors(errors);
+    fn write_error_json(
+        writer: &mut impl Write,
+        relative_to: &Path,
+        errors: &[Error],
+    ) -> anyhow::Result<()> {
+        let legacy_errors = LegacyErrors::from_errors(relative_to, errors);
         serde_json::to_writer_pretty(writer, &legacy_errors)?;
         Ok(())
     }
 
-    fn buffered_write_error_json(writer: impl Write, errors: &[Error]) -> anyhow::Result<()> {
+    fn buffered_write_error_json(
+        writer: impl Write,
+        relative_to: &Path,
+        errors: &[Error],
+    ) -> anyhow::Result<()> {
         let mut writer = BufWriter::new(writer);
-        Self::write_error_json(&mut writer, errors)?;
+        Self::write_error_json(&mut writer, relative_to, errors)?;
         writer.flush()?;
         Ok(())
     }
 
-    fn write_error_json_to_file(path: &Path, errors: &[Error]) -> anyhow::Result<()> {
-        fn f(path: &Path, errors: &[Error]) -> anyhow::Result<()> {
+    fn write_error_json_to_file(
+        path: &Path,
+        relative_to: &Path,
+        errors: &[Error],
+    ) -> anyhow::Result<()> {
+        fn f(path: &Path, relative_to: &Path, errors: &[Error]) -> anyhow::Result<()> {
             let file = File::create(path)?;
-            OutputFormat::buffered_write_error_json(file, errors)
+            OutputFormat::buffered_write_error_json(file, relative_to, errors)
         }
-        f(path, errors)
+        f(path, relative_to, errors)
             .with_context(|| format!("while writing JSON errors to `{}`", path.display()))
     }
 
-    fn write_error_json_to_console(errors: &[Error]) -> anyhow::Result<()> {
-        Self::buffered_write_error_json(stdout(), errors)
+    fn write_error_json_to_console(relative_to: &Path, errors: &[Error]) -> anyhow::Result<()> {
+        Self::buffered_write_error_json(stdout(), relative_to, errors)
     }
 
-    fn write_errors_to_file(&self, path: &Path, errors: &[Error]) -> anyhow::Result<()> {
+    fn write_errors_to_file(
+        &self,
+        path: &Path,
+        relative_to: &Path,
+        errors: &[Error],
+    ) -> anyhow::Result<()> {
         match self {
-            Self::MinText => Self::write_error_text_to_file(path, errors, false),
-            Self::FullText => Self::write_error_text_to_file(path, errors, true),
-            Self::Json => Self::write_error_json_to_file(path, errors),
+            Self::MinText => Self::write_error_text_to_file(path, relative_to, errors, false),
+            Self::FullText => Self::write_error_text_to_file(path, relative_to, errors, true),
+            Self::Json => Self::write_error_json_to_file(path, relative_to, errors),
             Self::OmitErrors => Ok(()),
         }
     }
 
-    fn write_errors_to_console(&self, errors: &[Error]) -> anyhow::Result<()> {
+    fn write_errors_to_console(&self, relative_to: &Path, errors: &[Error]) -> anyhow::Result<()> {
         match self {
-            Self::MinText => Self::write_error_text_to_console(errors, false),
-            Self::FullText => Self::write_error_text_to_console(errors, true),
-            Self::Json => Self::write_error_json_to_console(errors),
+            Self::MinText => Self::write_error_text_to_console(relative_to, errors, false),
+            Self::FullText => Self::write_error_text_to_console(relative_to, errors, true),
+            Self::Json => Self::write_error_json_to_console(relative_to, errors),
             Self::OmitErrors => Ok(()),
         }
     }
@@ -353,64 +390,65 @@ impl OutputFormat {
 pub struct Handles {
     /// A mapping from a file to all other information needed to create a `Handle`.
     /// The value type is basically everything else in `Handle` except for the file path.
-    path_data: HashMap<PathBuf, (ModuleName, SysInfo)>,
+    path_data: HashSet<ModulePath>,
 }
 
 impl Handles {
-    pub fn new(files: Vec<PathBuf>, config_finder: &ConfigFinder) -> Self {
+    pub fn new(files: Vec<PathBuf>) -> Self {
         let mut handles = Self {
-            path_data: HashMap::new(),
+            path_data: HashSet::new(),
         };
         for file in files {
-            handles.register_file(file, config_finder);
+            handles.path_data.insert(ModulePath::filesystem(file));
         }
         handles
     }
 
-    fn register_file(
-        &mut self,
-        path: PathBuf,
+    pub fn all(
+        &self,
         config_finder: &ConfigFinder,
-    ) -> &(ModuleName, SysInfo) {
-        let module_path = ModulePath::filesystem(path.clone());
-        let unknown = ModuleName::unknown();
-        let config = config_finder.python_file(unknown, &module_path);
+    ) -> (Vec<Handle>, SmallSet<ArcId<ConfigFile>>, Vec<ConfigError>) {
+        let mut configs = SmallMap::new();
+        for path in &self.path_data {
+            let unknown = ModuleName::unknown();
+            configs
+                .entry(config_finder.python_file(unknown, path))
+                .or_insert_with(SmallSet::new)
+                .insert(path.dupe());
+        }
 
-        let search_path = config.search_path();
-        let module_name = module_from_path(&path, search_path).unwrap_or(unknown);
-
-        self.path_data
-            .entry(path)
-            .or_insert((module_name, config.get_sys_info()))
-    }
-
-    pub fn all(&self, specified_require: Require) -> Vec<(Handle, Require)> {
-        self.path_data
+        let mut errors = Vec::new();
+        let mut reloaded_configs = SmallSet::new();
+        for (config, files) in &configs {
+            match config.requery_source_db(files) {
+                Ok(reload) if reload => {
+                    reloaded_configs.insert(config.dupe());
+                }
+                Err(error) => {
+                    errors.push(ConfigError::error(error));
+                }
+                _ => (),
+            }
+        }
+        let result = configs
             .iter()
-            .map(|(path, (module_name, runtime_metadata))| {
-                (
-                    Handle::new(
-                        module_name.dupe(),
-                        ModulePath::filesystem(path.to_path_buf()),
-                        runtime_metadata.dupe(),
-                    ),
-                    specified_require,
-                )
-            })
-            .collect()
+            .flat_map(|(c, files)| files.iter().map(|p| c.handle_from_module_path(p.dupe())))
+            .collect();
+        (result, reloaded_configs, errors)
     }
 
     fn update<'a>(
         &mut self,
         created_files: impl Iterator<Item = &'a PathBuf>,
         removed_files: impl Iterator<Item = &'a PathBuf>,
-        config_finder: &ConfigFinder,
     ) {
         for file in created_files {
-            self.register_file(file.to_path_buf(), config_finder);
+            self.path_data
+                .insert(ModulePath::filesystem(file.to_path_buf()));
         }
         for file in removed_files {
-            self.path_data.remove(file);
+            self.path_data
+                .remove(&ModulePath::filesystem(file.to_path_buf()));
         }
     }
 }
@@ -422,7 +460,7 @@ struct RequireLevels {
 
 async fn get_watcher_events(watcher: &mut Watcher) -> anyhow::Result<CategorizedEvents> {
     loop {
-        let events = CategorizedEvents::new(
+        let events = CategorizedEvents::new_notify(
             watcher
                 .wait()
                 .await
@@ -499,23 +537,10 @@ impl Timings {
 }
 
 impl CheckArgs {
-    pub fn get_handles(
-        self,
-        files_to_check: FilteredGlobs,
-        config_finder: &ConfigFinder,
-    ) -> anyhow::Result<Vec<(Handle, Require)>> {
-        let handles = Handles::new(
-            config_finder.checkpoint(files_to_check.files())?,
-            config_finder,
-        );
-        Ok(handles.all(self.get_required_levels().specified))
-    }
-
     pub fn run_once(
         self,
-        files_to_check: FilteredGlobs,
+        files_to_check: Box<dyn Includes>,
         config_finder: ConfigFinder,
-        allow_forget: bool,
     ) -> anyhow::Result<(CommandExitStatus, Vec<Error>)> {
         let mut timings = Timings::new();
         let list_files_start = Instant::now();
@@ -530,19 +555,22 @@ impl CheckArgs {
             return Ok((CommandExitStatus::Success, Vec::new()));
         }
 
-        let holder = Forgetter::new(State::new(config_finder), allow_forget);
-        let handles = Handles::new(expanded_file_list, holder.as_ref().config_finder());
+        let holder = Forgetter::new(State::new(config_finder), true);
+        let handles = Handles::new(expanded_file_list);
         let require_levels = self.get_required_levels();
         let mut transaction = Forgetter::new(
             holder
                 .as_ref()
                 .new_transaction(require_levels.default, None),
-            allow_forget,
+            true,
         );
+        let (loaded_handles, _, sourcedb_errors) = handles.all(holder.as_ref().config_finder());
         self.run_inner(
             timings,
             transaction.as_mut(),
-            &handles.all(require_levels.specified),
+            &loaded_handles,
+            sourcedb_errors,
+            require_levels.specified,
         )
     }
 
@@ -550,14 +578,13 @@ impl CheckArgs {
         self,
         code: String,
         config_finder: ConfigFinder,
-        allow_forget: bool,
     ) -> anyhow::Result<(CommandExitStatus, Vec<Error>)> {
         // Create a virtual module path for the snippet
         let path = PathBuf::from_str("snippet")?;
         let module_path = ModulePath::memory(path);
         let module_name = ModuleName::from_str("__main__");
 
-        let holder = Forgetter::new(State::new(config_finder), allow_forget);
+        let holder = Forgetter::new(State::new(config_finder), true);
 
         // Create a single handle for the virtual module
         let sys_info = holder
@@ -572,41 +599,49 @@ impl CheckArgs {
             holder
                 .as_ref()
                 .new_transaction(require_levels.default, None),
-            allow_forget,
+            true,
         );
 
         // Add the snippet source to the transaction's memory
         transaction.as_mut().set_memory(vec![(
             PathBuf::from(module_path.as_path()),
-            Some(Arc::new(code)),
+            Some(Arc::new(FileContents::from_source(code))),
         )]);
 
         self.run_inner(
             Timings::new(),
             transaction.as_mut(),
-            &[(handle, require_levels.specified)],
+            &[handle],
+            vec![],
+            require_levels.specified,
         )
     }
 
     pub async fn run_watch(
         self,
         mut watcher: Watcher,
-        files_to_check: FilteredGlobs,
+        files_to_check: Box<dyn Includes>,
         config_finder: ConfigFinder,
     ) -> anyhow::Result<()> {
         // TODO: We currently make 1 unrealistic assumptions, which should be fixed in the future:
         // - Config search is stable across incremental runs.
         let expanded_file_list = config_finder.checkpoint(files_to_check.files())?;
         let require_levels = self.get_required_levels();
-        let mut handles = Handles::new(expanded_file_list, &config_finder);
+        let mut handles = Handles::new(expanded_file_list);
         let state = State::new(config_finder);
         let mut transaction = state.new_committable_transaction(require_levels.default, None);
         loop {
             let timings = Timings::new();
+            let (loaded_handles, reloaded_configs, sourcedb_errors) =
+                handles.all(state.config_finder());
+            let mut_transaction = transaction.as_mut();
+            mut_transaction.invalidate_find_for_configs(reloaded_configs);
             let res = self.run_inner(
                 timings,
-                transaction.as_mut(),
-                &handles.all(require_levels.specified),
+                mut_transaction,
+                &loaded_handles,
+                sourcedb_errors,
+                require_levels.specified,
             );
             state.commit_transaction(transaction);
             if let Err(e) = res {
@@ -624,7 +659,6 @@ impl CheckArgs {
             handles.update(
                 events.created.iter().filter(|p| files_to_check.covers(p)),
                 events.removed.iter().filter(|p| files_to_check.covers(p)),
-                state.config_finder(),
             );
         }
     }
@@ -655,45 +689,78 @@ impl CheckArgs {
         &self,
         mut timings: Timings,
         transaction: &mut Transaction,
-        handles: &[(Handle, Require)],
+        handles: &[Handle],
+        mut sourcedb_errors: Vec<ConfigError>,
+        require: Require,
     ) -> anyhow::Result<(CommandExitStatus, Vec<Error>)> {
         let mut memory_trace = MemoryUsageTrace::start(Duration::from_secs_f32(0.1));
 
         let type_check_start = Instant::now();
         transaction.set_subscriber(Some(Box::new(ProgressBarSubscriber::new())));
-        transaction.run(handles);
+        transaction.run(handles, require);
         transaction.set_subscriber(None);
 
         let loads = if self.behavior.check_all {
             transaction.get_all_errors()
         } else {
-            transaction.get_errors(handles.iter().map(|(handle, _)| handle))
+            transaction.get_errors(handles)
         };
         timings.type_check = type_check_start.elapsed();
 
         let report_errors_start = Instant::now();
-        let config_errors = transaction.get_config_errors();
+        let mut config_errors = transaction.get_config_errors();
+        config_errors.append(&mut sourcedb_errors);
         let config_errors_count = config_errors.len();
         for error in config_errors {
             error.print();
         }
 
-        let errors = loads.collect_errors();
+        let relative_to = self.output.relative_to.as_ref().map_or_else(
+            || std::env::current_dir().ok().unwrap_or_default(),
+            |x| PathBuf::from_str(x.as_str()).unwrap(),
+        );
+
+        let errors = loads
+            .collect_errors_with_baseline(self.output.baseline.as_deref(), relative_to.as_path());
+
+        // We update the baseline file if requested, after reporting any new errors using the old baseline
+        if self.output.update_baseline
+            && let Some(baseline_path) = &self.output.baseline
+        {
+            let mut new_baseline = errors.shown.clone();
+            new_baseline.extend(errors.baseline);
+            new_baseline.sort_by_cached_key(|error| {
+                (
+                    error.path().to_string(),
+                    error.range().start(),
+                    error.range().end(),
+                    error.error_kind(),
+                )
+            });
+            OutputFormat::write_error_json_to_file(
+                baseline_path,
+                relative_to.as_path(),
+                &new_baseline,
+            )?;
+        }
+
         if let Some(path) = &self.output.output {
-            self.output
-                .output_format
-                .write_errors_to_file(path, &errors.shown)?;
+            self.output.output_format.write_errors_to_file(
+                path,
+                relative_to.as_path(),
+                &errors.shown,
+            )?;
         } else {
             self.output
                 .output_format
-                .write_errors_to_console(&errors.shown)?;
+                .write_errors_to_console(relative_to.as_path(), &errors.shown)?;
         }
         memory_trace.stop();
         if let Some(limit) = self.output.count_errors {
             print_error_counts(&errors.shown, limit);
         }
-        if let Some(path_index) = self.output.summarize_errors {
-            print_error_summary(&errors.shown, path_index);
+        if self.output.summarize_errors.is_some() {
+            print_error_summary(&errors.shown, self.output.only);
         }
         let mut shown_errors_count = config_errors_count;
         for error in &errors.shown {
@@ -703,27 +770,32 @@ impl CheckArgs {
         }
         timings.report_errors = report_errors_start.elapsed();
 
-        if self.output.summary != Summary::None && !self.output.no_summary {
-            let ignored = errors.disabled.len() + errors.suppressed.len();
-            if ignored == 0 {
+        if self.output.summary != Summary::None {
+            let suppress_count = errors.suppressed.len();
+            if suppress_count == 0 {
                 info!("{}", count(shown_errors_count, "error"))
             } else {
                 info!(
-                    "{} ({} ignored)",
+                    "{} ({} suppressed)",
                     count(shown_errors_count, "error"),
-                    number_thousands(ignored)
+                    number_thousands(suppress_count)
                 )
             };
         }
         if self.output.summary == Summary::Full {
+            let user_handles: HashSet<&Handle> = handles.iter().collect();
+            let (user_lines, dep_lines) = transaction.split_line_count(&user_handles);
             info!(
-                "{} ({}, {}); took {timings}; memory ({})",
+                "{} ({}); {} ({} in your project, {} in dependencies); \
+                took {timings}; memory ({})",
                 count(handles.len(), "module"),
                 count(
                     transaction.module_count() - handles.len(),
                     "dependent module"
                 ),
-                count(transaction.line_count(), "line"),
+                count(user_lines + dep_lines, "line"),
+                count(user_lines, "line"),
+                count(dep_lines, "line"),
                 memory_trace.peak()
             );
         }
@@ -737,18 +809,16 @@ impl CheckArgs {
             let is_javascript = debug_info.extension() == Some("js".as_ref());
             fs_anyhow::write(
                 debug_info,
-                report::debug_info::debug_info(
-                    transaction,
-                    &handles.map(|x| x.0.dupe()),
-                    is_javascript,
-                ),
+                report::debug_info::debug_info(transaction, handles, is_javascript),
             )?;
         }
         if let Some(glean) = &self.output.report_glean {
             fs_anyhow::create_dir_all(glean)?;
-            for (handle, _) in handles {
+            for handle in handles {
+                // Generate a safe filename using hash to avoid OS filename length limits
+                let module_hash = blake3::hash(handle.module().to_string().as_bytes());
                 fs_anyhow::write(
-                    &glean.join(format!("{}.json", handle.module())),
+                    &glean.join(format!("{}.json", &module_hash)),
                     report::glean::glean(transaction, handle),
                 )?;
             }
@@ -763,39 +833,10 @@ impl CheckArgs {
             fs_anyhow::write(path, report::trace::trace(transaction))?;
         }
         if self.behavior.suppress_errors {
-            let mut errors_to_suppress: SmallMap<PathBuf, Vec<Error>> = SmallMap::new();
-            let shown_errors = errors.shown.clone();
-            for e in shown_errors {
-                if e.severity() >= Severity::Warn
-                    && let ModulePathDetails::FileSystem(path) = e.path().details()
-                {
-                    errors_to_suppress.entry(path.clone()).or_default().push(e);
-                }
-            }
-            suppress::suppress_errors(&errors_to_suppress);
+            suppress::suppress_errors(errors.shown.clone());
         }
         if self.behavior.remove_unused_ignores {
-            let mut all_ignores: SmallMap<&PathBuf, SmallSet<LineNumber>> = SmallMap::new();
-            for (module_path, ignore) in loads.collect_ignores() {
-                if let ModulePathDetails::FileSystem(path) = module_path.details() {
-                    all_ignores.insert(path, ignore.get_pyrefly_ignores());
-                }
-            }
-
-            let mut suppressed_errors: SmallMap<&PathBuf, SmallSet<LineNumber>> = SmallMap::new();
-            for e in &errors.suppressed {
-                if e.is_ignored(false)
-                    && let ModulePathDetails::FileSystem(path) = e.path().details()
-                {
-                    suppressed_errors
-                        .entry(path)
-                        .or_default()
-                        .insert(e.display_range().start.line);
-                }
-            }
-
-            let unused_ignores = suppress::find_unused_ignores(all_ignores, suppressed_errors);
-            suppress::remove_unused_ignores(unused_ignores);
+            suppress::remove_unused_ignores(&loads, self.behavior.all);
         }
         if self.behavior.expectations {
             loads.check_against_expectations()?;
