@@ -10,21 +10,19 @@ use std::path::Path;
 use std::path::PathBuf;
 
 use clap::Parser;
+use pyrefly_config::args::ConfigOverrideArgs;
 use pyrefly_config::file_kind::ConfigFileKind;
 use pyrefly_config::migration::run::config_migration;
 use pyrefly_config::pyproject::PyProject;
 use pyrefly_util::absolutize::Absolutize as _;
-use pyrefly_util::display;
 use pyrefly_util::fs_anyhow;
 use tracing::error;
 use tracing::info;
-use tracing::warn;
 
 use crate::commands::check;
 use crate::commands::files::FilesArgs;
 use crate::commands::util::CommandExitStatus;
 use crate::config::config::ConfigFile;
-use crate::error::summarise;
 
 const MAX_ERRORS_TO_PROMPT_SUPPRESSION: usize = 100;
 
@@ -89,10 +87,6 @@ impl InitArgs {
                     else if error_count <= MAX_ERRORS_TO_PROMPT_SUPPRESSION {
                         return self.prompt_error_suppression(config_path, error_count);
                     }
-                    // 3b. Prompt error suppression in specific subdirectories if there are more than the maximum number of errors
-                    else {
-                        return self.prompt_init_on_subdirectory(config_path, errors);
-                    }
                 }
                 Ok(CommandExitStatus::Success)
             }
@@ -110,10 +104,10 @@ impl InitArgs {
 
         // Use get to get the filtered globs and config finder
         let (filtered_globs, config_finder) =
-            FilesArgs::get(Vec::new(), config_path, &check_args.config_override)?;
+            FilesArgs::get(Vec::new(), config_path, ConfigOverrideArgs::default())?;
 
         // Run the check directly
-        match check_args.run_once(filtered_globs, config_finder, true) {
+        match check_args.run_once(filtered_globs, config_finder) {
             Ok((status, errors)) => Ok((status, errors)),
             Err(e) => {
                 error!("Failed to run pyrefly check: {}", e);
@@ -127,8 +121,9 @@ impl InitArgs {
         config_path: Option<PathBuf>,
         error_count: usize,
     ) -> anyhow::Result<CommandExitStatus> {
-        let prompt =
-            format!("Found {error_count} errors. Would you like to suppress them? (y/N): ");
+        let prompt = format!(
+            "Found {error_count} errors. We can add suppression comments (e.g., `pyrefly: ignore`) to silence them for you. Would you like to suppress them? (y/N): "
+        );
 
         if Self::prompt_user_confirmation(&prompt) {
             info!("Running pyrefly check with suppress-errors flag...");
@@ -144,10 +139,10 @@ impl InitArgs {
 
             // Use get to get the filtered globs and config finder
             let (suppress_globs, suppress_config_finder) =
-                FilesArgs::get(Vec::new(), config_path, &suppress_args.config_override)?;
+                FilesArgs::get(Vec::new(), config_path, ConfigOverrideArgs::default())?;
 
             // Run the check with suppress-errors flag
-            match suppress_args.run_once(suppress_globs, suppress_config_finder, true) {
+            match suppress_args.run_once(suppress_globs, suppress_config_finder) {
                 Ok(_) => return Ok(CommandExitStatus::Success),
                 Err(e) => {
                     error!("Failed to run pyrefly check with suppress-errors: {}", e);
@@ -159,193 +154,6 @@ impl InitArgs {
         Ok(CommandExitStatus::Success)
     }
 
-    /// Disables typechecking for all repos except the selected directories by:
-    /// Setting project_includes to only include the selected directories
-    fn disable_typechecking_for_repos_other_than_selected_files(
-        &self,
-        config_path: &Option<PathBuf>,
-        selected_dirs: &[PathBuf],
-    ) -> anyhow::Result<()> {
-        if let Some(root_config_path) = config_path {
-            let mut include_patterns = Vec::new();
-
-            for dir in selected_dirs {
-                let config_dir = root_config_path.parent().unwrap_or(Path::new("/"));
-
-                let rel_path = if dir.starts_with(config_dir) {
-                    match dir.strip_prefix(config_dir) {
-                        Ok(rel) => rel.to_string_lossy().to_string(),
-                        Err(_) => dir.to_string_lossy().to_string(), // Fallback to absolute path
-                    }
-                } else {
-                    dir.to_string_lossy().to_string()
-                };
-
-                let glob_pattern = if rel_path.is_empty() {
-                    "**".to_owned()
-                } else {
-                    format!("{rel_path}/**")
-                };
-
-                include_patterns.push(glob_pattern);
-            }
-
-            // Read the existing config to preserve any other settings
-            let (mut existing_config, _) = ConfigFile::from_file(root_config_path);
-
-            // Update only the project_includes field
-            existing_config.project_includes = pyrefly_util::globs::Globs::new(include_patterns)?;
-
-            // Handle differently based on config file type
-            if root_config_path.ends_with(ConfigFile::PYPROJECT_FILE_NAME) {
-                // For pyproject.toml, use write_pyproject to update only the pyrefly section
-                // This preserves other tool configurations in the file
-                PyProject::update(root_config_path, existing_config)?;
-                info!(
-                    "Updated pyrefly section in pyproject.toml to focus typechecking on selected directories"
-                );
-            } else {
-                // For pyrefly.toml, write the updated config back to the file
-                let serialized = toml::to_string_pretty(&existing_config)?;
-                fs_anyhow::write(root_config_path, serialized)?;
-                info!("Updated pyrefly.toml to focus typechecking on selected directories");
-            }
-
-            info!(
-                "Updated root config at {} to focus typechecking on selected directories",
-                root_config_path.display()
-            );
-        }
-
-        Ok(())
-    }
-
-    fn prompt_init_on_subdirectory(
-        &self,
-        config_path: Option<PathBuf>,
-        errors: Vec<crate::error::error::Error>,
-    ) -> anyhow::Result<CommandExitStatus> {
-        // Get the top directories by error count
-        let (best_path_index, dirs_to_show) =
-            summarise::get_top_error_dirs_for_init(&errors, config_path.as_ref());
-
-        // Print the top directories
-        info!("Top 10 Directories by Error Count:");
-        if !dirs_to_show.is_empty() {
-            info!("  (Using path_index = {} for grouping)", best_path_index);
-
-            // Take the top 10 directories with <= 100 errors
-            for (i, (dir, error_count)) in dirs_to_show.iter().enumerate() {
-                info!(
-                    "  {}) {}: {}",
-                    i + 1,
-                    dir.display(),
-                    display::count(*error_count, "error")
-                );
-            }
-        } else {
-            error!("  No directories found with <= 100 errors.");
-            return Ok(CommandExitStatus::Success);
-        }
-
-        // Prompt user to select directories to suppress errors in
-        let prompt = "Enter directory numbers to suppress errors (comma-separated, e.g. 1,3,5), or press Enter to skip: ";
-        let selected_indices = Self::read_input_comma_separated_values(prompt, 1, 10);
-
-        // If no valid directories were selected, return success
-        if selected_indices.is_empty() {
-            error!("No valid directory numbers provided. Skipping error suppression.");
-            return Ok(CommandExitStatus::Success);
-        }
-
-        // Get the selected directories
-        let selected_dirs: Vec<PathBuf> = selected_indices
-            .iter()
-            .filter_map(|&idx| {
-                if idx < dirs_to_show.len() {
-                    Some(dirs_to_show[idx - 1].0.clone())
-                } else {
-                    None
-                }
-            })
-            .collect();
-
-        // Print selected directories
-        let dirs_str = if selected_dirs.len() == 1 {
-            format!("directory {}", selected_dirs[0].display())
-        } else {
-            format!(
-                "directories {}",
-                selected_dirs
-                    .iter()
-                    .map(|d| d.display().to_string())
-                    .collect::<Vec<_>>()
-                    .join(", ")
-            )
-        };
-
-        // Run check with suppress-errors for the selected directories
-        info!(
-            "Running pyrefly check with suppress-errors flag for selected directories: {}",
-            dirs_str
-        );
-
-        // Create check args with suppress-errors flag
-        let suppress_args = check::CheckArgs::parse_from([
-            "check",
-            "--suppress-errors",
-            "--output-format",
-            "omit-errors",
-            "--summary=none",
-        ]);
-
-        // Collect file paths from errors in selected directories
-        let mut files_to_check: Vec<String> = Vec::new();
-        for error in &errors {
-            let error_path = PathBuf::from(error.path().to_string());
-            if selected_dirs.iter().any(|dir| error_path.starts_with(dir)) {
-                // Convert PathBuf to String
-                files_to_check.push(error_path.to_string_lossy().into_owned());
-            }
-        }
-
-        // If there are no files to check in the selected directories, return success
-        if files_to_check.is_empty() {
-            error!("No errors found in the selected directories.");
-            return Ok(CommandExitStatus::Success);
-        }
-
-        // Use get to get the filtered globs and config finder, passing the files to check
-        let (suppress_globs, suppress_config_finder) = FilesArgs::get(
-            files_to_check,
-            config_path.clone(),
-            &suppress_args.config_override,
-        )?;
-
-        // Run the check with suppress-errors flag
-        match suppress_args.run_once(suppress_globs, suppress_config_finder, true) {
-            Ok(_) => {}
-            Err(e) => {
-                error!("Failed to suppress errors: {}", e);
-            }
-        };
-
-        // Disable typechecking for all repos except the selected ones
-        if let Err(e) = self
-            .disable_typechecking_for_repos_other_than_selected_files(&config_path, &selected_dirs)
-        {
-            error!("Failed to configure typechecking: {}", e);
-        }
-        info!(
-            "Disabled typechecking for all directories except the selected {}",
-            dirs_str
-        );
-        info!(
-            "To enable typechecking in other directories in the future, please expand project-include in pyproject.toml."
-        );
-        Ok(CommandExitStatus::Success)
-    }
-
     fn create_config(&self) -> anyhow::Result<(CommandExitStatus, Option<PathBuf>)> {
         let path = self.path.absolutize();
 
@@ -380,7 +188,6 @@ impl InitArgs {
         // This prevents us from simply outputting an empty file, and gives the user somewhere to start if they want to customize.
         let cfg = ConfigFile {
             project_includes: ConfigFile::default_project_includes(),
-            project_excludes: ConfigFile::default_project_excludes(),
             ..Default::default()
         };
 
@@ -426,50 +233,6 @@ impl InitArgs {
         std::io::stdin().read_line(&mut input).ok();
         input
     }
-
-    fn read_input_comma_separated_values(prompt: &str, min: usize, max: usize) -> Vec<usize> {
-        let input = Self::read_from_stdin(prompt);
-        Self::parse_comma_separated_values(&input, min, max)
-    }
-
-    /// Parses comma-separated values from a string and returns a vector of parsed values
-    /// within the specified range.
-    ///
-    /// # Arguments
-    /// * `input` - The input string containing comma-separated values
-    /// * `min` - The minimum valid value (inclusive)
-    /// * `max` - The maximum valid value (inclusive)
-    ///
-    /// # Returns
-    /// A vector of parsed values within the specified range
-    fn parse_comma_separated_values(input: &str, min: usize, max: usize) -> Vec<usize> {
-        let input = input.trim();
-
-        // If input is empty, return empty vector
-        if input.is_empty() {
-            return Vec::new();
-        }
-
-        // Parse comma-separated values
-        let selected_indices: Vec<usize> = input
-            .split(',')
-            .filter_map(|s| {
-                let trimmed = s.trim();
-                match trimmed.parse::<usize>() {
-                    Ok(num) if (min..=max).contains(&num) => Some(num),
-                    _ => {
-                        warn!(
-                            "'{}' is not a valid number ({}-{}), skipping.",
-                            trimmed, min, max
-                        );
-                        None
-                    }
-                }
-            })
-            .collect();
-
-        selected_indices
-    }
 }
 
 #[cfg(test)]
@@ -822,39 +585,4 @@ k = [\"v\"]
         assert_user_error(status);
         Ok(())
     }
-
-    // Test for parse_comma_separated_values with multiple test cases
-    #[test]
-    fn test_parse_comma_separated_values() {
-        // Define test cases as (input, min, max, expected_result)
-        let test_cases = [
-            // Empty input
-            ("", 1, 10, vec![]),
-            // Single valid number
-            ("5", 1, 10, vec![5]),
-            // Multiple valid numbers
-            ("1,3,5,10", 1, 10, vec![1, 3, 5, 10]),
-            // Input with spaces
-            (" 1, 3 , 5,  10 ", 1, 10, vec![1, 3, 5, 10]),
-            // Out of range values
-            ("0,11,15", 1, 10, vec![]),
-            // Mixed valid and invalid values
-            ("0,3,11,5", 1, 10, vec![3, 5]),
-            // Non-numeric values
-            ("a,b,c", 1, 10, vec![]),
-            // Mixed numeric and non-numeric values
-            ("1,a,3,b,5", 1, 10, vec![1, 3, 5]),
-            // Custom range
-            ("4,5,10,15,16", 5, 15, vec![5, 10, 15]),
-        ];
-
-        // Run each test case
-        for (i, (input, min, max, expected)) in test_cases.iter().enumerate() {
-            let result = InitArgs::parse_comma_separated_values(input, *min, *max);
-            assert_eq!(
-                result, *expected,
-                "Test case {i} failed: input='{input}', min={min}, max={max}",
-            );
-        }
-    }
 }
