@@ -6,15 +6,20 @@
  */
 
 pub mod all;
-pub mod autotype;
 pub mod buck_check;
 pub mod check;
 pub mod config_finder;
 pub mod dump_config;
 pub mod files;
 #[cfg(not(target_arch = "wasm32"))]
+pub mod infer;
+#[cfg(not(target_arch = "wasm32"))]
 pub mod init;
 #[cfg(not(target_arch = "wasm32"))]
 pub mod lsp;
 #[cfg(not(target_arch = "wasm32"))]
+pub mod report;
+#[cfg(not(target_arch = "wasm32"))]
+pub mod tsp;
+#[cfg(not(target_arch = "wasm32"))]
 pub mod util;
