@@ -5,8 +5,6 @@
  * LICENSE file in the root directory of this source tree.
  */
 
-use std::cmp;
-
 use pyrefly_python::ast::Ast;
 use pyrefly_python::docstring::Docstring;
 use pyrefly_python::dunder;
@@ -16,6 +14,7 @@ use pyrefly_python::short_identifier::ShortIdentifier;
 use pyrefly_python::symbol_kind::SymbolKind;
 use pyrefly_python::sys_info::SysInfo;
 use pyrefly_util::visit::Visit;
+use ruff_python_ast::Decorator;
 use ruff_python_ast::ExceptHandler;
 use ruff_python_ast::Expr;
 use ruff_python_ast::ExprAttribute;
@@ -25,26 +24,45 @@ use ruff_python_ast::Identifier;
 use ruff_python_ast::Operator;
 use ruff_python_ast::Pattern;
 use ruff_python_ast::Stmt;
+use ruff_python_ast::StmtClassDef;
 use ruff_python_ast::StmtExpr;
+use ruff_python_ast::StmtFunctionDef;
 use ruff_python_ast::name::Name;
 use ruff_text_size::TextRange;
 use starlark_map::small_map::Entry;
 use starlark_map::small_map::SmallMap;
 use starlark_map::small_set::SmallSet;
 
-use crate::types::globals::Global;
+use crate::export::special::SpecialExport;
+use crate::types::globals::ImplicitGlobal;
+
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, PartialOrd, Ord)]
+pub enum MutableCaptureKind {
+    /// Mutable capture coming from a `global` statement
+    Global,
+    /// Mutable capture coming from a `nonlocal` statement
+    Nonlocal,
+}
 
 /// How a name is defined. If a name is defined outside of this
-/// module, we additionally store the module we got it from
-#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, PartialOrd, Ord)]
+/// module, we additionally store the module we got it from.
+///
+/// This type is ordered - if there are multiple statements defining
+/// the name, then the minimal style is the one we track.
+#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
 pub enum DefinitionStyle {
-    /// Defined in this module, e.g. `x = 1` or `def x(): ...`
-    /// We also store what kind of symbol it is
-    Local(SymbolKind),
+    /// A name defined by a mutable capture. In valid code, the current scope
+    /// must be nested in some enclosing scope that defines the name.
+    MutableCapture(MutableCaptureKind),
+    /// An annotated definition or declaration, e.g. `x: int = 1` or `x: int`
+    Annotated(SymbolKind, ShortIdentifier),
+    /// An unannotated definition in this module, e.g. `x = 1` or `def x(): ...`
+    Unannotated(SymbolKind),
     /// Defined as an implicit global like `__name__`.
-    Global,
+    ImplicitGlobal,
     /// Imported with an alias, e.g. `from x import y as z`
-    ImportAs(ModuleName),
+    /// Name is the previous name before the alias
+    ImportAs(ModuleName, Name),
     /// Imported with an alias, where the alias is identical, e.g. `from x import y as y`
     ImportAsEq(ModuleName),
     /// Imported from another module, e.g. `from x import y`
@@ -53,35 +71,62 @@ pub enum DefinitionStyle {
     ImportModule(ModuleName),
     /// A relative import that does not exist: we do `....` more than we have depth
     ImportInvalidRelative,
+    /// A statement like `del x` defines `x` in the current scope, even if `x` has no
+    /// other definition.
+    Delete,
 }
 
 #[derive(Debug, Clone)]
 pub struct Definition {
-    /// A location where the definition is defined, there is no guarantee it is the first/last or otherwise.
-    pub range: TextRange,
     /// If the definition occurs multiple times, the lowest `DefinitionStyle` is used (e.g. prefer `Local`).
     pub style: DefinitionStyle,
-    /// The location of the first annotated name for this definition, if any.
-    pub annot: Option<ShortIdentifier>,
-    /// The number is the distinct times this variable was defined.
-    pub count: usize,
+    /// A location where the name is defined. Always matches the source of `self.style`.
+    pub range: TextRange,
+    /// Does this definition require an `Anywhere` binding at binding time? Typically yes if there
+    /// are multiple definitions, but mutable captures and `del` both require special handling.
+    pub needs_anywhere: bool,
     /// If the first statement in a definition (class, function) is a string literal, PEP 257 convention
-    /// states that is is the docstring.
+    /// states that is the docstring.
     pub docstring_range: Option<TextRange>,
 }
 
+impl Definition {
+    pub fn annotation(&self) -> Option<ShortIdentifier> {
+        match &self.style {
+            DefinitionStyle::Annotated(_, ann) => Some(*ann),
+            _ => None,
+        }
+    }
+
+    fn merge(&mut self, other: DefinitionStyle, range: TextRange) {
+        // To ensure binding code cannot produce invalid lookups, we ensure that
+        // `self.style` and `self.range` always match.
+        if other < self.style {
+            self.style = other;
+            self.range = range;
+        }
+        // If we've merged a Definition, then there are multiple definition sites.
+        //
+        // We want an Anywhere at bindings time unless either:
+        // - it is defined only by `del` statements (in which case we have to
+        //   avoid `Anywhere` because no one will ever actually populate it - this
+        //   is an implementation detail but can lead to panics if mis-handled).
+        // - this is a mutable capture (in which case it's not actually owned
+        //   by the current scope)
+        self.needs_anywhere = match &self.style {
+            DefinitionStyle::MutableCapture(..) | DefinitionStyle::Delete => false,
+            _ => true,
+        };
+    }
+}
+
 /// Find the definitions available in a scope. Does not traverse inside classes/functions,
 /// since they are separate scopes.
 #[derive(Debug, Clone, Default)]
 pub struct Definitions {
-    /// All the things defined in this module.
+    /// All the names defined in this scope, including mutable captures
+    /// (`global` and `nonlocal` declarations)
     pub definitions: SmallMap<Name, Definition>,
-    /// All the global names declared in this scope.
-    /// TODO(grievejia): This field is currently unused, but may be useful in the future.
-    pub globals: SmallSet<Name>,
-    /// All the nonlocal names declared in this scope.
-    /// TODO(grievejia): This field is currently unused, but may be useful in the future.
-    pub nonlocals: SmallSet<Name>,
     /// All the modules that are imported with `from x import *`.
     pub import_all: SmallMap<ModuleName, TextRange>,
     /// The `__all__` variable contents.
@@ -90,6 +135,10 @@ pub struct Definitions {
     /// that are guaranteed to be imported under `foo` when `foo` is itself imported in downstream
     /// files.
     pub implicitly_imported_submodules: SmallSet<Name>,
+    /// Deprecated names that are defined in this module.
+    pub deprecated: SmallSet<Name>,
+    /// Special exports defined in this module
+    pub special_exports: SmallMap<Name, SpecialExport>,
 }
 
 #[derive(Debug, Clone, PartialEq, Eq, Hash)]
@@ -154,6 +203,23 @@ fn implicitly_imported_submodule(
         .cloned()
 }
 
+fn is_deprecated_decorator(decorator: &Decorator) -> bool {
+    decorator.expression.as_call_expr().is_some_and(|x| {
+        x.func.as_name_expr().is_some_and(|x| {
+            x.id == "deprecated"
+                || x.id == "warnings.deprecated"
+                || x.id == "typing_extensions.deprecated"
+        })
+    })
+}
+
+fn is_overload_decorator(decorator: &Decorator) -> bool {
+    decorator
+        .expression
+        .as_name_expr()
+        .is_some_and(|x| x.id == "overload" || x.id == "typing.overload")
+}
+
 impl Definitions {
     pub fn new(x: &[Stmt], module_name: ModuleName, is_init: bool, sys_info: &SysInfo) -> Self {
         let mut builder = DefinitionsBuilder {
@@ -167,19 +233,22 @@ impl Definitions {
     }
 
     /// Add an implicit `from builtins import *` to the definitions.
+    /// Additional user-defined builtins are imported from `__builtins__.pyi`
     pub fn inject_builtins(&mut self) {
         self.import_all.entry(ModuleName::builtins()).or_default();
+        self.import_all
+            .entry(ModuleName::extra_builtins())
+            .or_default();
     }
 
-    pub fn inject_globals(&mut self) {
-        for global in Global::globals(false) {
+    pub fn inject_implicit_globals(&mut self) {
+        for global in ImplicitGlobal::implicit_globals(false) {
             self.definitions.insert(
                 global.name().clone(),
                 Definition {
                     range: TextRange::default(),
-                    style: DefinitionStyle::Global,
-                    annot: None,
-                    count: 1,
+                    style: DefinitionStyle::ImplicitGlobal,
+                    needs_anywhere: false,
                     docstring_range: None,
                 },
             );
@@ -202,7 +271,9 @@ impl Definitions {
                 && (style == ModuleStyle::Executable
                     || matches!(
                         def.style,
-                        DefinitionStyle::Local(_) | DefinitionStyle::ImportAsEq(_)
+                        DefinitionStyle::Annotated(..)
+                            | DefinitionStyle::Unannotated(..)
+                            | DefinitionStyle::ImportAsEq(_)
                     ))
             {
                 self.dunder_all
@@ -211,7 +282,7 @@ impl Definitions {
         }
     }
 
-    /// Add these names to `duner_all`, if they are defined in the module.
+    /// Add these names to `dunder_all`, if they are defined in the module.
     pub fn extend_dunder_all(&mut self, extra: &[Name]) {
         for name in extra {
             if let Some(def) = self.definitions.get(name) {
@@ -234,41 +305,37 @@ impl<'a> DefinitionsBuilder<'a> {
         x: &Name,
         range: TextRange,
         style: DefinitionStyle,
-        annot: Option<ShortIdentifier>,
         body: Option<&[Stmt]>,
     ) {
         match self.inner.definitions.entry(x.clone()) {
             Entry::Occupied(mut e) => {
-                e.get_mut().style = cmp::min(e.get().style, style);
-                if e.get().annot.is_none() {
-                    e.get_mut().annot = annot;
-                }
-                e.get_mut().count += 1;
+                e.get_mut().merge(style, range);
             }
             Entry::Vacant(e) => {
                 e.insert(Definition {
                     range,
                     style,
-                    annot,
-                    count: 1,
+                    needs_anywhere: false,
                     docstring_range: body.and_then(Docstring::range_from_stmts),
                 });
             }
         }
     }
 
-    fn add_name(
-        &mut self,
-        x: &Name,
-        range: TextRange,
-        style: DefinitionStyle,
-        annot: Option<ShortIdentifier>,
-    ) {
-        self.add_name_with_body(x, range, style, annot, None)
+    fn add_name(&mut self, x: &Name, range: TextRange, style: DefinitionStyle) {
+        if matches!(
+            style,
+            DefinitionStyle::Annotated(..) | DefinitionStyle::Unannotated(..)
+        ) && let Some(special_export) = SpecialExport::new(x)
+            && special_export.defined_in(self.module_name)
+        {
+            self.inner.special_exports.insert(x.clone(), special_export);
+        }
+        self.add_name_with_body(x, range, style, None)
     }
 
     fn add_identifier(&mut self, x: &Identifier, style: DefinitionStyle) {
-        self.add_name(&x.id, x.range, style, None);
+        self.add_name(&x.id, x.range, style);
     }
 
     fn add_identifier_with_body(
@@ -277,7 +344,7 @@ impl<'a> DefinitionsBuilder<'a> {
         style: DefinitionStyle,
         body: Option<&[Stmt]>,
     ) {
-        self.add_name_with_body(&x.id, x.range, style, None, body);
+        self.add_name_with_body(&x.id, x.range, style, body);
     }
 
     fn expr_lvalue(&mut self, x: &Expr) {
@@ -285,8 +352,7 @@ impl<'a> DefinitionsBuilder<'a> {
             self.add_name(
                 &x.id,
                 x.range,
-                DefinitionStyle::Local(SymbolKind::Variable),
-                None,
+                DefinitionStyle::Unannotated(SymbolKind::Variable),
             )
         };
         Ast::expr_lvalue(x, &mut add_name);
@@ -295,7 +361,7 @@ impl<'a> DefinitionsBuilder<'a> {
 
     fn pattern(&mut self, x: &Pattern) {
         Ast::pattern_lvalue(x, &mut |x| {
-            self.add_identifier(x, DefinitionStyle::Local(SymbolKind::Variable))
+            self.add_identifier(x, DefinitionStyle::Unannotated(SymbolKind::Variable))
         });
     }
 
@@ -315,14 +381,13 @@ impl<'a> DefinitionsBuilder<'a> {
                             &imported_module.first_component(),
                             a.name.range,
                             DefinitionStyle::ImportModule(imported_module),
-                            None,
                         ),
                         Some(alias) => self.add_identifier(
                             alias,
                             if alias.id == a.name.id {
                                 DefinitionStyle::ImportAsEq(imported_module)
                             } else {
-                                DefinitionStyle::ImportAs(imported_module)
+                                DefinitionStyle::ImportAs(imported_module, a.name.id.clone())
                             },
                         ),
                     };
@@ -353,38 +418,59 @@ impl<'a> DefinitionsBuilder<'a> {
                                 if a.asname.as_ref().map(|x| &x.id) == Some(&a.name.id) {
                                     DefinitionStyle::ImportAsEq(name)
                                 } else if a.asname.is_some() {
-                                    DefinitionStyle::ImportAs(name)
+                                    DefinitionStyle::ImportAs(name, a.name.id.clone())
                                 } else {
                                     DefinitionStyle::Import(name)
                                 }
                             }
                         };
-                        self.add_identifier(a.asname.as_ref().unwrap_or(&a.name), style);
-                        if matches!(style, DefinitionStyle::ImportAsEq(_))
+                        if matches!(&style, &DefinitionStyle::ImportAsEq(_))
                             && a.name.id == dunder::ALL
                             && let Some(module) = name
                         {
                             self.inner.dunder_all = vec![DunderAllEntry::Module(x.range, module)]
                         }
+                        self.add_identifier(a.asname.as_ref().unwrap_or(&a.name), style);
                     }
                 }
             }
-            Stmt::ClassDef(x) => {
+            Stmt::ClassDef(StmtClassDef {
+                name,
+                body,
+                decorator_list,
+                ..
+            }) => {
+                // If the class is decorated with `@deprecated`, we mark it as deprecated.
+                let mut is_deprecated = false;
+                for d in decorator_list {
+                    is_deprecated = is_deprecated || is_deprecated_decorator(d);
+                }
+                if is_deprecated {
+                    self.inner.deprecated.insert(name.id.clone());
+                }
                 self.add_identifier_with_body(
-                    &x.name,
-                    DefinitionStyle::Local(SymbolKind::Class),
-                    Some(&x.body),
+                    name,
+                    DefinitionStyle::Unannotated(SymbolKind::Class),
+                    Some(body),
                 );
                 return; // These things are inside a scope
             }
             Stmt::Nonlocal(x) => {
                 for name in &x.names {
-                    self.inner.nonlocals.insert(name.id.clone());
+                    self.add_name(
+                        &name.id,
+                        name.range,
+                        DefinitionStyle::MutableCapture(MutableCaptureKind::Nonlocal),
+                    );
                 }
             }
             Stmt::Global(x) => {
                 for name in &x.names {
-                    self.inner.globals.insert(name.id.clone());
+                    self.add_name(
+                        &name.id,
+                        name.range,
+                        DefinitionStyle::MutableCapture(MutableCaptureKind::Global),
+                    );
                 }
             }
             Stmt::Assign(x) => {
@@ -396,21 +482,52 @@ impl<'a> DefinitionsBuilder<'a> {
                     }
                 }
             }
+            Stmt::AnnAssign(x) => {
+                if let Some(value) = &x.value {
+                    self.named_in_expr(value);
+                }
+                if let Some(v) = &x.value
+                    && DunderAllEntry::is_all(&x.target)
+                {
+                    self.inner.dunder_all = DunderAllEntry::as_list(v.as_ref());
+                }
+                match &*x.target {
+                    Expr::Name(x) => {
+                        self.add_name(
+                            &x.id,
+                            x.range,
+                            DefinitionStyle::Annotated(
+                                SymbolKind::Variable,
+                                ShortIdentifier::expr_name(x),
+                            ),
+                        );
+                    }
+                    _ => self.expr_lvalue(&x.target),
+                }
+            }
             Stmt::AugAssign(x) => {
                 self.named_in_expr(&x.value);
                 if DunderAllEntry::is_all(&x.target) && x.op == Operator::Add {
                     self.inner
                         .dunder_all
                         .extend(DunderAllEntry::as_list(&x.value));
-                } else if let Expr::Name(name) = &*x.target {
+                }
+                if let Expr::Name(name) = &*x.target {
                     self.add_name(
                         &name.id,
                         name.range,
-                        DefinitionStyle::Local(SymbolKind::Variable),
-                        None,
+                        DefinitionStyle::Unannotated(SymbolKind::Variable),
                     )
                 }
             }
+            Stmt::Delete(x) => {
+                for target in &x.targets {
+                    self.named_in_expr(target);
+                    if let Expr::Name(name) = target {
+                        self.add_name(&name.id, name.range, DefinitionStyle::Delete)
+                    }
+                }
+            }
             Stmt::Expr(StmtExpr { value, .. }) => {
                 self.named_in_expr(value);
                 if let Expr::Call(
@@ -446,33 +563,33 @@ impl<'a> DefinitionsBuilder<'a> {
                     }
                 }
             }
-            Stmt::AnnAssign(x) => {
-                if let Some(value) = &x.value {
-                    self.named_in_expr(value);
-                }
-                match &*x.target {
-                    Expr::Name(x) => {
-                        self.add_name(
-                            &x.id,
-                            x.range,
-                            DefinitionStyle::Local(SymbolKind::Variable),
-                            Some(ShortIdentifier::expr_name(x)),
-                        );
-                    }
-                    _ => self.expr_lvalue(&x.target),
-                }
-            }
             Stmt::TypeAlias(x) => {
                 self.named_in_expr(&x.value);
                 if matches!(&*x.name, Expr::Name(_)) {
                     self.expr_lvalue(&x.name)
                 }
             }
-            Stmt::FunctionDef(x) => {
+            Stmt::FunctionDef(StmtFunctionDef {
+                name,
+                body,
+                decorator_list,
+                ..
+            }) => {
+                let mut is_overload = false;
+                let mut is_deprecated = false;
+                for d in decorator_list {
+                    is_overload = is_overload || is_overload_decorator(d);
+                    is_deprecated = is_deprecated || is_deprecated_decorator(d);
+                }
+                // If the function is not an overload and decorated with
+                // `@deprecated`, we mark it as deprecated.
+                if is_deprecated && !is_overload {
+                    self.inner.deprecated.insert(name.id.clone());
+                }
                 self.add_identifier_with_body(
-                    &x.name,
-                    DefinitionStyle::Local(SymbolKind::Function),
-                    Some(&x.body),
+                    name,
+                    DefinitionStyle::Unannotated(SymbolKind::Function),
+                    Some(body),
                 );
                 return; // don't recurse because a separate scope
             }
@@ -501,7 +618,7 @@ impl<'a> DefinitionsBuilder<'a> {
                             if let Some(name) = &x.name {
                                 self.add_identifier(
                                     name,
-                                    DefinitionStyle::Local(SymbolKind::Variable),
+                                    DefinitionStyle::Unannotated(SymbolKind::Variable),
                                 );
                             }
                         }
@@ -533,7 +650,6 @@ impl<'a> DefinitionsBuilder<'a> {
                 }
             }
             Stmt::Return(..)
-            | Stmt::Delete(..)
             | Stmt::Pass(..)
             | Stmt::Break(..)
             | Stmt::Continue(..)
@@ -548,7 +664,11 @@ impl<'a> DefinitionsBuilder<'a> {
             Expr::Named(expr_named) => {
                 self.expr_lvalue(&expr_named.target);
             }
-            Expr::Lambda(..) | Expr::SetComp(..) | Expr::DictComp(..) | Expr::ListComp(..) => {
+            Expr::Lambda(..)
+            | Expr::SetComp(..)
+            | Expr::DictComp(..)
+            | Expr::ListComp(..)
+            | Expr::Generator(..) => {
                 // These expressions define a scope, so walrus operators only define a name
                 // within that scope, not in the surrounding statement's scope.
             }
@@ -560,6 +680,7 @@ impl<'a> DefinitionsBuilder<'a> {
 #[cfg(test)]
 mod tests {
     use pyrefly_util::prelude::SliceExt;
+    use ruff_python_ast::PySourceType;
 
     use super::*;
 
@@ -611,7 +732,7 @@ mod tests {
         is_init: bool,
     ) -> Definitions {
         let mut res = Definitions::new(
-            &Ast::parse(contents).0.body,
+            &Ast::parse(contents, PySourceType::Python).0.body,
             module_name,
             is_init,
             &SysInfo::default(),
@@ -712,7 +833,10 @@ assert (x9 := 42), (x10 := "oops")
 type y = (x11 := int)
 # Named expressions inside expression-level scopes should not appear in definitions.
 lambda x: (z := 42)
+{z := "str" for _ in [1]}
+{(z := "str"):1 for _ in [1]}
 [z for x in [1, 2, 3] if z := x > 2]
+(z := "str" for _ in [1])
 "#,
         );
         assert_definition_names(
@@ -744,12 +868,18 @@ def bar(x: str) -> str: ...
         assert_definition_names(&defs, &["overload", "foo", "bar"]);
 
         let foo = defs.definitions.get(&Name::new_static("foo")).unwrap();
-        assert_eq!(foo.style, DefinitionStyle::Local(SymbolKind::Function));
-        assert_eq!(foo.count, 3);
+        assert_eq!(
+            foo.style,
+            DefinitionStyle::Unannotated(SymbolKind::Function)
+        );
+        assert!(foo.needs_anywhere);
 
         let bar = defs.definitions.get(&Name::new_static("bar")).unwrap();
-        assert_eq!(bar.style, DefinitionStyle::Local(SymbolKind::Function));
-        assert_eq!(bar.count, 2);
+        assert_eq!(
+            bar.style,
+            DefinitionStyle::Unannotated(SymbolKind::Function)
+        );
+        assert!(bar.needs_anywhere);
     }
 
     #[test]
@@ -784,6 +914,23 @@ __all__.remove('r')
         );
     }
 
+    #[test]
+    fn test_all_annotated() {
+        let defs = calculate_unranged_definitions_with_defaults(
+            r#"
+from foo import *
+a = 1
+b = 1
+__all__: list[str] = ["a", "b"]
+        "#,
+        );
+        assert_definition_names(&defs, &["a", "b", "__all__"]);
+        let loc = TextRange::default();
+        let a = &DunderAllEntry::Name(loc, Name::new_static("a"));
+        let b = &DunderAllEntry::Name(loc, Name::new_static("b"));
+        assert_eq!(defs.dunder_all.map(|x| x), vec![a, b]);
+    }
+
     #[test]
     fn test_all_reexport() {
         // Not in the spec, but see collections.abc which does this.
@@ -833,4 +980,51 @@ import derp.c.d
         );
         assert_implicitly_imported_submodules(&defs, &["b", "c"]);
     }
+
+    #[test]
+    fn test_named_in_del() {
+        let defs = calculate_unranged_definitions(
+            r#"
+del (y := {"x": 42})["x"]
+"#,
+            ModuleName::from_str("derp"),
+            true,
+        );
+        assert_definition_names(&defs, &["y"]);
+    }
+
+    #[test]
+    fn test_unused_mutable_captures() {
+        // These are illegal at the top-level, but they can occur in functions
+        // and the definitions extraction works the same way.
+        let defs = calculate_unranged_definitions(
+            r#"
+global x
+nonlocal y
+x = 5
+"#,
+            ModuleName::from_str("derp"),
+            true,
+        );
+        assert_definition_names(&defs, &["x", "y"]);
+        let x = defs.definitions.get(&Name::new_static("x")).unwrap();
+        assert!(!x.needs_anywhere);
+    }
+
+    #[test]
+    fn test_unused_del() {
+        // These are illegal at the top-level, but they can occur in functions
+        // and the definitions extraction works the same way.
+        let defs = calculate_unranged_definitions(
+            r#"
+del x
+del x
+"#,
+            ModuleName::from_str("derp"),
+            true,
+        );
+        assert_definition_names(&defs, &["x"]);
+        let x = defs.definitions.get(&Name::new_static("x")).unwrap();
+        assert!(!x.needs_anywhere);
+    }
 }
