@@ -5,10 +5,11 @@
  * LICENSE file in the root directory of this source tree.
  */
 
+pub mod bundled;
 pub mod finder;
-pub mod from_path;
 pub mod parse;
 pub mod typeshed;
+pub mod typeshed_third_party;
 
 // Temporary reexports. Given these are such central types going
 // to land the move with forwards, then clean up the forwards separately.
