@@ -11,11 +11,13 @@ use std::path::Path;
 use std::path::PathBuf;
 use std::sync::Arc;
 
+use crate::state::load::FileContents;
+
 #[derive(Debug, Clone, Default)]
-pub struct MemoryFiles(HashMap<PathBuf, Arc<String>>);
+pub struct MemoryFiles(HashMap<PathBuf, Arc<FileContents>>);
 
 #[derive(Debug, Default, Clone)]
-pub struct MemoryFilesOverlay(HashMap<PathBuf, Option<Arc<String>>>);
+pub struct MemoryFilesOverlay(HashMap<PathBuf, Option<Arc<FileContents>>>);
 
 impl MemoryFiles {
     pub fn apply_overlay(&mut self, overlay: MemoryFilesOverlay) {
@@ -29,7 +31,7 @@ impl MemoryFiles {
 }
 
 impl MemoryFilesOverlay {
-    pub fn set(&mut self, path: PathBuf, contents: Option<Arc<String>>) {
+    pub fn set(&mut self, path: PathBuf, contents: Option<Arc<FileContents>>) {
         self.0.insert(path, contents);
     }
 }
@@ -44,7 +46,7 @@ impl<'a> MemoryFilesLookup<'a> {
         Self { base, overlay }
     }
 
-    pub fn get(&self, path: &Path) -> Option<&'a Arc<String>> {
+    pub fn get(&self, path: &Path) -> Option<&'a Arc<FileContents>> {
         match self.overlay.0.get(path) {
             Some(contents) => contents.as_ref(), // Might be a None if deleted
             None => self.base.0.get(path),
