@@ -8,12 +8,12 @@
 pub mod dirty;
 pub mod epoch;
 pub mod errors;
-pub mod handle;
 pub mod ide;
 pub mod load;
 pub mod loader;
 pub mod lsp;
 pub mod memory;
+pub mod notebook;
 pub mod require;
 pub mod semantic_tokens;
 pub mod state;
