@@ -8,6 +8,43 @@
 use crate::test::util::TestEnv;
 use crate::testcase;
 
+// Test case for various edge cases where a name isn't in the flow, and we might
+// or might not decide an attribute has been defined.
+testcase!(
+    test_semantics_for_when_class_body_defines_attributes,
+    r#"
+from typing import assert_type, Any
+def condition() -> bool: ...
+class A:
+    # Annotated with no value
+    b: int
+    # Annotated twice with no value
+    c: str
+    c: str
+    # Defined in conditional control flow, with and without annotation
+    if condition():
+        d = 42
+        e: int = 42
+    # Defined (with or without annotation) but only in terminating control flow
+    if condition():
+        f = 42
+        g: int = 42
+        exit()
+    h = 42
+    i: int = 42
+    del h
+    del i
+assert_type(A.b, int)
+assert_type(A.c, str)
+assert_type(A.d, int)
+assert_type(A.e, int)
+assert_type(A.f, Any)  # E: Class `A` has no class attribute `f`
+assert_type(A.g, Any)  # E: Class `A` has no class attribute `g`
+assert_type(A.h, int)
+assert_type(A.i, int)
+    "#,
+);
+
 testcase!(
     test_set_attribute,
     r#"
@@ -76,6 +113,28 @@ del super(B, B()).a # E: Cannot delete field `a`
     "#,
 );
 
+testcase!(
+    test_super_retain_self,
+    r#"
+from typing import Self
+class A:
+    def m(self) -> Self:
+        return self
+
+    @classmethod
+    def classm(cls) -> type[Self]:
+        return cls
+
+class B(A):
+    def m(self) -> Self:
+        return super().m()
+
+    @classmethod
+    def classm(cls) -> type[Self]:
+        return super().classm()
+    "#,
+);
+
 testcase!(
     test_self_attribute_assign_twice,
     r#"
@@ -89,7 +148,8 @@ class A:
 );
 
 testcase!(
-    test_self_attribute_in_unrecognized_method,
+    test_self_attribute_in_unrecognized_method_enabled,
+    TestEnv::new().enable_implicitly_defined_attribute_error(),
     r#"
 from typing import assert_type
 class A:
@@ -100,6 +160,18 @@ def f(a: A):
     "#,
 );
 
+testcase!(
+    test_self_attribute_in_unrecognized_method_default_disabled,
+    r#"
+from typing import assert_type
+class A:
+    def f(self, x: int):
+        self.x = x
+def f(a: A):
+    assert_type(a.x, int)
+    "#,
+);
+
 testcase!(
     test_inherited_attribute_in_unrecognized_method,
     r#"
@@ -197,7 +269,7 @@ class C3:
     def f(x: int, /) -> str:
         return ""
 def foo(x: Callable[[int], str], c: C, c2: C2, c3: C3):
-    C.f = x  # E: `(int) -> str` is not assignable to attribute `f` with type `(self: Self@C, x: int, /) -> str`
+    C.f = x  # E: `(int) -> str` is not assignable to attribute `f` with type `(self: C, x: int, /) -> str`
     c.f = x
     C2.f = x
     c2.f = x
@@ -206,6 +278,27 @@ def foo(x: Callable[[int], str], c: C, c2: C2, c3: C3):
     "#,
 );
 
+testcase!(
+    bug = "classmethod bound object w/o targs is default-instantiated, solves T to Any",
+    test_bound_classmethod_explicit_targs,
+    r#"
+from typing import assert_type
+class A[T]:
+    x: T
+    def __init__(self, x: T):
+        self.x = x
+    @classmethod
+    def m(cls: 'type[A[T]]', x: T) -> 'A[T]':
+        return cls(x)
+
+assert_type(A[int].m(0), A[int])
+assert_type(A.m(0), A[int]) # TODO # E: assert_type(A[Any], A[int]) failed
+
+def test_typevar_bounds[T: A[int]](x: type[T]):
+    assert_type(x.m(0), A[int])
+    "#,
+);
+
 testcase!(
     test_use_of_class_body_scope_in_class_body_statement,
     r#"
@@ -356,6 +449,7 @@ def f(c: C):
 // TODO: Should we implement simple control-flow heuristics so `C.x` is recognized here?
 testcase!(
     test_set_attribute_in_init_indirect,
+    TestEnv::new().enable_implicitly_defined_attribute_error(),
     r#"
 class C:
     def __init__(self):
@@ -427,7 +521,7 @@ assert_type(C().f(None, 1), int)
 // because the default value can be overridden by instance assignment.
 //
 // Our behavior is compatible, but the underlying implementation is not, we are
-// behaving this way based on how we treate the Callable type rather than based
+// behaving this way based on how we treat the Callable type rather than based
 // on the absence of `ClassVar`.
 //
 // See https://discuss.python.org/t/when-should-we-assume-callable-types-are-method-descriptors/92938
@@ -455,7 +549,7 @@ class C:
     def __init__(self, y: str):
         self.x = 0
         self.y = y
-assert_type(C.x, Any)  # E: Instance-only attribute `x` of class `C` is not visible on the class
+assert_type(C.x, int)
 assert_type(C.y, Any)  # E: Instance-only attribute `y` of class `C` is not visible on the class
 c = C("y")
 assert_type(c.x, int)
@@ -474,10 +568,10 @@ def f1(c: Callable[[int], None]):
     pass
 def f2(c: Callable[[C, int], None]):
     pass
-f1(C.f)  # E: Argument `(self: Self@C, x: int) -> None` is not assignable to parameter `c` with type `(int) -> None`
+f1(C.f)  # E: Argument `(self: C, x: int) -> None` is not assignable to parameter `c` with type `(int) -> None`
 f1(C().f)
 f2(C.f)
-f2(C().f)  # E: Argument `BoundMethod[C, (self: Self@C, x: int) -> None]` is not assignable to parameter `c` with type `(C, int) -> None`
+f2(C().f)  # E: Argument `BoundMethod[C, (self: C, x: int) -> None]` is not assignable to parameter `c` with type `(C, int) -> None`
     "#,
 );
 
@@ -557,7 +651,7 @@ testcase!(
     test_class_generic_attribute_lookup,
     r#"
 class C[T]:
-    x = T
+    x: list[T] = []
 
 C.x  # E: Generic attribute `x` of class `C` is not visible on the class
 "#,
@@ -648,6 +742,21 @@ assert_type(C.x, int)
     "#,
 );
 
+testcase!(
+    test_metaclass_method_cls_typetype,
+    r#"
+from typing import assert_type
+
+class Meta(type):
+    def m[T](cls: type[T]) -> T: ...
+
+class C(metaclass=Meta):
+    pass
+
+assert_type(C.m(), C)
+"#,
+);
+
 fn env_with_stub() -> TestEnv {
     let mut t = TestEnv::new();
     t.add_with_path(
@@ -763,7 +872,7 @@ from typing import assert_type
 
 class Foo:
     def __getattr__(self, name: int) -> int: ...
-    def __setattr__(self, name: int, value: int) -> None: ...
+    def __setattr__(self, name: int, value: int) -> None: ...  # E: `Foo.__setattr__` overrides parent class `object` in an inconsistent manner
 
 def test(foo: Foo) -> None:
     foo.x = 1  # E: Argument `Literal['x']` is not assignable to parameter `name` with type `int`
@@ -835,13 +944,43 @@ from typing import assert_type, Any
 
 class C:
     def __init__[R](self, field: R):
-        self.field = field  # E: Cannot initialize attribute `field` to a value that depends on method-scoped type variable `R`
+        self.field = field  # E: Attribute `field` cannot depend on type variable `R`, which is not in the scope of class `C`
 
 c = C("test")
 assert_type(c.field, Any)
 "#,
 );
 
+// Note the difference between this and test_set_attribute_to_class_scope_type_variable.
+// `R` in `__init__` here refers to a method-scoped type variable that shadows a class-scoped one.
+testcase!(
+    test_illegal_type_variable_with_name_shadowing,
+    r#"
+class C[R]:
+    def __init__[R](self, field: R):
+        self.field = field  # E: Attribute `field` cannot depend on type variable `R`, which is not in the scope of class `C`
+"#,
+);
+
+// Note the difference between this and test_illegal_type_variable_with_name_shadowing.
+// `R` in `__init__` here refers to the class-scoped `R``.
+testcase!(
+    test_set_attribute_to_class_scope_type_variable,
+    r#"
+from typing import Generic, TypeVar
+
+R = TypeVar("R")
+
+class C1(Generic[R]):
+    def __init__(self, field: R):
+        self.field = field
+
+class C2[R]:
+    def __init__(self, field: R):
+        self.field = field
+"#,
+);
+
 testcase!(
     test_attr_unknown,
     r#"
@@ -1004,14 +1143,30 @@ testcase!(
     r#"
 from typing import assert_type, Any
 class Foo:
-    x: int = 1
-def f[T: Foo](y: T) -> T:
-    assert_type(y.x, int)
-    assert_type(T.x, int)
+    def m(self) -> int:
+        return 0
+def f[T: Foo](y: T, z: type[T]) -> T:
+    assert_type(y.m(), int)
+    assert_type(z.m(y), int)
+    assert_type(T.m(y), Any) # E: Object of class `TypeVar` has no attribute `m`
     return y
     "#,
 );
 
+testcase!(
+    test_classmethod_on_instance_typevar_bound,
+    r#"
+from typing import Self
+
+class A:
+    @classmethod
+    def bar(cls) -> Self: ...
+
+def test[T: A](a: T) -> T:
+    return a.bar()
+    "#,
+);
+
 testcase!(
     test_attribute_access_on_quantified_bound_by_union,
     r#"
@@ -1027,6 +1182,153 @@ def f[T: Foo | Bar](y: T, z: Foo | Bar) -> T:
     "#,
 );
 
+testcase!(
+    bug = "type[None] should be types.NoneType",
+    test_attribute_access_on_type_none,
+    r#"
+# handy hack to get a type[X] for any X
+def ty[T](x: T) -> type[T]: ...
+
+ty(None).__bool__(None) # E: Expr::attr_infer_for_type attribute base undefined
+"#,
+);
+
+testcase!(
+    test_attribute_access_on_type_literal,
+    r#"
+# handy hack to get a type[X] for any X
+def ty[T](x: T) -> type[T]: ...
+
+ty(0).bit_length(0)
+"#,
+);
+
+testcase!(
+    test_attribute_access_on_type_literalstring,
+    r#"
+from typing import LiteralString
+
+# handy hack to get a type[X] for any X
+def ty[T](x: T) -> type[T]: ...
+
+def test(x: LiteralString):
+    ty(x).upper(x)
+"#,
+);
+
+testcase!(
+    bug = "type[<<callable>>] should be... types.FunctionType, probably. type[object] if that's unagreeable",
+    test_attribute_access_on_type_callable,
+    r#"
+from typing import Callable
+
+# handy hack to get a type[X] for any X
+def ty[T](x: T) -> type[T]: ...
+
+def test_callable(x: Callable[[], None]):
+    ty(x).__call__(x) # E: Expr::attr_infer_for_type attribute base undefined
+"#,
+);
+
+testcase!(
+    bug = "type[<<function>>] should be types.FunctionType",
+    test_attribute_access_on_type_function,
+    r#"
+# handy hack to get a type[X] for any X
+def ty[T](x: T) -> type[T]: ...
+
+def foo(): ...
+
+ty(foo).__call__(foo) # E: Expr::attr_infer_for_type attribute base undefined
+"#,
+);
+
+testcase!(
+    bug = "type[<<boundmethod>>] should be types.FunctionType",
+    test_attribute_access_on_type_boundmethod,
+    r#"
+# handy hack to get a type[X] for any X
+def ty[T](x: T) -> type[T]: ...
+
+class X:
+    def m(self): ...
+
+ty(X().m).__call__(X().m) # E: Expr::attr_infer_for_type attribute base undefined
+"#,
+);
+
+testcase!(
+    bug = "type[<<overload>>] should be types.FunctionType",
+    test_attribute_access_on_type_overload,
+    r#"
+from typing import overload
+
+# handy hack to get a type[X] for any X
+def ty[T](x: T) -> type[T]: ...
+
+@overload
+def bar(x: int) -> int: ...
+@overload
+def bar(x: str) -> str: ...
+def bar(x: int | str) -> int | str: ...
+
+ty(bar).__call__(bar) # E: Expr::attr_infer_for_type attribute base undefined
+"#,
+);
+
+testcase!(
+    test_attribute_access_on_type_union,
+    r#"
+# handy hack to get a type[X] for any X
+def ty[T](x: T) -> type[T]: ...
+
+class A:
+    x = 0
+class B:
+    x = "foo"
+
+def test_union(x: A  | B):
+    ty(x).x
+"#,
+);
+
+testcase!(
+    bug = "type[ClassDef(..)] and type[ClassType(..)] should be type (or the direct metaclass?)",
+    test_attribute_access_on_type_class,
+    r#"
+# handy hack to get a type[X] for any X
+def ty[T](x: T) -> type[T]: ...
+
+class C:
+    @staticmethod
+    def m(x: int): ...
+
+class D[T]:
+    @classmethod
+    def m(cls, x: T): ...
+
+ty(C).m(0) # E: Expr::attr_infer_for_type attribute base undefined
+ty(D[int]).m(0) # E: Expr::attr_infer_for_type attribute base undefined
+"#,
+);
+
+testcase!(
+    bug = "type[TypedDict()] should be type",
+    test_attribute_access_on_type_typeddict,
+    r#"
+from typing import TypedDict
+
+# handy hack to get a type[X] for any X
+def ty[T](x: T) -> type[T]: ...
+
+class TD(TypedDict):
+    x: int
+
+ty(TD(x = 0))(x = 0)
+ty(TD).mro() # E: Expr::attr_infer_for_type attribute base undefined
+"#,
+);
+
 testcase!(
     test_type_magic_dunder_compare,
     r#"
@@ -1052,9 +1354,10 @@ def test(x: type[int], y: type[int]) -> None:
 testcase!(
     test_access_method_using_class_param_on_class,
     r#"
-from typing import assert_type, Any
+from typing import assert_type, reveal_type, Any
 class A[T]:
     def f(self) -> T: ...
+reveal_type(A.f) # E: revealed type: [T](self: A[T]) -> T
 assert_type(A.f(A[int]()), int)
     "#,
 );
@@ -1062,9 +1365,10 @@ assert_type(A.f(A[int]()), int)
 testcase!(
     test_access_generic_method_using_class_param_on_class,
     r#"
-from typing import assert_type, Any
+from typing import assert_type, reveal_type, Any
 class A[T]:
     def f[S](self, x: S) -> tuple[S, T]: ...
+reveal_type(A.f) # E: revealed type: [T, S](self: A[T], x: S) -> tuple[S, T]
 assert_type(A.f(A[int](), ""), tuple[str, int]) # E: assert_type(tuple[str, Any], tuple[str, int])
     "#,
 );
@@ -1072,17 +1376,37 @@ assert_type(A.f(A[int](), ""), tuple[str, int]) # E: assert_type(tuple[str, Any]
 testcase!(
     test_access_overloaded_method_using_class_param_on_class,
     r#"
-from typing import assert_type, overload, Any
+from typing import assert_type, reveal_type, overload, Any
 class A[T]:
     @overload
     def f(self) -> T: ...
     @overload
     def f(self, x: T | None) -> T: ...
     def f(self, x=None) -> Any: ...
+reveal_type(A.f) # E: revealed type: Overload[[T](self: A[T]) -> T, [T](self: A[T], x: T | None) -> T]
 assert_type(A.f(A[int]()), int)
     "#,
 );
 
+testcase!(
+    test_access_overloaded_staticmethod_using_class_param_on_class,
+    r#"
+from typing import assert_type, reveal_type, overload, Any
+class A[T]:
+    @overload
+    @staticmethod
+    def f(x: None = ...) -> None: ...
+    @overload
+    @staticmethod
+    def f(x: T) -> T: ...
+    @staticmethod
+    def f(x = None) -> Any: ...
+reveal_type(A.f) # E: revealed type: Overload[(x: None = ...) -> None, [T](x: T) -> T]
+assert_type(A.f(), None)
+assert_type(A.f(0), int)
+    "#,
+);
+
 testcase!(
     test_invalid_augmented_assign_in_init,
     r#"
@@ -1154,7 +1478,6 @@ def f(c: Config):
 );
 
 testcase!(
-    bug = "We probably should disallow reassignment of class objects",
     test_nested_class_mutability,
     r#"
 class Backend:
@@ -1202,7 +1525,6 @@ C.x = 43  # E: This field is marked as Final
 );
 
 testcase!(
-    bug = "other.output type is too general. Also, there should be no errors.",
     test_attr_cast,
     r#"
 from typing import Self, cast, Any, assert_type
@@ -1212,34 +1534,218 @@ class C:
     def f(self, other):
         other = cast(Self, other)
         assert_type(other, Self)
-        assert_type(other.outputs, Any) # E: TODO: Expr::attr_infer_for_type
-        len(self.outputs) == len(other.outputs) # E: TODO: Expr::attr_infer_for_type attribute base undefined for type: Self
+        assert_type(other.outputs, list[Any])
+        len(self.outputs) == len(other.outputs)
     "#,
 );
 
 testcase!(
-    bug = "There should be no errors here.",
     test_attr_tuple,
     r#"
 from typing import Any, Tuple
 
 def g(ann) -> None:
     if ann is Tuple: ...
-    ann.__module__ # E: TODO: Expr::attr_infer_for_type attribute base undefined for type: type[Tuple] | Unknown (trying to access __module__)
+    ann.__module__
     "#,
 );
 
 testcase!(
-    bug = "PyTorch TODO: First error message can be improved and there should be  no error on obj.__name__",
-    test_attr,
+    test_tuple_attribute_example,
     r#"
 def f(obj, g, field_type, my_type,):
     assert issubclass(obj, tuple) and hasattr(obj, "_fields")
-    for f in obj._fields: # E: TODO: Expr::attr_infer_for_type attribute base undefined for type: type[tuple[Unknown, ...]]
+    for f in obj._fields:
         if isinstance(field_type, my_type) and g is not None:
             if g is None:
                 raise ValueError(
-                    f"{obj.__name__}." # E: TODO: Expr::attr_infer_for_type attribute base undefined for type: @_ (trying to access __name__)
+                    f"{obj.__name__}."
                 )
     "#,
 );
+
+testcase!(
+    test_set_attr_in_child_class,
+    r#"
+from typing import assert_type
+
+class A:
+    def __init__(self):
+        self.x = 0
+
+class B(A):
+    def f(self):
+        self.x = ""  # E: `Literal['']` is not assignable to attribute `x` with type `int`
+
+class C(A):
+    def f(self):
+        self.x: str = ""  # E: Class member `C.x` overrides parent class `A` in an inconsistent manner
+    "#,
+);
+
+testcase!(
+    test_method_sets_inherited_generic_field,
+    r#"
+# Regression test for a bug tracked in https://github.com/facebook/pyrefly/issues/774
+from typing import assert_type, Any
+class A[T]:
+    x: T
+class B(A[int]):
+    def __init__(self, x: int):
+        # The test is primarily verifying that we handle this implicit definition
+        # correctly in the class field logic, when this is actually an inherited field.
+        self.x = x
+assert_type(B(42).x, int)
+    "#,
+);
+
+testcase!(
+    test_private_attr_assignment_in_constructor,
+    r#"
+from typing import assert_type
+
+class Config:
+    pass
+
+class DerivedConfig(Config):
+    def foo(self) -> None:
+        print("hello")
+
+class B:
+    def __init__(self, config: Config) -> None:
+        self.__config = config
+
+class C(B):
+    def __init__(self, config: DerivedConfig) -> None:
+        self.__config = config
+
+    def bar(self) -> None:
+        assert_type(self.__config, DerivedConfig)
+        self.__config.foo()
+    "#,
+);
+
+testcase!(
+    test_crtp_example, // CRTP = Curiously recurring template pattern
+    r#"
+from typing import Any, assert_type
+class Node[T: "Node[Any]"]:
+    children : tuple[T, ...]
+class Expr(Node["Expr"]):
+    ...
+class Singleton(Expr):
+    def __init__(self, v: Expr):
+        self.children = (v,)
+assert_type(Singleton(Expr()).children, tuple[Expr, ...])
+    "#,
+);
+
+testcase!(
+    test_mro_method,
+    r#"
+().mro()  # E: no attribute `mro`
+tuple.mro()
+type.mro()  # E: Missing argument `self`
+    "#,
+);
+
+// How special forms are represented in typing.py is an implementation detail, but in practice,
+// some of the representations are stable across Python versions. In particular, user code
+// sometimes relies on some special forms being classes and Type behaving like builtins.type.
+testcase!(
+    test_special_forms,
+    r#"
+from typing import Callable, Generic, Protocol, Tuple, Type
+def f1(cls):
+    if cls is Callable:
+        return cls.mro()
+def f2(cls):
+    if cls is Generic:
+        return cls.mro()
+def f3(cls):
+    if cls is Protocol:
+        return cls.mro()
+def f4(cls):
+    if cls is Tuple:
+        return cls.mro()
+def f5(cls, x: type):
+    if cls is Type:
+        return cls.mro(x)
+    "#,
+);
+
+testcase!(
+    test_get_type_new,
+    r#"
+from typing import cast, reveal_type
+def get_type_t[T]() -> type[T]:
+    return cast(type[T], 0)
+def foo[T](x: type[T]):
+    # mypy reveals the same thing we do (the type of `type.__new__`), while pyright reveals `Unknown`.
+    reveal_type(get_type_t().__new__)  # E: Overload[(cls: type[type], o: object, /) -> type, [Self](cls: type[Self], name: str, bases: tuple[type, ...], namespace: dict[str, Any], /, **kwds: Any) -> Self]
+    "#,
+);
+
+testcase!(
+    test_any_value_lookup,
+    r#"
+from typing import Any
+Any.foo.bar # E: Class `Any` has no class attribute `foo`
+    "#,
+);
+
+// T, P, and Ts are values of type TypeVar, ParamSpec, and TypeVarTuple respectively.
+// They should behave like values when we try to access attributes on them.
+testcase!(
+    test_typevar_value_lookups,
+    r#"
+from typing import Callable, TypeVar, ParamSpec, TypeVarTuple
+
+def ty[T](x: T) -> type[T]: ...
+
+T = TypeVar("T")
+P = ParamSpec("P")
+Ts = TypeVarTuple("Ts")
+
+T.__name__
+P.__name__
+P.args.__origin__
+P.kwargs.__origin__
+Ts.__name__
+
+ty(T).__name__
+ty(P).__name__
+ty(P.args).__origin__
+ty(P.args).__origin__
+ty(Ts).__name__
+
+def f(g: Callable[P, T], ts: tuple[*Ts], *args: P.args, **kwargs: P.kwargs):
+    args.count(1)
+    kwargs.keys()
+
+    ty(args).count(args, 1)
+    ty(kwargs).keys(kwargs)
+
+    T.__name__
+    P.__name__
+    P.args.__origin__
+    P.kwargs.__origin__
+    Ts.__name__
+
+    ty(T).__name__
+    ty(P).__name__
+    ty(P.args).__origin__
+    ty(P.args).__origin__
+    ty(Ts).__name__
+"#,
+);
+
+testcase!(
+    test_type_never,
+    r#"
+from typing import Never, assert_type, reveal_type
+def f() -> type[Never]: ...
+reveal_type(f().mro) # E: BoundMethod[type, (self: type) -> list[type]]
+assert_type(f().wut, Never)
+    "#,
+);
