@@ -814,6 +814,50 @@ zoo(partial(bar, b=99))
 "#,
 );
 
+testcase!(
+    bug = "Self in Metaclass should be error, treated as Any. Any in metaclass call should act like no annot.",
+    test_callable_class_substitute_self,
+    r#"
+from typing import Any, Callable, Self, assert_type
+
+def ret[T](f: Callable[[], T]) -> T: ...
+
+class Meta(type):
+    def __call__(self, *args, **kwargs) -> Self: ... # TODO: Error here and treat `Self` as `Any`
+
+# metaclass __call__
+class A(metaclass=Meta):
+    pass
+
+# __new__
+class B:
+    def __new__(cls, *args, **kwargs) -> Self: ...
+
+# __init__
+class C:
+    def __init__(self, *args, **kwargs) -> None: ...
+
+assert_type(ret(A), A) # TODO # E: assert_type(type[A], A) failed
+assert_type(ret(B), B)
+assert_type(ret(C), C)
+"#,
+);
+
+testcase!(
+    test_callable_class_self_confusion,
+    r#"
+from typing import Callable, Self, assert_type
+
+class A:
+    def __new__(cls) -> Self: ...
+
+class B[T]:
+    def __new__(self, f: Callable[[], T]) -> Self: ...
+
+assert_type(B(A), B[A])
+"#,
+);
+
 testcase!(
     test_call_self,
     r#"
@@ -868,3 +912,139 @@ def g(f: Callable[..., Any]):
 g(f)
     "#,
 );
+
+testcase!(
+    test_generic_bounds_to_callable,
+    r#"
+from typing import Callable
+
+class A: ...
+class B(A): ...
+class C(B): ...
+
+def f[T: B](x: T) -> T: ...
+
+c1: Callable[[A], A] = f # E: `[T](x: T) -> T` is not assignable to `(A) -> A`
+c2: Callable[[B], B] = f # OK
+c3: Callable[[C], C] = f # OK
+    "#,
+);
+
+testcase!(
+    test_return_generic_callable,
+    r#"
+from typing import assert_type, Callable
+def f[T]() -> Callable[[T], T]:
+    return lambda x: x
+
+g = f()
+assert_type(g(0), int)
+assert_type(g(""), str)
+
+@f()
+def h(x: int) -> int:
+    return x
+assert_type(h(0), int)
+    "#,
+);
+
+testcase!(
+    test_generic_callable_union,
+    r#"
+from typing import assert_type, Callable
+def f[T]() -> Callable[[T], T] | Callable[[T], list[T]]: ...
+g = f()
+assert_type(g(0), int | list[int])
+assert_type(g(""), str | list[str])
+    "#,
+);
+
+testcase!(
+    test_callable_returns_callable_returns_callable,
+    r#"
+from typing import assert_type, Callable
+
+def f[T]() -> Callable[[], Callable[[T], T]]:
+    def f():
+        return lambda x: x
+    return f
+
+g = f()()
+assert_type(g(0), int)
+assert_type(g(""), str)
+
+    "#,
+);
+
+testcase!(
+    test_return_substituted_callable,
+    r#"
+from typing import assert_type, Callable
+def f[T](x: T) -> Callable[[T], T]: ...
+g = f(0)
+assert_type(g(0), int)
+assert_type(g(""), int)  # E: `Literal['']` is not assignable to parameter with type `int`
+    "#,
+);
+
+testcase!(
+    test_generic_callable_or_none,
+    r#"
+from typing import assert_type, Callable
+def f[T]() -> Callable[[T], T] | None: ...
+g = f()
+if g:
+    assert_type(g(0), int)
+    assert_type(g(""), str)
+    "#,
+);
+
+testcase!(
+    test_pass_literals_through_identity,
+    r#"
+from typing import Callable, Literal, reveal_type
+def f(x: Literal[1]) -> Literal[1]:
+    return x
+def g[T](x: T) -> T:
+    return x
+h = g(f)
+reveal_type(h)  # E: revealed type: (x: Literal[1]) -> Literal[1]
+    "#,
+);
+
+testcase!(
+    test_boundmethod_union,
+    r#"
+from typing import assert_type
+def _(flag: bool):
+    class C:
+        if flag:
+            def __getitem__(self, key: int) -> str:
+                return str(key)
+        else:
+            def __getitem__(self, key: int) -> bytes:
+                return bytes()
+    c = C()
+    assert_type(c[0], bytes | str)
+    "#,
+);
+
+testcase!(
+    test_unknown_varargs_kwargs,
+    r#"
+from typing import Any, assert_type
+def f(*args, **kwargs):
+    assert_type(args, tuple[Any, ...])
+    assert_type(kwargs, dict[str, Any])
+    "#,
+);
+
+testcase!(
+    test_bad_varargs_kwargs,
+    r#"
+from typing import Annotated, Any, assert_type
+def f(*args: Annotated, **kwargs: Annotated): # E: # E:
+    assert_type(args, tuple[Any, ...])
+    assert_type(kwargs, dict[str, Any])
+    "#,
+);
