@@ -1177,7 +1177,7 @@ class Bad1:
     x: int
     y: InitVar[str]
     z: InitVar[bytes]
-    def __post_init__(self, y: bytes, z: str): ...  # E: `__post_init__` type `BoundMethod[Bad1, (self: Self@Bad1, y: bytes, z: str) -> None]` is not assignable to expected type `(y: str, z: bytes) -> object` generated from the dataclass's `InitVar` fields
+    def __post_init__(self, y: bytes, z: str): ...  # E: `__post_init__` type `BoundMethod[Bad1, (self: Bad1, y: bytes, z: str) -> None]` is not assignable to expected type `(y: str, z: bytes) -> object` generated from the dataclass's `InitVar` fields
 @dataclass
 class Bad2:
     x: int
@@ -1219,3 +1219,64 @@ class C2(C1):
 C2('', 0.2, b=3)
     "#,
 );
+
+testcase!(
+    test_assign_to_field_in_child,
+    r#"
+from dataclasses import dataclass
+
+@dataclass
+class Animal:
+    name: str | None = None
+    def speak(self) -> str: ...
+
+@dataclass
+class Dog(Animal):
+    def speak(self) -> str:
+        self.name = "dog"
+        return "woof"
+
+hdog = Dog(name="hdog")
+    "#,
+);
+
+testcase!(
+    test_fields_function,
+    r#"
+from typing import assert_type, Any
+from dataclasses import dataclass, fields, Field
+
+@dataclass
+class Person:
+    name: str
+    age: int
+
+# Test fields() on the class type
+f1 = fields(Person)
+assert_type(f1, tuple[Field[Any], ...])
+
+# Test fields() on an instance
+p = Person("Alice", 30)
+f2 = fields(p)
+assert_type(f2, tuple[Field[Any], ...])
+    "#,
+);
+
+testcase!(
+    test_final_field_no_modification,
+    r#"
+from typing import Final
+from dataclasses import dataclass
+@dataclass
+class C:
+    x: Final[int]
+    y: Final[int] = 42
+
+c = C(x=0)
+c.x = 1  # E: Cannot set field `x`
+c.y = 1  # E: Cannot set field `y`
+
+C.x = 1  # E: Cannot set field `x`
+C.y = 1  # E: Cannot set field `y`
+"#,
+);
