@@ -10,6 +10,26 @@ use crate::testcase;
 testcase!(
     test_literal_dict,
     r#"
-dict(x = 1, y = "test") # E:  No matching overload found for function `dict.__init__`
+dict(x = 1, y = "test")
+    "#,
+);
+
+testcase!(
+    test_unpack_empty,
+    r#"
+from typing import assert_type
+x = {**{}}
+x['x'] = 0
+assert_type(x, dict[str, int])
+    "#,
+);
+
+testcase!(
+    test_typeddict_interaction,
+    r#"
+from typing import TypedDict
+class C(TypedDict):
+    x: int
+x: C | dict[str, int] = {"y": 0}
     "#,
 );
