@@ -5,6 +5,11 @@
  * LICENSE file in the root directory of this source tree.
  */
 
+use pyrefly_build::handle::Handle;
+use pyrefly_python::module_name::ModuleName;
+use pyrefly_python::module_path::ModulePath;
+use pyrefly_util::fs_anyhow;
+
 use crate::test::util::TestEnv;
 use crate::testcase;
 
@@ -763,3 +768,311 @@ testcase!(
 from . import foo  # E: Could not find import of `.`
     "#,
 );
+
+#[test]
+fn test_interface_has_more() {
+    let temp = tempfile::tempdir().unwrap();
+    let mut env = TestEnv::new();
+    let foo_py = temp.path().join("foo.py");
+    let foo_pyi = temp.path().join("foo.pyi");
+
+    fs_anyhow::write(&foo_py, "import foo as X\nX.Extra").unwrap();
+    fs_anyhow::write(&foo_pyi, "class Extra: pass").unwrap();
+    env.add_real_path("foo", foo_py);
+    env.add_real_path("foo", foo_pyi);
+    let _ = env.to_state();
+}
+
+#[test]
+fn test_interface_disagree() {
+    let temp = tempfile::tempdir().unwrap();
+    let mut env = TestEnv::new();
+    let foo_py = temp.path().join("foo.py");
+    let foo_pyi = temp.path().join("foo.pyi");
+
+    fs_anyhow::write(
+        &foo_py,
+        "class Foo:\n  def method(self): pass\nFoo().method()",
+    )
+    .unwrap();
+    fs_anyhow::write(&foo_pyi, "").unwrap();
+    env.add_real_path("foo", foo_py.clone());
+    env.add_real_path("foo", foo_pyi);
+    let h_py = Handle::new(
+        ModuleName::from_str("foo"),
+        ModulePath::filesystem(foo_py),
+        env.sys_info(),
+    );
+    let (state, _) = env.to_state();
+    let errs = state
+        .transaction()
+        .get_errors([&h_py])
+        .collect_errors()
+        .shown;
+    assert_eq!(errs.len(), 0);
+}
+
+fn env_class_x_deprecated() -> TestEnv {
+    TestEnv::one(
+        "foo",
+        r#"
+from warnings import deprecated
+@deprecated("Don't use this")
+class X: ...
+x: X = X()
+"#,
+    )
+}
+
+testcase!(
+    test_import_deprecated_class_warn,
+    env_class_x_deprecated(),
+    r#"
+from foo import X # E: `X` is deprecated
+
+x = X()
+"#,
+);
+
+testcase!(
+    test_import_star_deprecated_class_warn,
+    env_class_x_deprecated(),
+    r#"
+from foo import * # E: `X` is deprecated
+
+x = X()
+"#,
+);
+
+fn env_func_x_deprecated() -> TestEnv {
+    TestEnv::one(
+        "foo",
+        r#"
+from warnings import deprecated
+@deprecated("Don't use this")
+def x(): ...
+"#,
+    )
+}
+
+testcase!(
+    test_import_deprecated_func_warn,
+    env_func_x_deprecated(),
+    r#"
+from foo import x # E: `x` is deprecated
+
+x()  # E: `foo.x` is deprecated
+"#,
+);
+
+testcase!(
+    test_import_as_deprecated_func_warn,
+    env_func_x_deprecated(),
+    r#"
+from foo import x as y # E: `x` is deprecated
+
+y()  # E: `foo.x` is deprecated
+"#,
+);
+
+testcase!(
+    test_import_star_deprecated_func_warn,
+    env_func_x_deprecated(),
+    r#"
+from foo import * # E: `x` is deprecated
+
+x()  # E: `foo.x` is deprecated
+"#,
+);
+
+fn env_func_x_deprecated_conditionally() -> TestEnv {
+    TestEnv::one(
+        "foo",
+        r#"
+from warnings import deprecated
+import sys
+
+if sys.version_info >= (3, 10):
+    @deprecated("Don't use this")
+    def x(): ...
+else:
+    def x(): ...
+"#,
+    )
+}
+
+fn env_func_x_deprecated_conditionally_no_deprecation() -> TestEnv {
+    TestEnv::one(
+        "foo",
+        r#"
+from warnings import deprecated
+import sys
+
+if sys.version_info < (3, 10):
+    @deprecated("Don't use this")
+    def x(): ...
+else:
+    def x(): ...
+"#,
+    )
+}
+
+testcase!(
+    test_import_conditionally_deprecated_func_warn,
+    env_func_x_deprecated_conditionally(),
+    r#"
+from foo import x # E: `x` is deprecated
+
+x()  # E: `foo.x` is deprecated
+"#,
+);
+
+testcase!(
+    test_import_conditionally_deprecated_func_no_warn,
+    env_func_x_deprecated_conditionally_no_deprecation(),
+    r#"
+from foo import x
+# No warning for import, since the function is not deprecated in this context
+
+x()
+"#,
+);
+
+fn env_func_x_deprecated_overload_only() -> TestEnv {
+    TestEnv::one(
+        "foo",
+        r#"
+from warnings import deprecated
+from typing import Any, overload
+@overload
+def x(y: int) -> int: ...
+
+@deprecated("Don't use this")
+@overload
+def x(y: str) -> str: ...
+
+def x(y: Any) -> Any: ...
+"#,
+    )
+}
+
+testcase!(
+    test_import_deprecated_overload_no_warn,
+    env_func_x_deprecated_overload_only(),
+    r#"
+from foo import x
+# No warning for import, since only the overload is deprecated
+
+x("hello")  # E: Call to deprecated overload `foo.x`
+"#,
+);
+
+fn env_submodule_attribute_name_collision() -> TestEnv {
+    let mut t = TestEnv::new();
+    t.add_with_path("foo", "foo/__init__.py", "from .bar import bar");
+    t.add_with_path("foo.bar", "foo/bar.py", "def bar(): pass");
+    t
+}
+
+testcase!(
+    test_submodule_attribute_name_collision,
+    env_submodule_attribute_name_collision(),
+    r#"
+from typing import assert_type, Callable
+import foo
+assert_type(foo.bar, Callable[[], None])
+    "#,
+);
+
+testcase!(
+    test_unittest_main,
+    r#"
+import unittest
+unittest.main()
+    "#,
+);
+
+testcase!(
+    bug = "In the presence of collisions, the runtime result can be order-dependent",
+    test_ambiguous_pkg_attribute_vs_submodule,
+    r#"
+# The behavior when a package `__init__` attribute and a submodule collide can
+# be ambiguous. We currently do not model this fully, and instead use the distinction
+# between `import pkg.xyz` vs `from pkg import xyz` (which is correlated with the
+# runtime behavior but not entirely the same) to approximate.
+#
+# One example where this leads us astray is with unittest.main; the following
+# snippet works at runtime:
+import unittest.main
+unittest.main()  # E: Expected a callable, got `Module[unittest.main]`
+    "#,
+);
+
+fn env_re_export_special() -> TestEnv {
+    TestEnv::one(
+        "foo",
+        r#"
+from typing import TypeVar
+from typing import TypeVarTuple as _TypeVarTuple
+from typing import TypeAlias
+from typing import Annotated as MyAnnotated
+"#,
+    )
+}
+
+testcase!(
+    test_re_exported_special,
+    env_re_export_special(),
+    r#"
+from foo import TypeVar
+import foo
+from foo import TypeAlias as TA, MyAnnotated
+
+T = TypeVar("T")
+Ts = foo._TypeVarTuple("Ts")
+
+X: TA = list[int]
+
+x: X = [1, 2, 3]
+def test(x: T, ys: tuple[*Ts]) -> tuple[T, *Ts]:
+    return (x, *ys)
+
+y: MyAnnotated[int, "hello"] = 1
+"#,
+);
+
+fn env_extra_builtins() -> TestEnv {
+    TestEnv::one_with_path(
+        "__builtins__",
+        "__builtins__.pyi",
+        r#"
+class X: ...
+"#,
+    )
+}
+
+testcase!(
+    test_extra_builtins,
+    env_extra_builtins(),
+    r#"
+x: X = X()
+"#,
+);
+
+fn env_extra_builtins_in_typings() -> TestEnv {
+    TestEnv::one_with_path(
+        "__builtins__",
+        "typings/__builtins__.pyi",
+        r#"
+class X: ...
+"#,
+    )
+}
+
+testcase!(
+    test_extra_builtins_in_typings,
+    env_extra_builtins_in_typings(),
+    r#"
+x: X = X()
+"#,
+);
