@@ -7,10 +7,10 @@
 
 use itertools::Itertools as _;
 use pretty_assertions::assert_eq;
+use pyrefly_build::handle::Handle;
 use pyrefly_python::module::TextRangeWithModule;
 use ruff_text_size::TextSize;
 
-use crate::state::handle::Handle;
 use crate::state::state::State;
 use crate::test::util::code_frame_of_source_at_range;
 use crate::test::util::get_batched_lsp_operations_report;
@@ -135,7 +135,7 @@ Definition Result:
 4 | def f(x: list[int], y: str, z: Literal[42]):
                                     ^
 Definition Result:
-249 | Literal: _SpecialForm
+253 | Literal: _SpecialForm
       ^^^^^^^
 
 8 | yyy = f([1, 2, 3], "test", 42)
@@ -216,6 +216,33 @@ Definition Result:
     );
 }
 
+#[test]
+fn branch_with_pinning_test() {
+    let code = r#"
+def test(flag: bool, x: str | None) -> None:
+    if flag:
+        y = "derp"
+        x = y
+    x
+#   ^
+"#;
+    let report = get_batched_lsp_operations_report(&[("main", code)], get_test_report);
+    // Why didn't it go to the parameter `x`? That's because we don't explore all branches of Phi nodes
+    // in gotodef yet
+    assert_eq!(
+        r#"
+# main.py
+6 |     x
+        ^
+Definition Result:
+5 |         x = y
+            ^
+"#
+        .trim(),
+        report.trim(),
+    );
+}
+
 #[test]
 fn shadowed_def_test0() {
     let code = r#"
@@ -476,7 +503,7 @@ Definition Result:
                               ^
 Definition Result:
 6 |     def foo(self, y: str, x: int) -> None:
-                      ^       
+                      ^
 "#
         .trim(),
         report.trim(),
@@ -485,7 +512,7 @@ Definition Result:
 
 #[test]
 fn keyword_argument_multi_file() {
-    let code_fuction_provider = r#"
+    let code_function_provider = r#"
 def foo(x: int, y: str) -> None:
     pass
 "#;
@@ -495,7 +522,7 @@ foo(0, y="foo")
 #      ^
 "#;
     let report = get_batched_lsp_operations_report(
-        &[("main", code), ("my_func", code_fuction_provider)],
+        &[("main", code), ("my_func", code_function_provider)],
         get_test_report,
     );
     assert_eq!(
@@ -563,7 +590,7 @@ Definition Result:
 6 | foo: Literal[1] = 1
              ^
 Definition Result:
-249 | Literal: _SpecialForm
+253 | Literal: _SpecialForm
       ^^^^^^^
 
 8 | bar = f([1], "", 42)
@@ -884,6 +911,42 @@ Definition Result: None
     );
 }
 
+#[test]
+fn legacy_typevar_test() {
+    let code = r#"
+from pathlib import Path
+
+class A[T, U]:
+    pass
+
+class B:
+    pass
+
+def f(x: A[B, Path]) -> None:
+#          ^  ^
+    pass
+"#;
+    let report = get_batched_lsp_operations_report(&[("main", code)], get_test_report);
+    assert_eq!(
+        r#"
+# main.py
+10 | def f(x: A[B, Path]) -> None:
+                ^
+Definition Result:
+7 | class B:
+          ^
+
+10 | def f(x: A[B, Path]) -> None:
+                   ^
+Definition Result:
+173 | class Path(PurePath):
+            ^^^^
+"#
+        .trim(),
+        report.trim(),
+    );
+}
+
 #[test]
 fn multi_definition_test() {
     let code = r#"
@@ -1223,3 +1286,310 @@ Definition Result:
         report.trim(),
     );
 }
+
+// todo(kylei) go-to definition on x should go to the definition
+#[test]
+fn global_keyword() {
+    let code = r#"
+x = 5
+def test():
+    global x
+    #      ^
+"#;
+    let report = get_batched_lsp_operations_report(&[("main", code)], get_test_report);
+    assert_eq!(
+        r#"
+# main.py
+4 |     global x
+               ^
+Definition Result: None
+"#
+        .trim(),
+        report.trim(),
+    );
+}
+
+#[test]
+fn module_attr_access_should_follow_exports_not_dunder_all_test() {
+    let code_name_provider = r#"
+__all__ = []
+x: int = 42
+"#;
+    let code = r#"
+import foo
+foo.x
+#   ^
+"#;
+    let report = get_batched_lsp_operations_report(
+        &[("main", code), ("foo", code_name_provider)],
+        get_test_report,
+    );
+    assert_eq!(
+        r#"
+# main.py
+3 | foo.x
+        ^
+Definition Result:
+3 | x: int = 42
+    ^
+
+
+# foo.py
+"#
+        .trim(),
+        report.trim(),
+    );
+}
+
+#[test]
+fn renamed_reexport() {
+    let lib2 = r#"
+def foo() -> None: ...
+"#;
+    let lib = r#"
+from lib2 import foo as foo_renamed
+"#;
+    let code = r#"
+from lib import foo_renamed
+#                    ^
+"#;
+    let report = get_batched_lsp_operations_report(
+        &[("main", code), ("lib", lib), ("lib2", lib2)],
+        get_test_report,
+    );
+    assert_eq!(
+        r#"
+# main.py
+2 | from lib import foo_renamed
+                         ^
+Definition Result:
+2 | def foo() -> None: ...
+        ^^^
+
+
+# lib.py
+
+# lib2.py
+"#
+        .trim(),
+        report.trim(),
+    );
+}
+
+#[test]
+fn renamed_assignment_reexport() {
+    let lib2 = r#"
+def foo() -> None: ...
+"#;
+    let lib = r#"
+from lib2 import foo
+foo_renamed = foo
+"#;
+    let code = r#"
+from lib import foo_renamed
+#                    ^
+"#;
+    let report = get_batched_lsp_operations_report(
+        &[("main", code), ("lib", lib), ("lib2", lib2)],
+        get_test_report,
+    );
+    assert_eq!(
+        r#"
+# main.py
+2 | from lib import foo_renamed
+                         ^
+Definition Result:
+3 | foo_renamed = foo
+    ^^^^^^^^^^^
+
+
+# lib.py
+
+# lib2.py
+"#
+        .trim(),
+        report.trim(),
+    );
+}
+
+#[test]
+fn operator_comparison_overload() {
+    let code = r#"
+class My:
+    def __lt__(self, other: object) -> bool:
+        return True
+    def __le__(self, other: object) -> bool:
+        return True
+    def __gt__(self, other: object) -> bool:
+        return True
+    def __ge__(self, other: object) -> bool:
+        return True
+
+a = My()
+b = My()
+result1 = a < b
+#           ^
+result2 = a <= b
+#            ^
+result3 = a > b
+#           ^
+result4 = a >= b
+#            ^
+"#;
+    let report = get_batched_lsp_operations_report(&[("main", code)], get_test_report);
+    assert_eq!(
+        r#"
+# main.py
+14 | result1 = a < b
+                 ^
+Definition Result:
+3 |     def __lt__(self, other: object) -> bool:
+            ^^^^^^
+
+16 | result2 = a <= b
+                  ^
+Definition Result:
+5 |     def __le__(self, other: object) -> bool:
+            ^^^^^^
+
+18 | result3 = a > b
+                 ^
+Definition Result:
+7 |     def __gt__(self, other: object) -> bool:
+            ^^^^^^
+
+20 | result4 = a >= b
+                  ^
+Definition Result:
+9 |     def __ge__(self, other: object) -> bool:
+            ^^^^^^
+"#
+        .trim(),
+        report.trim(),
+    );
+}
+
+#[test]
+fn unreachable_branch() {
+    let code = r#"
+x = 5
+if False:
+    print(x)
+    #     ^
+"#;
+    let report = get_batched_lsp_operations_report(&[("main", code)], get_test_report);
+    assert_eq!(
+        r#"
+# main.py
+4 |     print(x)
+              ^
+Definition Result: None
+
+"#
+        .trim(),
+        report.trim(),
+    );
+}
+
+#[test]
+fn operator_binop_overload() {
+    let code = r#"
+class My:
+    def __add__(self, other: object) -> "My":
+        return self
+    def __sub__(self, other: object) -> "My":
+        return self
+    def __mul__(self, other: object) -> "My":
+        return self
+    def __truediv__(self, other: object) -> "My":
+        return self
+
+a = My()
+b = My()
+result1 = a + b
+#           ^
+result2 = a - b
+#           ^
+result3 = a * b
+#           ^
+result4 = a / b
+#           ^
+"#;
+    let report = get_batched_lsp_operations_report(&[("main", code)], get_test_report);
+    assert_eq!(
+        r#"
+# main.py
+14 | result1 = a + b
+                 ^
+Definition Result:
+3 |     def __add__(self, other: object) -> "My":
+            ^^^^^^^
+
+16 | result2 = a - b
+                 ^
+Definition Result:
+5 |     def __sub__(self, other: object) -> "My":
+            ^^^^^^^
+
+18 | result3 = a * b
+                 ^
+Definition Result:
+7 |     def __mul__(self, other: object) -> "My":
+            ^^^^^^^
+
+20 | result4 = a / b
+                 ^
+Definition Result:
+9 |     def __truediv__(self, other: object) -> "My":
+            ^^^^^^^^^^^
+"#
+        .trim(),
+        report.trim(),
+    );
+}
+
+#[test]
+fn operator_unaryop_overload() {
+    let code = r#"
+class My:
+    def __neg__(self) -> "My":
+        return self
+    def __pos__(self) -> "My":
+        return self
+    def __invert__(self) -> "My":
+        return self
+
+a = My()
+result1 = -a
+#         ^
+result2 = +a
+#         ^
+result3 = ~a
+#         ^
+"#;
+    let report = get_batched_lsp_operations_report(&[("main", code)], get_test_report);
+    assert_eq!(
+        r#"
+# main.py
+11 | result1 = -a
+               ^
+Definition Result:
+3 |     def __neg__(self) -> "My":
+            ^^^^^^^
+
+13 | result2 = +a
+               ^
+Definition Result:
+5 |     def __pos__(self) -> "My":
+            ^^^^^^^
+
+15 | result3 = ~a
+               ^
+Definition Result:
+7 |     def __invert__(self) -> "My":
+            ^^^^^^^^^^
+"#
+        .trim(),
+        report.trim(),
+    );
+}
