@@ -7,6 +7,7 @@
 
 #![cfg(test)]
 
+mod abstract_methods;
 mod assign;
 mod attribute_narrow;
 mod attributes;
@@ -25,17 +26,23 @@ mod decorators;
 mod delayed_inference;
 mod descriptors;
 mod dict;
+mod django;
 mod enums;
-mod flow;
+mod flow_branching;
+mod flow_looping;
 mod generic_basic;
+mod generic_legacy;
 mod generic_restrictions;
+mod generic_sub;
 mod imports;
 mod incremental;
+mod inference;
 mod literal;
 mod lsp;
 mod mro;
 mod named_tuple;
 mod narrow;
+mod natural;
 mod new_type;
 mod operators;
 mod overload;
@@ -43,6 +50,9 @@ mod paramspec;
 mod pattern_match;
 mod perf;
 mod protocol;
+mod pydantic;
+mod pysa;
+mod redundant_cast;
 mod returns;
 mod scope;
 mod simple;
@@ -50,6 +60,7 @@ mod state;
 mod subscript_narrow;
 mod suppression;
 mod sys_info;
+mod tsp;
 mod tuple;
 mod type_alias;
 mod type_var_tuple;
