@@ -52,19 +52,6 @@ def f(x: str | None):
     "#,
 );
 
-testcase!(
-    bug = "We don't identify bool() as builtins.bool if it's used elsewhere as a type annotation, instead the special exports system thinks it's defined by the current file (main.bool)",
-    test_bool_special_exports_bug,
-    r#"
-from typing import assert_type
-def f(x: bool):
-    if bool(x):
-        assert_type(x, bool)  # should be Literal[True]
-    else:
-        assert_type(x, bool)  # should be Literal[False]
-    "#,
-);
-
 testcase!(
     test_eq,
     r#"
@@ -229,6 +216,32 @@ def f(x: bool | None):
     "#,
 );
 
+testcase!(
+    test_or_multiple_vars,
+    r#"
+from typing import assert_type
+def f(x: int | None, y: int | None) -> None:
+    if x is None or y is None:
+        assert_type(y, int | None)
+        assert_type(x, int | None)
+    else:
+        assert_type(y, int)
+        assert_type(x, int)
+    "#,
+);
+
+testcase!(
+    test_or_walrus_multiple_vars,
+    r#"
+from typing import assert_type
+def f(x: None | int, y: None | int) -> None:
+    if (z := x is None) or y is None:
+        assert_type(x, None | int)
+        assert_type(y, None | int)
+        assert_type(z, bool)
+    "#,
+);
+
 testcase!(
     test_elif,
     r#"
@@ -249,14 +262,14 @@ def f(x: str | None, y: int | None):
 testcase!(
     test_not,
     r#"
-from typing import assert_type
+from typing import assert_type, Literal
 def f(x: str | None):
     if not x is None:
         assert_type(x, str)
     else:
         assert_type(x, None)
     if not x:
-        assert_type(x, str | None)
+        assert_type(x, None | Literal[""])
     else:
         assert_type(x, str)
     "#,
@@ -307,6 +320,22 @@ def f(x: str | None):
     "#,
 );
 
+testcase!(
+    test_prod_assert,
+    r#"
+from typing import assert_type
+def prod_assert(x: object, msg: str | None = None): ...
+
+def test_only(x: str | None) -> None:
+    prod_assert(x is not None)
+    assert_type(x, str)
+
+def test_and_message(x: str | None) -> None:
+    prod_assert(x is not None, "x is None")
+    assert_type(x, str)
+    "#,
+);
+
 testcase!(
     test_while_else,
     r#"
@@ -324,7 +353,7 @@ assert_type(x, str)
 );
 
 testcase!(
-    test_while_break,
+    test_while_break_no_else,
     r#"
 from typing import assert_type
 def f() -> str | None: ...
@@ -422,6 +451,20 @@ if x := f():
     "#,
 );
 
+testcase!(
+    test_walrus_with_rhs_narrowing,
+    r#"
+from typing import assert_type, Literal
+def f(x: None | int) -> int:
+    # this is: y := (x is None)
+    # `y` being truthy means `x is None` is also true
+    if y := x is None:
+        assert_type(y, Literal[True])
+        x = 1
+    return x
+    "#,
+);
+
 testcase!(
     test_walrus_value,
     r#"
@@ -562,6 +605,24 @@ def f(x: str | int):
     "#,
 );
 
+testcase!(
+    test_dunder_bool_truthy_narrow,
+    r#"
+from typing import assert_type, Literal
+class Falsey:
+    def __bool__(self) -> Literal[False]:
+        return False
+class Truthy:
+    def __bool__(self) -> Literal[True]:
+        return True
+def f(x: Falsey | Truthy):
+    if x:
+        assert_type(x, Truthy)
+    else:
+        assert_type(x, Falsey)
+    "#,
+);
+
 testcase!(
     test_type_eq,
     r#"
@@ -812,6 +873,7 @@ class B: ...
 def test(x: A | B):
     y = isinstance(x, A) and (z := True)
     assert_type(x, A | B)
+    # Intended false negative for uninitialized local check.
     assert_type(z, Literal[True])
     "#,
 );
@@ -1210,7 +1272,7 @@ class B: pass
 class C(A, B): pass  # not used, but demonstrates why the narrow is not Never
 def f(x: A):
     if isinstance(x, B):
-        # In theory we could use `A & B` here; any common sublcass like `C` is possible.
+        # In theory we could use `A & B` here; any common subclass like `C` is possible.
         # Given that we don't have intersections, we follow Pyre's lead and use `B`
         assert_type(x, B)
 "#,
@@ -1569,3 +1631,176 @@ class C:
         return isinstance(x, int)
 "#,
 );
+
+testcase!(
+    test_while_try_except,
+    r#"
+from typing import assert_type
+class Test:
+    x: dict[str, str] | None
+    def test(self) -> None:
+        assert self.x is not None
+        while True:
+            try:
+                assert_type(self.x, dict[str, str])
+                x = self.x.get("asdf")
+            except:
+                pass
+    "#,
+);
+
+testcase!(
+    test_discriminated_union_key,
+    r#"
+from typing import TypedDict, assert_type, Literal
+
+class UserDict(TypedDict):
+    kind: Literal["user"]
+    is_admin: Literal[False]
+
+class AdminDict(TypedDict):
+    kind: Literal["admin"]
+    is_admin: Literal[True]
+
+def test(x: UserDict | AdminDict):
+    if x["kind"] == "user":
+        assert_type(x, UserDict)
+    else:
+        assert_type(x, AdminDict)
+    if x["is_admin"] is True:
+        assert_type(x, AdminDict)
+    else:
+        assert_type(x, UserDict)
+    "#,
+);
+
+testcase!(
+    test_discriminated_union_attr,
+    r#"
+from typing import assert_type, Literal
+
+class User:
+    kind: Literal["user"]
+    is_admin: Literal[False]
+
+class Admin:
+    kind: Literal["admin"]
+    is_admin: Literal[True]
+
+def test(x: User | Admin):
+    if x.kind == "user":
+        assert_type(x, User)
+    elif x.kind == "admin":
+        assert_type(x, Admin)
+
+    if x.is_admin is True:
+        assert_type(x, Admin)
+    else:
+        assert_type(x, User)
+    "#,
+);
+
+testcase!(
+    test_discriminated_union_index,
+    r#"
+from typing import assert_type, Literal
+
+def test(x: tuple[Literal["user"], Literal[None]] | tuple[Literal["admin"], int]):
+    if x[0] == "user":
+        assert_type(x, tuple[Literal["user"], Literal[None]])
+    else:
+        assert_type(x, tuple[Literal["admin"], int])
+
+    if x[1] is None:
+        assert_type(x, tuple[Literal["user"], Literal[None]])
+    else:
+        assert_type(x, tuple[Literal["admin"], int])
+
+    if x[1] is not None:
+        assert_type(x, tuple[Literal["admin"], int])
+    else:
+        assert_type(x, tuple[Literal["user"], Literal[None]])
+    "#,
+);
+
+testcase!(
+    test_narrow_and_placeholder,
+    r#"
+from typing import assert_type
+
+class A: pass
+class B: pass
+
+def test1(flag: bool, x: A | B) -> None:
+    if isinstance(x, A) and flag:
+        pass
+    else:
+        assert_type(x, A | B)
+
+def test2(flag: bool, x: A | B) -> None:
+    if flag and isinstance(x, A):
+        pass
+    else:
+        assert_type(x, A | B)
+
+def test3(x: A | B, y: A | B):
+    if isinstance(x, A) and isinstance(y, B):
+        pass
+    else:
+        assert_type(x, A | B)
+
+def foo() -> bool: ...
+def test4(x: int | None) -> None:
+    if x is not None and foo():
+        return
+    assert_type(x, int | None)
+    "#,
+);
+
+testcase!(
+    test_truthy_falsy_builtins,
+    r#"
+from typing import assert_type, Literal
+def test(a: int, b: str, c: bytes):
+    if not a:
+        assert_type(a, Literal[0])
+    else:
+        assert_type(a, int)
+    assert_type(a, int)
+
+    if not b:
+        assert_type(b, Literal[""])
+    else:
+        assert_type(b, str)
+    assert_type(b, str)
+
+    if not c:
+        assert_type(c, Literal[b""])
+    else:
+        assert_type(c, bytes)
+    assert_type(c, bytes)
+
+    if a:
+        assert_type(a, int)
+    else:
+        assert_type(a, Literal[0])
+
+    if b:
+        assert_type(b, str)
+    else:
+        assert_type(b, Literal[""])
+
+    if c:
+        assert_type(c, bytes)
+    else:
+        assert_type(c, Literal[b""])
+    "#,
+);
+
+testcase!(
+    test_do_not_narrow_class_name,
+    r#"
+assert issubclass(list, object)
+x: list[int] = [1]
+    "#,
+);
