@@ -31,6 +31,31 @@ def foo(x: A | B) -> None:
 "#,
 );
 
+testcase!(
+    test_union_not_callable,
+    r#"
+from typing import assert_type
+class A:
+    def foo(self) -> int: ...
+class B:
+    foo: str
+def foo(x: A | B) -> None:
+    x.foo()  # E: Expected a callable, got `str`
+"#,
+);
+
+testcase!(
+    test_distribute_type_over_union,
+    r#"
+from typing import assert_type
+def f(x: type[int] | type[str], y: type[int | str]) -> None:
+    assert_type(x, type[int] | type[str])
+    assert_type(x, type[int | str])
+    assert_type(y, type[int] | type[str])
+    assert_type(y, type[int | str])
+"#,
+);
+
 testcase!(
     test_simple_call,
     r#"
@@ -200,13 +225,14 @@ x: "list" "[int]" = [] # E: Expected a type form
 testcase!(
     test_globals,
     r#"
-from typing import assert_type, Any
+from typing import assert_type, Any, MutableSequence
 assert_type(__file__, str)
 assert_type(__name__, str)
 assert_type(__debug__, bool)
 assert_type(__package__, str | None)
 assert_type(__annotations__, dict[str, Any])
 assert_type(__spec__, Any)
+assert_type(__path__, MutableSequence[str])
 "#,
 );
 
@@ -267,22 +293,6 @@ assert_type(f(), str)
 "#,
 );
 
-testcase!(
-    test_class_rebind_attribute,
-    r#"
-from typing import assert_type
-def f(x: str) -> int:
-    return 42
-
-attribute = "test"
-
-class C:
-    attribute = f(attribute)
-
-assert_type(C().attribute, int)
-"#,
-);
-
 testcase!(
     test_class_attribute_lookup,
     r#"
@@ -327,7 +337,7 @@ testcase!(
 from typing import Final, assert_type, Literal, cast
 x: Final[int] = 1
 y: Final = "test"
-z: Final[str] = cast(str, "")
+z: Final[str] = cast(str, "")  # No warning - Literal[''] is not exactly the same as str
 w: Final[int] = "bad"  # E: `Literal['bad']` is not assignable to `int`
 
 assert_type(x, int)
@@ -425,17 +435,18 @@ assert_type(0 if derp() else 1, Literal[0] | Literal[1])
 testcase!(
     test_raise,
     r#"
-def test_raise() -> None:
+def test_raise(exception_or_none: BaseException | None) -> None:
     raise
-    raise None  # E: does not derive from BaseException
+    raise None  # E: expected `BaseException`
     raise BaseException
     raise BaseException()
-    raise 42  # E: does not derive from BaseException
+    raise 42  # E: expected `BaseException`
     raise BaseException from None
     raise BaseException() from None
     raise BaseException() from BaseException
     raise BaseException() from BaseException()
-    raise BaseException() from 42   # E: does not derive from BaseException
+    raise BaseException() from 42   # E: expected `BaseException` or `None`
+    raise BaseException() from exception_or_none
 "#,
 );
 
@@ -472,7 +483,7 @@ testcase!(
 from typing import Any, assert_type
 x = [[], [], [[]]]
 # Not too important it is precisely this type, but detect changes
-assert_type(x, list[list[list[Any]]])
+assert_type(x, list[list[list[Any]] | list[Any]])
 "#,
 );
 
@@ -517,7 +528,7 @@ def f(x: list[int], y: dict[str, bool]) -> None:
     c = C()
     c[0]  # E: Cannot index into `C`\n  Object of class `C` has no attribute `__getitem__`
     d = D()
-    d[0]  # E: Expected `__getitem__` to be a callable, got int
+    d[0]  # E: Expected `__getitem__` to be a callable, got `int`
 "#,
 );
 
@@ -724,6 +735,31 @@ def f(a: A):
     "#,
 );
 
+testcase!(
+    test_literal_string_subscript_precision,
+    r#"
+from typing import Literal, assert_type
+s: Literal["abcde"] = "abcde"
+ss: Literal["こんにちは"] = "こんにちは"
+em: Literal["\U0001F44D\U0001F3FC"] = "\U0001F44D\U0001F3FC"
+assert_type(s[0], Literal["a"])
+assert_type(s[0:2], Literal["ab"])
+assert_type(s[-1], Literal["e"])
+assert_type(s[2:], Literal["cde"])
+assert_type(s[::-1], Literal["edcba"])
+assert_type(s[3:0:-2], Literal["db"])
+assert_type(ss[0], Literal["こ"])
+assert_type(ss[-1], Literal["は"])
+assert_type(ss[2:], Literal["にちは"])
+assert_type(ss[:1], Literal["こ"])
+assert_type(ss[::-1], Literal["はちにんこ"])
+assert_type(em[0], Literal["\U0001F44D"])
+assert_type(em[-1], Literal["\U0001F3FC"])
+assert_type(em[:1], Literal["\U0001F44D"])
+assert_type(em[::-1], Literal["\U0001F3FC\U0001F44D"])
+    "#,
+);
+
 testcase!(
     test_invalid_annotation,
     r#"
@@ -734,7 +770,7 @@ def test(
     b: lambda: None,  # E: lambda definition cannot be used in annotations
     c: [foo(arg=val)], # E: list literal cannot be used in annotations
     d: (1, 2), # E: tuple literal cannot be used in annotations
-    e: 1 + 2,  # E: expression cannot be used in annotations
+    e: 1 + 2,  # E: binary operation `+` cannot be used in annotations
 ): ...
 "#,
 );
@@ -770,6 +806,28 @@ def f(x: Annotated[int, "test"], y: Annotated[int, "test", "test"]):
     "#,
 );
 
+testcase!(
+    test_nested_string_annotation,
+    r#"
+x: "'int'" = 1  # E: Expected a type form, got instance of `Literal['int']`
+y: "list['int']" = [1]  # E: Expected a type form, got instance of `Literal['int']`
+    "#,
+);
+
+testcase!(
+    test_nested_subscript_annotation,
+    r#"
+# we'll emit parse errors for invalid string literal annotations that aren't inside a subscript
+# for annotations inside a subscript, we already emit an "expected a type" annotation at solving time
+import typing
+# type: ignore
+PyTree: typing.TypeAlias = getattr(typing, "foo" + "bar")
+x: PyTree[int, " T"] = 1  # this is fine because the base of the subscript is some invalid type so it's all `Any`
+y: list[" T"] = []  # E: Expected a type form, got instance of `Literal[' T']
+z: " T" = 1  # E: Expected a type form, got instance of `Literal[' T']`
+    "#,
+);
+
 testcase!(
     test_no_backtracking,
     r#"
@@ -794,7 +852,8 @@ def foo(x: int):
 testcase!(
     test_builtins_type_constructor,
     r#"
-from typing import assert_type
+from typing import assert_type, LiteralString, Literal, TypedDict
+from types import NoneType
 class Foo:
     @classmethod
     def method(cls, x: str) -> int:
@@ -803,6 +862,17 @@ class Foo:
         assert_type(type(self).method("tst"), int)
 x = Foo()
 assert_type(type(x), type[Foo])
+x2: LiteralString = "test"
+assert_type(type(x2), type[str])
+x3 = (1, 2)
+assert_type(type(x3), type[tuple])
+x4 = None
+assert_type(type(x4), type[NoneType])
+x5: Literal[""] = ""
+assert_type(type(x5), type[str])
+class TD(TypedDict): ...
+x6: TD = {}
+assert_type(type(x6), type[dict])
 "#,
 );
 
@@ -916,7 +986,7 @@ testcase!(
     test_empty_if,
     r#"
 # This parse error results in two identical Identifiers, which previously caused a panic.
-a = True if # E: Parse 
+a = True if # E: Parse
 "#,
 );
 
@@ -984,7 +1054,7 @@ import typing
 typing.assert_type(0, str)  # E: assert_type(Literal[0], str) failed
 # Make sure that calling by bare name without importing performs the assertion, as this is very convenient for debugging.
 # It's fine if a name error is also generated.
-assert_type(0, str)  # E: assert_type(Literal[0], str) failed  # E: Could not find name `assert_type`
+assert_type(0, str)  # E: assert_type(Literal[0], str) failed  # E: `assert_type` must be imported from `typing` for runtime usage
     "#,
 );
 
@@ -996,7 +1066,7 @@ import typing
 typing.reveal_type(0)  # E: revealed type: Literal[0]
 # Make sure that calling by bare name without importing reveals the type, as this is very convenient for debugging.
 # It's fine if a name error is also generated.
-reveal_type(0)  # E: revealed type: Literal[0]  # E: Could not find name `reveal_type`
+reveal_type(0)  # E: revealed type: Literal[0]  # E: `reveal_type` must be imported from `typing` for runtime usage
     "#,
 );
 
@@ -1060,7 +1130,7 @@ testcase!(
 from typing import Literal, assert_type, reveal_type
 
 a = assert_type(0, str) # E: assert_type(Literal[0], str) failed
-assert_type(a, Literal[0]) 
+assert_type(a, Literal[0])
 
 b = reveal_type(0) # E: revealed type: Literal[0]
 assert_type(b, Literal[0])
@@ -1330,69 +1400,76 @@ testcase!(
     r#"
 class NotBoolable:
     __bool__: int = 3
-
-# bool()
-y = bool(NotBoolable())  # E: `NotBoolable.__bool__` has type `int`, which is not callable
-
-# if expressions
-x = 0 if NotBoolable() else 1  # E: `NotBoolable.__bool__` has type `int`, which is not callable
-
-# if statements
-if NotBoolable(): ...  # E: `NotBoolable.__bool__` has type `int`, which is not callable
-
-# while statements
-while NotBoolable(): ...  # E: `NotBoolable.__bool__` has type `int`, which is not callable
-
-# expression evaluating to NotBoolable
 def f() -> NotBoolable:
   return NotBoolable()
 
-if (f() if True else None): ...  # E: `NotBoolable.__bool__` has type `int`, which is not callable
+y = bool(NotBoolable())  # E: has type `int`, which is not callable
+z = not NotBoolable()  # E: has type `int`, which is not callable
+x = 0 if NotBoolable() else 1  # E: has type `int`, which is not callable  # E: Expected `__bool__` to be a callable, got `int`
+if NotBoolable(): ...  # E: has type `int`, which is not callable
+while NotBoolable(): ...  # E: has type `int`, which is not callable
+if f(): ...  # E: has type `int`, which is not callable
+
+# We don't treat `__getattr__` as implying a `__bool__`.
+class C:
+    def __getattr__(self, x: str) -> object: ...
+def test(o: C):
+    if o: # should not fail
+        pass
 "#,
 );
 
-// Check that we don't raise false positives (or true positives that we want to avoid failing on)
-// for some corner cases
 testcase!(
     test_valid_dunder_bool,
     r#"
-
 from typing import Any, Literal, Never
 
-# Any is always assumed to be valid
+# We always allow `Any` to be used as a bool
 def f(x):
-  if x: ...
-
+    if x: ...
 def g(x: Any):
-  if x: ...
+    if x: ...
 
-# Never does not raise a type error
+# We always allow `Never` to be used as a bool
 def g() -> Never:
     raise Exception()
-
 if g(): ...
 
-# Union types are not checked due to risk of false positives
-class B:
-  def __bool__(self) -> bool:
-    return True
-
+# We don't treat `__getattr__` as implying a `__bool__`
 class C:
-  def __bool__(self) -> Literal[False]:
-    return False
+    def __getattr__(self, x: str) -> object: ...
+def test(o: C):
+    if o: # should not fail
+        pass
+"#,
+);
 
+testcase!(
+    test_union_dunder_bool,
+    r#"
+from typing import Literal
+
+class B:
+    def __bool__(self) -> bool: return True
+class C:
+    def __bool__(self) -> Literal[False]: return False
 class D:
     __bool__ = 42
+class E:
+    pass
 
-def j(x: B | C):
-    if x: ...
+def f(ok: B | C, bad: B | D):
+    if ok: ...
+    if bad: ...  # E: The `__bool__` attribute of `D` has type `int`, which is not callable
 
-# Invalid, but we don't check
-def k(x: B | D):
-    if x: ...
-
-def l(x: int | None):
-    if x: ...
+# Regression tests for https://github.com/facebook/pyrefly/issues/1364
+def g(ok: B | None, bad1: D | None, bad2: D | E):
+    if ok: ...
+    if bad1 is None: ...  # This is okay - we are not using a truthiness check
+    if bad1: ...  # E: The `__bool__` attribute of `D` has type `int`, which is not callable
+    if not bad1: ... # E: The `__bool__` attribute of `D` has type `int`, which is not callable
+    if bad2: ...  # E: The `__bool__` attribute of `D` has type `int`, which is not callable
+    if not bad2: ...  # E: The `__bool__` attribute of `D` has type `int`, which is not callable
 "#,
 );
 
@@ -1429,19 +1506,6 @@ def g(f: Callable[[Any], int], inputs: Any) -> None:
     "#,
 );
 
-testcase!(
-    test_legacy_typevar_revealed_type,
-    r#"
-from typing import reveal_type, TypeVar
-
-T = TypeVar("T")
-TypeForm = type[T]
-
-reveal_type(T)  # E: type[TypeVar[T]]
-reveal_type(TypeForm)  # E: revealed type: type[type[T]]
-    "#,
-);
-
 testcase!(
     test_union_function_exponential,
     r#"
@@ -1500,6 +1564,22 @@ t: "õ" # E: Could not find name `õ`
 "#,
 );
 
+testcase!(
+    test_unused_coroutine,
+    r#"
+from typing import Any
+def not_async() -> Any:
+    pass
+async def foo():
+    return 1
+async def bar():
+    foo()  # E: Result of async function call is unused. Did you forget to `await`?
+    await foo()  # ok
+    x = foo()  # ok
+    not_async()  # ok
+"#,
+);
+
 testcase!(
     test_loop_forever,
     r#"
@@ -1669,7 +1749,7 @@ testcase!(
     test_bin_op_bool_num,
     r#"
 def f(x:float, y:bool) -> float:
-    return x * y 
+    return x * y
 "#,
 );
 
@@ -1689,6 +1769,19 @@ def f(condition: bool):
     "#,
 );
 
+testcase!(
+    test_class_getitem_magic_dunder,
+    r#"
+from typing import assert_type
+
+class Foo:
+    def __class_getitem__(cls, item: int) -> str:
+        return str(item)
+
+assert_type(Foo[0], str)
+"#,
+);
+
 testcase!(test_panic_docstring, "\"\"\" F\n\u{85}\"\"\"",);
 
 testcase!(
@@ -1704,3 +1797,110 @@ testcase!(
 [a for [a for # E: Could # E: Parse # E: Could # E: Parse # E: Parse
 "#,
 );
+
+// Added to stress test cases that at one point panicked from duplicate
+// first-use linkage keys due to overlapping ranges.
+testcase!(
+    test_overlapping_link_keys,
+    r#"
+def f():
+    del yield  # E: # E:
+    assert False, yield  # E:
+    "#,
+);
+
+testcase!(
+    test_lambda_args,
+    r#"
+from typing import reveal_type
+f1 = lambda x, *args, **kwargs: x
+reveal_type(f1) # E: revealed type: (x: Unknown, *args: Unknown, **kwargs: Unknown) -> Unknown
+    "#,
+);
+
+testcase!(
+    test_missing_name_in_dunder_all,
+    r#"
+__all__ = ["x", "y"]
+x = 5
+    "#,
+);
+
+testcase!(
+    test_ellipsis_default_source,
+    r#"
+def f(x: bool = ...):  # E: Default `Ellipsis` is not assignable to parameter `x` with type `bool`
+    pass
+    "#,
+);
+
+testcase!(
+    test_ellipsis_default_stub,
+    TestEnv::one_with_path("foo", "foo.pyi", "def f(x: bool = ...): pass"),
+    r#"
+import foo
+    "#,
+);
+
+testcase!(
+    test_functional_type_creation,
+    r#"
+from typing import assert_type
+X = type("X", (object,), {"foo": "bar"})
+assert_type(X, type)
+    "#,
+);
+
+testcase!(
+    test_typing_annotated_validation,
+    r#"
+from typing import Annotated, ClassVar
+X = Annotated[int, 'ok']
+class A:
+    x: Annotated[ClassVar[int], 'also ok']
+Y = Annotated[ClassVar[int], 'wrong context']  # E: `ClassVar` is not allowed in this context
+Z = Annotated[0, 'not a type']  # E: Expected a type form, got instance of `Literal[0]`
+    "#,
+);
+
+testcase!(
+    test_starred_empty_tuple_no_panic,
+    r#"
+(),*()
+    "#,
+);
+
+testcase!(
+    test_asyncio_gather,
+    r#"
+import asyncio
+from typing import assert_type
+
+async def wait(n: int) -> int:
+    await asyncio.sleep(0.1)
+    return n
+
+async def main() -> None:
+    a = await asyncio.gather(*[wait(i) for i in range(10)])
+    assert_type(a, list[int])
+    "#,
+);
+
+testcase!(
+    test_int_supports_index,
+    r#"
+from typing import SupportsIndex
+x: SupportsIndex = 3
+    "#,
+);
+
+// A request from Apache Airflow: make sure we don't add logic that complains on
+// an unused bitshift expression - some linters dislike this, but Airflow uses it
+// as a side-effecting operation to express DAG dependencies.
+testcase!(
+    test_bitshift_as_side_effect_does_not_error,
+    r#"
+def f(x: int, y: int):
+    x >> y
+    "#,
+);
