@@ -17,6 +17,11 @@ use std::time::Instant;
 use anstream::ColorChoice;
 use anyhow::anyhow;
 use dupe::Dupe;
+use pyrefly_build::handle::Handle;
+use pyrefly_build::source_db::map_db::MapDatabase;
+use pyrefly_config::error::ErrorDisplayConfig;
+use pyrefly_config::error_kind::ErrorKind;
+use pyrefly_config::error_kind::Severity;
 use pyrefly_python::module_name::ModuleName;
 use pyrefly_python::module_path::ModulePath;
 use pyrefly_python::module_path::ModulePathDetails;
@@ -35,7 +40,6 @@ use ruff_source_file::PositionEncoding;
 use ruff_source_file::SourceLocation;
 use ruff_text_size::TextRange;
 use ruff_text_size::TextSize;
-use starlark_map::small_map::SmallMap;
 
 use crate::binding::binding::KeyExport;
 use crate::config::base::UntypedDefBehavior;
@@ -43,7 +47,8 @@ use crate::config::config::ConfigFile;
 use crate::config::finder::ConfigFinder;
 use crate::error::error::print_errors;
 use crate::module::finder::find_import;
-use crate::state::handle::Handle;
+use crate::state::errors::Errors;
+use crate::state::load::FileContents;
 use crate::state::require::Require;
 use crate::state::state::State;
 use crate::state::subscriber::TestSubscriber;
@@ -92,18 +97,42 @@ fn default_path(module: ModuleName) -> PathBuf {
     PathBuf::from(format!("{}.py", module.as_str().replace('.', "/")))
 }
 
-#[derive(Debug, Default, Clone, PartialEq, Eq)]
+#[derive(Debug, Clone, PartialEq, Eq)]
 pub struct TestEnv {
-    modules: SmallMap<ModuleName, (ModulePath, Option<Arc<String>>)>,
+    modules: Vec<(ModuleName, ModulePath, Option<Arc<FileContents>>)>,
     version: PythonVersion,
     untyped_def_behavior: UntypedDefBehavior,
+    infer_with_first_use: bool,
+    site_package_path: Vec<PathBuf>,
+    implicitly_defined_attribute_error: bool,
+    implicit_any_error: bool,
+    implicit_abstract_class_error: bool,
+    open_unpacking_error: bool,
+    default_require_level: Require,
 }
 
 impl TestEnv {
     pub fn new() -> Self {
         // We aim to init the tracing before now, but if not, better now than never
         init_test();
-        Self::default()
+        TestEnv {
+            modules: Vec::new(),
+            version: PythonVersion::default(),
+            untyped_def_behavior: UntypedDefBehavior::default(),
+            infer_with_first_use: true,
+            site_package_path: Vec::new(),
+            implicitly_defined_attribute_error: false,
+            implicit_any_error: false,
+            implicit_abstract_class_error: false,
+            open_unpacking_error: false,
+            default_require_level: Require::Exports,
+        }
+    }
+
+    pub fn new_with_site_package_path(path: &str) -> Self {
+        let mut res = Self::new();
+        res.site_package_path = vec![PathBuf::from(path)];
+        res
     }
 
     pub fn new_with_version(version: PythonVersion) -> Self {
@@ -118,27 +147,62 @@ impl TestEnv {
         res
     }
 
+    pub fn new_with_infer_with_first_use(infer_with_first_use: bool) -> Self {
+        let mut res = Self::new();
+        res.infer_with_first_use = infer_with_first_use;
+        res
+    }
+
+    pub fn enable_implicitly_defined_attribute_error(mut self) -> Self {
+        self.implicitly_defined_attribute_error = true;
+        self
+    }
+
+    pub fn enable_implicit_any_error(mut self) -> Self {
+        self.implicit_any_error = true;
+        self
+    }
+
+    pub fn enable_implicit_abstract_class_error(mut self) -> Self {
+        self.implicit_abstract_class_error = true;
+        self
+    }
+
+    pub fn enable_open_unpacking_error(mut self) -> Self {
+        self.open_unpacking_error = true;
+        self
+    }
+
+    pub fn with_default_require_level(mut self, level: Require) -> Self {
+        self.default_require_level = level;
+        self
+    }
+
+    pub fn with_version(mut self, version: PythonVersion) -> Self {
+        self.version = version;
+        self
+    }
+
     pub fn add_with_path(&mut self, name: &str, path: &str, code: &str) {
         assert!(
             path.ends_with(".py") || path.ends_with(".pyi") || path.ends_with(".rs"),
             "{path} doesn't look like a reasonable path"
         );
-        self.modules.insert(
+        self.modules.push((
             ModuleName::from_str(name),
-            (
-                ModulePath::memory(PathBuf::from(path)),
-                Some(Arc::new(code.to_owned())),
-            ),
-        );
+            ModulePath::memory(PathBuf::from(path)),
+            Some(Arc::new(FileContents::from_source(code.to_owned()))),
+        ));
     }
 
     pub fn add(&mut self, name: &str, code: &str) {
         let module_name = ModuleName::from_str(name);
         let relative_path = ModulePath::memory(default_path(module_name));
-        self.modules.insert(
+        self.modules.push((
             module_name,
-            (relative_path, Some(Arc::new(code.to_owned()))),
-        );
+            relative_path,
+            Some(Arc::new(FileContents::from_source(code.to_owned()))),
+        ));
     }
 
     pub fn one(name: &str, code: &str) -> Self {
@@ -156,18 +220,18 @@ impl TestEnv {
     pub fn add_real_path(&mut self, name: &str, path: PathBuf) {
         let module_name = ModuleName::from_str(name);
         self.modules
-            .insert(module_name, (ModulePath::filesystem(path), None));
+            .push((module_name, ModulePath::filesystem(path), None));
     }
 
     pub fn sys_info(&self) -> SysInfo {
         SysInfo::new(self.version, PythonPlatform::linux())
     }
 
-    pub fn get_memory(&self) -> Vec<(PathBuf, Option<Arc<String>>)> {
+    pub fn get_memory(&self) -> Vec<(PathBuf, Option<Arc<FileContents>>)> {
         self.modules
-            .values()
-            .filter_map(|(path, contents)| match path.details() {
-                ModulePathDetails::Memory(path) => Some((path.clone(), contents.dupe())),
+            .iter()
+            .filter_map(|(_, path, contents)| match path.details() {
+                ModulePathDetails::Memory(path) => Some(((**path).clone(), contents.dupe())),
                 _ => None,
             })
             .collect()
@@ -177,11 +241,30 @@ impl TestEnv {
         let mut config = ConfigFile::default();
         config.python_environment.python_version = Some(self.version);
         config.python_environment.python_platform = Some(PythonPlatform::linux());
-        config.python_environment.site_package_path = Some(Vec::new());
+        config.python_environment.site_package_path = Some(self.site_package_path.clone());
         config.root.untyped_def_behavior = Some(self.untyped_def_behavior);
-        for (name, (path, _)) in self.modules.iter() {
-            config.custom_module_paths.insert(*name, path.clone());
+        config.root.infer_with_first_use = Some(self.infer_with_first_use);
+        if config.root.errors.is_none() {
+            config.root.errors = Some(ErrorDisplayConfig::new(HashMap::new()));
+        };
+        let errors = config.root.errors.as_mut().unwrap();
+        if self.implicitly_defined_attribute_error {
+            errors.set_error_severity(ErrorKind::ImplicitlyDefinedAttribute, Severity::Error);
         }
+        if self.implicit_any_error {
+            errors.set_error_severity(ErrorKind::ImplicitAny, Severity::Error);
+        }
+        if self.implicit_abstract_class_error {
+            errors.set_error_severity(ErrorKind::ImplicitAbstractClass, Severity::Error);
+        }
+        if self.open_unpacking_error {
+            errors.set_error_severity(ErrorKind::OpenUnpacking, Severity::Error);
+        }
+        let mut sourcedb = MapDatabase::new(config.get_sys_info());
+        for (name, path, _) in self.modules.iter() {
+            sourcedb.insert(*name, path.dupe());
+        }
+        config.source_db = Some(Arc::new(Box::new(sourcedb)));
         config.interpreters.skip_interpreter_query = true;
         config.configure();
         ArcId::new(config)
@@ -200,19 +283,21 @@ impl TestEnv {
             // Reverse so we start at the last file, which is likely to be what the user
             // would have opened, so make it most faithful.
             .rev()
-            .map(|(x, (path, _))| Handle::new(*x, path.clone(), config.dupe()))
+            .map(|(x, path, _)| Handle::new(*x, path.dupe(), config.dupe()))
             .collect::<Vec<_>>();
         let state = State::new(self.config_finder());
         let subscriber = TestSubscriber::new();
-        let mut transaction =
-            state.new_committable_transaction(Require::Exports, Some(Box::new(subscriber.dupe())));
+        let mut transaction = state.new_committable_transaction(
+            self.default_require_level,
+            Some(Box::new(subscriber.dupe())),
+        );
         transaction.as_mut().set_memory(self.get_memory());
-        transaction
-            .as_mut()
-            .run(&handles.map(|x| (x.dupe(), Require::Everything)));
+        transaction.as_mut().run(&handles, Require::Everything);
         state.commit_transaction(transaction);
         subscriber.finish();
+        let project_root = PathBuf::new();
         print_errors(
+            project_root.as_path(),
             &state
                 .transaction()
                 .get_errors(handles.iter())
@@ -223,7 +308,7 @@ impl TestEnv {
             let name = ModuleName::from_str(module);
             Handle::new(
                 name,
-                find_import(&config_file, name, None).unwrap(),
+                find_import(&config_file, name, None).finding().unwrap(),
                 config.dupe(),
             )
         })
@@ -309,13 +394,16 @@ pub fn extract_cursors_for_test(source: &str) -> Vec<TextSize> {
 
 pub fn mk_multi_file_state(
     files: &[(&'static str, &str)],
+    default_require_level: Require,
     assert_zero_errors: bool,
 ) -> (HashMap<&'static str, Handle>, State) {
     let mut test_env = TestEnv::new();
     for (name, code) in files {
         test_env.add(name, code);
     }
-    let (state, handle) = test_env.to_state();
+    let (state, handle) = test_env
+        .with_default_require_level(default_require_level)
+        .to_state();
     let mut handles = HashMap::new();
     for (name, _) in files {
         handles.insert(*name, handle(name));
@@ -340,8 +428,9 @@ pub fn mk_multi_file_state(
 
 pub fn mk_multi_file_state_assert_no_errors(
     files: &[(&'static str, &str)],
+    default_require_level: Require,
 ) -> (HashMap<&'static str, Handle>, State) {
-    mk_multi_file_state(files, true)
+    mk_multi_file_state(files, default_require_level, true)
 }
 
 fn get_batched_lsp_operations_report_helper(
@@ -349,7 +438,7 @@ fn get_batched_lsp_operations_report_helper(
     assert_zero_errors: bool,
     get_report: impl Fn(&State, &Handle, TextSize) -> String,
 ) -> String {
-    let (handles, state) = mk_multi_file_state(files, assert_zero_errors);
+    let (handles, state) = mk_multi_file_state(files, Require::indexing(), assert_zero_errors);
     let mut report = String::new();
     for (name, code) in files {
         report.push_str("# ");
@@ -389,7 +478,7 @@ pub fn get_batched_lsp_operations_report_no_cursor(
     files: &[(&'static str, &str)],
     get_report: impl Fn(&State, &Handle) -> String,
 ) -> String {
-    let (handles, state) = mk_multi_file_state(files, true);
+    let (handles, state) = mk_multi_file_state(files, Require::indexing(), true);
     let mut report = String::new();
     for (name, _code) in files {
         report.push_str("# ");
@@ -412,7 +501,7 @@ pub fn init_test() {
     init_thread_pool(ThreadCount::NumThreads(NonZeroUsize::new(3).unwrap()));
 }
 
-/// Shared state with all the builtins already initialised (by a dummy module).
+/// Shared state with all the builtins already initialized (by a dummy module).
 static SHARED_STATE: LazyLock<State> =
     LazyLock::new(|| TestEnv::one("_shared_state", "").to_state().0);
 
@@ -424,19 +513,23 @@ pub fn testcase_for_macro(
     line: u32,
 ) -> anyhow::Result<()> {
     init_test();
-    let is_empty_env = env == TestEnv::default();
+    let is_empty_env = env == TestEnv::new();
     let mut start_line = line as usize + 1;
-    if !env.modules.is_empty() {
+    if !env.modules.is_empty() || !env.site_package_path.is_empty() {
         start_line += 1;
     }
     let contents = format!("{}{}", "\n".repeat(start_line), contents);
     env.add_with_path("main", file, &contents);
-    // If any given test regularly takes > 10s, that's probably a bug.
+    // If any given test regularly takes > 20s, that's probably a bug.
     // Currently all are less than 3s in debug, even when running in parallel.
-    let limit = 10;
+    let limit = 20;
+    let check = |errors: Errors| {
+        errors.check_against_expectations()?;
+        errors.check_var_leak()?;
+        Ok::<(), anyhow::Error>(())
+    };
     for _ in 0..3 {
         let start = Instant::now();
-        // Disable the empty test optimisation for now, because it doesn't catch breakages
         if is_empty_env {
             // Optimisation: For simple tests, just reuse the base state, to avoid rechecking stdlib.
             let mut t = SHARED_STATE.transaction();
@@ -447,28 +540,21 @@ pub fn testcase_for_macro(
             );
             t.set_memory(vec![(
                 PathBuf::from(file),
-                Some(Arc::new(contents.clone())),
+                Some(Arc::new(FileContents::from_source(contents.clone()))),
             )]);
-            t.run(&[(h.dupe(), Require::Everything)]);
+            t.run(&[h.dupe()], Require::Everything);
             let errors = t.get_errors([&h]);
-            print_errors(&errors.collect_errors().shown);
-            errors.check_against_expectations()?;
+            let project_root = PathBuf::new();
+            print_errors(project_root.as_path(), &errors.collect_errors().shown);
+            check(errors)?;
         } else {
             let (state, handle) = env.clone().to_state();
             let t = state.transaction();
             // First check against main, so we can capture any import order errors.
-            t.get_errors([&handle("main")])
-                .check_against_expectations()?;
+            check(t.get_errors(&[handle("main")]))?;
             // THen check all handles, so we make sure the rest of the TestEnv is valid.
-            let handles = env
-                .modules
-                .keys()
-                .map(|x| handle(x.as_str()))
-                .collect::<Vec<_>>();
-            state
-                .transaction()
-                .get_errors(handles.iter())
-                .check_against_expectations()?;
+            let handles = env.modules.map(|(x, _, _)| handle(x.as_str()));
+            check(state.transaction().get_errors(handles.iter()))?;
         }
         if start.elapsed().as_secs() <= limit {
             return Ok(());
